!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file MainLoop_sda_pmf.f90
!! Loop for sda_pmf

!> Main function for sda_pmf
!!
!! See the ProMetCS paper for the theory of the method:
!!
!!   ProMetCS: An Atomistic Force Field for Modeling Protein−Metal Surface Interactions in a Continuum Aqueous Solvent
!!   Daria B. Kokh, Stefano Corni, Peter J. Winn, Martin Hoefling, Kay E. Gottschalk, and Rebecca C. Wade
!!   Journal of Chemical Theory and Computation 2010 6 (5), 1753-1768.
!!
!! The entropic terms are computed using the method described in:
!!
!!   Association Entropy in Adsorption Processes
!!   Nir Ben-Tal, Barry Honig, Carey K. Bagdassarian and Avinoam Ben-Shaul
!!   Biophysical Journal 2000 79, 1180-1187.
!!
!!
!! This version tries to run everything in one parallel loop to \n
!! avoid the allocation/deallocation/copy of tab_proteins and array energy
!!
!! In this version use section and nested loop (to activate in main.f90)\n
!! use nested parallel do in comp_pmf\n
!! Loose energymin to avoid CRITICAL section, provided in trajectory
!!
!! Flexibility is not implemented.
!!
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param o_complexes : instance of \ref mod_record::record "record" for complexes output
!! @param trajectories : instance of \ref mod_record::record "record" for trajectory output
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param param_analytic : structure \ref com_sda::parameter_analytic "parameter_analytic"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param p_restart : pointer to the restart file
subroutine MainLoop_sda_pmf ( tab_protein, o_complexes, trajectories, &
                              param_force_energy, param_analytic, param_metaldesolv, param_pmf, geom, type_calc )

!$  USE omp_lib
    USE com_sda
    USE mod_set_parameters
    USE mod_gridtype
    USE mod_array_protein ! include mod_setofgrid, which includes mod_grid
    USE mod_force_energy
    USE mod_compute_energy
    USE mod_compute_image_charge
    USE mod_compute_metaldesolv
    USE mod_record
    USE mod_geometry
    USE AllTimer

    IMPLICIT NONE

    interface

            subroutine convert_to_record_format ( array_energy, array_energy_1d, max_bit_energy )
                IMPLICIT NONE
                real(kind=8), dimension(:,:), intent ( in ) :: array_energy
                real(kind=8), dimension(:,:), intent ( out ) :: array_energy_1d
                integer, intent ( in ) :: max_bit_energy
            end subroutine convert_to_record_format

            subroutine compute_pmf(ent, n_theta, n_phi, n_psi, n_x, n_y, dA, dH_z, dS_z, partitionF_zcoord )
                IMPLICIT NONE
                integer, intent (in) :: n_theta, n_phi, n_psi, n_x, n_y
                real ( kind=8 ), dimension (:,:,:,:,:), intent (in) :: ent
                real ( kind=8 ), intent (in) :: partitionF_zcoord, dA
                real ( kind=8 ), intent (out) :: dH_z, dS_z
             end subroutine compute_pmf

    end interface

    type ( array_protein_type ), target :: tab_protein
    type ( record ) :: o_complexes, trajectories
    type ( type_force_energy ) :: param_force_energy
    type ( parameter_analytic ) :: param_analytic
    type ( geometry ) :: geom
    type ( type_calculation ) :: type_calc
    type ( parameter_metaldesolv ) :: param_metaldesolv
    type ( parameter_pmf ) :: param_pmf


    real ( kind=8 ), dimension(:,:), pointer:: array_energy
    real ( kind=8 ), dimension( : ), pointer :: dummy_array

    ! make a general pointer ptab_protein, allows to use one variable only, needed for openmp
    type ( array_protein_type ), pointer :: ptab_protein
    ! local variables, used for simplicity
    type ( protein ), pointer :: protein1, protein2

    integer :: max_energy_terms, size_array_energy
    real ( kind=8 ), dimension(:,:), pointer :: array_energy_1d
    real ( kind=8 ), dimension ( 3,3 ) :: rot

    real ( kind=8 ) :: energy
    real ( kind=4 ) :: Cut2, dist

    integer, dimension(4) :: array_int_complexe ! For recording conformation
    ! Normally allocatable array, 1 for trajectory or 3 for complexes.
    ! If fixed better for old OpenMP ?
    real (kind=8), dimension(1) :: other_real

    logical :: iflex, overlap
	
    !! PMF declarations
    ! Number of phi, theta and psi angles and, x y and z coordinates to be sampled
    integer :: n_phi, n_theta, n_psi, n_x, n_y, n_z
    integer :: i, ierr, ient, ix, iy, ips, iph, ith
    ! Elements in x, y and z directions and area element (dA)
    real ( kind=8 ) :: dz, dx, dy, dA
    ! Sampling begins at "z_begin" and ends at "z_end" spanning a length of "length_z"
    real ( kind=8 ) :: length_z, z_begin, z_end
    ! Volume element = d_omega * sin_theta
    real ( kind=8 ) :: d_omega, sin_theta
    ! Rotation angles: rot_psi_angle, rot_theta_angle, rot_phi_angle
    real ( kind=8 ) :: rot_psi_angle, rot_theta_angle, rot_phi_angle
    real ( kind=8 ) :: partitionF_zcoord, partitionF_add, partitionF_total
    real ( kind=8 ) :: energymin, local_energymin, global_emin
    ! dSl and dSz stands for orientational and translational entropies (?)
    real ( kind=8 ) :: dG, dH, dH_z, dS_z, dSl, dSz
    real ( kind=8 ) :: pmf_z, global_min_pmf, state_probability
    real ( kind=8 ) :: sum_pmf, sum_state_probability

    ! Store the local partition_function sum (qqz) and entropy (ssz) for each dz to be sampled.
    real ( kind=8 ), dimension (:), allocatable :: qqz, ssz
    ! 3 rotation angles + x and y
    real ( kind=8 ), dimension (:,:,:,:,:), allocatable :: ent
    ! array pmf,... only for tests of consistency (?)
    real ( kind=8 ), dimension (:), allocatable :: array_pmf

    ! Energy penalty when there is a sterical clash between the structures
    ! Following SDA6, penalty =  10.0 kT
    real ( kind=8 ), parameter :: PENALTY_OVERLAP = 10.d0
    real ( kind=8 ), parameter :: PI_C = acos(-1.0) ! pi value as a constant parameter
    !$  integer::tid

    !! end of declarations

    iflex = .false.
    size_array_energy = 0
    max_energy_terms = 0
    nullify ( array_energy, array_energy_1d )
    Cut2 = 0.

    write(*,*)
    write(*,*) 'Entry MainLoop_sda_pmf'
    ! write(*,*) 'exp(- PENALTY_OVERLAP)=', exp(-1.d0 * PENALTY_OVERLAP)

    !We need the trajectory for the minimum energy orientations but should it always be required?
    if ( associated ( trajectories % ptrajectory ) .eqv. .false. ) then
        write(*,*) 'ERROR A trajectory must be defined to run a sda_pmf calculation'
        stop 1
    end if

    ! intialize the array, use only protein 1 information
    call init_force_energy ( param_force_energy, tab_protein, type_calc, param_metaldesolv % imetaldesolv )

    ! new function, set up only shared variable
    call set_dummy_array_and_logical ( tab_protein, dummy_array, iflex )

    !NJB: Setup cut array
    call set_cut_array(tab_protein, geom, Cut2)
    write(*,*) 'Interaction cutoff square, Cut2 ', Cut2

    ! Print out the summary data
    call print_info_param_fe ( param_force_energy )

    if (param_pmf % ipmf .eqv. .true.) then
        n_theta = param_pmf % mentth
        n_phi   = param_pmf % mentfi
        n_psi   = param_pmf % mentom
        n_x     = param_pmf % mentx
        n_y     = param_pmf % menty
        z_end   = param_pmf % pfinish
     else
        write(*,*) " WARNING! PMF group is empty!!! "
        STOP 1
     end if

    ! Allocate memory for energy matrix, shared array, but writen only by one thread at a place
    allocate(ent(n_theta, n_phi, n_psi, n_x, n_y))
    ent = 0.d0

#ifdef VERBOSE
     write(*,*) 'initial position of prot2'
     write(*,*) 'pos ', tab_protein % array ( 2 ) %  position
     write(*,*) 'orientx ', tab_protein % array ( 2 ) % orientation_x
     write(*,*) 'orienty ', tab_protein % array ( 2 ) % orientation_y
     write(*,*) 'orientz ', tab_protein % array ( 2 ) % orientation_z
#endif

!!!!!!!!!!!!!!!!!
     if ( param_pmf % pstart  > 0.d0 ) then
         z_begin  = param_pmf % pstart
     else
         !z_begin = tab_protein % array ( 2 ) % p_sogrid % distance_min + 3.0
         z_begin = 2.d0
     endif

     !Only a 6A x 6A area of the surface is sampled.
     !This can be changed depending on your need.
     dx = 6.0/(real(n_x))
     dy = 6.0/(real(n_y))

     ! Element of surface
     dA = dx*dy

     ! Volume element for the sampling of the Euler space:
     !
     ! d_omega = sin(theta)* dphi * dtheta * dpsi
     ! where:
     ! dphi   = 2.d0 * (PI_C/n_phi)
     ! dtheta = 2.d0 * (PI_C/n_theta)
     ! dpsi   = 2.d0 * (PI_C/n_psi)

     ! Note that 'd_omega' variable represents the real volume element
     ! when multiplied by the 'sin_theta' variable!
     d_omega = ( 2.d0*PI_C/n_phi ) * ( PI_C/n_theta ) * ( 2.d0*PI_C/n_psi )

     ! Length of the profile
     length_z = z_end - z_begin

     if ( param_pmf % dZ < 0.d0 ) then
        dz = 0.2d0
     else
        dz = param_pmf % dZ
     end if

     n_z = ceiling( length_z/dz )
     allocate( qqz(n_z) )
     allocate( ssz(n_z) )
     allocate( array_pmf(n_z) )

     qqz = 0.d0
     ssz = 0.d0
     array_pmf = 0.d0

!#ifdef DEBUG
     write(*,*) 'Sampling begins at distance   : ', z_begin
     write(*,*) 'Sampling ends at distance     : ', z_end
     write(*,*) 'Total distance of z           : ', length_z
     write(*,*) 'Number of steps along z coord : ', n_z
     write(*,*) 'Step size (dz)                : ', dz
!#endif

    ! open output file
    open(iopmf,file='pmf.dat',status='UNKNOWN')
    WRITE(iopmf,*) '         z          Q(z)           dH(z)         dS(z)         dG(z)'

    ! shared variable not modified in OpenMP loop
    global_min_pmf=0.d0
    global_emin=0.d0
    dH=0.d0
    ! inserted in the main loop, sum over all z
    partitionF_total=0.d0
    partitionF_zcoord = 0.d0
    energymin = PENALTY_OVERLAP


!$OMP PARALLEL  DEFAULT(NONE) &
!$OMP SHARED (tab_protein, o_complexes, trajectories,  dummy_array, type_calc, &
!$OMP         ent,n_theta,n_phi,n_psi, n_x, n_y, n_z, &
!$OMP         partitionF_zcoord, partitionF_total, energymin, &
!$OMP         param_force_energy,param_metaldesolv, &
!$OMP         dx,dy,dz,dA,d_omega, &
!$OMP         global_emin, global_min_pmf, pmf_z, qqz, ssz, dH, &
!$OMP         z_begin, array_pmf, EXIT_PROG ) &
!$OMP PRIVATE (ips,iph,ith,rot_theta_angle,rot_phi_angle,rot_psi_angle, ient, &
!$OMP          max_energy_terms, size_array_energy, local_energymin, &
!$OMP          array_energy, array_energy_1d, &
!$OMP          energy, rot, sin_theta, &
!$OMP          ptab_protein,protein1,protein2, &
!$OMP          ix,iy,dist, partitionF_add, tid, ierr, &
!$OMP          dH_z, dS_z, overlap, &
!$OMP          array_int_complexe, other_real) !timer_thread)


!$  tid = omp_get_thread_num()
    nullify ( protein1, protein2 )
    nullify ( ptab_protein )
    nullify ( array_energy, array_energy_1d)

    ! reinitialize private variables
    array_int_complexe = 0
    other_real = 0.d0


! local copy for tab_protein, master points to the shared
! associate pointer ptab_protein to a copy of initial tab_protein if OpenMP, master points to the original
!$  if ( tid > 0 ) then
!! !$      write (*,*) 'openmp activated make copy_array_protein for tid', tid
!$      call copy_array_protein ( tab_protein, ptab_protein, type_calc, param_metaldesolv )
    ! normal if not OpenMP or master
!$  else
        ptab_protein => tab_protein
!$  end if

    protein1 => ptab_protein % array ( 1 )
    protein2 => ptab_protein % array ( 2 )

    ! need trajectory to set array_energy. use mod_force_energy would be better ?
    call set_array_energy ( type_calc, trajectories, size ( ptab_protein % array ), &
                            array_energy, max_energy_terms )
    ! only need for recording complexes/trajectories
    if ( trajectories % ntraj_rec /= 0 ) then
        size_array_energy = get_nbterms_energy_compatible ( trajectories, .true. )
        allocate ( array_energy_1d ( size_array_energy, 1 ), stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'Error in allocation of array_energy size ', size_array_energy
            stop 1
        endif
        array_energy_1d = 0.d0
    end if


#ifdef VERBOSE
!$OMP CRITICAL( crit_check_pos )
    !write(*,*) 'fix position of solute 1'
    !write(*,*) 'pos ', tab_protein % array ( 1 ) %  position
    !write(*,*) 'orientx ', tab_protein % array ( 1 ) % position
    !write(*,*) 'orienty ', tab_protein % array ( 1 ) % orientation_y
    !write(*,*) 'orientz ', tab_protein % array ( 1 ) % orientation_z
    write(*,*) 'init position of solute 2'
    write(*,*) 'pos ', protein2 % position
    write(*,*) 'orientx ', protein2 % orientation_x
    write(*,*) 'orienty ', protein2 % orientation_y
    write(*,*) 'orientz ', protein2 % orientation_z
    write(*,*) 'pos atom surf1', protein1 % p_list_atom_surface(:,1), &
            protein1 % p_list_atom_surface( :,protein1 % nb_atom_surface )
    write(*,*) 'pos charge2', protein2 % p_pos_charge(:,1) , &
                protein2 % p_pos_charge( :,protein2 % nb_charge )
    write(*,*) 'pos atom surf2', protein2 % p_list_atom_surface(:,1) , &
            protein2 % p_list_atom_surface( :,protein2 % nb_atom_surface )
!$OMP END CRITICAL( crit_check_pos )
#endif

    ! start sampling over z, serial(?) part here, 'ient' must be private!
LOOP_Z: do ient = 1, n_z

        ! signal handling, will stop the job. Maybe here can print properly the available data for restart
!        if ( EXIT_PROG .eqv. .true. ) write(*,*) 'thread got the message exit tid', tid
        !NJB: Removed reference to OMP only variable 
        if ( EXIT_PROG .eqv. .true. ) write(*,*) 'One thread got the message exit '

        !private variable, reinitialize
        local_energymin = PENALTY_OVERLAP

        ! only when change step
        array_int_complexe(2) = ient

        ! Set initial position vector
        protein2 % position(1) = 0.0
        protein2 % position(2) = 0.0
        protein2 % position(3) = real(ient) * dz + z_begin

#ifdef VERBOSE
    !$OMP CRITICAL(crit_init_pos)
!$  write(*,*) 'tid ',tid
    write(*,*) 'position after copy movable array', protein2 % position
    write(*,*) 'rot1 ', protein2 % orientation_x
    write(*,*) 'rot2 ', protein2 % orientation_y
    write(*,*) 'rot3 ', protein2 % orientation_z
    !$OMP END CRITICAL(crit_init_pos)
#endif

! barrier may be necessary here, seems implict barrier only at the end
!$OMP BARRIER

!$OMP DO SCHEDULE( DYNAMIC )
LOOP_OMP:  do ips=1,n_psi
        ! loop over psi
        rot_psi_angle = real( ips-0.5d0 ) * 2.d0 * PI_C / real(n_psi)

        ! loop over theta
        do ith=1,n_theta
            rot_theta_angle = real( ith-0.5d0 ) * PI_C / real(n_theta)

            sin_theta = sin( rot_theta_angle )

            ! loop over phi
            innermost: do iph=1,n_phi
                rot_phi_angle = real( iph-0.5d0 ) * 2.d0 * PI_C / real(n_phi)

                ! Michael's note:
                ! Tested this function in test_module/test_orientation rotation at start,
                ! should reproduce sda 6
                ! need to invert the angles, intrinsic->extrinsic
                call rotation_z1x2z3(rot, rot_psi_angle, rot_theta_angle, rot_phi_angle)

                ! apply the rotation on the original position (1,0,0),(0,1,0),(0,0,1)
                call copy_movable_array (protein2, rot)

                ! distance will not change
                call compute_distance ( protein1 , protein2 , dist )

                ! check for overlap do not apply the rotation here, already correct
                call check_exclusion ( protein1, protein2, protein2 % position, &
                                       rot, overlap, .false. )

                ! now can keep the same orientation, sample area (x,y)
                ! distance computed from the surface, only z is accounted
                ! Fill array 'ent' even if there is an overlap!
                do ix=1, n_x
                    do iy=1, n_y

                        ! Penalty if overlap, everything will be filled, correct normalisation factor
                        if ( overlap .eqv. .true. ) then
                            energy = PENALTY_OVERLAP

                        else
                            ! update the position of the center of the protein for x and y coordinates
                            protein2 % position(1) = real(ix-1) * dx
                            protein2 % position(2) = real(iy-1) * dy

                            energy = 0.d0
                            array_energy = 0.d0

                            !compute interaction energy and image charge energy
                            call energy_epedhdlj_2proteins ( protein1, protein2, param_force_energy, &
                                dist, array_energy, dummy_array )

                            if ( param_force_energy % iimg_chg ) then
                                call energy_image_charge ( ptab_protein, param_force_energy, array_energy )
                            endif

                            if ( param_force_energy % imetaldesolv ) then
                                call compute_energy_metaldesolv_2prot( ptab_protein, param_metaldesolv, array_energy )
                            end if

                            energy = sum ( array_energy(:,2) ) + sum ( array_energy( :, 1 ) )

                        endif

!#ifdef VERBOSE
!        write(*,*) '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
!        write(*,*) ' ENERGY '
!        write(*,*) ' total energy ', energy !array_energy( 1, 2 )
!        write(*,*) ' potential energy', array_energy_local ( 2 , 1 ), array_energy_local (2, 2)
!        write(*,*) ' desol energy', array_energy_local( 3, 1 ), array_energy_local( 3, 2 )
!        write(*,*) ' hydro. energy', array_energy_local( 4, 1 ), array_energy_local( 4 , 2 )
!        write(*,*) ' lj energy', array_energy_local( 5, 1 ), array_energy_local( 5 , 2 )
!        write(*,*) ' img_chg energy', array_energy( 7, 1 ), array_energy ( 7 , 2 )
!        write(*,*) ' metal desol energy', array_energy( 8, 1 ), array_energy( 8, 2 )
!#endif

                        ! update the energy matrix, each thread different ips
                        ent(ith,iph,ips,ix,iy) = energy

#ifdef VERBOSE
                        if ( sin(rot_theta_angle) < 0.d0 ) then
                            write(*,*) 'Sinus theta is negative !!'
                        endif
#endif
                        ! Partition function
                        partitionF_add = sin_theta * d_omega * dA * exp( -1d0 * energy )

                        ! compute partition function for each z distance
                        !$OMP ATOMIC
                        partitionF_zcoord = partitionF_zcoord + partitionF_add

                        partitionF_add = partitionF_add * dz

                        ! partitionF_total is the total partition function
                        !$OMP ATOMIC
                        partitionF_total = partitionF_total + partitionF_add

                        ! Keep record of the details of the minimum energy orientation
                        ! find min energy at each z and save conformation in the file
                        if (local_energymin .gt. energy) then
                            local_energymin = energy

                            ! if no trajectory to ouput, no need this part
                            if ( trajectories % ntraj_rec /= 0 ) then
                                !write(*,*) 'got energymin, tid/energy ', tid, local_energymin

                                ! put total energy here, could change, in energy of prot2
                                array_energy(1,2) = energy
                                call convert_to_record_format ( array_energy, array_energy_1d, max_energy_terms )

                                ! only this part really need to be in CRITICAL, need energymin private
!$OMP CRITICAL( write_onecomplexe )
                                ! access to the current pointer to trajectory, never changed during the full loop
                                ! will overwritte the current one_complexe each time a minimum energy is found
                                ! getNext is called after the loop
                                call set_onecomplexe_wrap( trajectories, protein2 % position, protein2 % box, &
                                    protein2 % orientation_x, protein2 % orientation_y, &
                                    array_int_complexe, array_energy_1d(:,1), other_real )
!$OMP END CRITICAL( write_onecomplexe )

                            end if ! trajectories % ntraj_rec
                        end if ! energy smaller

                    end do ! y
                end do ! x

            end do innermost !theta
        end do !phi

    end do LOOP_OMP !psi

    ! write(*,*) 'wait in barrier after LOOP_OMP tid ', tid
    !$OMP END DO

    ! all threads must have updated energymin, before single section starts
    ! Now want only one thread doing this, or parrallelize inside again

    !$OMP SINGLE

    !write(*,*) 'tid in section write trajectory', tid
    ! compute PMF and dA at current z
    ! compute parameters integrated over z: probability, lowest PMF and lowest energy

#ifdef VERBOSE
    write(*,*) 'Partition function (z): ', partitionF_zcoord
    write(*,*) 'Partition function: ', partitionF_total
#endif

    ! write only if some minenergy have been found, if only overlap onecomplexe is not valid
    if ( trajectories % ntraj_rec /= 0 ) then
    ! moved to next after insertion, total always available ??
        if ( trajectories % p_onecompl % real_ocompl(1) /= 0.d0 ) then
            call getNext ( trajectories, ierr )
            ! here we know the size =
            ! check if array is full, write to disk and return at the beginning of the array
            ! ierr == 1 if getNext outside array, force to write and reset at first
            if ( ierr /= 0 ) then
                call write_array ( trajectories % ptrajectory, trajectories )
                call reset_position ( trajectories )
            end if
        end if
    end if

    ! compute dG
    qqz(ient)=partitionF_zcoord
    ! compute local dG(z), pmf
    ! problem if U=0, should be zero ? bad sampling (to get partitionF_zcoord=denominator)
    ! Yes sampling effect ! incresase samplng it tends to 0
    pmf_z = -log(partitionF_zcoord/(8*PI_C*PI_C*dA*n_x*n_y))
    ! store for later test integration
    array_pmf(ient) = pmf_z

    ! global_min_pmf/global_min_pmf
    if(pmf_z .lt. global_min_pmf) global_min_pmf = pmf_z

    !write(*,*) 'section compute entropy', tid
    ! can use open mp inside ?
    ! only for entropy/dS_z) and enthalpy(dH_z), pmf need only Zs,omega = partitionF_zcoord
    call compute_pmf(ent, n_theta, n_phi, n_psi, n_x, n_y, dA, dH_z, dS_z, partitionF_zcoord)


    write(iopmf,*) (real(ient)-0.5)*dz+z_begin, partitionF_zcoord, dH_z, dS_z, pmf_z
    flush(iopmf)


    ! entropy for each term z
    ssz(ient)=dS_z
    ! Sum <U> along z, multiplied by partitionF_zcoord here and divided at the end by partitionF_total(sum partitionF_zcoord.DZ)
    ! This gives different results than only summing dz*dH_z
    !dH = dH + dz * dH_z * partitionF_zcoord !qqz(ient)
    ! will be normalized by the full partition function partitionF_total at the end
    dH = dH + dz * dH_z

    !write(*,*) 'single reinitialize variables', tid
    ! more secure here, only one thread must reinitialize before the next loop
    energymin = PENALTY_OVERLAP
    partitionF_zcoord = 0.d0
    ent = 0.d0
    !$OMP END SINGLE

end do LOOP_Z

    ! need to delete all private variables
    if ( associated ( array_energy ) ) then
        deallocate ( array_energy, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocating array_energy '
            STOP 1
        end if
        nullify ( array_energy )
    end if

    if ( associated (array_energy_1d )) then
        deallocate( array_energy_1d, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocating array_energy '
            STOP 1
        end if
        nullify ( array_energy )
    endif

!$OMP END PARALLEL

    ! correct final division by volume(Qb) * 8 pi^2(Qf)
    dG = -log( partitionF_total/ (length_z*dA*n_x*n_y * 8*PI_C*PI_C) )
    ! normalize by the full partition function only at the end (we do not do it before?!)
    dH = dH / partitionF_total

    ! compute adsorption entropy change, set to 0 in the code ?
    dSl=0.0
    dSz=0.0
    do i=1, n_z
        state_probability=qqz(i)/partitionF_total
        dSz=dSz-dz*state_probability*log(8*PI_C*PI_C*state_probability*(dA*n_x*n_y)*length_z)
        dSl=dSl+dz*state_probability*ssz(i)
    end do

    write(*,*) 'dG; dH; TdS(=dH-dG); dSz; dSl',dG, dH, dH-dG, dSz, dSl
    write(*,*) 'Min PMF: ', global_min_pmf

    ! add for test, consistent results Probability(z) <=> PMF(z)
    ! Probability if no interaction = 1/L

    write(*,*) '========='

    ! expects to get the Delta G^0. Due to penalty last number are wrong
    write(*,*) 'Test PMF integration'
    sum_pmf = 0.d0

    do i= n_z, 1, -1
        sum_pmf = sum_pmf + dz * array_pmf(i)
        write(*,*) 'i, sum pmf', i, sum_pmf
    enddo

    write(*,*) 'Test probability / PMF'
    write(*,*) 'Probability P(z),        PMF,      P(z)= exp(-beta PMF(z))/L'

    ! to check sum of probability = 1, ok independent sampling
    ! if no interaction (or constant) P(z) = 1/L, ok, independent sampling
    ! P(z) = exp(-beta PMF(z)) / L
    ! if U=0 dG=0, PMF(z)=0 ?
    sum_state_probability = 0.d0
    do i = 1, n_z
        ! probability to be at z (qqz = Zs,omega(z) = partitionF_zcoord) / (Zz,s,omega = full partition function)
        state_probability = qqz(i) / partitionF_total
        ! only for test should give 1
        sum_state_probability = sum_state_probability + dz * state_probability

        write(*,*) i, state_probability*dz, array_pmf(i),exp( -1.d0 * array_pmf(i)) / ( length_z )
    end do

    write(*,*) 'Sum of partition(z) / partition =1', sum_state_probability

    deallocate(ent)

    ! NEED to explicitly call write_record with complexes
    ! should NOT for trajectory done when deleted, could do the same for complexe ( but option comput_dispersion or not )
    ! use default value true, recompute dispersion
    !call write_record ( o_complexes )
    !call write_record( trajectories )

    close (iopmf)
    deallocate ( dummy_array )

    write(*,*) 'Exit MainLoop_sda_energy'

end subroutine MainLoop_sda_pmf

! can keep in function ?, or need to include n the same function

!> Subroutine to compute the pmf value for a given z above the surface.\n
!! rotational and translation configurational distributions c and
!! write occupation map (translational and rotational)
!! norm to 1; dV=dT*dz- 3-dimentional space volume = zmax*4pi
!!
!! state_probability*dT*dz is the probability of  finding the protein at Y+dT, Z+dz
!! dH- change of enthalpy, dSt - change of translational entropy; dSr- rot entropy
!!
!! @param dA surface area sampled along x/y, dA = dx * dy
!! @param dH_z return the local enthalpy, H = <U>
!! @param dS_z return the local entropy
subroutine compute_pmf(ent,n_theta,n_phi,n_psi,n_x,n_y,dA,dH_z,dS_z,partitionF_zcoord )
!$  USE omp_lib
    implicit none
    integer, intent (in) :: n_theta, n_phi, n_psi, n_x, n_y
    real ( kind=8 ), dimension (:,:,:,:,:), intent (in) :: ent
    real ( kind=8 ), intent (in) :: partitionF_zcoord, dA
    real ( kind=8 ), intent (out) :: dH_z, dS_z

    integer :: ix, iy, p, k, j
    real ( kind=8 ) :: rot_theta_angle, rot_phi_angle, rot_psi_angle
    real ( kind=8 ) :: dT, state_probability, tmp_dH_z, tmp_dS_z
    !integer :: tid

    real (kind=8), parameter :: PI_C = acos(-1.0)

    dH_z = 0.d0
    dS_z = 0.d0
    state_probability = 0.d0

#ifdef VERBOSE
    write(*,*) 'Entry comp_pmf '
    write(*,*) 'partitionF_zcoord ', partitionF_zcoord
    write(*,*) 'dA', dA
#endif

    !$OMP PARALLEL DO DEFAULT( none ) &
    !$OMP SHARED (n_theta, n_psi, n_phi, n_x, n_y, partitionF_zcoord,dA, &
    !$OMP         dH_z, dS_z, ent ) &
    !$OMP PRIVATE ( rot_theta_angle, rot_psi_angle, rot_phi_angle, dT, p, j, ix, iy, state_probability, tmp_dH_z, tmp_dS_z)
    do k=1,n_theta
        rot_theta_angle = (k-0.5)*PI_C/n_theta
        dT = sin(rot_theta_angle) * (2*PI_C/real(n_phi)) * (PI_C/real(n_theta)) * (2*PI_C/real(n_psi))

        do p=1,n_psi
            rot_psi_angle = (p-0.5)*2*PI_C/n_psi

            do j=1,n_phi
                rot_phi_angle = (j-0.5)*2*PI_C/n_phi

                do ix=1,n_x ! energy sampled over xy
                    do iy=1,n_y ! energy sampled over xy

                        ! probability factor
                        state_probability = exp(-ent(k,j,p,ix,iy))

                        tmp_dH_z = dT*dA * ent(k,j,p,ix,iy) * exp(-ent(k,j,p,ix,iy))
                        !$OMP ATOMIC
                        dH_z = dH_z + tmp_dH_z

                        ! local entropy
                        if (state_probability.gt.1.e-20)   then
                            tmp_dS_z = dT*(state_probability/partitionF_zcoord)*dA*&
                            log(8*PI_C*PI_C*state_probability*(dA*real(n_x*n_y))/partitionF_zcoord)
                            !$OMP ATOMIC
                            dS_z = dS_z - tmp_dS_z
                        end if
                    end do
                end do
            end do
        end do
    end do  ! energy sampled over xy
   !$OMP END PARALLEL DO

#ifdef VERBOSE
    write(*,*) 'out conf_pmf'
    write(*,*) 'dH_z =<U>', dH_z
    write(*,*) 'dS_z ', dS_z
#endif

    !return
end subroutine compute_pmf

!> compute the rotation matrix for a given set of angles.\n
! This function was used in SDA6 and should be obsolete in SDA7!
! Which format does this function follow?
subroutine rotation_pmf_sda6(rot, rotentfi,rotentth,rotentom)
    IMPLICIT NONE
    real ( kind=8 ), intent (IN) :: rotentfi, rotentth, rotentom
    real ( kind=8 ) PI_C, c1, c2, c3, s1, s2, s3
    real ( kind=8 ), dimension ( 3 ) :: om
    real ( kind=8 ), dimension ( 3,3 ), intent (INOUT) :: rot

    PI_C = acos(-1.0)
    om(1)=rotentfi
!   rotation around x
    om(2)=rotentth
!   rotation around old z
    om(3)=rotentom
!   rot around new z
    c1 = dcos(om(1))
    s1 = dsin(om(1))
    c2 = dcos(om(2))
    s2 = dsin(om(2))
    c3 = dcos(om(3))
    s3 = dsin(om(3))
    rot(1,1) = c1*c3-s1*c2*s3
    rot(1,2) = c1*s3+s1*c2*c3
    rot(1,3) = s1*s2
    rot(2,1) = -s1*c3-c1*c2*s3
    rot(2,2) = -s1*s3+c1*c2*c3
    rot(2,3) = s2*c1
    rot(3,1) = s2*s3
    rot(3,2) = -s2*c3
    rot(3,3) = c2
    return
end subroutine rotation_pmf_sda6


!> To get an extrinsic rotation with correct angles.
!
! Give rotation(phi around X), then rotation(theta arond Y) and rotation(psi around Z) in a fixed frame, as expected.
!
! \note wikipedia 'the intrinsic rotations x-y’-z″ by angles α, β, γ are equivalent\n
!       to the extrinsic rotations z-y-x by angles γ, β, α'\n
!
! This subroutine must be called rotation_z1y2x3_tbryan(rot, rotentom, rotentth, rotentfi)\n
! The order phi, theta, psi is inversed !\n
! We use intrinsic gamma, beta, alpha to get extrinsic alpha, beta, gamma.\n
!
! To generate a transformation which will apply (in the order):\n
! - rotation phi around y
! - rotation theta around y
! - rotation psi around z\n
! They will apply an extrinsic transformation (fixed frame) if called with left multiplication: new_position = rot . old_position\n
!
!
! Function tested manually (step of pi/2 for all angles)\n
! 122 (phi, theta, psi) : x->-z, y->-x, z->y \n
! 221 x-> -z, y->x, z-> -y \n
! 223 = 324  x-> -z, y-> -x, z-> y
subroutine rotation_z1y2x3_tbryan(rot, rotentfi,rotentth,rotentom)
    implicit none
    real ( kind=8 ), intent (in) :: rotentfi, rotentth, rotentom
    real ( kind=8 ), dimension ( 3,3 ), intent (out) :: rot

    real ( kind=8 ) :: c1, c2, c3, s1, s2, s3
    real ( kind=8 ), dimension ( 3 ) :: om

    om(1)=rotentfi
!   First rotation around x
    om(2)=rotentth
!   rotation around y
    om(3)=rotentom
!   rot around  z
    c1 = dcos(om(1))
    s1 = dsin(om(1))
    c2 = dcos(om(2))
    s2 = dsin(om(2))
    c3 = dcos(om(3))
    s3 = dsin(om(3))

    rot(1,1) = c1*c2
    rot(1,2) = c1*s2*s3 - c3*s1
    rot(1,3) = s1*s3 + c1*c3*s2
    rot(2,1) = c2*s1
    rot(2,2) = c1*c3 + s1*s2*s3
    rot(2,3) = c3*s1*s2 - c1*s3
    rot(3,1) = -s2
    rot(3,2) = c2*s3
    rot(3,3) = c2*c3

end subroutine rotation_z1y2x3_tbryan

!> Try to follow wikipedia, z1x2z3
!
! From first test / comparison wikipedia:
! - http://en.wikipedia.org/wiki/Euler_angles
! - http://en.wikipedia.org/wiki/Rotation_matrix
!
! Euler notation is Z1,X2,Z3, in active and intrinsic rotation
!
! To pass from intrinsic(phi, theta, psi) to extrinsic invert phi and psi !
subroutine rotation_z1x2z3(rot, rotentfi,rotentth,rotentom)
    implicit  none
    real ( kind=8 ), intent (in) :: rotentfi, rotentth, rotentom

    real ( kind=8 ) :: c1, c2, c3, s1, s2, s3
    real ( kind=8 ), dimension ( 3 ) :: om
    real ( kind=8 ), dimension ( 3,3 ), intent (out) :: rot

    om(1)=rotentfi
!   rotation around old z
    om(2)=rotentth
!   rotation around old y
    om(3)=rotentom
!   rot around old z
    c1 = dcos(om(1))
    s1 = dsin(om(1))
    c2 = dcos(om(2))
    s2 = dsin(om(2))
    c3 = dcos(om(3))
    s3 = dsin(om(3))
    rot(1,1) = c1*c3-c2*s1*s3
    rot(1,2) = -c1*s3-c2*c3*s1
    rot(1,3) = s1*s2
    rot(2,1) = s1*c3+c1*c2*s3
    rot(2,2) = -s1*s3+c1*c2*c3
    rot(2,3) = -s2*c1
    rot(3,1) = s2*s3
    rot(3,2) = s2*c3
    rot(3,3) = c2

end subroutine rotation_z1x2z3
