!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file MainLoop_sdamm.f90
!! Loop in the case of multiple proteins

!> Main function for sdamm
!!
!! In SDAMM ( SDA with Multiple Molecules ) we simulate crowding effects\n
!! Many proteins are inside a box ( with or without PBC (periodic boundary conditions ) and are simulated for a definite time\n
!! The main output is a trajectory file
!!
!! The case where a solute is a surface is treated slightly differently ( use 2 cutoffs, modify PBC )\n
!! Some, or all solutes may also be flexible
!!
!! Only parts of the computation is parrallelised :
!! - Force computation
!! - Energy computation
!! - Brownian Dynamics move
!! - Radial Distribution Function ( RDF ) in compute_rdf() ( to check efficient )
!!
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param rcriteria : reaction criteria
!! @param o_complexes : instance of \ref mod_record::record "record" for complex output
!! @param trajectories : instance of \ref mod_record::record "record" for trajectory output
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param resid_time : instance of \ref mod_residencetime::residence_time "residence_time", here used for radial distribution function
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param param_timestep : instance of \ref com_sda::parameter_timestep "parameter_timestep"
!! @param param_force_energy : structure \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param param_analytic : instance of \ref com_sda::parameter_analytic "parameter_analytic"
!! @param param_metaldesolv : instance of \ref com_sda::parameter_metaldesolv "parameter_metaldesolv"
!! @param timemax : length of the trajectory
subroutine MainLoop_sdamm ( tab_protein, rcriteria, o_complexes, trajectories, geom, resid_time, &
                            type_calc, param_timestep, param_force_energy, param_analytic, param_metaldesolv, timemax )

    USE mod_array_protein
    USE mod_ReactionCriteria
    USE mod_geometry
    USE mod_ResidenceTime
    USE mod_record
    USE mod_force_energy
    USE com_sda
    USE mod_set_parameters
    USE mod_compute_force_sdamm
    USE mod_compute_energy_sdamm
    USE mod_compute_energy_flex
    USE mod_compute_image_charge
    USE mod_compute_energy
    USE mod_compute_image_charge
    USE mod_hydrodynamics
    USE mod_compute_metaldesolv
    USE mod_compute_analytic_sdamm
    USE AllTimer

    IMPLICIT NONE

    interface

        subroutine change_conformation ( prot, tab_prot, geom, param_force_energy, param_metal, type_calc, dummy_array, &
                                           hit, accepted, nb_overlap_flex, ptrf_energy )
            USE mod_force_energy
            USE mod_array_protein
            USE mod_geometry
            USE mod_type_calc
            USE mod_compute_energy_flex
            USE com_sda
            IMPLICIT NONE
            interface
                subroutine ptrf_energy ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                                         dummy_array, energy, energy_img_chg )
                  USE mod_force_energy
                  USE mod_array_protein
                  USE mod_geometry
                  USE com_sda
                  implicit none
                  type ( protein ), intent(in), target :: prot
                  type ( array_protein_type ), intent( in ), target :: tab_prot
                  type ( geometry ), intent( in ) :: geom
                  type ( type_force_energy ), intent ( in ) :: param_force_energy
                  type ( parameter_metaldesolv ), intent ( in ) :: param_metal
                  real ( kind=8 ), dimension( : ), intent( in ),target :: dummy_array
                  real ( kind=8 ) :: energy, energy_img_chg
                  integer, intent ( in ) :: nb_prot
                end subroutine ptrf_energy
            end interface

            type ( protein ) :: prot
            type ( array_protein_type ), intent( in ), target :: tab_prot
            type ( geometry ), intent( in ) :: geom
            type ( type_force_energy ), intent ( in ) :: param_force_energy
            type ( parameter_metaldesolv ), intent ( in ) :: param_metal
            type ( type_calculation ), intent ( in ) :: type_calc
            real ( kind=8 ), dimension( : ), intent( in ),target :: dummy_array
            integer :: nb_overlap_flex
            logical, intent ( in ) :: hit
            logical, intent ( out ) :: accepted
        end subroutine change_conformation

        subroutine  make_bd_move_sdamm ( tab_prot, geom, nb_prot, &
                                         force, torque, dt )
            USE mod_array_protein
            USE mod_geometry
            IMPLICIT NONE
            type ( array_protein_type ), target :: tab_prot
            type ( geometry ), intent ( in ) :: geom
            real ( kind=8 ), dimension(3 , nb_prot), intent ( in ) :: force, torque
            integer, intent ( in ) :: nb_prot
            real ( kind=8 ), intent ( in ) :: dt
        end subroutine make_bd_move_sdamm

        subroutine compute_rdf ( resid_time, tab_protein, geom )
            USE mod_ResidenceTime
            USE mod_array_protein
            USE mod_geometry
         !$ USE omp_lib
            IMPLICIT NONE
            type ( residence_time ) :: resid_time
            type ( array_protein_type ), intent ( in ) :: tab_protein
            type ( geometry ) :: geom
        end subroutine compute_rdf

        subroutine check_reaction_pair_sdamm ( tab_protein, set1, set2, rcriteria, geom, sum_time, dtnow, &
                                               array_close, nclose, close )
            USE mod_array_protein
            USE mod_ReactionCriteria ! include mod_rate_calculation
            USE mod_geometry
            IMPLICIT NONE
            type ( array_protein_type ) :: tab_protein
            type ( react_criter ) :: rcriteria
            type ( geometry ) :: geom
            real ( kind=4 ), dimension(:,:) :: sum_time
            integer, dimension (:,:) :: array_close
            real ( kind=4 ) :: dtnow
            integer :: set1, set2, nclose
            logical, intent( out ) :: close
        end subroutine check_reaction_pair_sdamm

        subroutine convert_to_record_format ( array_energy, array_energy_1d, max_bit_energy )
                IMPLICIT NONE
                real(kind=8), dimension(:,:), intent ( in ) :: array_energy
                real(kind=8), dimension(:,:), intent ( out ) :: array_energy_1d
                integer, intent ( in ) :: max_bit_energy
        end subroutine convert_to_record_format

    end interface

    type ( array_protein_type ) :: tab_protein
    type ( react_criter ) :: rcriteria
    type ( record ) :: o_complexes, trajectories
    type ( geometry ) :: geom
    type ( residence_time ) :: resid_time
    type ( type_calculation ) :: type_calc
    type ( parameter_timestep ) :: param_timestep
    type ( type_force_energy ) :: param_force_energy
    type ( parameter_analytic ) :: param_analytic
    type ( parameter_metaldesolv ) :: param_metaldesolv
    real ( kind=8 ) :: timemax

    type ( mytimer ) :: timer_force, timer_energy, timer_bd, timer_flex, timer_record, timer_rdf

    integer :: nb_protein, ierr, i
#ifdef VERBOSE
    ! only use in verbose mode
    integer :: j
#endif
    integer :: current_step, n_rdfs
    ! used for repulsive lennardjones, contains only 1.0
    real ( kind=8 ), dimension( : ), pointer :: dummy_array
    real ( kind=8 ), dimension ( :, : ), pointer :: array_energy_sdamm

    real ( kind=8 ), dimension ( :, : ), allocatable :: force, torque
    ! Cutoff when computing force and torque and energy
    real ( kind=4 ) :: Cut2, half_box
    real ( kind=8 ) :: dt, time
    ! for flex
    logical :: iflex, print_step_traj, print_step_compl, accepted_new_conf
    integer :: nb_overlap_flex

    ! for printing steps.log
    character*32 :: steps_str

    integer :: nb_prot1, nb_prot2, size_array_energy
    real ( kind=8 ), dimension(:,:), pointer :: array_energy_sda2prot
    real ( kind=8 ), dimension(:,:), pointer :: array_energy_sda2prot_1d
    real (kind=8 ), dimension(3) :: pos_relat
    real ( kind=4 ) :: dist
    integer :: size_array_energy_sda2prot, size_array_energy_sda2prot_1d
    type ( array_protein_type ) :: dummy_array_protein
    real(kind=8), dimension(3,3) :: mrot

    nb_protein = 0
    ! set up to false by default
    iflex = .false.
    accepted_new_conf = .false.
    print_step_traj = .false.
    print_step_compl = .false.
    n_rdfs = 0
    Cut2 = 0.
    dt = 0.d0
    time = 0.d0
    current_step = 0
    size_array_energy = 0
    n_rdfs = 0
    ! for flexibility
    nb_overlap_flex = 0

    ! initialize correctly the dummy object for correct deletion, default null
    nullify ( dummy_array_protein % array, dummy_array_protein % set_of_protein )
    dummy_array_protein % nb_grids = 0

    nb_prot1 = 0
    nb_prot2 = 0
    size_array_energy_sda2prot = 0
    size_array_energy_sda2prot_1d = 0
    dist = 0.
    pos_relat = (/0.d0, 0.d0, 0.d0/)
    mrot = 0.d0

    ! create steps.log, take into account if fparrellel
    steps_str = 'steps.log'
    ! open file only for master
    ! to test if fparrallel modify name
    open ( io_steps, file = steps_str, form='formatted', status='replace', iostat = ierr )
    if ( ierr /= 0 ) then
        write ( io6, * ) 'ERROR In opening steps.log '
        stop 1
    end if

    ! initialise timer
    call init_mytimer ( timer_force )
    call init_mytimer ( timer_energy )
    call init_mytimer ( timer_bd )
    call init_mytimer ( timer_flex )
    call init_mytimer ( timer_record )
    call init_mytimer ( timer_rdf )

    write (*,*)
    write (*,*) '========================================================================='
    write (*,*) ' Entry main loop sdamm '
    write (*,*)

    ! start setup of param_force_energy, contains informations for force and energies calculations
    call init_force_energy ( param_force_energy, tab_protein, type_calc, &
                             param_metaldesolv % imetaldesolv )
    if ( param_analytic % hydro ) then
        write(*,*) 'Set param_force_energy % ihi to true'
        param_force_energy % ihi = .true.
    end if

    ! for test only if force asymetric, uncomment to force asymetric energy and force
    !param_force_energy % fast_computation = .false.

    ! general variable
    nb_protein = size ( tab_protein % array )

    ! unique function in mod_set_parameters to allocate array_energy
    if ( associated ( trajectories % ptrajectory ) ) then

        call set_array_energy ( type_calc, trajectories, nb_protein, &
                                array_energy_sdamm, opt_size_array_energy = size_array_energy )

    else
        write(*,*)
        write(*,*) 'ERROR Need a trajectory restart for sdamm'
        write(*,*)
        stop 1
    end if

    ! setup only shared variable
    call set_dummy_array_and_logical ( tab_protein, dummy_array, iflex )
    !write(*,*) 'iflex ', iflex
    !write(*,*) 'size_dummy_array ', size ( dummy_array )

#ifdef VERBOSE
    write(*,*) '=========='
    write(*,*) 'init pos prot 1'
    write(*,*) 'pos ', tab_protein % array ( 1 ) % position
    write(*,*) 'x ', tab_protein % array ( 1 ) % orientation_x
    write(*,*) 'y ', tab_protein % array ( 1 ) % orientation_y
    write(*,*) 'z ', tab_protein % array ( 1 ) % orientation_z
    write(*,*) '=========='
    write(*,*) 'init pos prot 2'
    write(*,*) 'pos ', tab_protein % array ( 2 ) % position
    write(*,*) 'x ', tab_protein % array ( 2 ) % orientation_x
    write(*,*) 'y ', tab_protein % array ( 2 ) % orientation_y
    write(*,*) 'z ', tab_protein % array ( 2 ) % orientation_z
    write(*,*) '=========='
    write(*,*) 'init pos prot 3'
    write(*,*) 'pos ', tab_protein % array ( 3 ) % position
    write(*,*) 'x ', tab_protein % array ( 3 ) % orientation_x
    write(*,*) 'y ', tab_protein % array ( 3 ) % orientation_y
    write(*,*) 'z ', tab_protein % array ( 3 ) % orientation_z

    write(*,*) 'pos atom surf', &
        tab_protein % array ( 1 ) % p_list_atom_surface (:,1)
    write(*,*) '             ', &
        tab_protein % array ( 1 ) % p_list_atom_surface ( :, tab_protein % array ( 1 ) % nb_atom_surface )

    write(*,*) 'pos charge prot', &
        tab_protein % array ( 1 ) % p_pos_charge (:,1)
    write(*,*) '               ', &
        tab_protein % array ( 1 ) % p_pos_charge ( :, tab_protein % array ( 1 ) % nb_charge )
    write(*,*) 'dr ', tab_protein % array ( 1 ) % p_sogrid % diffusion_rotat
    write(*,*) 'dm ', tab_protein % array ( 1 ) % p_sogrid % diffusion_trans
#endif

    ! allocate array for force and torque
    allocate ( force ( 3, nb_protein ), torque ( 3, nb_protein ), stat = ierr )
    if ( ierr /= 0 ) then
        write (*,*) 'ERROR Cannot allocate force and torque'
        stop 1
    end if
    ! init to zero
    force = 0.d0
    torque = 0.d0

    ! sdamm dt is constant by default
    dt = param_timestep % dt1

    !NJB: call set_cut_array, set also global cut2
    call set_cut_array(tab_protein, geom, Cut2)
    write(*,*) 'Maximum interaction cutoff square, Cut2 ', Cut2

    ! could be done in set_cutoff
    if ( param_force_energy % ihi ) then

       !NJB: Call function to determine radius of local volume in mean-field HI calculation
       call set_hi_rad (tab_protein, geom, param_analytic%lvol_rcut, param_force_energy % cutoff_hi)

!       !NJB: Local volume radius now read from input file
!       param_force_energy % cutoff_hi = param_analytic%lvol_rcut

       write (*,'(" Radius of local volume for hydrodynamics interactions = " F8.3)') param_force_energy % cutoff_hi


        write(*,*) 'min_sizebox = ', MINVAL ( geom % pbox % sizebox )
        if ( param_force_energy % cutoff_hi .gt. MINVAL ( geom % pbox % sizebox )/2. ) then
            write(*,*) 'INFO Local volume fraction cutoff too large. Rescaled to 1/2 of the simulation box size.'
            write(*,*) 'INFO Cutoff HI ', MINVAL ( geom % pbox % sizebox ) / 2.
            param_force_energy % cutoff_hi = MINVAL ( geom % pbox % sizebox ) / 2.
        end if
    end if


    !write(*,*) 'Compute first energy'
    ! initialisation, already done
    array_energy_sdamm = 0.d0

    ! initialize DH or crowders (not implemented)
    ! and hydrodynamic
    ! need more data to init param_force_energy
    ! todo move into the init of param_force_energy
    if ( param_analytic % ianalytic ) then
#ifdef DEBUG
        write(*,*)
        write(*,*) 'call init_analytic from Main loop'
#endif

        if ( param_analytic % debyeh ) then
            ! sda 2proteins, use the c-surf as limit
            ! sdamm uses half the box size - 5 Angstrom
            ! safer to use the minimum in case of PBC, in case of surface could be more specific
            ! minval never tested with intel/pgf
            half_box = minval ( geom % pbox % sizebox )/2. - 5
#ifdef DEBUG
            write(*,*) 'sdamm half box - 5 Angs. set into init_analytic ', half_box
#endif
            call init_analytic ( param_force_energy, tab_protein, &
                                 half_box, param_analytic % h_analytic, param_analytic % ionic )
        end if
    endif

    ! summary data, analytical included
    call print_info_param_fe ( param_force_energy )

    ! Compute energy of restart position, usefull for debugging SDA flex / SDAMM
    ! usefull for tests and implements asymetry also
    ! could be one asymetry (eletrostatic but no desolvation ) and still use the fast version
    ! but not working if electro and desolvation are asymetric
#ifdef DEBUG
    write(*,*) 'compute restart energy '
#endif
    if ( param_force_energy % fast_computation ) then
        call energy_epedhdlj_sdamm_fast ( tab_protein, geom, nb_protein, param_force_energy, &
                                          array_energy_sdamm, dummy_array )
    else
        call energy_epedhdlj_sdamm ( tab_protein, geom, nb_protein, param_force_energy, &
                                     array_energy_sdamm, dummy_array )
    end if

    !DEBUG, multiply by two the energy for SDAMM compliance
    ! works only for electrostatic
    !to comment later !
    !array_energy_sdamm = array_energy_sdamm * 2.d0
    !write(*,*) 'uhbd energies'
    !do i = 1, nb_protein
    !    write(*,*) 'potential energy', array_energy_sdamm (2: tab_protein % nb_grids + 2 - 1,i )
    !end do

    if ( param_force_energy % iimg_chg ) then
        call energy_image_charge_sdamm ( tab_protein, param_force_energy, array_energy_sdamm )
    endif

    if ( param_force_energy % imetaldesolv ) then
        call compute_energy_metaldesolv_sdamm( tab_protein, param_force_energy, param_metaldesolv, &
                                               array_energy_sdamm )
    end if

    if ( param_force_energy % ianalytic ) then
        if ( param_force_energy % idebye ) then
            call energy_analytic_dh_sdamm ( tab_protein, geom, nb_protein, param_force_energy, array_energy_sdamm )
        end if
    end if

    do i = 1, nb_protein
         array_energy_sdamm ( 1, i ) = sum ( array_energy_sdamm(:,i) )
    end do

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Notes SDAMM, as before electrostatic energy is 2 times lower is SDA flex, compared to SDAMM
    ! Should have the same factor with DH (DH splitted in SDAMM) !
#ifdef VERBOSE
    write(*,*)
    write(*,*) '=== Energy Restart ==='
    ! only for kdiff3 comparison with sdamm
    write(*,*) '=== All energies ==='

    do i = 1, nb_protein
        write(*,*) 'protein ', i
        !write(*,*) 'full array ', array_energy_sdamm(:,i)
        write(*,*) 'potential energy', array_energy_sdamm (2: tab_protein % nb_grids + 2 - 1,i )
        write(*,*) 'desol energy', array_energy_sdamm ( 2 + ( tab_protein % nb_grids * 1 ): ( tab_protein % nb_grids * 2 )+1,i )
        write(*,*) 'hydro. energy', array_energy_sdamm ( 2 + ( tab_protein % nb_grids * 2 ): ( tab_protein % nb_grids * 3 )+1,i)
        write(*,*) 'rep lj energy', array_energy_sdamm ( 2 + ( tab_protein % nb_grids * 3 ): ( tab_protein % nb_grids * 4 )+1,i)
        write(*,*) 'total energy  ',array_energy_sdamm(1,i)
!        write(*,*) 'lenjon energy', array_energy_sdamm ( 2 + ( tab_protein % nb_grids * 4 ): ( tab_protein % nb_grids * 5 ) + 1,i)
!        write(*,*) 'image charge energy', array_energy_sdamm(2+( tab_protein % nb_grids * 5 ):( tab_protein % nb_grids * 6 ) + 1,i)
    end do
!    stop 1
#endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! always print initial conformation as step 0 in trajectory, with recomputation of energy
    ! due to pre-computed grid of space 1 A, same number appears in output
    ! maybe interpolation later
    if ( associated ( trajectories % ptrajectory ) ) then
        call add_record ( trajectories, tab_protein % array, 1, 0, array_energy_sdamm, 0. )
    end if

    write(*,*)
    !write(*,*) '================='
    write(*,*) 'Start Main Loop'
    write(*,*) 'Timemax', timemax
    !write(*,*) 'time ', time

    flush ( io6 )

!! !$POMP INST BEGIN(loop_bd)

!!    .. Start of BD run, do-loop
BD_LOOP: &
    do while ( time < timemax )

#ifdef VERBOSE
        write(*,*) 'time ', time
#endif

        ! signal handling, stop the job
        if ( EXIT_PROG .eqv. .true. ) then
            !write(*,*) 'thread got the message exit '
            rewind ( io_steps )
            write ( io_steps, fmt='(f12.2)' ) time !current_steps - 1
            flush ( io_steps )
            EXIT BD_LOOP
        end if

        ! reinitialize force
        force = 0.
        torque = 0.
        ! if print initial as zero, first one will be dt
        time = time + dt
        current_step = current_step + 1
        !write(*,*) 'current_step ', current_step
        ! reset to false by default, tested just after
        print_step_traj = .false.
        print_step_compl = .false.
        ! reset boolean accept new_conf
        accepted_new_conf = .false.

        ! modulo to know if needed printed out and compute energy
        if ( associated ( trajectories % ptrajectory ) ) then
            !write(*,*) 'check print_step ', trajectories % freq_print, current_step
#ifdef WEBSDA
            if ( mod (current_step, 100) == 0 ) then
#else
            if ( ( trajectories % freq_print /= 0 ) .and. &
                 ( MOD ( current_step, trajectories % freq_print ) == 0 ) ) then
#endif

                 print_step_traj = .true.
                 rewind ( io_steps )
                 write ( io_steps, fmt='(f12.2)' ) time !current_steps - 1
                 flush ( io_steps )
            end if

        end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!  Compute forces
        call start_mytimer ( timer_force )

        force = 0.d0
        torque = 0.d0

        !write(*,*) 'before force'
        !write(*,*) 'tab_prot % array (2) position ', tab_protein % array (2) % position
        !write(*,*) 'tab prot array (1) position ', tab_protein % array (1) % position

        ! fastest, up to a factor 2
        if ( param_force_energy % fast_computation ) then
            !write(*,*) 'call fast force'
            call force_epedhdlj_sdamm_fast ( tab_protein, geom, nb_protein, param_force_energy, &
                                         force, torque, dummy_array )
        ! same comment than for energy
        else
             !write(*,*) 'call normal force'
             call force_epedhdlj_sdamm ( tab_protein, geom, nb_protein, param_force_energy, &
                                         force, torque, dummy_array )
        end if

#ifdef VERBOSE
        write(*,*) 'uhbd forces'
        write (*,*) 'force 1', force (:,1)
        write (*,*) 'torque 1', torque(:,1)
        write (*,*) 'force 2', force(:,2)
        write (*,*) 'torque 2', torque(:,2)
#endif

        ! need to write OpenMP, implement Cutoff
        if ( param_force_energy % iimg_chg ) then
            ! write(*,*) 'Will compute image_charge'
            ! need only Cut2_surf for image charge !
            call force_image_charge_sdamm ( tab_protein, param_force_energy, force, torque )
        endif

       !if ( param_force_energy % imetaldesolv ) then
            !write(*,*) 'compute force metal desolv'
       !     call compute_force_metaldesolv_sdamm  ( tab_protein, param_force_energy, param_metaldesolv, &
       !                                             force, torque )
       !end if

       if ( param_force_energy % ianalytic ) then

           if ( param_force_energy % idebye ) then
               call force_analytic_dh_sdamm ( tab_protein, geom, nb_protein, param_force_energy, force, torque )

           end if
       end if

#ifdef VERBOSE
        write(*,*)
        write(*,*) '==== all forces ===='
        do i = 1, nb_protein ! 2
            write(*,*) 'protein ',i
            write(*,*) 'force  ', force(:,i)
            write(*,*) 'torque ', torque(:,i)
            do j = 1, 3
                if (isnan( force(j,i))) then
                    write(*,*) 'force solute ',i
                    stop '"force" has a NaN'
                endif
                if (isnan(torque(j,i))) then
                    write(*,*) 'torque solute ',i
                    stop '"torque" has a NaN'
                endif
            end do
         end do
#endif
        call stop_mytimer ( timer_force )


!!!!!!!!!!!!!!!!!!!!!!!!!  Make BD move
        call start_mytimer ( timer_bd )

        if ( param_force_energy % ihi ) then
#ifdef VERBOSE
            !write(*,*) 'ihi activated,call scale diff coeff with cutoff_hi ', param_force_energy % cutoff_hi
            write(*,*)
            write(*,*) '=== Before t_diff_coeff, hydrodynamics part ==='
#endif
            ! rescale diffusion coeeficient with Weinstein method
            ! keep cutoff as 4 in memory, like other if later regrouped in one array
            call scale_diffusion_coefficients ( tab_protein, geom, real(param_force_energy % cutoff_hi,8) )
        end if

        call make_bd_move_sdamm ( tab_protein, geom, nb_protein, &
                                  force, torque, dt )

#ifdef VERBOSE
        write(*,*)
        write(*,*) '=== After BD move ==='
        do i = 1, nb_protein
            write(*,*) 'new pos protein ',i
            write(*,*) 'pos', tab_protein % array ( i ) % position
            write(*,*) 'x ', tab_protein % array ( i ) % orientation_x
            write(*,*) 'y ', tab_protein % array ( i ) % orientation_y
            write(*,*) 'z ', tab_protein % array ( i ) % orientation_z
        end do
#endif
        call stop_mytimer ( timer_bd )

!!!!!!!!!! update rdf after BD move

        if ( ( resid_time % enum_rt == rt_distance ) .and. MOD( current_step, 5 ) == 0 ) then
            call start_mytimer ( timer_rdf )
            call compute_rdf ( resid_time, tab_protein, geom )
            n_rdfs = n_rdfs + 1
            call stop_mytimer ( timer_rdf )
        end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!  Computation of energy if print into trajectory !!!!!!!!!!!!!!!!!!!!

        !! if need to compute energy
        if ( print_step_traj ) then
            !write(*,*) 'print_step_traj is true compute energy sdamm'
!!!!!!!!!!!!! Compute energy

    call start_mytimer ( timer_energy )

            ! reinitialize all terms to 0
            array_energy_sdamm = 0.d0
            !write(*,*) 'will call energy'

            ! to comment if asymetric interaction are not needed
            ! fast computation if no asymetric interaction, parrallel loop
            if ( param_force_energy % fast_computation ) then
                !write(*,*) 'call fast energy'
                call energy_epedhdlj_sdamm_fast ( tab_protein, geom, nb_protein, param_force_energy, &
                                                  array_energy_sdamm, dummy_array )
            ! if asymetric interaction present
            ! new implementation with explicit test of nqX and associated p_grid
            else
                !write(*,*) 'call normal energy'
                call energy_epedhdlj_sdamm ( tab_protein, geom, nb_protein, param_force_energy, &
                                             array_energy_sdamm, dummy_array )
            end if

            !DEBUG, to uncomment after
            !array_energy_sdamm = 2.d0 * array_energy_sdamm
            !write(*,*)
            !write(*,*) '===='
            !write(*,*) 'uhbd energies'
            !do i = 1, nb_protein
            !    write(*,*) 'potential energy', array_energy_sdamm (2: tab_protein % nb_grids + 2 - 1, i )
            !end do

            ! nedd to omplement OpenMP, working, only partially tested
            if ( param_force_energy % iimg_chg ) then
                call energy_image_charge_sdamm ( tab_protein, param_force_energy, array_energy_sdamm )
            endif

            ! same need openmp, even less tested
            if ( param_force_energy % imetaldesolv ) then
                 !write(*,*) 'Will compute metal desolvation energy'
                 !! energy updated inside
                 call compute_energy_metaldesolv_sdamm( tab_protein, param_force_energy, param_metaldesolv, array_energy_sdamm )
            end if

            if ( param_force_energy % ianalytic ) then

                if ( param_force_energy % idebye ) then
                    call energy_analytic_dh_sdamm ( tab_protein, geom, nb_protein, param_force_energy, array_energy_sdamm )

                end if
            end if

            ! total energy is the sum of all elements
            do i = 1, nb_protein
                !write(*,*) i, ' array_energy_sdamm before sum total ', array_energy_sdamm
                array_energy_sdamm ( 1, i ) = sum ( array_energy_sdamm(:,i) )
            end do
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!      write(*,*) 'Total Energy ', array_energy_sdamm(1,1)
#ifdef VERBOSE
      write(*,*)
      write(*,*) '=== All energies ==='
      do i = 1, nb_protein
          write(*,*) 'protein ', i
!         write(*,*) 'full array ', array_energy_sdamm(:,i)

          ! general ( 2 + nb_types*(grid_number-1) : nb_types*(grid_number)+1
          write(*,*) 'potential energy', array_energy_sdamm (2: tab_protein % nb_grids + 2 - 1,i )
          write(*,*) 'desol energy ', array_energy_sdamm ( 2 + ( tab_protein % nb_grids * 1 ): ( tab_protein % nb_grids * 2 )+1,i )
          write(*,*) 'hydro. energy',array_energy_sdamm ( 2 + ( tab_protein % nb_grids * 2 ): ( tab_protein % nb_grids * 3 )+1,i)
          write(*,*) 'rep lj energy',array_energy_sdamm ( 2 + ( tab_protein % nb_grids * 3 ): ( tab_protein % nb_grids * 4 ) + 1,i)
!          write(*,*) 'lenjon energy',array_energy_sdamm ( 2 + ( tab_protein % nb_grids * 4 ): ( tab_protein % nb_grids * 5 ) + 1,i)
!          write(*,*) 'image charge energy',array_energy_sdamm(2+(tab_protein % nb_grids *5 ): ( tab_protein % nb_grids * 6 ) + 1,i)
          ! always first
          write(*,*) 'total energy    ',array_energy_sdamm(1,i)
      end do
#endif

    call stop_mytimer ( timer_energy )

!!!!!!!!!!!!!!!!!!!!!!!!!! Update trajectory file
    call start_mytimer ( timer_record )

            ! 1 is the current_run, sdamm always 1 run by default
            call add_record ( trajectories, tab_protein % array, 1, current_step, &
                              array_energy_sdamm, real(time,4) )

    call stop_mytimer ( timer_record )

        end if ! print_step_traj

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  Change of conformation

        ! general switch to avoid testing all proteins
        if ( iflex ) then
            ! make loop over protein first, should be asynchronous change in sdamm,
            ! cannot change conf of 2 proteins at the same time. Not sure now, but order will have influence
            ! time_flex belongs to flexible, make random choice of prot, gillepsie ??
            do i = 1, nb_protein
                !write(*,*) 'prot ',i ,' flex ', tab_protein % array ( i ) % iflex

                ! apply only to flexible proteins
                if ( tab_protein % array ( i ) % iflex ) then
                    ! update statistics for the conformations
                    call update_visited_conf ( tab_protein % array ( i ), real (dt,4) )

                    ! update time_next_flex
                    tab_protein % array ( i ) % p_flex % time_next_flex = &
                        tab_protein % array ( i ) % p_flex % time_next_flex - real(dt,4)
                    !write(*,*) 'prot i=',i, ' update time ', tab_protein % array ( i ) % p_flex % time_next_flex

                    ! check if one protein should change conformation, and ONLY 1 in a step for sdamm (to check)
                    ! maybe not, but order is limiting
                    if ( ( tab_protein % array ( i ) % p_flex % time_next_flex <= 0. ) .and. &
                         ( accepted_new_conf .eqv. .false. ) ) then
                         !write(*,*)
                         !write(*,*) 'prot i= ', i, 'test new_conf current ', tab_protein % array ( i ) % p_flex % current_conf
                         !write(*,*) 'prot ',i, ' current_conf ', tab_protein % array ( i ) % p_flex % current_conf

                  call start_mytimer ( timer_flex )

                        ! add function as parameter, use sdamm version for energy_flex
                        call change_conformation ( tab_protein % array ( i ), tab_protein, &
                                                   geom, param_force_energy, param_metaldesolv, type_calc, dummy_array, &
                                                   ! hit = 'false' in sdamm, accepted,, nb_overlap_flex shoudl be always 0
                                                   .false., accepted_new_conf, nb_overlap_flex, energy_epedhdlj_flex_sdamm )
                        ! write(*,*) 'prot ',i, ' current_conf ', tab_protein % array ( i ) % p_flex % current_conf

                        ! if change accepted / or not ?? reinit change_conf
                        ! new commented, if chosen and rejected still reset the time
                        ! if ( accepted_new_conf .eqv. .true. ) then
                            !write(*,*) 'prot i= ', i, 'new_conf accepted '
                            call init_flexible_freq ( tab_protein % array ( i ) % p_flex )
                            ! write(*,*) 'prot ', i,' new conf accepted, reinit time ', &
                            !            tab_protein % array ( i ) % p_flex % time_next_flex
                        !end if

                  call stop_mytimer ( timer_flex )
                    end if

                end if ! end if protein flex
            end do ! end do nb_protein

        end if ! end if iflex

    ! end main loop
    end do BD_LOOP
    !write(*,*) 'nb_overlap_flex ', nb_overlap_flex

!! !$POMP INST END(loop_bd)

    ! write restart file, default name restart.ascii/.bin
    call write_restart ( trajectories, io_restart )

    if ( associated ( o_complexes % pcomplexe ) ) then
        write(*,*) 'final writing of complexes '
        call write_record ( o_complexes )
    end if

    if ( resid_time % enum_rt /= rt_off ) then
        !print_residence_time ( resid_time, geom, nrun )
        write(*,*) 'print rdf n_rdfs ', n_rdfs
        call print_rdf ( resid_time, tab_protein, geom, n_rdfs )
    end if

    ! print statistics for each flexible protein
    if ( iflex ) then
        do i = 1, nb_protein
            if ( tab_protein % array ( i ) % iflex ) then
                write ( io6, * )
                write ( io6, * ) 'Protein ', i
                call print_visited_conf ( tab_protein % array ( i ) )
            end if
        end do
    end if

    if ( rcriteria % type_reaction /= off_rc ) then
        ! ok should normalize from 0 to 1, in pourcent of time
        call print_sumtime ( rcriteria % p_assoc_rate, real( current_step,4 ) * real (dt,4) )
    end if

    if ( associated ( array_energy_sda2prot ) ) then
        deallocate ( array_energy_sda2prot, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocating array_energy_sda2prot '
        end if
        nullify ( array_energy_sda2prot )
    end if

    if ( associated ( array_energy_sda2prot_1d ) ) then
        write(*,*) 'deallocate array_energy_1d '
        deallocate ( array_energy_sda2prot_1d, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In deallocating array_energy_sda2prot_1d '
        end if
        nullify ( array_energy_sda2prot_1d )
    end if

    if ( associated ( dummy_array_protein % array ) ) then
        call delete_array_protein ( dummy_array_protein )
    end if

    deallocate ( dummy_array )
    deallocate ( array_energy_sdamm )

    ! Print Timer
    write(*,*) 'Mytimer force'
    call print_mytimer (timer_force)
    write(*,*) 'Mytimer energy'
    call print_mytimer (timer_energy)
    write(*,*) 'Mytimer record'
    call print_mytimer (timer_record)
    write(*,*) 'Mytimer bd'
    call print_mytimer (timer_bd)
    write(*,*) 'Mytimer flex'
    call print_mytimer (timer_flex)
    write(*,*) 'Mytimer rdf'
    call print_mytimer (timer_rdf)

end subroutine MainLoop_sdamm

!> Call during the simulation to compute the radial distribution function
!!
!! Certainly not optimal
!!
!! @param resid_time : instance \ref mod_residencetime::residence_time "residence_time"
!! @param tab_protein : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
subroutine compute_rdf ( resid_time, tab_protein, geom )
    USE mod_ResidenceTime
    USE mod_array_protein
    USE mod_geometry
 !$ USE omp_lib

    IMPLICIT NONE
    type ( residence_time ) :: resid_time
    type ( array_protein_type ), intent ( in ) :: tab_protein
    type ( geometry ) :: geom

    real ( kind=8 ), dimension ( 3 ) :: pos_relat
    real ( kind=4) :: dist
    integer :: i, j, size_array
    integer :: type_indice_i, type_indice_j

    !write(*,*) 'Entry compute_rdf '

    size_array = size ( tab_protein % array )

    !$OMP PARALLEL DEFAULT ( NONE ) SHARED ( geom, tab_protein, resid_time, size_array ) &
    !$OMP PRIVATE ( i, j, type_indice_i, type_indice_j, dist, pos_relat )
    !$OMP DO SCHEDULE( DYNAMIC )
    do i = 1, size_array - 1

        type_indice_i = get_type (  tab_protein, i )
        !write(*,*) 'type_indice_i ', type_indice_i
        do j = i + 1,  size_array

            type_indice_j = get_type ( tab_protein, j )

            ! can avoid recomputation of position, in saving in an array
            ! or saving all distance pairs...
            !call position_pbc ( geom, tab_protein % array ( i ) % position, &
            !                    tab_protein % array ( j ) % position, &
            !                    pos_relat )
            ! avoid use of position_pbc, replaced by
            call compute_dist_pbc ( geom, tab_protein % array ( i ), tab_protein % array ( j ), &
                                     dist, pos_relat )

            call update_rdf ( resid_time, real(dist,8), type_indice_i, type_indice_j )

        end do
    end do
    !$OMP END DO
    !$OMP END PARALLEL

end subroutine compute_rdf
