!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file bd_move.f90
!! Implement functions for BD translation and rotation, boost

!------------------------------------------------------------------------
!
! random and forced bd move of the protein_2's center
!
!
!     r orig is the position relative
!     rtry the new position to test
!     rmsnow, amplitude random translation
!     use of dmd, translation from force

!! contains bd_translation, bd_rotation, bd_rotation_simple and boost

!> Return a random vector for translation\n
!! Used by both sda and sdamm\n
!! PBC applied by update translation in mod_geometry\n
!! Thread-safe function
!! @param rorig : original position
!! @param cell :
!! @param rtry : position after translation
!! @param geom : instance of mod_geometry
!! @param force : force on the solute
!! @param dmd : diffusional translation
!! @param rmsnow : fluctuation term in translation
subroutine bd_translation ( rorig, cell, rtry, geom, force, dmd, rmsnow )
    USE mod_geometry
    IMPLICIT NONE

    real ( kind=8 ), dimension ( 3 ) :: rtry
    real ( kind=8 ), dimension ( 3 ), intent ( in ) :: rorig, force
    integer, dimension ( 3 ), intent ( inout ) :: cell
    type ( geometry ), intent ( in ) :: geom
    real ( kind=8 ), intent ( in ) :: dmd, rmsnow

    real ( kind=8 ), dimension ( 3 ) :: gaus, rtry_tmp

    !write(*,*) 'rorig ', rorig
    !write(*,*) 'dmd ', dmd
    !write(*,*) 'rmsnow ', rmsnow
    call ggnml_thread_safe(gaus)
    !write(*,*) 'gaus ', gaus
    !write(*,*) 'force ', force

    rtry_tmp = rorig + rmsnow * gaus
    !! if f = 0 => rtry = rtry
    rtry_tmp = rtry_tmp + dmd * force

    !! call correct function depending of the geometry, could apply pbc
    call update_translation ( geom, rtry_tmp, rtry, cell )
    !write(*,*) 'rtry2: ', rtry

end subroutine bd_translation


!> Return a random rotation matrix rot_try\n
!! Function used by sda_2proteins, which corrects rotation of protein 1 into a translation of protein 2
!! @param torque : torque between proteins
!! @param rot_try : rotation after BD, matrix form
!! @param rot_displ : rotational diffusion coefficient
!! @param rmsang : fluctuation for rotation
!! @param prot_id : index of the protein, if 1 apply the inverse transformation to protein 2
subroutine bd_rotation ( torque, rot_try, rot_displ, rmsang,  prot_id )
    IMPLICIT NONE

! drd ? belong to protein ??
    real ( kind=8 ), dimension(3), intent (in) :: torque
    real ( kind=8 ), dimension (3,3), intent (out) :: rot_try
    real ( kind=8 ) :: rot_displ, rmsang
    integer, intent (in) :: prot_id

    real ( kind=8 ), dimension(3) :: gaus, om

!--- compute rotation ---------------------------
      ! before send dr or dr1 and compute rmsang
      !tc = drd*rotime
      !rmsang = dsqrt(2.d0*tc)

    !write(*,*) 'bdrot drd, rotime ' !, drd, rotime
    call ggnml_thread_safe(gaus)
    !write(*,*) 'gaus ', gaus
    !write(*,*) 'rmsang ', rmsang
    om = rmsang * gaus
    om = om + rot_displ * torque

    ! modification for protein1
    ! do not change timing if deleted, may be to test more
    if (prot_id == 1) om = - om
!      write (*,*) 'om ', (om(m),m=1,3)

! converts rotation angle to rotation matrix
!      write(*,*) 'torq ', ( torque(m),m=1,3 )
!      write(*,*) 'om ', (om(m),m=1,3)
    call matrix(om,rot_try)

end subroutine bd_rotation

!> Return a random rotation matrix rot_try\n
!! Simple version without correction. Used by sdamm
!! @param torque : total torque on one protein
!! @param rot_try : rotational BD in matrix form
!! @param rot_displ : rotational coeffcient
!! @param rmsang : fluctuation for rotation
subroutine bd_rotation_simple ( torque, rot_try, rot_displ,  rmsang )
!! !$  USE omp_lib
    IMPLICIT NONE
! drd ? belong to protein ??
    real ( kind=8 ), dimension (3), intent (in) :: torque
    real ( kind=8 ), dimension (3,3), intent (out) :: rot_try
    real ( kind=8 ) :: rot_displ, rmsang

    real ( kind=8 ), dimension(3) :: gaus, om
    !write(*,*) 'rmsang ', rmsang, 'rot_displ ', rot_displ

!--- compute rotation ---------------------------
      ! before send dr or dr1 and compute rmsang
      !tc = drd*rotime
      !rmsang = dsqrt(2.d0*tc)

    !call ggnml(3,gaus)
    call ggnml_thread_safe(gaus)

      !write(*,*) 'ok gaus', (gaus(m),m=1,3)
      !om = dfloat(rmsang) * gaus
    om = rmsang * gaus
      !om = om + tc * torque
      !om = om + dfloat(rot_displ) * torque
    om = om + rot_displ * torque

! converts rotation angle to rotation matrix
      !write(*,*) 'torq ', ( torque(m),m=1,3 )
      !write(*,*) 'om ', (om(m),m=1,3)
    call matrix (om,rot_try)

end subroutine bd_rotation_simple

!> Apply boost if too many overlaps, only for sda_2protein
!! @param r : relative position, updated buring the boost
!! @param geom : instance of mod_geometry
!! @param dist : distance before the boost
!! @param rboost : boost value
subroutine boost ( r, geom, dist , rboost )
    USE mod_geometry
    IMPLICIT NONE
    real ( kind=8 ), dimension( 3 ) :: r
    type ( geometry ), intent ( in ) :: geom
    real ( kind=4 ), intent ( in ) :: rboost
    real ( kind=4 ), intent ( in ) :: dist

    real ( kind=4 ) :: fbd1

    fbd1 = 1. + rboost/dist
    !write(*,*) 'fbd1 ', fbd1

    !! implicit if surface geometry is box, may be not for john ??
    !! consider r(3) always > 0
    if ( geom % surface ) then
        r ( 3 ) = r ( 3 ) * fbd1
    else
        r = r * fbd1
    end if

end subroutine boost
