!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file change_conformation.f90
!! Functions used for the change of conformation
!!
!! Implement the different algorithms\n
!! Initialize the conformation and make the change\n



!> Change the conformation according to the correct method.\n
!!
!! This method is thread-safe, but in evaluating energy differences,
!! the other proteins should not change conformation at the same time\n
!!
!! Accept a function as last argument, function from mod_compute_energy_flex
!!
!! Implements 3 methods for changing conformation :
!! - random : choose a random position, can be the same
!! - minimum : chose the minimum position with nearest neighbour
!! - metropolis : choose random neighbour and apply the metropolis algorithm
!!
!! Make a copy of movable atoms if protein rotates before the energy computation. better ??
!!
!! In keeping trace of the last conf in memory, try to minimise the number of call to copy_movable ( make explicit copy and may need to rotate correctly),
!! but increase call to change_conf_flex ( pointer association ), could be removed in some case ( when sure to exist in Minimum )
!!
!! add type calc for exclusion check, add hit can delete type_calc to do later
!!
!! @param prot : instance of \ref mod_protein::protein "protein" which is tested for a change in conformation
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param param_force_energy : instance of \ref mod_force_energy "mod_force_energy"
!! @param param_metal : instance of "param_metaldesolvation"
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param dummy_array : array of 1. for repulsive jennard-jones calculation
!! @param hit : logical, if check overlap is needed
!! @param accepted : results indicates if the change is accepted
!! @param nb_overlap_flex : count the number of overlap when changing conformation
!! @param ptrf_energy : function for computing energy / energy_epedhdl_flex/sdamm signature
subroutine change_conformation ( prot, tab_prot, geom, param_force_energy, param_metal, type_calc, dummy_array, &
                                 hit, accepted, nb_overlap_flex, ptrf_energy )
    USE mod_array_protein
    USE mod_geometry
    USE mod_type_calc
    USE mod_force_energy
    USE mod_compute_energy_flex
    USE com_sda
    IMPLICIT NONE

    interface

        !> Interface for the function argument, expects compute_energy_flex ( sda or sdamm version )
        subroutine ptrf_energy ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                                 dummy_array, energy, energy_img_chg )
            USE mod_force_energy
            USE mod_array_protein
            USE mod_geometry
            USE com_sda
            IMPLICIT NONE
            type ( protein ), intent( in ), target :: prot
            type ( array_protein_type ), intent( in ), target :: tab_prot
            type ( geometry ), intent( in ) :: geom
            type ( type_force_energy ), intent ( in ) :: param_force_energy
            type ( parameter_metaldesolv ), intent ( in ) :: param_metal
            real ( kind=8 ), dimension( : ), intent( in ),target :: dummy_array
            real ( kind=8 ) :: energy, energy_img_chg
            integer, intent ( in ) :: nb_prot
        end subroutine ptrf_energy

        subroutine check_exclusion ( prot1, prot2, rtry, rot_try, overlap, do_rot )
            USE mod_protein
            IMPLICIT NONE
            type ( protein ), intent( in ), target :: prot1, prot2
            real ( kind=8 ), dimension(3), intent( in ) :: rtry
            real ( kind=8  ), dimension (3,3), intent( in ) :: rot_try
            logical , intent ( out ) :: overlap
            logical, intent ( in ) :: do_rot
        end subroutine check_exclusion

    end interface

    type ( protein ) :: prot
    type ( array_protein_type ), intent(in), target :: tab_prot
    type ( geometry ), intent ( in ) :: geom
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    type ( parameter_metaldesolv ), intent ( in ) :: param_metal
    type ( type_calculation ), intent ( in ) :: type_calc
    real ( kind=8 ), dimension(:), intent ( in ), target :: dummy_array
    integer :: nb_overlap_flex
    logical, intent ( in ) :: hit
    logical, intent ( out ) :: accepted

    real ( kind=8  ), dimension(3,3) :: null_rot
    logical :: bool_last_movable_inmemory, overlap, chk_exclusion
    integer :: last_inmemory, nb_prot
    !! to store initial conformation
    type ( NodeFlex ), pointer :: memo_nodeflex
    integer :: memo_current_conf

    real ( kind=8 ) :: unif !, rt
    real ( kind = 8 ) :: tot_energy0, tot_energy1, tot_energy2, energy_img_chg
    integer :: shift

    nullify ( memo_nodeflex )
    null_rot = 0.d0
    tot_energy0 = 1000.d0
    tot_energy1 = 1000.d0
    tot_energy2 = 1000.d0
    shift = 0
    accepted = .false.
    overlap = .false.
    chk_exclusion = .false.
    !! default boolean for last call to copy_movable
    bool_last_movable_inmemory = .false.
    !! default conformation in memory the current one
    last_inmemory = 0
    nb_prot = size ( tab_prot % array )

    ! never used, introduced only for tests, keep image charge energy in a separate variable
    energy_img_chg = 0.d0

#ifdef VERBOSE_FLEX
    write(*,*) 'Entry change conformation solute'
    write(*,*) 'current conformation ', prot % p_flex % current_conf
    !write(*,*) 'nearest ', prot % p_flex % plist_flex % nearest_ngb
    !write(*,*) 'nb prot ', nb_prot
#endif
    !! check for check_exclusion
    if ( ( type_calc %  enum_type_calc /= sdamm ) .and. ( hit ) ) then
        chk_exclusion = .true.
        !write(*,*) 'set chk_exclusion TRUE'
    end if

!!!!!!!!!!!! Random     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !! random, just pick up one randomly among all conformation
    !! if nearest, random +/- 1: can be rejected
    !! else random, can be the same one
    if ( prot % p_flex % plist_flex % method_flex == flex_random ) then

        !write (*,*) 'random change, conf before ', prot % p_flex % current_conf
        !! memorize initial conformation
        memo_nodeflex => prot % p_flex % p_current_node
        memo_current_conf = prot % p_flex % current_conf
        !write(*,*) 'current_conf before change', prot % p_flex % current_conf

        !! update p_current_node and current_conf
        !! -1 randomly and nearest false : randon among all conformations
        !! the same conformation can be chosen, would need more code here if want to re-test
        call change_conf_flex ( prot % p_flex, -1 )

        !! check if accepted and different as initial, always accepted if random and nearest is false
        if ( prot % p_flex % current_conf /= memo_current_conf ) then

            !! update the set of grid of the protein
            prot % p_sogrid => prot % p_flex % p_current_node % p_data
            !! copy_movable atoms before check_exclusion, done here and not last_copy_movable
            call copy_movable_array ( prot )

            !! check exclusion if sda_2proteins, always prot1 with exclusion
            if ( chk_exclusion ) then
                call check_exclusion ( tab_prot % array ( 1 ), tab_prot % array ( 2 ), tab_prot % array ( 2 ) % position, &
                                       null_rot, overlap, .false. )
            end if

            if ( overlap .eqv. .true. ) then
                write(*,*) 'overlap random', overlap
                nb_overlap_flex = nb_overlap_flex + 1
                !! come back to initial, will need to recopy movable_array
                prot % p_flex % p_current_node => memo_nodeflex
                prot % p_sogrid => prot % p_flex % p_current_node % p_data
                prot % p_flex % current_conf = memo_current_conf
                !! will need to reset up the initial conformation in this case

            !! else overlap is false, accepted and last in memory
            else
                ! no overlap and different
                accepted = .true.
                !! need copy_movable before check_exclusion, so last_in_memory if accepted
                bool_last_movable_inmemory = .true.
                !write(*,*) 'conf changed to ', prot % p_flex % current_conf
            end if

        !! else associated to the same p_current_node, no need of copy movable array
        else
            ! if not changed, movable_array has not been changed
            bool_last_movable_inmemory = .true.
            !! here can be considered as accepted ? not changed
            !accepted = .true.
        end if

!!!!!!!!!!!! Flex_time    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !! time always the next one, use optional shift with + 1
!    else if  ( prot % p_flex % plist_flex % method_flex == flex_time ) then
!        !write (*,*) 'time change'
!        memo_nodeflex => prot % p_flex % p_current_node
!
!        !! may add a test status return, replace by test same association for the moment
!        call change_conf_flex ( prot % p_flex, 0, 1 )
!
!        if ( associated ( prot % p_flex % p_current_node, memo_nodeflex ) .eqv. .false. ) then
!            !! update p_sogrid, should be always true true in this case
!            prot % p_sogrid => prot % p_flex % p_current_node % p_data
!
!            !! needed before exlcusion check
!            call copy_movable_array ( prot )
!
!            if ( chk_exclusion ) then
!                !call copy_movable_array ( prot )
!                call check_exclusion ( tab_prot % array ( 1 ), tab_prot % array ( 2 ), tab_prot % array ( 2 ) % position, &
!                                       null_rot, overlap, .false. )
!            end if
!
!            !! possible only if chk_exclusion
!            if ( overlap .eqv. .true. ) then
!
!                write(*,*) 'Overlap time', overlap
!                nb_overlap_flex = nb_overlap_flex + 1
!
!                !! come back to initial, will need to recopy movable_array
!                prot % p_flex % p_current_node => prot % p_flex % p_current_node % p_previous
!                prot % p_sogrid => prot % p_flex % p_current_node % p_data
!                prot % p_flex % current_conf = prot % p_flex % current_conf - 1
!
!            !! default overlap is false
!            else
!                !!not necessary, to delete later
!                accepted = .true.
!                !! need copy_movable before check_exclusion, so last_in_memory if accepted
!                bool_last_movable_inmemory = .true.
!            end if
!
!        !! same association, nothing to change, Here it is an ERROR
!        else
!            bool_last_movable_inmemory = .true.
!        end if

!!!!!!!!!!!! Minimum    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !! minimum, evaluate all energies: actual only previous and next conf
    !! actual may be not needed for sda_2proteins, only for sdamm
    !! method uses nearest neighbour as basis, check +1/-1
    !! would need to loop over all grids ??
    !! DO NOT work with nearest == 0 !
    else if ( prot % p_flex % plist_flex % method_flex == flex_minimum ) then
        !write(*,*) 'Minimum '
        !write(*,*) 'current_conf before change', prot % p_flex % current_conf
        !write(*,*) 'pos atom1 pflex', prot % pos_charge_prot(:,1)
        !! memorize initial conformation
        memo_nodeflex => prot % p_flex % p_current_node
        memo_current_conf = prot % p_flex % current_conf

        !! compute energy of prot with all others in its original conformation
        !! could be extracted from array_energy ??

        !! version with pointer, will call appropriate openmp if sdamm
        call ptrf_energy ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                           dummy_array, tot_energy0, energy_img_chg )
        !write (*,*) 'tot_energy0', tot_energy0

        !! update pointer of p_current_node and update current_conf
        call change_conf_flex ( prot % p_flex, 0, -1 )
        !write(*,*) 'pos atom1 newpflex avant copy', prot % p_pos_charge(:,1)
        !write(*,*) 'after change_conf_flex -1, current_conf ', prot % p_flex % current_conf

        !! they are different if change conf has been accepted
        !if ( associated ( prot % p_flex % p_current_node, memo_nodeflex ) .eqv. .false. ) then
        if ( prot % p_flex % current_conf /= memo_current_conf ) then

            prot % p_sogrid => prot % p_flex % p_current_node % p_data
            !! update movable atoms before energy calculation

            call copy_movable_array ( prot )
            !write(*,*) 'pos atom1 newpflex apres copy', prot % p_pos_charge(:,1)

            if ( chk_exclusion ) then
                call check_exclusion ( tab_prot % array ( 1 ), tab_prot % array ( 2 ), tab_prot % array ( 2 ) % position, &
                                       null_rot, overlap, .false. )
            end if

            if ( overlap .eqv. .true. ) then
                !write(*,*) 'Overlap minimum ', overlap
                nb_overlap_flex = nb_overlap_flex + 1
            !! overlap false, compute energy1.
            else
                call ptrf_energy ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                             dummy_array, tot_energy1,  energy_img_chg )
                !write (*,*) 'tot_energy1', tot_energy1
            end if

            !! in this method always comes back to initial conf before next change
            !! but do not make the copy
            prot % p_flex % p_current_node => memo_nodeflex
            !! needed ?
            prot % p_sogrid => prot % p_flex % p_current_node % p_data
            !! here is needed, for sure
            prot % p_flex % current_conf = memo_current_conf

            !! if +1 and energy0 are higher, will be the last one in memory
            last_inmemory = 1
        end if

        !write(*,*) 'Before second change_conf_flex current_conf ', prot % p_flex % current_conf
        call change_conf_flex ( prot % p_flex, 0, +1 )
        !write(*,*) 'after change_conf_flex +1, current_conf ', prot % p_flex % current_conf

        if ( prot % p_flex % current_conf /= memo_current_conf ) then
            prot % p_sogrid => prot % p_flex % p_current_node % p_data
            !! update movable atoms
            !write(*,*) 'second copy_movable'
            call copy_movable_array ( prot )
            !write(*,*) 'pos atom1 newpflex apres copy', prot % p_pos_charge(:,1)

            !! check_exclusion
            if ( chk_exclusion ) then
                call check_exclusion ( tab_prot % array ( 1 ), tab_prot % array ( 2 ), tab_prot % array ( 2 ) % position, &
                                       null_rot, overlap, .false. )
            end if

            if ( overlap .eqv. .true. ) then
                !write(*,*) 'Overlap metropolis', overlap
                nb_overlap_flex = nb_overlap_flex + 1
            !! overlap false, can comppute energy.
            else
                call ptrf_energy ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                             dummy_array, tot_energy2, energy_img_chg )
                !write (*,*) 'tot_energy2', tot_energy2
            end if

            ! this method always come back to initial conf
            prot % p_flex % p_current_node => memo_nodeflex
            !! both needed
            prot % p_sogrid => prot % p_flex % p_current_node % p_data
            prot % p_flex % current_conf = memo_current_conf
            !! last in memory, even if overlap == true, but tot_energy2 default is higher
            last_inmemory = 2
        end if

        !write(*,*) 'energy: ', tot_energy0,' ',tot_energy1,' ',tot_energy2

        !! initial use egal or < for test, no change to < only ( if 0 stays )
        !! if overlap default energy1 and 2 are higher
        if ( ( tot_energy1 < tot_energy0 ) .or. ( tot_energy2 < tot_energy0 ) ) then

            !! general has been accepted
            accepted = .true.

            !! choose shift to -1
            if ( tot_energy1 < tot_energy2 ) then
                !write(*,*) 'tot_energy1 is minimum need to re-change conf flex to -1'
                !! update p_current_node and current_conf
                !! could avoid and do it explictely, no need to call change_conf_flex, because sure to exist
                call change_conf_flex ( prot % p_flex, 0, -1 )
                prot % p_sogrid => prot % p_flex % p_current_node % p_data

                !write(*,*) 'choose shift -1 final conformation ', prot % p_flex % current_conf

                if ( last_inmemory == 1 ) bool_last_movable_inmemory = .true.
            !! make shift +1
            else
                !write(*,*) 'tot_energy2 is minimum, it is the last in memo, last_movable in memory is true'
                !! update p_current_node and current_conf
                call change_conf_flex ( prot % p_flex, 0, +1 )
                prot % p_sogrid => prot % p_flex % p_current_node % p_data

                !write(*,*) 'choose shift +1 final conformation ', prot % p_flex % current_conf

                !! if here, it is the last in memory for sure, test last not necessary
                !! only for test to comment/delete
                !if ( last_inmemory /= 2 ) then
                !    write(*,*) 'ERROR in LOGIC'
                !    stop
                !endif
                bool_last_movable_inmemory = .true.
            end if
        ! else energy0 is minimum: do nothing, stay to the same,
        ! but may need to recopy the movable array which may have been overwritten
        else
            !write(*,*) 'initial conf is better, keep default memo_pointer'
            !! if any other conformation have been tried, keep default value and do not need to copy movable atoms
            if ( last_inmemory == 0 ) bool_last_movable_inmemory = .true.
        end if

!!!!!!!!!!!! Metropolis    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    else if ( prot % p_flex % plist_flex % method_flex == flex_metropolis ) then
#ifdef VERBOSE_FLEX
        write(*,*) 'Metropolis method'
        !write(*,*) 'prot % position ', prot % position
#endif

        memo_nodeflex => prot % p_flex % p_current_node
        memo_current_conf = prot % p_flex % current_conf

        !! compute energy in initial conformation
        !call ptrf_energy ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
        !             dummy_array, tot_energy0,  energy_img_chg )
        call energy_epedhdlj_flex_sdamm( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                     dummy_array, tot_energy0,  energy_img_chg )
#ifdef VERBOSE_FLEX
        write(*,*) 'total energy0 ', tot_energy0
#endif

        !! here use always random (-1), test only one different conformation
        !! depends of nearest, but always random to get a better symetry
        call change_conf_flex ( prot % p_flex, -1 )
#ifdef VERBOSE_FLEX
        write(*,*) 'new conf ', prot % p_flex % current_conf
#endif

        !! one way to compare, or need return value/ status from change_conf
        !! Should check than at least one has been accepted, do not seem necessary, but safer
        if ( prot % p_flex % current_conf /= memo_current_conf ) then

            !! associate protein set of grid to the new one before computing energy
            prot % p_sogrid => prot % p_flex % p_current_node % p_data
            !! only for tests, to comment
            !if ( associated ( prot % p_sogrid ) .eqv. .false. ) then
            !    write(*,*) 'problem p_sogrid not associated'
            !    stop
            !end if

            !! need to copy_movable atoms, or send send rot_try to check_exclusion
            !! advantage if accepted
            call copy_movable_array ( prot )

            if ( chk_exclusion ) then
                !write(*,*) 'check exclusion'
                call check_exclusion ( tab_prot % array ( 1 ), tab_prot % array ( 2 ), tab_prot % array ( 2 ) % position, &
                                       null_rot, overlap, .false. )
            end if

            if ( overlap .eqv. .true. ) then
                !write(*,*) 'Overlap metropolis', overlap
                nb_overlap_flex = nb_overlap_flex + 1
            else
                !! test energy, only one neigbour has been chosen randomly, except min and max
                !write(*,*) 'nb_prot ', nb_prot
                call ptrf_energy ( prot, tab_prot, geom, nb_prot, param_force_energy, param_metal, &
                             dummy_array, tot_energy1,  energy_img_chg )
                !write(*,*) 'total energy1 ', tot_energy1
            end if
#ifdef VERBOSE_FLEX
            write(*,*) 'energy0/1 ', tot_energy0, tot_energy1
#endif

        !! else identical, no test to do, happens only with second implementation
        else
#ifdef VERBOSE_FLEX
            write(*,*) 'pick up same conformation same energy can return'
            write(*,*) 'return with current conf ', prot % p_flex % current_conf
#endif
            return
        end if

        ! Metropolis algorithm, if overlap, tot_energy1 = 1000 so exp is almost 0 ??
        ! or include in overlap == false
        if ( tot_energy1 < tot_energy0 ) then
            !! accept the move
            !write(*,*) 'tot_ener1 < tot_ener0, move accepted '
            accepted = .true.
            !! last in memory, correct conformation
            bool_last_movable_inmemory = .true.

        else
            ! generate uniform random number between 0 and 1
            call ggubs_thread_safe ( unif )
            ! not sure about kT, already included when reading the grid ?
#ifdef VERBOSE_FLEX
            write(*,*) 'exp(e0-e1) ', exp(tot_energy0 - tot_energy1)
#endif
            !if ( unif < exp( ( total_energy0 - total_energy1 ) / rt ) ) then
            if ( unif < exp ( tot_energy0 - tot_energy1  )  ) then
                !! accept the move
#ifdef VERBOSE_FLEX
                write(*,*) 'move accepted after random unif, tot0-tot1', unif, tot_energy0 - tot_energy1, &
                            exp ( tot_energy0 - tot_energy1  )
#endif
                accepted = .true.
                !! same could return directly grids correctly assigned
                bool_last_movable_inmemory = .true.

            else
                !! rejected
                !write(*,*) 'rejected with energy ',tot_energy0, tot_energy1
                !!recopy initial value
                prot % p_flex % p_current_node => memo_nodeflex
                prot % p_sogrid => memo_nodeflex % p_data
                prot % p_flex % current_conf = memo_current_conf
                !write(*,*) 'return to original conf ', prot % p_flex % current_conf
            end if
        end if

    ! end choice of method
    end if

    !! need to have correct movable atoms at the exit of the function
    if ( bool_last_movable_inmemory .eqv. .false. ) then
        !write(*,*) 'last copy_movable is needed'
        call copy_movable_array ( prot )
    !else
    !    write(*,*) 'do not need the last copy_movable'
    end if

    !write(*,*)  'Exit change conf current_conf ', prot % p_flex % current_conf
    !write(*,*)  'accepted ', accepted
    !write (*,*) 'state p_current_node next and previous', prot % p_flex % p_current_node % p_next % state, &
    !              prot % p_flex % p_current_node % p_previous % state

end subroutine change_conformation
