!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file check_reaction_pair.f90
!! Check if the reaction pairs are satisfied between protein 1 and 2.\n
!! Include all cases : docking, association rate and electron transfert

! new version with minimum visited windows
! need to keep escape(:,:) because of electron transfert

!> Check the reaction criteria from the solute's position for all calculations.\n
!! Account for any contact number in association rate calculation.\n
!! Modification of sda6 when checking independent distance :\n
!! Use same as in bootstrap : record only the minimum window visited for each contact.\n
!!
!! @param prot1 : protein1, fixed in sda
!! @param prot2 : protein2, mobil
!! @param rcriteria : instance of a reaction_criteria object, \ref mod_reactioncriteria "mod_ReactionCriteria" include \ref mod_rate_calculation
!! @param visited_windows : array for assocaiton and first passage time
!! @param escape : update escape array for one trajectory
!! @param sum_time : array use for
!! @param first_time : array for first passage time
!! @param time : actual time of one trajectory
!! @param dtnow : timestep ( variable in sda )
!! @param close : bolean true if the distance < rhit
subroutine check_reaction_pair ( prot1, prot2, rcriteria, visited_windows, escape, sum_time, first_time, time, dtnow, close )
    USE mod_protein
    USE mod_ReactionCriteria ! include mod_rate_calculation

    IMPLICIT NONE

    type ( protein ), intent ( in ) :: prot1, prot2
    type ( react_criter ) :: rcriteria
    real ( kind=4 ), dimension(:,:) :: first_time, sum_time
    real ( kind=8 ), dimension (:,:) :: escape
    integer, dimension(:) :: visited_windows
    real ( kind=4 ) :: time, dtnow
    logical, intent( out ) :: close

    real ( kind=4 ), dimension(:), allocatable :: dist2
    integer, dimension(:), allocatable :: irxa
    type ( association_rate ), pointer :: p_assoc_rate
    real ( kind=4 ), dimension(3) :: tmp_pos_relat
    real ( kind=8 ) :: tmp_e

    integer :: ir, iless, less, iap, irxn, nnons_min, ntotal, nb_cont
    integer :: i, iw, nwin
    logical :: accept

    !! default false
    close = .false.
    less = 0
    irxn = 0
    nwin = 0
    nb_cont = 0

    nullify ( p_assoc_rate )

    nnons_min = rcriteria % nnons_min
    ntotal = rcriteria % ntotal
    !write (*,*) 'nnons_min', nnons_min
    !write (*,*) 'nwrec', p_reaction % nwrec
    !write(*,*) 'p_reaction % type_reaction ', p_reaction % type_reaction
    !write(*,*) 'ntotal reaction ', ntotal
    !write(*,*) 'nspec reaction ', rcriteria % nspec
    !write(*,*) 'time ', time

    ! could be a global array to avoid allocation
    allocate ( dist2 ( ntotal ) )
    dist2 = 0.

    !! cannot send real *8 to real*4 function, use intermediate
    !! here works only for sda_2proteins
    tmp_pos_relat =  real ( prot2 % position, 4 )
    !write(*,*) 'tmp_pos_relat ', tmp_pos_relat

    ! cannot, missing geometry as parameter
    ! call position_pbc ( geom, pprot1 % position, pprot2 % position, tmp_pos_relat )


    !!compute all distance square, use real position, do not care about surface or z only distance
    !! should find better, and generalize to both / three now
    call compute_distance_criteria_v2 ( prot1, prot2, tmp_pos_relat, 1, ntotal, dist2 )

    if ( rcriteria % type_reaction == assoc_rate ) then

        !write(*,*) 'type assoc_rate'
        ! no care about spec or not, loop over all reaction pair nr2

        ! variable of rate_c are used only here
        nwin = rcriteria % p_assoc_rate % nb_window

        ! assign p_assoc_rate, only if assoc_rate
        p_assoc_rate => rcriteria % p_assoc_rate

        !write(*,*) 'reaction dist2 ', dist2 ( 1 ), dist2( rcriteria % ntotal )

        ! allocate max temp array, for independent pairs
        allocate ( irxa ( ntotal ) )
        irxa = 0

        do iw = 1, nwin
            less = 0
            do ir = 1, ntotal

              !! dependent pairs check
              !write(*,*) '1 ',dist2 ( ir ), ' 2 ', p_assoc_rate % window2 ( iw )
              if ( dist2 ( ir ) < p_assoc_rate % window2 ( iw ) ) then

                  !! check independent pairs
                  if ( rcriteria % ind_pairs ) then
                      ! default accept it
                      accept = .true.
                      ! check if independent to previously accepted
                      ! order matters because first one always accepted
                      do iless = 1, less
                          iap = irxa( iless )
                          if ( rcriteria % dependent_pairs( iap, ir ) == 1 ) accept = .false.
                      end do

                      if ( accept ) then
                          less = less + 1
                          irxa ( less ) = ir
                      end if

                  !! else no check of dependence, increment less directly
                  else
                      less = less + 1
                  end if ! dependent or not pairs

              end if ! distance2 < win2
            end do ! loop over ir

            !write(*,*) 'iw ', iw,' less ', less !' nnons_min', nnons_min
            !write(*,*) 'less, irxa ', less, irxa(1:less)
            !write (*,*) ' p_assoc_rate % nb_contact ', p_assoc_rate % nb_contact
            !! here should more clever, variable min first 0
            !! because if 0 in window 10 , 0 in all other windows

            ! new code
            do nb_cont = 0, p_assoc_rate % nb_contact - 1

                ! if pass this test, nb_cont < less, so all arrays can be filled
                if ( less == nb_cont ) GOTO 101 ! could be cycle

                ! update escape, if here nb_cont < less
                ! escape ( iw, nb_cont + 1 ) = 0.d0
                ! need to test if minimum window at this contact
                ! should use something else than -1, nb_window + 1 would be more performant
                ! todo optimize without -1 test
                if ( ( visited_windows( nb_cont + 1 ) .gt. iw ) .or. ( visited_windows( nb_cont + 1 ) == -1 ) ) &
                    visited_windows( nb_cont + 1 ) = iw

                ! again here for sure nb_cont < less. Apply only for this snapshot, ok
                sum_time ( iw, nb_cont + 1 ) = sum_time ( iw, nb_cont + 1 ) + dtnow

                ! update fpt
                if ( p_assoc_rate % fpt ) then
                    !write(*,*) 'update first_time_array'
                    ! check it is the first update
                    if ( first_time ( iw, nb_cont+1 ) == -1. ) then
                        !write(*,*) 'update iw= ', iw,' with time ', time
                        first_time ( iw, nb_cont+1 ) = time
                    end if
                end if
                ! write(*,*) 'first_time ', first_time

                ! test if docking, old one
                ! if ( ( iw == rcriteria % nwrec ) .and. ( nnons_min == nb_cont + 1 ) ) irxn = 1
                ! better one ?
                if ( ( iw .le. rcriteria % nwrec ) .and. ( nnons_min .le. nb_cont + 1 ) ) irxn = 1

            end do ! nb_cont

 101      continue

        end do ! nb_window

        if (irxn == 1) close = .true.

        !write(*,*) 'state of visited_windows '
        !write(*,*) 'min ', visited_windows


!!!!!!!!!!!!!!! here electron transfer,
! easy to split at first, only one contact

!
! doc from sda6
!. win2 here must be actually win (see sda.f)
!. et rate each step ~ 10**5*Hda**2*exp(-Ea)/sqrt(L) ps**(-1)
!. Hda is coupling constant in eV, for each pair its log is read in
!. rxnreadet as scou and should be multiplied by 0.6*exp(-1.7*(d-1.4))
!. L and Ea - reorganisation and activation energies, resp., in kT
!. final formula is
!. exp(11.51+3.74+scou*2-3.4*d+Ea-0.5*ln(L)) ==
!. exp(scou*2-3.4*d+w)
!. where w=15.25+Ea-ln(L)/2, i.e. Ea = 15.25-w-ln(L)/2
!. ln(L)/2 ~ 1.15-1.50 for L=10-20 kT
!
    else if ( rcriteria % type_reaction == elec_transfer ) then


!c.. the routine checks if N of d(i) less than w(k)
!c.. then makes e(l)=0 if is.
!c.. here N=1: l=(N-1)*nw+k

        !nwin = rcriteria % p_assoc_rate % nb_window
        !! assign p_assoc_rate, only if assoc_rate
        p_assoc_rate => rcriteria % p_assoc_rate

        !write(*,*) 'check elec_tranfert ntotal', ntotal, p_assoc_rate % nb_window

        do i = 1, ntotal

          ! original code from sda6
          !IF (d(ia).GT.20.0d0) GO TO 101
          !DO 100 iw = 1,nw
          !    ee = 1.d0 - dexp( scou(ia)*2.d0-d(ia)*3.4d0+w(iw) )
 ! why ?? !    IF (ee.LT.eemin(iw)) eemin(iw) = ee
          !    IF (ee.LT.0.d0) ee = 0.d0
          !    e(iw) = e(iw)*ee ! e is escape
!         100     !CONTINUE

          !write(*,*) 'dist2 ', dist2 ( i )
          ! hard coded, should use window( nb_window ) ??
          ! \bug value of the criteria distance for electron transfert is hard-coded !
          if ( dist2 ( i ) < 400. ) then

            do iw = 1, p_assoc_rate % nb_window

                ! to test, costs nothing
                sum_time ( iw, 1 ) = sum_time ( iw, 1 ) + dtnow
                ! dist_rc contains in this case the value of scout
                tmp_e = 1.d0 - dexp( 2.0d0 * rcriteria % dist_rc ( i ) &
                                   - 3.4d0 * sqrt ( dist2 ( i ) ) &
                                   + p_assoc_rate % window ( iw ) )
                !write(*,*) 'tmp_e ', tmp_e
                !!if ( tmp_e <  p_assoc_rate % escape ( iw, 1 ) ) eemin ?? why
                if ( tmp_e < 0.d0 ) tmp_e = 0.d0
                ! always 1 contact in electron transfer
                escape ( iw, 1 ) =  escape ( iw, 1 ) * tmp_e
            end do

          end if
        end do
        ! can modify close easily, just need to check nwrec
        close = .false.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Here docking

    !> is docking, may add more like electron transfer
    else if ( rcriteria % type_reaction == docking ) then
!        write(*,*) 'type docking'
        !! docking option 6, 7

        close = .false.

        !! check spec first, to test docking with spec
        if ( rcriteria % nspec > 0 ) then

            !write(*,*) 'reaction nspec ', dist2(1), dist2( rcriteria % nspec )

            !! check distance of specific, if at least one is greater test rejected
            do i = 1, rcriteria % nspec
                !write (*,*) 'imax ', imax, ' max',dist2(imax(1))
                if ( dist2 ( i ) > rcriteria % dist_rc2 ( i ) ) then
                    deallocate ( dist2 )
                    !write(*,*) 'specific not validated, return'
                    return
                end if
            end do
        endif

        !! check aspecific, should be more than some predefined criteria
        if ( rcriteria % naspec > 0 ) then
            !write(*,*) 'reaction naspec ', dist2(1), dist2( rcriteria % naspec )
            do i = 1, rcriteria % naspec
                ! if less than criteria exit
                if ( dist2 ( i + rcriteria % nspec ) < rcriteria % dist_rc2 ( i + rcriteria % nspec ) ) then
                    deallocate ( dist2 )
                    !write(*,*) 'Anti Specific not validated, return'
                    return
                end if
            end do
        endif

        if ( rcriteria % nnonspec == 0 ) then
            deallocate ( dist2 )
            !! cspec tested before,ok so close is true
            close = .true.
            return
        end if

        less = 0
        allocate ( irxa( rcriteria % nnonspec ) )

!        write(*,*) 'rcriteria % naspec: ', rcriteria % naspec
!        write(*,*) 'rcriteria % nnonspec: ', rcriteria % nnonspec
!        write(*,*) 'rcriteria % nspec: ', rcriteria % nspec
        
NNONS:  do i = 1, rcriteria % nnonspec
!        write(*,*) 'dist2, criteria ', dist2( i + rcriteria % nspec ), rcriteria % dist_rc2 ( i + rcriteria % nspec )
            if ( dist2( i + rcriteria % nspec + rcriteria % naspec ) < &
                       rcriteria % dist_rc2 ( i + rcriteria % nspec + rcriteria % naspec ) ) then

!                write(*,*) 'criteria satisfied'
                !! no dependent pairs check
                if ( .not. rcriteria % ind_pairs ) then
                    less = less + 1
                    CYCLE NNONS
                end if


                !! check if they are independent
                accept = .true.
                do iless = 1, less
                    iap = irxa ( iless )
                    !write(*,*) 'iap ', iap
                    !write(*,*) 'dependent pairs (iap,i)', rcriteria % dependent_pairs( iap, i )
                    !! if independent, increment less, need to enter the loop
                    if ( rcriteria % dependent_pairs( iap, i ) == 1 ) accept = .false.

                end do

                if ( accept ) then
                    less = less + 1
                    irxa( less ) = i
                    !write(*,*) 'increment less', less
                end if

            end if
        end do NNONS

!        write(*,*) 'accept: ', accept
        !! check minimun nonspec is satisfied
        !write(*,*) 'less ', less
        if ( less >= rcriteria % nnons_min ) close = .true.

        !need to deallocate before leaving function ? not in this case
        deallocate ( irxa )

    end if ! end if docking

    deallocate ( dist2 )
    !write(*,*) 'clos1 ',close

end subroutine check_reaction_pair

!> Make separate version for testing, maybe merged into the common function later on
!!
!! similar code than normal version, only recopy part for association rate, sum_time not needed as input
!! Record sum_time for all windows, contacts specified in input
!! Check if close ( criteria nwrec and nnons ), will save complexess and write in record
subroutine check_reaction_pair_sdamm ( tab_protein, set1, set2, rcriteria, geom, sum_time, dtnow, array_close, nclose, close )
    USE mod_array_protein
    USE mod_ReactionCriteria ! include mod_rate_calculation
    USE mod_geometry
    IMPLICIT NONE

    interface

!        subroutine position_pbc ( geom, r1, r2, r_corrected )
!            USE mod_geometry
!            IMPLICIT NONE
!            type ( geometry ), intent ( in ) :: geom
!            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: r1, r2
!            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
!        end subroutine position_pbc
        subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( geometry ) :: geom
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
         end subroutine compute_dist_pbc

    end interface

    type ( array_protein_type ) :: tab_protein
    type ( react_criter ) :: rcriteria
    type ( geometry ) :: geom
    real ( kind=4 ), dimension(:,:) :: sum_time
    integer, dimension (:,:) :: array_close
    real ( kind=4 ) :: dtnow
    integer :: set1, set2, nclose
    logical, intent( out ) :: close

    type ( protein ), pointer :: pprot1, pprot2
    real ( kind=4 ), dimension(:), allocatable :: dist2
    integer, dimension(:), allocatable :: irxa
    type ( association_rate ), pointer :: p_assoc_rate
    real ( kind=8 ), dimension(3) :: tmp_pos_relat8
    real ( kind=4 ), dimension(3) :: tmp_pos_relat
    real ( kind=4 ) :: norm_factor, tmp_dist
    integer :: ir, iless, less, iap, irxn, nnons_min, ntotal, nb_cont
    integer :: iw, nwin, i1, i2 ! ierr
    logical :: accept

    !! default false
    close = .false.
    less = 0
    irxn = 0
    nwin = 0
    nb_cont = 0
    ! normalisation factor
    norm_factor = 0.
    tmp_dist = 0.

    !write(*,*) 'check reaction for sdamm sets ', set1, set2

    nnons_min = rcriteria % nnons_min
    ntotal = rcriteria % ntotal
    !write(*,*) 'ntotal ', ntotal

    ! could be a global array
    allocate ( dist2 ( ntotal ) )
    dist2 = 0.

    ! reinitilialize array_close
    nclose = 0
    array_close = 0

    ! consider it exists and is correct
    p_assoc_rate => rcriteria % p_assoc_rate

    ! allocate max temp array, for independent pairs
    allocate ( irxa ( ntotal ) )

    ! normalization factor, to adapt
    norm_factor = real ( tab_protein % set_of_protein ( set2, 2 ) - tab_protein % set_of_protein ( set2, 1 ) + 1, 4 ) * &
                  real ( tab_protein % set_of_protein ( set1, 2 ) - tab_protein % set_of_protein ( set1, 1 ) + 1, 4 )
    !write(*,*) 'norm_factor ', norm_factor


  !! loop over all pairs of the 2 first groups
  do i1 = tab_protein % set_of_protein ( set1, 1 ), tab_protein % set_of_protein ( set1, 2 )
    pprot1 => tab_protein % array ( i1 )

  do i2 = tab_protein % set_of_protein ( set2, 1 ), tab_protein % set_of_protein ( set2, 2 )
    !write(*,*) 'i1, i2 ', i1, i2
    pprot2 => tab_protein % array ( i2 )

    irxn = 0

    !write(*,*) 'i1, i2 ', i1, i2
    !write(*,*) 'position 1', pprot1 % position
    !write(*,*) 'position 2', pprot2 % position
    !! not same type real*8 and real*4
    !call position_pbc ( geom, pprot1 % position, pprot2 % position, tmp_pos_relat8 )
    call compute_dist_pbc ( geom, pprot1, pprot2, tmp_dist, tmp_pos_relat8 )
    !write(*,*) 'tmp_pos_relat8 ', tmp_pos_relat8
    tmp_pos_relat = real( tmp_pos_relat8, 4 )

    !! compute all distance square, use real position, do not care about surface or z only distance
    !! should find better, and generalize to both / three now
    call compute_distance_criteria_v2 ( pprot1, pprot2, tmp_pos_relat, 1, ntotal, dist2 )
    !write(*,*) 'dist2 ', dist2

    nwin = rcriteria % p_assoc_rate % nb_window
    !write(*,*) 'reaction dist2 ', dist2 ( 1 ), dist2( rcriteria % ntotal )

    ! need to reinitialize
    irxa = 0

    do iw = 1, nwin
        less = 0
        do ir = 1, ntotal

          !! dependent pairs check
          !write(*,*) '1 ',dist2 ( ir ), ' 2 ', p_assoc_rate % window2 ( iw )
          if ( dist2 ( ir ) < p_assoc_rate % window2 ( iw ) ) then

              !! check independent pairs
              if ( rcriteria % ind_pairs ) then
                  ! default accept it
                  accept = .true.
                  ! check if independent to previously accepted
                  ! order matters because first one always accepted
                  do iless = 1, less
                      iap = irxa( iless )
                      if ( rcriteria % dependent_pairs( iap, ir ) == 1 ) accept = .false.
                  end do

                  if ( accept ) then
                      less = less + 1
                      irxa ( less ) = ir
                  end if

              !! else no check of dependence, increment less directly
              else
                  less = less + 1
              end if ! dependent or not pairs

          end if ! distance2 < win2
        end do ! loop over ir

        !write(*,*) 'iw ', iw,' less ', less !' nnons_min', nnons_min
        !write(*,*) 'less, irxa ', less, irxa(1:less)
        !write (*,*) ' p_assoc_rate % nb_contact ', p_assoc_rate % nb_contact
        !! here should more clever, variable min first 0
        !! because if 0 in window 10 , 0 in all other windows

        ! new code
        do nb_cont = 0, p_assoc_rate % nb_contact - 1

            ! if pass this test, nb_cont < less, so all arrays can be filled
            if ( less == nb_cont ) GOTO 101 ! could be cycle

            ! again here for sure nb_cont < less. Apply only for this snapshot, ok
            !sum_time ( iw, nb_cont + 1 ) = sum_time ( iw, nb_cont + 1 ) + dtnow
            sum_time ( iw, nb_cont + 1 ) = sum_time ( iw, nb_cont + 1 ) + dtnow / norm_factor

           ! latest version
           if ( ( iw .le. rcriteria % nwrec ) .and. ( nnons_min .le. nb_cont + 1 ) ) irxn = 1

        end do ! nb_cont

101      continue

    end do ! nb_window

    ! add inside the loop, need to store closed conformation
    if  ( irxn == 1 ) then
        nclose = nclose + 1
        array_close ( nclose, 1 ) = i1
        array_close ( nclose, 2 ) = i2
    end if

  end do ! i2
  end do ! i1

    ! keep for further add of complexes output, not so easy, which one ?? need array of clos ??
    !if (irxn == 1) close = .true.
    if ( nclose > 0 ) close = .true.

    !! all cases, no need explicti deallocate, local to function
    !deallocate ( dist2 )
    !deallocate ( irxa )

 end subroutine check_reaction_pair_sdamm
