!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file compute_distance.f90
!! General functions to compute the distance between 2 proteins


!> Compute distance and relative position between 2 solutes.\n
!! Always r2-r1
!!
!! Applied with sdamm (should work with sda_2prot).
!! Take into account sphere/box, PBC, surface
!!
!! It returns the correct relative position (after PBC)\n
!! but the distance (dist) is only along the z-axis if one of the solute is a surface.\n
!! So sqrt( dot(r_corrected,r_corrected) ) is not always equal to dist.
!!
!! This function should always be called in force or energy calculation in the case of sdamm.
!! With sda_2proteins, pos_relat is always = prot2 % position, and dist is setup correctly
!! by compute_distance_test
!!
!! @param geom : instance of \ref mod_geometry::geometry "mod_geometry"
!! @param prot1 : solute 1, instance of \ref mod_protein::protein "protein"
!! @param prot2 : solute 2, instance of \ref mod_protein::protein "protein"
!! @param dist : return distance between the 2 solutes
!! @param r_corrected : return corrected relative position
subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )

    USE mod_geometry
    USE mod_protein
    IMPLICIT NONE
    type ( geometry ) :: geom
    type ( protein ), intent ( in ) :: prot1, prot2
    real ( kind=4 ), intent ( out ) :: dist
    real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected

    real (kind=8), dimension(3) :: r1, r2
    integer :: k

    dist = 0.
    r_corrected = 0.d0

    r1 = prot1 % position
    r2 = prot2 % position

    ! first compute the correct relative position
    ! if sphere (pbc not implemented for sphere) OR box without pbc
    if ( ( geom % type == enum_sphere ) .or. ( geom % pbc .eqv. .false. ) ) then
        r_corrected = r2 - r1

    ! else box AND pbc
    else
        ! pbc along x and y
        do k = 1 ,2
                r_corrected ( k ) = r2 ( k ) - r1 ( k )
                r_corrected ( k ) = r_corrected ( k ) - &
                                    geom % pbox % sizebox (k ) * nint( r_corrected ( k ) / geom % pbox % sizebox ( k ) )
        end do
        ! real distance for z, if surface no pbc along z
        r_corrected ( 3 ) = r2 ( 3 ) - r1 ( 3 )

        ! if no  surface, apply pbc on z also
        if ( geom % surface .eqv. .false. ) then
            r_corrected ( 3 ) = r_corrected ( 3 ) - &
                                geom % pbox % sizebox (3 ) * nint( r_corrected ( 3 ) / geom % pbox % sizebox ( 3 ) )
        end if

    end if

    ! second update the distance. If a surface is present, return only the z distance.
    !call compute_distance ( prot1, prot2, dist )

    ! if one surface is present, take only the distance along z
    if ( ( prot1 % surface ) .or. ( prot2 % surface) ) then
        ! just substract !! take absolute value
        !dist = real ( abs( prot2 % position( 3 ) - prot1 % position ( 3 ) ), 4 )
        dist = real ( abs ( r_corrected(3) ), 4 )

    ! only proteins
    else
        dist = real ( sqrt ( dot_product (r_corrected, r_corrected) ), 4 )
    end if

end subroutine compute_dist_pbc

!> Compute the distance between two proteins\n
!! Take into if a surface is present, but not for pbc ( call compute_dist_pbc instead )
!!
!! @todo See if some function should be moved into mod_geometry.\n
!!
!! @param prot1 : solute 1, instance of \ref mod_protein::protein "protein"
!! @param prot2 : solute 2, instance of \ref mod_protein::protein "protein"
!! @param dist : return distance between the 2 solutes
subroutine compute_distance ( prot1, prot2, dist )
    USE mod_geometry
    USE mod_protein
    IMPLICIT NONE
    type ( protein ), intent ( in ) :: prot1, prot2
    real ( kind=4 ), intent ( out ) :: dist

    ! if one of the protein is a surface, use only z
    if ( ( prot1 % surface ) .or. ( prot2 % surface) ) then
         ! just substract !! take absolute value
         dist = real ( abs( prot2 % position( 3 ) - prot1 % position ( 3 ) ), 4 )
    ! only proteins
    else
        dist = real ( sqrt ( dot_product ( prot2 % position - prot1 % position, &
                                    prot2 % position - prot1 % position ) ), 4 )
    end if

end subroutine compute_distance

!> Version used by make_bd_2prot to update dist.\n
!!
!! Used only by sda_2proteins, check if one of the solute is a surface
!!
!! Here tricky, because move not yet accepted before test escape\n
!! Particular to sda_2proteins, rtry is the potential new position\n
!!
!! @todo prot_pos1, always 0 (only call by 2proteins) ?? prot1_pos to delete ?
!!
!! @param geom : instance of \ref mod_geometry
!! @param rtry : potential new position of the solute
!! @param prot1_pos : position of the solute 1
!! @param dist : return the new distance if the move rtry is accepted
subroutine compute_distance_test ( geom, rtry, prot1_pos, dist )
    USE mod_geometry
    IMPLICIT NONE
    type ( geometry ) :: geom
    real ( kind = 8 ), dimension ( 3 ) :: rtry, prot1_pos
    real ( kind=4 ) :: dist

    ! if one of the protein is a surface and geom contains a surface, use only z
    if ( ( geom % surface ) .and. ( geom % type == enum_box ) ) then
         dist = real ( sqrt ( ( rtry ( 3 ) - prot1_pos ( 3 )) * ( rtry ( 3 ) - prot1_pos ( 3 )) ), 4 )
    ! only proteins
    else
        dist = real ( sqrt ( dot_product ( rtry - prot1_pos, rtry - prot1_pos ) ), 4 )
    end if

end subroutine compute_distance_test

! More logical r2-r1 like in sda_2protein

! Transform the position with PBC\n
!
! Depreceated, shoud delete this function.
!
! Account for box/sphere\n
! If a surface is present, PBC on x/y, refelxion on z_max
! param geom : instance of \ref mod_geometry::geometry "mod_geometry"
! param r1, r2 : initial position of solute 1 and 2
! param r_corrected : return corrected position (maybe after PBC)
!subroutine position_pbc ( geom, r1, r2, r_corrected )
!    USE mod_geometry
!    IMPLICIT NONE
!    type ( geometry ), intent ( in ) :: geom
!    real ( kind=8 ), dimension ( 3 ), intent ( in ) :: r1, r2
!    real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
!
!    integer :: k
!
!    !write(*,*) 'entry position_pbc'
!
!    ! no pbc implemented for sphere, and box without pbc
!    if ( ( geom % type == enum_sphere ) .or. ( geom % pbc .eqv. .false. ) ) then
!        r_corrected = r2 - r1
!
!    ! else box AND pbc
!    else
!        ! pbc along x and y
!        do k = 1 ,2
!                r_corrected ( k ) = r2 ( k ) - r1 ( k )
!                r_corrected ( k ) = r_corrected ( k ) - &
!                                    geom % pbox % sizebox (k ) * nint( r_corrected ( k ) / geom % pbox % sizebox ( k ) )
!        end do
!        ! real distance for z, if surface no pbc along z
!        r_corrected ( 3 ) = r2 ( 3 ) - r1 ( 3 )
!
!        ! if no  surface, apply pbc on z also
!        if ( geom % surface .eqv. .false. ) then
!            r_corrected ( 3 ) = r_corrected ( 3 ) - &
!                                geom % pbox % sizebox (3 ) * nint( r_corrected ( 3 ) / geom % pbox % sizebox ( 3 ) )
!        end if
!
!    end if
!
!end subroutine position_pbc

!> Compute the distance need by reaction criteria\n
!!
!! version changed because the movable atoms belong to protein now.
!! checked if prot1 is a surface - then, taking only the z-distance
!! @param prot1 : solute 1
!! @param prot2 : solute 2
!! @param pos_relat : relative position
!! @param n1 : which reaction criterium to start from
!! @param n2 : which reaction criterium to end on
!! @param dist : array with values of squared distance for each reaction criterium
subroutine compute_distance_criteria_v2 ( prot1, prot2, pos_relat, n1, n2, dist )
    USE mod_protein
    IMPLICIT NONE
    type ( protein ) :: prot1, prot2
    real ( kind=4 ), dimension ( n2-n1+1 ) :: dist
    real ( kind = 4 ), dimension( 3 ), intent( in ) :: pos_relat
    integer, intent( in ) :: n1, n2

    integer :: i

!            write(*,*) 'n1, n2', n1, n2
!            write(*,*) 'pos atom1 p1',  prot1 % p_pos_rcriteria(:,1)
!            write(*,*) 'pos atom1 p2',  prot2 % p_pos_rcriteria(:,1)

    if ( prot1 % surface ) then
!                write(*,*) 'prot1 is a surface!'
        do i = n1, n2
            dist( i ) = ( prot1 % p_pos_rcriteria(3,i) - pos_relat(3) - prot2 % p_pos_rcriteria(3,i) )**2
        end do
    else
!                write(*,*) 'prot1 is NOT a surface!'
        do i = n1, n2
            dist( i ) =( prot1 % p_pos_rcriteria(1,i) - pos_relat(1) - prot2 % p_pos_rcriteria(1,i) )**2 + &
                       ( prot1 % p_pos_rcriteria(2,i) - pos_relat(2) - prot2 % p_pos_rcriteria(2,i) )**2 + &
                       ( prot1 % p_pos_rcriteria(3,i) - pos_relat(3) - prot2 % p_pos_rcriteria(3,i) )**2
            !write(*,*) ' i dist ', i, dist ( i )
            !dot_vector ( pos_atom(:,i) - r - y(:,i) , pos_atom(:,i) - r - y(:,i) )
        end do
    end if
!            write(*,*) 'dist: ', dist

end subroutine compute_distance_criteria_v2

!> Compute the minimum distance between the surface atoms (i.e., solvent accessible atoms)
!! of the two given proteins (prot1 and prot2).
!! @param prot1 : input, solute 1
!! @param prot2 : input, solute 2
!! @param max_contact_dist : input, the maximum distance between two atoms from the two proteins that is required to define a contact
!! @param min_contact_dist : input, the minimum ditance between atoms on the same proteins to define two different contacts
!! @return contact_atoms : output, array of pairs of numbers, pointing to the accessible atoms lists of prot1 and prot2, respectively
function get_atoms_min_dist( prot1, prot2, max_contact_dist, min_contact_dist ) result ( contact_atoms )
    use mod_protein
    IMPLICIT NONE
    
    type ( protein ), intent (in) :: prot1, prot2
    real ( kind = 8 ), intent (in) :: max_contact_dist, min_contact_dist
    integer, dimension ( :,: ), allocatable :: contact_atoms
    ! 'helper' array
    integer, dimension ( :,: ), allocatable :: contact_test
    ! counters:
    integer :: i, j, k
    integer :: nb_contacts, max_nb_contacts, err
    real ( kind = 8 ) :: cur_dist
    real ( kind = 8 ), dimension (3) :: cur_atom1, cur_atom2, prev_atom
    logical :: new_contact
    
    nb_contacts = 0
    max_nb_contacts = 20
    new_contact = .true.
    
    if ( max_contact_dist .le. 1.0 ) then
        write(*,*) "ERROR The contact (inter) definition may be too strict: ", max_contact_dist
        STOP 1
    end if
    if ( min_contact_dist .le. 2.0 ) then
        write(*,*) "ERROR the contact (intra) definition may be too strict: ", min_contact_dist
        STOP 1
    end if
    
    if( .not.associated( prot1 % p_list_atom_surface ) .or. .not.associated( prot2 % p_list_atom_surface ) ) then
        write (*,*) "ERROR one of the proteins doesn't have accessibility computed!"
        STOP 1
    end if
    
    allocate( contact_test( max_nb_contacts,2 ), contact_atoms( max_nb_contacts,2 ), stat = err )
    if ( err .ne. 0 ) then
        write(*,*) "ERROR Subroutine compute_min_dist, cannot allocate 'contact_atoms', size: ",max_nb_contacts
        STOP 1
    end if
    contact_atoms( :,: ) = 0
    contact_test( :,: ) = 0
    
!    write(*,*) ' (compute distance) nb_atoms1: ',prot1 % nb_atom_surface,'; nb_atoms2: ',prot2 % nb_atom_surface
    do i = 1, prot1 % nb_atom_surface
        cur_atom1(:) = prot1 % p_list_atom_surface( :,i )
        do j = 1, prot2 % nb_atom_surface
            cur_atom2(:) = prot2 % p_list_atom_surface( :,j )
            
            cur_dist = sqrt( ( cur_atom1(1) - cur_atom2(1) )**2 + ( cur_atom1(2) - cur_atom2(2) )**2 &
                       + ( cur_atom1(3) - cur_atom2(3) )**2 )
!            write(*,*) '                cur dist: ',cur_dist
            if ( cur_dist .le. max_contact_dist ) then
            ! we found a contact!
                if ( nb_contacts .gt. 0 ) then
                ! but this might be not different from any previous one...
                    do k = 1, nb_contacts
                        prev_atom(:) = prot1 % p_list_atom_surface( :,contact_atoms( k,1 ) )
                        cur_dist = sqrt( ( cur_atom1(1) - prev_atom(1) )**2 + ( cur_atom1(2) - prev_atom(2) )**2 &
                           + ( cur_atom1(3) - prev_atom(3) )**2 )
                        if ( cur_dist .le. min_contact_dist ) then
                            ! this contact is not really a new one
                            new_contact = .false.
                            EXIT
                        end if

                        prev_atom(:) = prot2 % p_list_atom_surface( :,contact_atoms( k,2 ) )
                        cur_dist = sqrt( ( cur_atom2(1) - prev_atom(1) )**2 + ( cur_atom2(2) - prev_atom(2) )**2 &
                           + ( cur_atom2(3) - prev_atom(3) )**2 )
                        if ( cur_dist .le. min_contact_dist ) then
                            ! this contact is not really a new one
                            new_contact = .false.
                            EXIT
                        end if
                    end do
                end if
                if ( .not.new_contact ) then
                    new_contact = .true.
                    CYCLE
                end if
                
                ! yes, this was a true, new contact
                nb_contacts = nb_contacts + 1
                if ( nb_contacts .gt. max_nb_contacts ) then
                    ! we need to reallocate!
                    ! first, copy all the contacts found till now
                    contact_test( :,: ) = contact_atoms(:,:)
                    deallocate( contact_atoms )
                    allocate( contact_atoms( 2*max_nb_contacts,2 ), stat = err )
                    if ( err .ne. 0 ) then
                        write(*,*) "ERROR Subroutine compute_min_dist, cannot allocate 'contact_atoms', size: ",2*max_nb_contacts
                        STOP 1
                    end if
                    ! copy back
                    contact_atoms( :,: ) = 0
                    contact_atoms( 1:max_nb_contacts,: ) = contact_test( :,: )
                    deallocate( contact_test )
                    allocate( contact_test( 2*max_nb_contacts,2 ), stat = err )
                    if ( err .ne. 0 ) then
                        write(*,*) "ERROR Subroutine compute_min_dist, cannot allocate 'contact_test', size: ",2*max_nb_contacts
                        STOP 1
                    end if
                    contact_test( :,: ) = 0
                    max_nb_contacts = max_nb_contacts*2
                end if
                
!                write(*,*) ' TRUE CONTACT FOUND between atoms: ',i,' and ',j
                ! update the information: which atoms are in contact
                contact_atoms( nb_contacts,1 ) = i
                contact_atoms( nb_contacts,2 ) = j
            end if
        end do
    end do
    
    if ( nb_contacts .eq. 0 ) then
        deallocate( contact_atoms )
        allocate( contact_atoms( 1,1 ) )
        contact_atoms(1,1) = 0
    else if ( nb_contacts .lt. max_nb_contacts ) then
    ! cleanup - reallocate arrays
        contact_test( :,: ) = contact_atoms(:,:)
        deallocate( contact_atoms )
        allocate( contact_atoms( nb_contacts,2 ), stat = err )
        if ( err .ne. 0 ) then
            write(*,*) "ERROR Subroutine compute_min_dist, cannot allocate 'contact_atoms', size: ",nb_contacts
            STOP 1
        end if
        ! copy back
        contact_atoms( :,: ) = contact_test( 1:nb_contacts,: )
    end if
    deallocate( contact_test )
    
end function get_atoms_min_dist

