!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file make_bd_move_2prot.f90
!!
!! Apply the BD translation and rotation to the solutes


!> Main function to call from the main loop, version without timer_exclusion
!!
!! In addition to the deterministic forces ( electrostatic, etc ) a random translational and rotational displacement is applied to all solutes
!!
!! Diffusion coefficients are from the protein class, not from the set_of_grid.
!! It allows to rescale the coefficients independently in every thread, set_of_grid keep the original value.
!!
!! Adapted to the 2 solutes case with exclusion check if the distance is smaller than a criteria (see rhit and hit )\n
!! If the solutes overlap, the move rejected and an other is tested until it accepted or total_overlap reach max_overlap
!!
!!
!! The move are not applied directly, but rather tested. Only at the end, if still all move were accepted the atoms are rotated\n
!! Delete the timer for exclusion_check in this version. see make_bd_move_2prot_timer_excl.f90\n
!! Trick : the solute 1 never rotates, it is the solute 2 which is translated. It keeps force calculation simpler
!!
!! @param prot1 : solute 1, instance of \ref mod_protein::protein "protein"
!! @param prot2 : solute 2, instance of \ref mod_protein::protein "protein"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param force : array of the deterministic forces on the solutes
!! @param torque : torque applied by the solute 1 on the solute 2
!! @param torque1 : torque applied by the solute 2 on the solute 1
!! @param dtnow : current timestep
!! @param dist : distance between the solutes
!! @param rhit : distance at which hit must be set to true
!! @param hit : if the exclusion must be checked
!! @param accept : return if the BD move has been accepted
!! @param escape : return if the solute 2 escapes from the box
!! @param total_overlap : return, count the number of overlap ( rejected move ). Use before running boost
subroutine make_bd_move_2prot ( prot1, prot2, geom, &
                                force, torque, torque1, &
                                dtnow, dist, rhit, hit, &
                                accept, escape, total_overlap )

    USE mod_protein
    USE mod_geometry
    IMPLICIT NONE

    interface

        subroutine bd_translation ( rorig, cell, rtry, geom, force, dmd, rmsnow )
            USE mod_geometry
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ) :: rtry
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: rorig, force
            integer, dimension ( 3 ), intent ( inout ) :: cell
            type ( geometry ), intent ( in ) :: geom
            real ( kind=8 ), intent ( in ) :: dmd, rmsnow
        end subroutine bd_translation

        subroutine bd_rotation_simple ( torque, rot_try, rot_displ,  rmsang )
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent (in) :: torque
            real ( kind=8 ), dimension (3,3), intent (out) :: rot_try
            real ( kind=8 ) :: rot_displ, rmsang
        end subroutine bd_rotation_simple

        subroutine check_exclusion ( prot1, prot2, rtry, rot_try, overlap, do_rot )
            USE mod_protein
            IMPLICIT NONE
            type ( protein ), intent( in ), target :: prot1, prot2
            real ( kind=8 ), dimension(3), intent( in ) :: rtry
            real ( kind=8  ), dimension (3,3), intent( in ) :: rot_try
            logical , intent ( out ) :: overlap
            logical, intent ( in ) :: do_rot
        end subroutine check_exclusion

        subroutine eschk ( geom, dtnow, disto4, dist4, dm, escape )
            USE mod_geometry
            IMPLICIT NONE
            type ( geometry ), intent ( in ) :: geom
            real ( kind=8 ), intent ( in ) :: dtnow, dm
            real ( kind=4 ), intent ( in ) :: disto4, dist4
            logical, intent ( out ) :: escape
        end subroutine eschk

    end interface

    type ( protein ) :: prot1, prot2
    type ( geometry ), intent ( in ) :: geom
    real ( kind=8 ), dimension ( 3 ), intent ( in ) :: force, torque, torque1
    real ( kind=8 ), intent ( in ) :: dtnow
    real ( kind=4 ) :: dist, rhit
    integer :: total_overlap
    ! accept true if at least one motion has been accepted
    logical, intent ( out ) :: hit, accept, escape

    real ( kind=8 ), dimension(3) :: rtry, pos_relat
    real ( kind=8 ), dimension(3,3) :: rot_try, rot_try1, rot_try_total

    real ( kind=8 ) :: dm_d, dm4, dr_d, dr1_d
    real ( kind=8 ) :: rms_trans_now, rms_angle_p2, rms_angle_p1
    real ( kind=4 ) :: disto
    logical :: overlap

    !write(*,*) 'make bd_move hit ', hit
    !write ( *,* ) 'dist hit', dist, hit
    !write ( *,* ) 'dtnow', dtnow
    !write(*,*) 'hit, rhit ', hit, rhit

#ifdef VERBOSE
    write(*,*) 'make bd_move hit ', hit
    write(*,*) 'dist / dtnow ', dist, dtnow
    write(*,*) 'trans so_grid ',  prot2 % p_sogrid % diffusion_trans
    write(*,*) 'trans protein ',  prot2 % diff_trans
    write(*,*) 'prot2 so_grid diff trans ',  prot2 % p_sogrid % diffusion_rotat
    write(*,*) 'prot2 diff_trans ',  prot2 % diff_rotat
#endif

    ! initialize constant,
    ! protein2 these parameters still belong to set of grid
    dm_d = prot2 % diff_trans * dtnow
    dm4 =  prot2 % diff_trans * 4.
    dr_d = prot2 % diff_rotat * dtnow
    !write(*,*) 'dm_d ', dm_d
    !write(*,*) 'dr_d ', dr_d

    ! protein1
    dr1_d = prot1 % diff_rotat * dtnow

    rms_trans_now = sqrt( 2.d0 * dm_d )
    rms_angle_p2 =  sqrt( 2.d0 * dr_d )
    rms_angle_p1 =  sqrt( 2.d0 * dr1_d )

    ! make a copy of previous distance, along z if correct
    pos_relat = prot2 % position
    disto = dist

    overlap = .false.
    accept = .false.

!!!!! Make Translation of protein 2 !!!!!!!!!

    !write(*,*) dist, rms_trans_now
    !! here random choice of displacement and update the position
    call bd_translation ( pos_relat, prot2 % box, rtry, geom, force, dm_d, rms_trans_now )
    !write(*,*) 'after bdmov rtry ', rtry

    ! can use interface here protein or position
    ! test version, because is not applied directly
    call compute_distance_test ( geom, rtry, prot1 % position, dist )

    !write(*,*) 'rtry ', rtry
    !write ( *,*) ' new_dist after translation dist disto ', dist, disto

    ! check first if protein2 escape from c surface
    call eschk ( geom, dtnow, disto, dist, dm4, escape )
    if ( escape ) then
             !write (*,*) 'eschk escape ', escape
             return
    end if

    !! first update of hit should be here !
    !write (*,*) 'before update hit dist hit', dist, hit
    if ( dist < rhit ) then
        hit = .true.
    !! in case true before and becomes false after translation
    else
        hit = .false.
    end if
    !write (*,*) 'update hit dist hit', dist, hit

!    write(*,*) 'hit first', hit
    ! test could be done only here, hit not yet updated here !!
    overlap = .false.
    if ( hit ) &
        call check_exclusion ( prot1, prot2, rtry, rot_try, overlap, .false. )
    !write(*,*) 'overlap trans ', overlap

    !! here if bias dynamic, could be a function
    !! if space between .not. and overlap, bug only with flag -O2
    if ( overlap .eqv. .false. ) then
        pos_relat = rtry
        ! apply here, even if rotation are rejected
        prot2 % position = rtry
!       write(*,*) 'accept translation r ', rtry
        !write(*,*) 'accept trans'

        ! at least one move accpeted
        accept = .true.

        ! here maybe
    !! else overlap true
    else
        !write(*,*) 'reject translation'
        dist = disto
        total_overlap = total_overlap + 1
        

        !! test for hit deleted before rotation, here should be reset
        !! if here check_has been done, it means it was already true before
        hit = .true.
        !!if ( disto > rhit ) hit = .false.

    end if

!!!!!!!! Make Rotation of protein 2 !!!!!!!

    call bd_rotation ( torque, rot_try, dr_d, rms_angle_p2, 2 )
    !write(*,*) 'after bdrot ytry ',  rot_try
    !write(*,*) 'rot matrix ',  rot_try
!      write(*,*) 'rtry ', ( pos_relat(m),m=1,3 )
!      write(*,*) 'rot_matrix', (rot_try(m,3),m=1,3)
!      write(*,*) 'rot_matrix', (rot_try(m,2),m=1,3)
!      write(*,*) 'rot_matrix', (rot_try(m,1),m=1,3)
!      write(*,*) 'rtry ', ( rtry(m),m=1,3 )

    !! hit used in case of possible overlap, may change after translation
    !! should be checked after tranlsation, only if accepted
    !! move before rotation, to update just after translation, rotation will no change dist !!
    !! cancel here, test done after translation
    !if (dist < rhit) hit = .true.


!    write(*,*) 'dist', dist, ' rhit ', rhit
    !write( *,*) ' hit rot2', hit
    !! need to check overlap
    ! version explicit
    overlap = .false.
    !write(*,*) 'prot1 ', prot1 % position
    !write(*,*) 'prot2 ', prot2 % position
    !write(*,*) 'pos_relat ', pos_relat
    !write(*,*) 'rot_try ', rot_try

    if ( hit ) &
        call check_exclusion ( prot1, prot2, pos_relat, rot_try, overlap, .true. )

    !write(*,*) 'after exchk overlap', overlap
    if ( overlap .eqv. .false. ) then
        ! if bias dynamics
        accept = .true.
        !write(*,*) 'accept rot'
   ! need to reinitialze rot_try to unity if overlap, otherwise keep for second rotation
   ! problem appears if rot2 is rejected but rot1 is ok when multiply rotation
    else
        !write(*,*) 'reject rot'
        total_overlap = total_overlap + 1
        rot_try(1,:) = (/1., 0., 0./)
        rot_try(2,:) = (/0., 1., 0./)
        rot_try(3,:) = (/0., 0., 1./)
        !! accept may be true if translat accepted
    end if

!!!!! Make rotation of protein 1 !!!!!!!!!!!!!
    if ( prot1 % rotate ) then
      !write(*,*) 'tworot ', prot1 % rotate, 'make rotation protein 1 !!!!'

   ! trick for rotation, cm of protein2 is modified
   ! rather than rotating protein1

   ! now split in 2
   ! now send back a new random matrix rotation, before rot used only internally
   ! need prot_id for difference between protein
   !              write(*,*) 'pos_relat ', (pos_relat(m),m=1,3)

        call bd_rotation ( torque1, rot_try1, dr1_d, rms_angle_p1, 1 )
    ! here use the last rot_try1, pos_relat should be the new position
    !    rtry = MATMUL( rot_try1, pos_relat )
        call rotate31 ( rtry, rot_try1, pos_relat )
! write(*,*) 'rot1 rtry ', ( pos_relat(m),m=1,3 )
! write(*,*) 'rot_matrix1', (rot_try1(m,3),m=1,3)
! write(*,*) 'rot_matrix1', (rot_try1(m,2),m=1,3)
! write(*,*) 'rot_matrix1', (rot_try1(m,1),m=1,3)
! write(*,*) 'rtry ', ( rtry(m),m=1,3 )

!      first rotation has not been applied to list_atom surface and charges
!      in case rot_try was accepted, we need to count for it
!       rot_try_total = MATMUL( rot_try1, rot_try )
        call rotate33 ( rot_try_total, rot_try1, rot_try )

        overlap = .false.
        if ( hit ) &
            call check_exclusion ( prot1, prot2, rtry, rot_try_total, overlap, .true.)

        if ( overlap .eqv. .false. ) then
            !! will need to rotate, and apply the translation to prot 2 when prot1 rotates
            prot2 % position = rtry
            accept = .true.
            ! if bias to add
            rot_try = rot_try_total
        else
            total_overlap = total_overlap + 1
        end if

    end if ! end if tworot

    ! finally apply the rotation(s) if necessary
    if ( accept .eqv. .true. ) then

        !write(*,*) 'final accept is true'

        ! version2 rotate movable atoms in proteins
        call update_rotation_protein ( prot2, rot_try )

        ! update orientation of axis, explicit terms
        call update_rot_orientation ( prot2, rot_try )

   endif

   !write(*,*) 'accept ', accept, 'last hit', hit
   !write(*,*) 'last hit', hit

   !write(*,*) 'update pos '
   !write(*,*) 'pos ', prot2 % position
   !write(*,*) 'x ',   prot2 % orientation_x
   !write(*,*) 'y ',   prot2 % orientation_y
   !write(*,*) 'z ',   prot2 % orientation_z
   !write(*,*) 'pos atom surf2', prot2 % p_list_atom_surface(:,1) , &
   !           prot2 % p_list_atom_surface( :,prot2 % nb_atom_surface )

  ! write(*,*) ''
  ! write(*,*) '==============================================='
  ! write(*,*) ''


end subroutine make_bd_move_2prot
