!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file make_bd_move_sdamm.f90
!!
!! Make the BD motion in the case of a sdamm simulation

!> Make BD motion for sdamm
!!
!! In addition to the deterministic forces ( electrostatic, etc ) a random translational and rotational displacement is applied to all solutes
!!
!! No exclusion check is performed with sdamm. Rather a repulsive lennard-jones interaction is included\n
!! Move are always accepted\n
!! This function is parralelised with OpenMP. But the scaling is limited\n
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param nb_prot : total number of solutes
!! @param force : array containing the total deterministic forces applied to all solutes
!! @param torque : array containing the total deterministic torques applied to all solutes
!! @param dt : timestep, constant in sdamm
subroutine  make_bd_move_sdamm ( tab_prot, geom, nb_prot, &
                                  force, torque, dt )

    USE mod_array_protein
    USE mod_geometry
    USE omp_lib
    IMPLICIT NONE

    interface
        subroutine bd_translation ( rorig, cell, rtry, geom, force, dmd, rmsnow )
            USE mod_geometry
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ) :: rtry
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: rorig, force
            integer, dimension ( 3 ), intent ( inout ) :: cell
            type ( geometry ), intent ( in ) :: geom
            real ( kind=8 ), intent ( in ) :: dmd, rmsnow

        end subroutine bd_translation

        subroutine bd_rotation_simple ( torque, rot_try, rot_displ,  rmsang )
            IMPLICIT NONE
            real ( kind=8 ), dimension (3), intent (in) :: torque
            real ( kind=8 ), dimension (3,3), intent (out) :: rot_try
            real ( kind=8 ) :: rot_displ, rmsang
        end subroutine bd_rotation_simple
    end interface


    type ( array_protein_type ), target :: tab_prot
    type ( geometry ), intent ( in ) :: geom
    real ( kind=8 ), dimension(3 , nb_prot), intent ( in ) :: force, torque
    integer, intent ( in ) :: nb_prot
    real ( kind=8 ), intent ( in ) :: dt

    type ( protein ), pointer :: pprot
    real ( kind=8 ) :: dm_d, dr_d
    real ( kind=8 ) :: rms_trans_now, rms_angle

    real ( kind=8 ), dimension(3) :: rtry
    real ( kind=8 ), dimension(3,3) :: rot_try
    integer :: n !,i

    !write(*,*) 'Entry make_bd_sdamm'
    !write(*,*) 'dt ',dt

!$OMP PARALLEL DEFAULT ( NONE ) SHARED ( tab_prot, nb_prot, force, torque, geom, dt ) &
!$OMP PRIVATE ( dm_d, rms_trans_now, dr_d, rms_angle, rtry, rot_try, pprot )

!! !$OMP DO SCHEDULE( DYNAMIC )
!! improve if using static here, particularly for gfortran, but still not very well scalable
!$OMP DO SCHEDULE ( STATIC )
NPROT : &
    do n = 1, nb_prot

        pprot => tab_prot % array ( n )

        !! test here, everytime. or before can make loop from 2 to nb_prot
        if ( pprot % surface ) then
            !write(*,*) 'prot ', n,' is a surface and does not move'
            CYCLE NPROT
        end if

        !! get parameters for move
        !dm_d = pprot % p_sogrid % diffusion_trans * dt
        !! movei into protein use in case of HI
        dm_d = pprot % diff_trans * dt
        rms_trans_now = sqrt( 2.d0 * dm_d )
        !dr_d = pprot % p_sogrid % diffusion_rotat * dt
        dr_d = pprot % diff_rotat * dt
        rms_angle =  sqrt( 2.d0 * dr_d )
        !write(*,*) 'prot ', n
        !write(*,*) 'diffusion_trans ', pprot % p_sogrid % diffusion_trans
        !write(*,*) 'diffusion_rot ',   pprot % p_sogrid % diffusion_rotat
        !write(*,*) 'before trans dm_d rms_trans_now, dr_d, rms_angle ', dm_d, rms_trans_now, dr_d, rms_angle
        !write(*,*) 'before bdmov rmsnow ', rms_trans_now
        !write(*,*) 'before bdrot rms_angle ', rms_angle
        !write(*,*) 'before translation n=',n
        !write(*,*) 'position ', pprot % position

        ! update position and cell information
        call bd_translation ( pprot % position, pprot % box, rtry, &
                              geom, force ( :,n ), dm_d, rms_trans_now )

        !! simple no exclusion check
        call bd_rotation_simple ( torque ( :, n ), rot_try, dr_d, rms_angle )

        !! update, no check here, may add one for large translation, rotation
        pprot % position = rtry
        !write(*,*) 'rtry ', rtry


        ! version with explicit rotation
        !call update_rotation2 ( pprot % p_sogrid, rot_try )
        !! alternative version, grid charge and atom surface in protein
        call update_rotation_protein ( pprot, rot_try )
        ! update orientation of axis, explicit terms
        call update_rot_orientation ( pprot, rot_try )

    end do NPROT
!$OMP END DO
!$OMP END PARALLEL

!   do i=1,nb_prot
!     write(*,*) 'update pos prot i', i
!     write(*,*) 'pos', tab_prot % array ( i ) % position
!     write(*,*) 'x ', tab_prot % array ( i ) % orientation_x
!     write(*,*) 'y ', tab_prot % array ( i ) % orientation_y
!     write(*,*) 'z ', tab_prot % array ( i ) % orientation_z
!   end do


end subroutine make_bd_move_sdamm
