!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_ReactionCriteria.f90
!!
!! Define reaction criteria

!> \defgroup ReactionCriteria Reaction Criteria
!!
!! More information about the reaction criteria ( defined in \ref mod_ReactionCriteria.f90 ).\n
!!
!! It is necessary to define a set of reaction criteria for :
!!  - Association rate calculations, \ref mod_reactioncriteria::react_criter::type_reaction "type_reaction" = \ref mod_reactioncriteria::assoc_rate "assoc_rate"
!!  - Docking,  \ref mod_reactioncriteria::react_criter::type_reaction "type_reaction" = \ref mod_reactioncriteria::docking "docking"
!!  - Electron transfer,\ref mod_reactioncriteria::react_criter::type_reaction "type_reaction" = \ref mod_reactioncriteria::docking "elec_transfert"
!!
!! By default, the object is switiched off with \ref mod_reactioncriteria::react_criter::type_reaction "type_reaction" = \ref mod_reactioncriteria::off_rc "off_rc"
!!
!! Both association rates and a limited docking simulations can be performed at the same time, but the rules for recording the complexes are simplified
!!
!! The reaction criteria input file consists of a set of definitions like :\n
!!
!! \verbatim
!! #keyword                              criteria for solute 1  | distance|                                 criteria for solute 2
!! CNONS ATOM    540  OE2 GLU A  60      26.527  32.135  -0.455 |    60.00| ATOM    332  N   LEU D  34      28.306  33.610   1.176
!! \endverbatim
!!
!! They consist of a keyword of 5 letters, an non read block of X letter, the position of the reaction criteria of the solute 1, the distance criteria, a non-read block and the position of the reaction criteria of the solute 2\n
!! \warning This is fortran code, the line should follow exactly the same formatting
!!
!! If solutes are not flexible, only the position of the atoms are read from the line. Otherwise the atom number is read and stored.
!! And then the position of the atoms, for each conformation, will be read from the pdb input file.
!!
!! Docking simulations need to define a set of reaction criteria with keywords :
!!   - CSPEC for a specific reaction. The distance between the atoms defining the reaction  must be less than the distance defined in order for a complex to be recorded.
!!   - CNONS for a non-specific reaction. Complexes are recorded when a certain number (or greater) of non-specific reaction criteria are met (as with specific reaction criteria, the distance between atoms defining the reaction criteria must be less than the defining distance). 
!! The number of non-specific criteria that must be met to record a complex is given by nnnons in the input file, and stored in the variable \ref mod_reactioncriteria::react_criter::nnons_min "nnons_min".
!!   - ASPEC for an anti-specific reaction. Here the distance between the atoms defining the reaction must be GREATER THAN the defined distance in order for a complex to be complex to be recorded.
!!
!! All theses requirements must be met for the complex to be accepted and recorded in the complex file. For example, if you define two specific reaction criteria, three non-specific reaction criteria, and set nnnons to two. The two defined specific criteria, and at least two of the non-specific criteria, must be met for a complex to be recorded.
!!
!! Association rate calculations will record the number of contacts which is formed at some distance.\n
!! You specify in the GROUP RateCalculation the maximum number of contacts with the keyword  nb_contact in the sda input file, \n
!! and the distance in defining the distance of the closest window ( first_window ), the number of windows ( nb_window ) and the width of the windows ( width_window ).\n
!! All theses parameters are stored as data members of \ref mod_rate_calculation::association_rate \n
!!
!! In this case you can still record complexes but all reaction criteria are considered as non specific ( CNONS ).\n
!! Only the complexes which satisfy both  :
!!  - nnnons : the minimum number of contatcs formed
!!  - nwrec : a fixed window number, at which the number of contacts will be computed
!! will be recorded in the complex file. The same rules for inserting complexes are applied
!!
!! Additionaly some pairs of independent contacts can be defined with the dind parameter.\n
!! If 2 reaction criteria of the same solute are at a distance smaller than dind, only 1 reaction will be considered.\n
!! You can see in reaction_criteria.log, how many of the criteria are considered as independent.\n
!!

! New version, add pointer to rate_calculation : 1 reaction criteria => 1 rate_calculation
!
! dist_rc and dist_rc2 use only for docking, distance of each pair is specified
! assoc_rate use rate_calculation % window, window2, all distance are computed and filled in escape...

!> Module to define reaction criteria\n
!! More informations on the \ref ReactionCriteria "reaction criteria page"
MODULE mod_ReactionCriteria

    USE mod_rate_calculation

    !> Enumeration, associated with  type_rection
    !! @{
    integer, parameter :: docking = 1, assoc_rate = 2, elec_transfer= 3, all = 4, off_rc = 5
    !! @}

    !> react_criter deals mainly with the reaction criteria positions and distances
    !!
    !! It owns a \ref mod_rate_calculation::association_rate "association_rate" object which will more specifically store statistics\n
    !! and compute association rates, electron_transfert and first passage time
    type react_criter

        !> position of the atom involved in reaction criteria
        real (kind=4), dimension(:,:), allocatable :: pos_atom1, pos_atom2

        !> Deal with independent pairs
        !! If dind > 0, keep in memory the dependent pairs \n
        !! Apply only to non specifi pairs
        integer ( kind=1 ), dimension(:,:), allocatable :: dependent_pairs

        !> Store the distance and the square distance associated to each reaction criteria
        !! @{
        real (kind=4), dimension(:), allocatable :: dist_rc, dist_rc2
        !> @}

        !> Store the pdb atom number ( read in \ref read_rxna )\n
        !! Apply only for flexible solutes, because the position may vary between the conformations
        !! @{
        integer, dimension(:), allocatable :: atom_nb1, atom_nb2
        !> @}


        !> parameter for distance independence
        real (kind=4) :: dind
        !> number of specific, anti-specific, non specific and total number of reaction criteria
        !! @{
        integer :: nspec, naspec, nnonspec, ntotal
        !> @}

        !> input parameter, equivalent to previous nnnons and ncec
        !!
        !! for docking, minimum number of nonspecif constrained to satisfy to record a complex
        !! for assoc_rate (and icmx>0) miminum number of constrained (all nonspecific in assoc_rate)
        !!     which satisfy a distance less than the one precised by nwrec
        !! @{
        integer :: nnons_min, nwrec
        !> @}

        !> define the type of calculation to perform, refer to \ref assoc_rate, \ref docking, \ref elec_transfer or \ref off_rc
        integer :: type_reaction

        !> set to true if we want a check of independent pairs, done if dind > 0
        logical :: ind_pairs

        !> pointer to a \ref mod_rate_calculation::association_rate "association_rate"
        type ( association_rate ), pointer :: p_assoc_rate

    end type react_criter

    contains

        !> Allocate arrays
        !!
        !! It sets up parameters and calls the function \ref read_rxna or \ref read_et_rxna for reading the reaction criteria
        !! The \ref mod_rate_calculation::association_rate "association_rate" pointer may be then initilized with \ref allocate_rate
        !! @param this : instance of react_criter
        !! @param type_reaction_string : "docking" / "asociation" / "electron_transfert" / "all" / "off_rc"
        !! @param filename : name of the file with the reaction criteria
        !! @param et1f, et2f : in case of electron transfert the reaction criteria are defined in theses files
        !! @param nnons : minimum number of nonspecific criteria to valid for accepting complexes
        !! @param nwrec : in case of association rate, the specific window where reaction criteria are checked
        !! @todo check initialization of dind
        subroutine allocate_react_criter ( this, type_reaction_string, filename, et1f, et2f, nnons, nwrec )
            IMPLICIT NONE
            type ( react_criter ) :: this
            character*128, intent ( in ) :: type_reaction_string, filename, et1f, et2f
            integer, intent( in ) :: nnons, nwrec

            ! initialize the variable of the module
            this % nnons_min = nnons
            this % nwrec = nwrec

            ! default ind_pairs is false by default, will be overwritten by make_independent
            this % ind_pairs = .false.
            ! init member variable before reading *rxna file
            this % nspec = 0
            this % naspec = 0
            this % nnonspec = 0
            this % ntotal = 0
            this % dind = 0.

            ! nullify pointer to association rate
            nullify ( this % p_assoc_rate )

            ! Assign correct type of reaction
            if ( type_reaction_string == 'association' ) then
                this % type_reaction = assoc_rate
            else if ( type_reaction_string == 'docking' ) then
                this % type_reaction = docking
            else if ( type_reaction_string == 'electron_transfer' ) then
                this % type_reaction = elec_transfer
            else if ( type_reaction_string == 'all' ) then
                this % type_reaction = all
            else if ( type_reaction_string == 'off_rc' ) then
                this % type_reaction = off_rc
            else
                write(*,*) 'This reaction criteria not yet implemented ', trim ( type_reaction_string )
                stop 1
            end if

            ! continue in reading file and allocating array
            ! if all no need
            if ( this % type_reaction == all ) return

            ! read rxna file, case assocaition or docking
            if ( ( this % type_reaction == assoc_rate ) .or. ( this % type_reaction == docking ) ) then
                call read_rxna ( this, filename )
            else if  ( this % type_reaction == elec_transfer ) then
                call read_et_rxna ( this, et1f, et2f )
            end if

        end subroutine allocate_react_criter

        !> Delete the arrays
        !!
        !! Delete all allocated arrays and the \ref mod_rate_calculation::association_rate "association_rate" object
        !! @param this : instance of react_criter
        subroutine delete_react_criter( this )
            IMPLICIT NONE
            type ( react_criter ) :: this
            integer :: ierr=0

            !! if rc not allocated, avoid to check undefined pointer p_react_assoc
            if ( this % type_reaction == off_rc ) return

            !! minimum if it has been initialized
            if ( allocated( this % dist_rc ) ) then
                deallocate( this % dist_rc, this % dist_rc2, &
                            this % pos_atom1, this % pos_atom2 , stat=ierr )
            endif
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR In deallocation memory of reaction criteria, dist_rc '
                stop 1
            endif

            !! if atom number from pdb, new for sda_flex
            if ( allocated( this % atom_nb1 ) ) then
                deallocate( this % atom_nb1, this % atom_nb2, stat=ierr )
            endif
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR In deallocation memory of reaction criteria, atom_nbX '
                stop 1
            endif

            !! for dependent pairs
            if ( allocated( this % dependent_pairs ) ) then
                deallocate ( this % dependent_pairs, stat = ierr )
            endif
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR In deallocation memory of dependent pairs '
                stop 1
            endif

            !!! reinitilaize variable
            this % type_reaction = off_rc
            this % nspec = 0
            this % naspec = 0
            this % nnonspec = 0
            this % ntotal = 0
            this % dind = 0.
            this % ind_pairs = .false.
            this % nnons_min = 0
            this % nwrec = 0

            !! delete p_assoc_rate, if needed
            if ( associated ( this % p_assoc_rate ) ) then
                call deallocate_rate_association ( this % p_assoc_rate )
                deallocate ( this % p_assoc_rate, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'Cannot deallocate  p_assoc_rate'
                end if
            end if

        end subroutine delete_react_criter

        !> Initilialize the association_rate pointer
        !!
        !! Create the object and initiliaze the internal pointer \ref react_criter::p_assoc_rate "p_assoc_rate"
        !! @param this : instance of react_criter
        !! @param win0, nwin, dwin : first_window, number and width of windows
        !! @param nb_contact : the maximum number of contact to record
        !! @param b,c : b and c surface
        !! @param dm : translational diffusion coefficient of the solute 2
        !! @param ibootsp, ifpt : if bootstrap for association rate or first tpassage time must be applied
        subroutine allocate_rate ( this, win0, nwin, dwin, nb_contact, &
                                   b, c, dm, ibootsp, ifpt )
            ! could replace by geom b,c sure, dm tedious...
            IMPLICIT NONE
            type ( react_criter ) :: this
            real ( kind=4 ), intent(in) :: win0, dwin
            integer, intent(in) :: nwin, nb_contact
            real ( kind=4 ), intent ( in ) :: b, c
            real ( kind=8 ), intent ( in ) :: dm
            logical :: ibootsp, ifpt

            integer :: ierr
            logical :: iboot_et

            !write(*,*) 'entry allocate rate b, c dm', b, c, dm
            !write(*,*) 'iboot fpt ', ibootsp, ifpt

            allocate ( this % p_assoc_rate, stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR in allocation of p_assoc_rate'
                stop 1
            end if

            ! set internal for elec_transfer, used only for bootstrap
            if ( this % type_reaction == elec_transfer ) then
                iboot_et = .true.
            ! default false, bootstrap for assoc_rate
            else
                iboot_et = .false.
            end if

            call allocate_rate_association ( this % p_assoc_rate, win0, nwin, dwin, nb_contact, &
                                             ! to replace by geom !! b,c sure, dm tedious...depends of sogrid
                                             b, c, dm, ibootsp, ifpt, iboot_et )

        end subroutine allocate_rate

        !> Read the reaction criteria file, usually *.rxna file
        !!
        !! Called by \ref allocate_react_criter\n
        !! Do not check for independence, default ind_pairs is false , done optionaly by make_independent\n
        !! Should be called the first time, if nb criteria are identical use update rxna\n
        !! Data stored in arrays drc and drc2 in the order 1,nspec and nspec+1, nspec+1+aspec,..nspec+aspec+1+nnonspec
        !! @param this : instance of react_criter
        !! @param filename : name of the reaction criteria file
        subroutine read_rxna ( this, filename )
            IMPLICIT NONE
            type ( react_criter ) :: this
            character*128, intent ( in ) :: filename

            ! for reading input file
            integer :: io_rxna=2111

            CHARACTER*128 string
            integer :: ierr
            integer :: io, tmp_spec, tmp_aspec, tmp_nonspec, tmp, m
            logical :: ex

            tmp_spec = 0
            tmp_aspec = 0
            tmp_nonspec = 0
            tmp = 0

            !write(*,*) 'Entry read rxna'

            !! check existence of filename
            inquire ( file=filename, exist = ex )
            !write(*,*) ' check filename ', trim ( filename ), ' ex:',ex
            if ( .NOT.ex ) then
                write( *, * ) 'ERROR File does not exist ', trim ( filename )
                stop 1
            endif
            open (io_rxna, file=filename , form= 'formatted', status='old', iostat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR Cannot open file, but present ???'
                stop 1
            end if
            !! check than not empty
            read ( io_rxna, '(a)', iostat=io ) string
            if ( io < 0 ) then
                write( *,* ) 'ERROR file rxna is empty', trim ( filename )
                stop 1
            endif

            !! make first loop to count spec and nonspec
            do while ( io == 0 )
                if (string(1:5).EQ.'CSPEC') then
                    tmp_spec = tmp_spec + 1
                else if (string(1:5).EQ.'ASPEC') then
                    tmp_aspec = tmp_aspec + 1
                else if (string(1:5).EQ.'CNONS') then
                    tmp_nonspec = tmp_nonspec + 1
                !else
                !    write(*,*) 'strange reading in read_rxna!!!'
                endif
                read ( io_rxna, '(a)', iostat=io ) string
            end do

            this % ntotal = tmp_spec + tmp_aspec + tmp_nonspec
            this % nspec = tmp_spec
            this % naspec = tmp_aspec
            this % nnonspec = tmp_nonspec


            !write(*,*) this % ntotal, this % nspec, this % naspec, this % nnonspec

            !write (*,*) 'total ', this%ntotal, ' tmp_spec ', tmp_spec, &
            !            ' tmp_nonspec ', tmp_nonspec
            ! allocate array
            ! HERE SHOULD CHECK TYPE, dist_rc/2 not needed if assoc_rate !
            ! atom_nb 1/2 used only if flex
            allocate ( this % dist_rc ( this % ntotal ), this % dist_rc2 ( this % ntotal ), &
                       this % atom_nb1 ( this % ntotal ), this % atom_nb2 ( this % ntotal ), &
                       this % pos_atom1 ( 3, this % ntotal ), this % pos_atom2 ( 3, this % ntotal ), stat=ierr )
            if ( ierr /= 0 ) then
                write (*,*) ' ERROR Couldn t allocate memory for reaction criteria '
                stop 1
            endif

            !! second loop, record dist and atom_nb array
            rewind( io_rxna )

            tmp_spec = 0
            tmp_aspec = this % nspec
            tmp_nonspec = this % nspec + this % naspec

            read ( io_rxna, '(a)', iostat=io ) string
LOOP_READ:  do while ( io == 0 )
                if (string(1:5).EQ.'CSPEC') then
                    tmp_spec = tmp_spec + 1
                    read ( string, fmt='(10x,i7,19x,3f8.3 )' ) &
                           this % atom_nb1 ( tmp_spec ), (this % pos_atom1 ( m, tmp_spec),m=1,3)

                    read ( string, fmt='(77x,i7,19x,3f8.3 )' ) &
                           this % atom_nb2 ( tmp_spec ), (this % pos_atom2 ( m, tmp_spec),m=1,3)

                    tmp = tmp_spec

                else if ( string(1:5) == 'ASPEC' ) then
                    tmp_aspec = tmp_aspec + 1
                    !write(*,*) 'read aspec tmp_aspec ', tmp_aspec
                    read ( string, fmt='(10x,i7,19x,3f8.3 )' ) &
                           this % atom_nb1 ( tmp_aspec ), (this % pos_atom1 ( m, tmp_aspec),m=1,3)

                    read ( string, fmt='(77x,i7,19x,3f8.3 )' ) &
                           this % atom_nb2 ( tmp_aspec ), (this % pos_atom2 ( m, tmp_aspec),m=1,3)

                    tmp = tmp_aspec
                    !write(*,*) 'read aspec tmp_aspec ', tmp_aspec

                else if (string(1:5).EQ.'CNONS') then

                    tmp_nonspec = tmp_nonspec + 1
                    read ( string, FMT='(10x,i7,19x,3f8.3 )' ) &
                           this % atom_nb1 ( tmp_nonspec ), (this % pos_atom1(m,tmp_nonspec),m=1,3)
                    !write(*,*) 'tmp ', this%atom_nb1( tmp_nonspec ), ' pos ', &
                    !         (this % pos_atom1(m,tmp_nonspec),m=1,3)

                    read ( string, FMT='(77x,i7,19x,3f8.3 )' )  &
                           this % atom_nb2 ( tmp_nonspec ), (this % pos_atom2(m,tmp_nonspec),m=1,3)
                    !write(*,*) 'tmp ', this%atom_nb2( tmp_nonspec ), ' pos ', &
                    !         (this % pos_atom2(m,tmp_nonspec),m=1,3)
                    tmp = tmp_nonspec
                !! else read the next line and come back to start of the loop
                else
                      !write(*,*) 'strange reading in read_rxna!!!'
                      read ( io_rxna, '(a)', iostat=io ) string
                      CYCLE LOOP_READ
                endif

                !! read associated distance to this reaction
                read (string(63:71),*) this % dist_rc ( tmp )
                this % dist_rc2( tmp ) = this % dist_rc ( tmp ) **2

                !! read next line
                read ( io_rxna, '(a)', iostat=io ) string
            end do LOOP_READ

            ! Close file
            close (io_rxna)

            !! print info
            !do m= 1, this % ntotal
            !    write(*,*) this % atom_nb1 (m), this % pos_atom1 (:,m)
            !    write(*,*) this % atom_nb2 (m), this % pos_atom2 (:,m)
            !end do

        end subroutine read_rxna

        !> Read reaction criteria for electron transfert calculation
        !!
        !! Called by \ref allocate_react_criter
        !! @param this : instance of react_criter
        !! @param et1f, et2f : file names where the reaction criteria are defined
        subroutine read_et_rxna ( this, et1f, et2f )
            IMPLICIT NONE
            type ( react_criter ) :: this
            character*128, intent ( in ) :: et1f, et2f

            ! for reading input file
            integer :: io_rxna=2111

            CHARACTER*128 string
            real ( kind=4 ) :: scout
            integer :: ierr
            integer :: io, tmp
            logical :: ex

#ifdef DEBUG
            write(*,*) 'Entry read_et_rxna ',trim(et1f), trim(et2f)
#endif

            !!!!!!!!! read file 1
            ! check existence of filename
            inquire ( file=et1f, exist = ex )
            !write(*,*) ' check filename ', trim ( filename ), ' ex:',ex
            if ( .not.ex ) then
                write( *, * ) 'ERROR File does not exist ', trim ( et1f )
                stop 1
            endif
            open (io_rxna, file=et1f , form= 'formatted', status='old', iostat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR Cannot open file: ', trim(et1f),', but present ???'
                stop 1
            end if
            ! check than not empty
            read ( io_rxna, '(a)', iostat=io ) string
            if ( io < 0 ) then
                write( *,* ) 'file: ', trim ( et1f ), ' is empty'
                stop 1
            endif

            tmp = 0
            !! make first loop to count the total number of line, no spec or nonspec for et
            do while ( io == 0 )
                if ( (string(1:4)/='ATOM').and.(string(1:6)/='HETATM') ) then
                    write(*,*) 'not ATOM or HETATM in modreaction criteria read et'
                else
                    tmp = tmp + 1
                end if
                read ( io_rxna, '(a)', iostat=io ) string
            end do

            !write(*,*) 'total number line ', tmp
            this % ntotal = tmp
            allocate ( this % dist_rc ( this % ntotal ), this % dist_rc2 ( this % ntotal ), &
                       this % atom_nb1 ( this % ntotal ), this % atom_nb2 ( this % ntotal ), &
                       this % pos_atom1 ( 3, this % ntotal ), this % pos_atom2 ( 3, this % ntotal ), stat=ierr )
            if ( ierr /= 0 ) then
                write (*,*) ' ERROR Couldn t allocate memory for reaction criteria, in read elect. transfer '
                stop 1
            endif

            !! second loop, record dist and atom_nb array
            rewind( io_rxna )

            ! read only last column
            read ( io_rxna, '(a)', iostat=io ) string
            tmp = 0
            do while ( io == 0 )

                if ( (string(1:4)/='ATOM').and.(string(1:6)/='HETATM') ) then
                    write(*,*) 'not ATOM or HETATM in modreaction criteria read_et'
                else
                    tmp = tmp + 1
                     !case HETATM
                     !read ( string(1:61),fmt='(6X,i5,19X,3f8.3,6x,f6.2)' ) &
                     ! format initial example cyt
                     read ( string(1:67),fmt='(6X,i5,19X,3f8.3,6x,f6.2)' ) &
                        this % atom_nb1( tmp ), this % pos_atom1(:,tmp), scout
                     ! Mic: to rerun example, need initial implementation with facotr *10.
                     !write(*,*) 'scout prot1', scout
                     this % dist_rc ( tmp ) = scout * 10.
                endif
                read ( io_rxna, '(a)', iostat=io ) string
            end do
            close ( io_rxna )

            !!!!!!!!!!!!!!!!!!!!!    read file 2  !!!!!!!!!!!!!!!!!!!!!!!!!!
            ! check existence of filename
            inquire ( file=et2f, exist = ex )
            !write(*,*) ' check filename ', trim ( filename ), ' ex:',ex
            if ( .NOT.ex ) then
                write( *, * ) 'ERROR File does not exist: ', trim ( et2f )
                stop 1
            endif
            open (io_rxna, file=et2f , form= 'formatted', status='old', iostat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR Cannot open file ', trim ( et2f ),' but present ???'
                stop 1
            end if
            ! check than not empty
            read ( io_rxna, '(a)', iostat=io ) string
            if ( io < 0 ) then
                write( *,* ) 'file: ',trim ( et2f ), ' is empty'
                stop 1
            endif

            tmp = 0
            do while ( io == 0 )

                if ( (string(1:4)/='ATOM').and.(string(1:6)/='HETATM') ) then
                    write(*,*) 'not ATOM or HETATM in modreaction criteria read_et'
                else
                    tmp = tmp + 1
                    !read ( string(1:61),fmt='(6X,i5,19X,3f8.3,6x,f6.2)' ) &
                    ! format initial example cyt
                    read ( string(1:67),fmt='(6X,i5,19X,3f8.3,6x,f6.2)' ) &
                        this % atom_nb2( tmp ), this % pos_atom2(:,tmp), scout
                        !write(*,*) 'scout prot2 ', scout
                    ! Mic to rerun example, original factor * 10.
                    this % dist_rc ( tmp ) = this % dist_rc ( tmp ) + scout * 10.
                endif
                read ( io_rxna, '(a)', iostat=io ) string
            end do
            close ( io_rxna )

        end subroutine read_et_rxna

        ! print to standart log file

        !> Make the list of independent pairs
        !!
        !! @todo Need to think about independent pairs in case of flexible/multi-proteins\n
        !!
        !! @bug To transform irxa array from static to dynamic. Crash if too large number of contacts.
        !! Bug appears in webSDA, with 2 overlapping structures.
        !! A check is now done before in webSDA, but not in SDA.\n
        !!
        !! Independent pairs apply only to associaiton rates\n
        !! independence is check only for each pos_atom 1 and 2 separately but stored in matrix ( 1, 2 ) => no problem of center of mass
        !! @param this : instance of react_criter
        !! @param dind : distance to use for indepence ( in Angstrom ), Default 6 A
        subroutine make_independent( this, dind )
            IMPLICIT NONE
            type ( react_criter ) :: this
            real (kind=4), intent(in) :: dind

            real (kind=4) :: dind2, dd1, dd2, ddmax=0., ddmin= 999.

            ! tmp variable, not conserved, bug from here
            ! integer, dimension(7)  :: irxa
            integer, dimension(30) :: irxa

            integer :: ndepend =0, nspec, nnonspec, i1, i2, iap, less
            integer :: a11, a12, a21, a22, ierr
            logical :: accept

            nspec = this % nspec + this % naspec
            nnonspec = this % nnonspec
            this % dind = dind
            dind2 = dind * dind

            ! if here should have been tested before
            if ( dind > 0. ) this % ind_pairs = .true.

            !write(*,*) 'first allocation ', nnonspec, nnonspec*nnonspec

            ! before irxp
            allocate ( this % dependent_pairs( nnonspec, nnonspec ), stat = ierr )
            if ( ierr /= 0 ) then
                write (*,*) ' ERROR Couldn t allocate memory for dependent_pairs size', nnonspec,' square'
                stop 1
                !return, use a status, not a huge problem with sda_flex
            endif

            !! irxp is used in common variable, not irxa ??
            do i1 = 1, nnonspec
                do i2 = 1, nnonspec
                    !irxp( i1, i2 ) = 0

                    ! 0 default independent
                    this % dependent_pairs ( i1, i2 ) = 0

                    a11 = nspec + i1
                    a12 = nspec + i2
                    dd1 = ( this % pos_atom1 ( 1, a11 ) - this % pos_atom1 ( 1, a12 ))**2 + &
                          ( this % pos_atom1 ( 2, a11 ) - this % pos_atom1 ( 2, a12 ))**2 + &
                          ( this % pos_atom1 ( 3, a11 ) - this % pos_atom1 ( 3, a12 ))**2
                    a21 = nspec + i1
                    a22 = nspec + i2
                    dd2 = ( this % pos_atom2 ( 1, a21 ) - this % pos_atom2 ( 1, a22 ))**2 + &
                          ( this % pos_atom2 ( 2, a21 ) - this % pos_atom2 ( 2, a22 ))**2 + &
                          ( this % pos_atom2 ( 3, a21 ) - this % pos_atom2 ( 3, a22 ))**2

                    if ( ( dd1 < dind2 ) .or. ( dd2 < dind2 ) ) then
                        !write(*,*) 'write 1 in ', i1, i2
                        this  % dependent_pairs ( i1, i2 ) = 1
                        ndepend = ndepend + 1
                    end if

                end do
            end do

            write( *,* )
            write( *,* ) ' Construct the independent matrix for nonspecific reactions only '
            write( *,* ) 'it has ',ndepend, ' dependent pairs out of', ( this % nnonspec )**2
            !write( *,* ) 'nspec ', nspec

            ! not used, except to print max pair-distance
            do i1 = 1, nnonspec
                a11 = nspec + i1
                dd1 = ( this % pos_atom1 ( 1, a11 ) - this % pos_atom2 ( 1, a11 ))**2 + &
                      ( this % pos_atom1 ( 2, a11 ) - this % pos_atom2 ( 2, a11 ))**2 + &
                      ( this % pos_atom1 ( 3, a11 ) - this % pos_atom2 ( 3, a11 ))**2
                if ( dd1 > ddmax ) then
                    ddmax = dd1
                    !write(*,*) 'ddmax ', ddmax, ' dd1' , dd1
                endif
                if ( dd1 < ddmin ) then
                    ddmin = dd1
                endif
                !! here influence the bug
                if ( dd1 < 1.0 ) then
                    write(*,*) 'i1  dd1 ', i1, dd1
                end if
            end do
            write ( *,* ) 'maximal pair-distance as read in is',sqrt(ddmax)
            write ( *,* ) 'minimal pair-distance as read in is',sqrt(ddmin)
            write ( *,* ) 'independence distance            is', this % dind !sqrt(dind2)

            ! not necessary, just give an approximate number of independent pairs

            !! strange here, not really understood irxa
            !! used in all, but not a common variable !!
            !! just a temporary tabular to get the independent contacts
            !! certainly useless here
            !! need dependent_pairs ( irxp before )
            less = 0

            ! result depends on the order ! cannot give the independent number of contact !
            do i1 = 1, nnonspec

                accept = .true.
                do i2 = 1, less
                   iap = irxa ( i2 )
                   !if ( irxp(iap, i1) == 1 ) accept = .false.
                   if ( this % dependent_pairs ( iap, i1 ) == 1 ) accept = .false.
                end do

                ! first less == 0, always accepted !
                if ( accept ) then
                    !write(*,*) 'accepted i1, i2, iap ', i1, i2, iap
                    less = less + 1
                    !
                    if ( less == 29 ) then
                        write(*,'("ERROR: Greater than",I3," independent contacts were found.")') , less
                        write(*,*) "  Check that you input solute coordinates form a correct complex"
                        stop 1
                    end if
                    irxa ( less ) = i1
                end if
           end do

           write ( *,* ) 'there are at least ', less, &
                ' independent contacts in ', nnonspec, ' reaction criteria'
           !write( *,*) 'irxa ', irxa
           !write( *,*) 'will quit'

        end subroutine make_independent

        !> Test when reading reaction criteria, called by \ref read_pdb
        !!
        !! Used when one solute is flexible, search for the correct atom number\n
        !! Check than the atom number of this particular conformation corresponds to the atom number of the initial reaction criteria
        !! @param this : instance of react_criter
        !! @param counter : identifier of the solute
        !! @param nb_atom : atom number
        !! @param xv : position of the atom
        !! @param tmp_rcriter_pos : return the position of the atom corresponding to the atom number
        !! @param status_incl : return true if the atom has been included in the array of reaction criteria
        subroutine check_add_criteria ( this, counter, nb_atom, xv, tmp_rcriter_pos, status_incl )
            IMPLICIT NONE
            type ( react_criter ) :: this
            integer, intent ( in ) :: nb_atom
            real ( kind=8 ), dimension( 3 ), intent ( in ) :: xv
            real ( kind=8 ), dimension(:,:), pointer :: tmp_rcriter_pos
            integer :: counter
            logical, intent ( out ) :: status_incl

            integer :: i

            !write(*,*) 'Entry check_add_criteria '
            !write(*,*) 'nb_atom ', nb_atom, 'xv ', xv
            status_incl = .false.

            do i= 1, this % ntotal

                if ( counter == 1 ) then
                    if ( this % atom_nb1 ( i ) == nb_atom ) then
                        write(*,*) 'found correspondance prot 1', nb_atom, xv
                        tmp_rcriter_pos ( :, i ) = xv
                        status_incl = .true.
                    end if

                else if ( counter == 2 ) then
                    if ( this % atom_nb2 ( i ) == nb_atom ) then
                        write(*,*) 'found correspondance prot 2', nb_atom, xv
                        tmp_rcriter_pos ( :, i ) = xv
                        status_incl = .true.
                    end if
                endif

            end do

        end subroutine check_add_criteria

        !> Write information to a log file
        !!
        !! io_unit is given by \ref read_input_file, egal to fort.52
        !! @param this : instance of react_criter
        !! @param io_unit : file descriptor of the log file
        subroutine write_to_logfile ( this, io_unit )
            IMPLICIT NONE
            type ( react_criter ) :: this
            integer, intent ( in ) :: io_unit

            integer :: i, io_err

            !write(*,*) 'entry write_to_logfile this % ntotal ', this % ntotal

            ! trick for false parrellel, maybe for real as well
            if ( io_unit == 0 ) then
                write(*,*) 'io_unit == 0 do not write the log file'
                return
            end if

            open (io_unit, file='reaction_criteria.log', form= 'formatted', status='replace', iostat = io_err )
            if ( io_err /= 0 ) then
                write (*,*) 'problem in opening the file reaction_criteria.log '
                stop 1
            end if

            write (io_unit,*)
            write (io_unit,*) 'Reaction criteria in memory'
            do i = 1, this % ntotal
                  write ( io_unit,* ) this % atom_nb1 ( i ), this % pos_atom1 (:, i), this % dist_rc ( i ), &
                           this % atom_nb2 ( i ), this % pos_atom2 (:, i)
            end do

            !! Print rxna in a pdb format
            write ( io_unit,* ) 'Write rxna files in a pdb format '
            !write ( io_unit,* ) 'Just copy and paste and superpose to your structure for checking rxna position, ', &
            !                    'take care of the initial structure if flexible '
            write ( io_unit,* ) 'This file can be directly loaded in pymol and VMD (load as PDB)'
            write ( io_unit,* ) 'generates 2 frames for solute 1 and 2'
            write ( io_unit,* ) 'Solute 1'
            write ( io_unit,fmt='(a)') 'MODEL'
            do i = 1, this % ntotal
                  write ( io_unit, fmt='("ATOM",26X,3f8.3)' ) this % pos_atom1 (:, i)
            end do
            write ( io_unit,fmt='(a)' ) 'ENDMDL'
            write ( io_unit,* )
            write ( io_unit,* ) 'Solute 2'
            write ( io_unit,fmt='(a)') 'MODEL'
            do i = 1, this % ntotal
                  write ( io_unit, fmt='("ATOM",26X,3f8.3)' ) this % pos_atom2 (:, i)
            end do
            write ( io_unit,fmt='(a)') 'ENDMDL'

            close ( io_unit )

        end subroutine write_to_logfile

END MODULE mod_ReactionCriteria
