!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_ResidenceTime.f90
!!
!! Compute, store and print all residence time and the radial distribution function with sdamm

! To extend if a spot on a surface, additional array
! default will be distance to surface

! tabular updated during BD loop
! related strongly to geometry, but particular parameters

!> residence time and rdf statistics
MODULE mod_ResidenceTime

    !! add new value rt_rdf for sdamm ?? or makes apart could combine later rdf with surface or other ??
    integer, parameter :: rt_off = 1, rt_distance = 2, rt_plan = 3, rt_resid3d = 4 !, rt_rdf = 5

    !> Arrays for 1, 2 and 3 dimensional residence time and radial distribution function with sdamm
    !!
    !! 3D arrays are printed in a UHBD grid format, others in the standard output file
    !!
    !! Arrays are updated every steps in sda_2proteins, only one step over 5 ( hardcoded with sdamm )\n
    !! Update must be thread-safe, use of ATOMIC section for a minimum overhead
    !!
    !! Radial distribution function can be computed with different types of solutes
    !! \todo correct the normalisation of rdf with different types of solutes
    !!
    !! If a 3 dimensional array is computed, 2 and 1 D residence time will be printed
    !! \todo check if resid 3d / 2d with sdamm is possible
    type residence_time

        !> type of the residence time : \ref rt_off, \ref rt_distance, \ref rt_plan or \ref rt_resid3d
        integer :: enum_rt

        !> array for 2D residence time, use x-y plan
        real ( kind=4 ), dimension ( :, : ), allocatable :: resid_plan

        ! could extend in sdamm between all types of proteins
        !> array for 1D, store only the distance
        real ( kind=4 ), dimension ( : ), allocatable :: resid_dist

        !> for radial distribution function, only sdamm
        !!
        !! different types possible ( ntype*(ntype+1)/2, nb_window )
        real ( kind=4 ), dimension( :, :, : ), allocatable :: rdf

        ! for PI3P case, want both distance to plan and along the plane
        ! make general 3d, need dynamic allocation, can be huge
        !> array for 3D residence time
        real ( kind=4 ), dimension(:,:,:), allocatable :: resid_3d


        !! 0 default fix the size of the cell in input, 1 fix the nb of cells, 2 fix both

        !> parameter for the grid
        !! @{
        integer :: fixed_bin_size
        real ( kind=4 ) :: xcell, ycell, zcell
        integer :: nb_xcell, nb_ycell, nb_zcell
        !> @}

        !> only for 3d, format of the grid, 0 : binary, 1 : ascii, default 0
        integer :: format_resid3d

        !> If more than one tpye of solutes is present in sdamm
        !! @{
        logical :: pair_proteins
        integer :: nb_pairs
        !> @}


        ! should be in geometry, get from geometry, local copy is usefull for update
        ! if sphere use xmin, xmax
        ! min value should be enought, because we have the number of cells
        !> origin of the grid
        !! @{
        real ( kind=4 ) :: x_origin, y_origin, z_origin
        !> @}

        ! should have equivalent b,c, let say xmin = 10 xmax = c

    end type residence_time

    contains

    ! geometry and fixe_bin_size should be enought, geom not needed here

    !> Initialize the instance
    !! @param this : instance of residence_time
    !! @param fixe_bin_size : if the size of the bin is fixed
    !! @param enum_str : "off" / "distance" / "plan" / "resid_3d"
    !! @param format_resid3d : ascii or binary for the output of the grid
    !! @param b_pair : if more than type of solute is present in sdamm
    !! @param nb_type : number of different type of solutes in sdamm, will set the number of pairs ( ( nb_type * ( nb_type + 1 ) ) / 2 )
    subroutine init_resid_time ( this, fixe_bin_size, enum_str, format_resid3d, b_pair, nb_type )
        USE mod_geometry
        IMPLICIT NONE
        type ( residence_time ) :: this
        integer, intent(in) :: fixe_bin_size
        !! for init should send the string
        character, intent(in) :: enum_str*(*)
        integer, intent(in) :: format_resid3d, nb_type
        logical, intent(in) :: b_pair

        !write(*,*) 'Entry init_resid_time'
        !write(*,*) 'b_pair ', b_pair, ' nb_type ', nb_type

        if ( enum_str == 'off' ) then
            this % enum_rt = rt_off
            return
        else if ( enum_str == 'distance' ) then
            this % enum_rt = rt_distance

        else if ( enum_str == 'plan' ) then
            this % enum_rt = rt_plan

        else if ( enum_str == 'resid_3d' ) then
            this % enum_rt = rt_resid3d
        else
            write(*,*)
            write(*,*) 'ERROR Value not known in residence time ', enum_str
            STOP 1
        end if

        this % fixed_bin_size = fixe_bin_size
        this % format_resid3d = format_resid3d
        this % pair_proteins = b_pair

        if ( b_pair .eqv. .true. ) then
            this % nb_pairs = ( nb_type * ( nb_type + 1 ) ) / 2
        else
            this % nb_pairs = 1
        end if
        !write(*,*) 'nb_pairs ', this % nb_pairs

        !initialize all array to null
        this % xcell = 0.
        this % ycell = 0.
        this % zcell = 0.
        this % nb_xcell = 0
        this % nb_ycell = 0
        this % nb_zcell = 0

        !! make own copy to avoid to test geometry every update
        !! even more general, can limit for small distance, role of origin
        this % x_origin = 0.
        this % y_origin = 0.
        this % z_origin = 0.

    end subroutine init_resid_time

    !> Delete arrays
    !! @param this : instance of residence_time
    subroutine delete_resid_time ( this )
        IMPLICIT NONE
        type ( residence_time ) :: this

        if ( allocated ( this % resid_dist ) ) then
            !write(*,*) 'Will deallocate resid_dist'
            deallocate ( this % resid_dist )
        end if

        if ( allocated ( this % resid_plan ) ) then
            !write(*,*) 'Will deallocate resid_plan'
            deallocate ( this % resid_plan )
        end if

        if ( allocated ( this % resid_3d ) ) then
            !write(*,*) 'Will deallocate resid_3d'
            deallocate ( this % resid_3d )
        end if

        if ( allocated ( this % rdf ) ) then
            !write(*,*) 'Will deallocate rdf'
            deallocate ( this % rdf )
        end if

    end subroutine delete_resid_time

    !> Not a real copy, data are modified when we want to create a 2D or a 1D form a 3D reisdence_time
    !!
    !! Only valable when resid3d and called by print_resid_time, NOT general\n
    !! only to use if this == resid_3d, want to produce new_rt as distance or plan residence time\n
    !!
    !! @param this : instance of residence_time, source of the copy
    !! @param new_rt : instance of residence_time, target of the copy
    !! @param new_enum_rt : type of the new_rt residence_time ( "plan" or "distance" )
    subroutine copy_resid_time ( this, new_rt, new_enum_rt )
        IMPLICIT NONE
        type ( residence_time ) :: this, new_rt
        integer :: new_enum_rt

        integer :: ierr
        !! if copy keep same opt_fixe_bin size, not really need if copy
        new_rt % fixed_bin_size = this % fixed_bin_size

        !! default value
        new_rt % nb_xcell = this % nb_xcell
        new_rt % nb_ycell = this % nb_ycell
        new_rt % nb_zcell = this % nb_zcell
        !! define other
        new_rt % xcell = this % xcell
        new_rt % ycell = this % ycell
        new_rt % zcell = this % zcell
        !! define last
        new_rt % x_origin = this % x_origin
        new_rt % y_origin = this % y_origin
        new_rt % z_origin = this % z_origin

        new_rt % enum_rt = new_enum_rt

        !write(*,*) 'copy resid_time'
        !write(*,*) 'this rt ', this % enum_rt
        !write(*,*) 'new_rt ', new_rt % enum_rt

        !! allocate array
        if ( new_rt % enum_rt == rt_distance ) then
            !write(*,*) 'allocate rt_distance with zcell', this % nb_zcell
            allocate ( new_rt % resid_dist ( new_rt % nb_zcell ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating resid_dist'
                STOP 1
            end if
            !write(*,*) 'after allocate shape ', shape ( new_rt % resid_dist )
            new_rt % resid_dist = 0.

            !! come from resid3d to distance, where only x used ( more logic for sphere...? )
            !! so overwrtties data z_data becomes x_data
            new_rt % nb_xcell = new_rt % nb_zcell
            new_rt % xcell = new_rt % zcell
            new_rt % x_origin = new_rt % z_origin
            !!other 0 or let defualt

        else if  ( new_rt % enum_rt == rt_plan ) then
            !write(*,*) 'allocate rt_distance with x, y', this % nb_xcell, this % nb_ycell
            allocate ( new_rt % resid_plan ( new_rt % nb_xcell, new_rt % nb_ycell ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating resid_plan'
                STOP 1
            end if
            new_rt % resid_plan = 0.
        !else
        !    write(*,*) 'error in copy_resid_time'
        end if

    end subroutine copy_resid_time

    !> Set grid parameters
    !!
    !! Can specify the size of the cells or the number of cells for each axis or both\n
    !! But it is an UHBD format which accepts only a fixed cell spacing for all 3 dimensions\n
    !! The grid can be produced correctly, but not visualized in VMD
    !!
    !! Modified origin, for the shift a grid spacing is missing
    !!
    !! @param this : instance of residence_time
    !! @param geom_resid : instance of \ref mod_geometry::geometry "geometry"
    !! @param nb_xcell, nb_ycell, nb_zcell : number of cell for each axis
    !! @param xcell, ycell, zcell : size of the cells for each axis
    subroutine set_array_rt ( this, geom_resid, nb_xcell, nb_ycell, nb_zcell, &
                              xcell, ycell, zcell )
        USE mod_geometry
        IMPLICIT NONE
        type ( residence_time ) :: this
        type ( geometry ), intent ( in ) :: geom_resid
        integer, intent ( in ) :: nb_xcell, nb_ycell, nb_zcell
        real ( kind=4 ), intent ( in ) :: xcell, ycell, zcell

        integer :: ierr

        !write(*,*) 'Entry set_array_rt fixed_bin_size ', this % fixed_bin_size, ' enum_rt ', this % enum_rt

        ! if 1 or 2 fix cell size
        if ( ( this % fixed_bin_size == 1 ) .or. ( this % fixed_bin_size == 2 ) ) then
                this % xcell = xcell
                this % ycell = ycell
                this % zcell = zcell
        end if

        ! if 0 or 2 fix nb cells, if 2 both are fixed
        if ( ( this % fixed_bin_size == 0 ) .or. ( this % fixed_bin_size == 2 ) ) then
                this % nb_xcell = nb_xcell
                this % nb_ycell = nb_ycell
                this % nb_zcell = nb_zcell
        end if

        !write (*,*) 'xcell ', this % xcell, this % ycell, this % zcell
        !write (*,*) 'nb_cell ', this % nb_xcell, this % nb_ycell, this % nb_zcell

        !! case of distance, origin is zero by default for both
        if ( this % enum_rt == rt_distance ) then
            !! case of sphere
            if ( geom_resid % type == enum_sphere ) then

                !! compute number of cells
                if  ( this % fixed_bin_size == 1 ) then
                    this % nb_xcell = int ( ( geom_resid % pshere % c_surface ) / this % xcell )

                else if ( this % fixed_bin_size == 0 ) then
                    this % xcell = ( geom_resid % pshere % c_surface ) / float ( this % nb_xcell )

                !! case  2 both fixed by input nothing to do
                end if

            else if ( geom_resid % type == enum_box ) then

                if ( this % fixed_bin_size == 1 ) then
                    this % nb_xcell = int ( ( geom_resid % pbox % z_max - this % z_origin ) / this % xcell )

                else if ( this % fixed_bin_size == 0 ) then
                    this % xcell = ( geom_resid % pbox % z_max - geom_resid % pbox % z_min ) / float ( this % nb_xcell )
                !! else 2 nothing to do
                end if

            end if

            !! allocate array
            allocate ( this % resid_dist ( this % nb_xcell ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating resid_dist'
                STOP 1
            end if
            this % resid_dist = 0.

            !!default origin = 0, nothing to do
            this % nb_ycell = 0
            this % nb_zcell = 0
            this % ycell = 0
            this % zcell = 0

        !! case plan
        else if ( this % enum_rt == rt_plan ) then

            if ( geom_resid % type == enum_sphere ) then
                write(*,*) 'Residence time in a plan with a sphere geometry is not implemented '
                stop 1

            else if ( geom_resid % type == enum_box ) then

                if ( this % fixed_bin_size == 1 ) then
                    this % nb_xcell = int ( ( geom_resid % pbox % x_max - geom_resid % pbox % x_min ) &
                                              / this % xcell ) + 1
                    this % nb_ycell = int ( ( geom_resid % pbox % y_max - geom_resid % pbox % y_min ) &
                                              / this % ycell ) + 1

                else if ( this % fixed_bin_size == 0 ) then
                    this % xcell = ( geom_resid % pbox % x_max - geom_resid % pbox % x_min ) &
                                              / float ( this % nb_xcell )
                    this % ycell = ( geom_resid % pbox % y_max - geom_resid % pbox % y_min ) &
                                              / float ( this % nb_ycell )
                end if

                !! default origin for a plan
                this % x_origin = - ( this % nb_xcell * this % xcell / 2. )
                this % y_origin = - ( this % nb_ycell * this % ycell / 2. )
                this % nb_zcell = 0
                this % zcell = 0
            end if

            allocate ( this % resid_plan ( this % nb_xcell, this % nb_ycell ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating resid_3d'
                STOP 1
            end if
            this % resid_plan = 0.

        ! case resid_3d
        else

            if ( geom_resid % type == enum_sphere ) then

                if ( this % fixed_bin_size == 1 ) then
                    this % nb_xcell = int ( 2. * geom_resid % pshere % c_surface  / this % xcell )
                    this % nb_ycell = int ( 2. * geom_resid % pshere % c_surface  / this % ycell )
                    this % nb_zcell = int ( 2. * geom_resid % pshere % c_surface  / this % zcell )

                else if ( this % fixed_bin_size == 0 ) then
                    this % xcell = 2. * geom_resid % pshere % c_surface / float ( this % nb_xcell )
                    this % ycell = 2. * geom_resid % pshere % c_surface / float ( this % nb_ycell )
                    this % zcell = 2. * geom_resid % pshere % c_surface / float ( this % nb_zcell )
                end if

                !! default origin, centeter around protein1
                this % x_origin = - ( this % nb_xcell * this % xcell / 2. )
                this % y_origin = - ( this % nb_ycell * this % ycell / 2. )
                this % z_origin = - ( this % nb_zcell * this % zcell / 2. )


            else if ( geom_resid % type == enum_box ) then

                if ( this % fixed_bin_size == 1 ) then
                    this % nb_xcell = int ( ( geom_resid % pbox % x_max - geom_resid % pbox % x_min ) / this % xcell ) + 1
                    this % nb_ycell = int ( ( geom_resid % pbox % y_max - geom_resid % pbox % y_min ) / this % ycell ) + 1
                    this % nb_zcell = int ( ( geom_resid % pbox % z_max - geom_resid % pbox % z_min ) / this % zcell ) + 1

                else if ( this % fixed_bin_size == 0 ) then
                    this % xcell = ( geom_resid % pbox % x_max - geom_resid % pbox % x_min ) / float ( this % nb_xcell )
                    this % ycell = ( geom_resid % pbox % y_max - geom_resid % pbox % y_min ) / float ( this % nb_ycell )
                    this % zcell = ( geom_resid % pbox % z_max - geom_resid % pbox % z_min ) / float ( this % nb_ycell )
                end if

                !! box, center 0 in xy plan, zmin is 0 by default, center of membrane
                this % x_origin = - ( this % nb_xcell * this % xcell / 2. )
                this % y_origin = - ( this % nb_ycell * this % ycell / 2. )

            end if

            ! check spacing for uhbd
            if ( ( xcell /= ycell ) .or. ( xcell /= zcell ) .or. ( ycell /= zcell ) ) then
                write(*,*)
                write(*,*) 'WARNING The spacing is different for the axis of the residence_3d grid : ', xcell, ycell, zcell
                write(*,*) 'WARNING Not supported by UHBD, you will not be able to visualize in VMD'
                write(*,*)
            end if

            !write(*,*) 'allocate array resid 3d', this % nb_xcell, this % nb_ycell, this % nb_zcell
            !! allocate array resid_3d
            allocate ( this % resid_3d ( this % nb_xcell, this % nb_ycell, this % nb_zcell ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating resid_3d'
                STOP 1
            end if
            this % resid_3d = 0.

        !! end if type residence
        end if

    end subroutine set_array_rt

    ! can group together later
    ! stupid to allocate nb_type * nb_type, need better..

    !> Similar to \ref set_array_rt but only for sdamm
    !!
    !! Additionaly allocate the array for the radial distribution functions
    !! @param this : instance of residence_time
    !! @param geom_resid : instance of \ref mod_geometry::geometry "geometry"
    !! @param nb_type : number of different type of solutes, needed for rdf
    !! @param nb_xcell, nb_ycell, nb_zcell : number of cell for each axis
    !! @param xcell, ycell, zcell : size of the cells for each axis
    subroutine set_array_rt_sdamm ( this, geom_resid, nb_type, nb_xcell, nb_ycell, nb_zcell, &
                              xcell, ycell, zcell )
        USE mod_geometry
        IMPLICIT NONE
        type ( residence_time ) :: this
        type ( geometry ), intent ( in ) :: geom_resid
        integer, intent ( in ) :: nb_xcell, nb_ycell, nb_zcell, nb_type
        real ( kind=4 ), intent ( in ) :: xcell, ycell, zcell

        integer :: ierr

        write(*,*) 'Entry set_array_rt_sdamm nb_type=', nb_type

        ! if 1 or 2 fix cell size
        if ( ( this % fixed_bin_size == 1 ) .or. ( this % fixed_bin_size == 2 ) ) then
                this % xcell = xcell
                this % ycell = ycell
                this % zcell = zcell
        end if

        ! if 0 or 2 fix nb cells, if 2 both are fixed
        if ( ( this % fixed_bin_size == 0 ) .or. ( this % fixed_bin_size == 2 ) ) then
                this % nb_xcell = nb_xcell
                this % nb_ycell = nb_ycell
                this % nb_zcell = nb_zcell
        end if

        write (*,*) 'xcell ', this % xcell, this % ycell, this % zcell
        write (*,*) 'nb_cell ', this % nb_xcell, this % nb_ycell, this % nb_zcell

        if ( this % enum_rt == rt_distance ) then

             if ( geom_resid % type == enum_box ) then

                ! keep same option as normal case
                if ( this % fixed_bin_size == 1 ) then
                    this % nb_xcell = int ( ( geom_resid % pbox % z_max - this % z_origin ) / this % xcell )

                else if ( this % fixed_bin_size == 0 ) then
                    this % xcell = ( geom_resid % pbox % z_max - geom_resid % pbox % z_min ) / float ( this % nb_xcell )
                !! else 2 nothing to do
                end if

             else
                write(*,*) 'sdamm in spehre not implemented'
            end if

            !! allocate array
            write(*,*) 'allocate rdf ', nb_type, nb_type, this % nb_xcell
            allocate ( this % rdf ( nb_type, nb_type, this % nb_xcell ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating rdf'
                STOP 1
            end if
            this % rdf = 0.

            !!default origin = 0, nothing to do
            this % nb_ycell = 0
            this % nb_zcell = 0
            this % ycell = 0
            this % zcell = 0

        else
            write(*,*) 'sdamm only option rt_distance implemented'
        end if

    end subroutine set_array_rt_sdamm

    !> Shift the origin of the residence_time grid
    !!
    !! Apply only to resid3d ??
    !! @param this : instance of residence_time
    !! @param shift_x, shift_y, shift_z : shift to apply to the origin
    subroutine apply_shift_to_rt ( this, shift_x, shift_y, shift_z )
        IMPLICIT NONE
        type ( residence_time ) :: this
        real ( kind=4 ) :: shift_x, shift_y, shift_z

        write(*,*) 'apply shift to rt ', shift_x, shift_y, shift_z

        this % x_origin = this % x_origin + shift_x
        this % y_origin = this % y_origin + shift_y
        this % z_origin = this % z_origin + shift_z

    end subroutine apply_shift_to_rt


    ! After bd_move accepted update the desired residence_time array
    ! should be dependent of the geometry, not really, fixed before

    ! dist was needed before, keep avoids to recompute the distance for sda_2proteins
    ! pos_prot need for plan and resid_3d only
    ! for rdf ?? certainly a different function used by sdamm, need to find wich type of proteins
    ! allow box smaller than available, need to check inside the box

    ! for use with sdamm, extend with array_protein, for test, makes appart
    ! do n

    ! version 3, use intial version of update_residence_time, modify only write_resid_3d
    !> Update residence_time
    !!
    !! Only used by sda_2proteins\n
    !! Update all arrays, done every steps in sda_2proteins\n
    !! This function must be thread-safe
    !! @param this : instance of residence_time
    !! @param pos_prot : position of one solute
    !! @param dist : distance, used only for sda_2proteins
    !! @param dtnow : current value of the timestep ( variable for sda_2proteins )
    subroutine update_residence_time ( this, pos_prot, dist, dtnow )
        USE mod_geometry
        IMPLICIT NONE
        type ( residence_time ) :: this
        !type ( geometry ) :: geom
        real ( kind=8 ), dimension ( 3 ) :: pos_prot
        real ( kind=8 ) :: dtnow
        real ( kind=4 ) :: dist

        real ( kind=4 ), dimension ( 3 ) :: pos_prot4
        integer :: dummy_r, dummy_x, dummy_y, dummy_z

        !write(*,*) 'Entry update_residence_time pos, dtnow ', pos_prot, dtnow

        !! should recompute dist or send as argument ??
        !! argument good, because dist can be distance protein-surface, sda_2proteins, already accounted in dist

        if ( this % enum_rt == rt_distance ) then
            !write(*,*) 'will fill resid_dist'

            !! here can trick with dist ?? if dist > threshold, problem no compute dummy_r
            !! certainly faster, and works for box geometry as well

            !! should substract xmin, 0 by default
            dummy_r = int ( ( dist / this % xcell ) - this % x_origin ) + 1

            !if ( dummy_r <= 0 ) then
            !    write(*,*) 'Error dummy_r ', dummy_r
            !    write(*,*) 'dist x_origin ', dist, this % x_origin
            !end if
            !if ( dummy_r > this % nb_xcell ) return

            if ( ( dummy_r <= 0 ) .or. ( dummy_r > this % nb_xcell ) ) return

            !$OMP  ATOMIC
            this % resid_dist ( dummy_r ) =  this % resid_dist ( dummy_r ) + real (dtnow,4)

            !end if
            !write(*,*) 'dummy_r ', dummy_r, ' dtnow ', dtnow
        !!end if

        else if ( this % enum_rt == rt_plan ) then
            !write(*,*) 'will fill bool_plan'
            pos_prot4 = real ( pos_prot, 4 )
            !! if some stuffs

            dummy_x = int ( ( pos_prot4( 1 ) - this % x_origin ) / this % xcell ) + 1
            dummy_y = int ( ( pos_prot4( 2 ) - this % y_origin ) / this % ycell ) + 1

            !! just for test to delete
            !write(*,*) 'min ', this % xmin,this % ymin,' dummy ', dummy_x, dummy_y
            if ( ( dummy_x > this % nb_xcell ) .or. ( dummy_x + 1 <= 0 ) ) then
                write(*,*) 'Problem with plan nb_xcell=', this % nb_xcell, dummy_x
                stop 1
            else if ( ( dummy_y > this % nb_ycell ) .or. ( dummy_y + 1 <= 0 ) ) then
                write(*,*) 'Problem with plan nb_ycell=', this % nb_ycell, dummy_y
                stop 1
            end if

            !$OMP ATOMIC
            this % resid_plan ( dummy_x, dummy_y ) =  this % resid_plan ( dummy_x, dummy_y ) + real(dtnow,4)

        else if ( this % enum_rt == rt_resid3d ) then
            !write(*,*) 'will fill resid_3d'
            ! intermediate pos_prot4, could skip ??
            pos_prot4 = real ( pos_prot, 4 )

            dummy_x = int ( ( pos_prot4( 1 ) - this % x_origin ) / this % xcell ) + 1
            dummy_y = int ( ( pos_prot4( 2 ) - this % y_origin ) / this % ycell ) + 1
            dummy_z = int ( ( pos_prot4( 3 ) - this % z_origin ) / this % zcell ) + 1
            !write(*,*) 'dummy_x, y, z ',dummy_x, dummy_y, dummy_z

            if ( ( dummy_x > this % nb_xcell ) .or. ( dummy_x <= 0 ) .or. &
                 ( dummy_y > this % nb_ycell ) .or. ( dummy_y <= 0 ) .or. &
                 ( dummy_z > this % nb_zcell ) .or. ( dummy_z <= 0 ) ) then
                    !write(*,*) 'out box return '
                    return
                    !stop 1
            end if

            !$OMP ATOMIC
            this % resid_3d ( dummy_x, dummy_y, dummy_z ) =  this % resid_3d ( dummy_x, dummy_y, dummy_z ) + real(dtnow,4)

        end if

    end subroutine update_residence_time

    ! slight modification from Myk comments
    ! two ways to modify, delete the shift from x_origin, always symetric ! ( not really if shift is applied )
    ! by default + 1 was applied before, no zero should be fine. Only 3d modified
!    subroutine update_residence_time ( this, pos_prot, dist, dtnow )
!        USE mod_geometry
!        IMPLICIT NONE
!        type ( residence_time ) :: this
!        !type ( geometry ) :: geom
!        real ( kind=8 ), dimension ( 3 ) :: pos_prot
!        real ( kind=8 ) :: dtnow
!        real ( kind=4 ) :: dist
!
!        real ( kind=4 ), dimension ( 3 ) :: pos_prot4
!        integer :: dummy_r, dummy_x, dummy_y, dummy_z
!
!        !write(*,*) 'Entry update_residence_time pos, dtnow ', pos_prot, dtnow
!
!        !! should recompute dist or send as argument ??
!        !! argument good, because dist can be distance protein-surface, sda_2proteins, already accounted in dist
!
!        if ( this % enum_rt == rt_distance ) then
!            !write(*,*) 'will fill resid_dist'
!
!            !! here can trick with dist ?? if dist > threshold, problem no compute dummy_r
!            !! certainly faster, and works for box geometry as well
!
!            !! should substract xmin, 0 by default
!            dummy_r = int ( ( dist / this % xcell ) - this % x_origin ) !+ 1
!
!            !if ( dummy_r <= 0 ) then
!            !    write(*,*) 'Error dummy_r ', dummy_r
!            !    write(*,*) 'dist x_origin ', dist, this % x_origin
!            !end if
!            !if ( dummy_r > this % nb_xcell ) return
!
!            if ( ( dummy_r <= 0 ) .or. ( dummy_r > this % nb_xcell ) ) return
!
!            !$OMP  ATOMIC
!            this % resid_dist ( dummy_r ) =  this % resid_dist ( dummy_r ) + real (dtnow,4)
!
!            !end if
!            !write(*,*) 'dummy_r ', dummy_r, ' dtnow ', dtnow
!        !!end if
!
!        else if ( this % enum_rt == rt_plan ) then
!            !write(*,*) 'will fill bool_plan'
!            pos_prot4 = real ( pos_prot, 4 )
!            !! if some stuffs
!
!            dummy_x = int ( ( pos_prot4( 1 ) - this % x_origin ) / this % xcell ) !+ 1
!            dummy_y = int ( ( pos_prot4( 2 ) - this % y_origin ) / this % ycell ) !+ 1
!
!            !! just for test to delete
!            !write(*,*) 'min ', this % xmin,this % ymin,' dummy ', dummy_x, dummy_y
!            if ( ( dummy_x > this % nb_xcell ) .or. ( dummy_x + 1 <= 0 ) ) then
!                write(*,*) 'Problem with plan nb_xcell=', this % nb_xcell, dummy_x
!                stop 1
!            else if ( ( dummy_y > this % nb_ycell ) .or. ( dummy_y + 1 <= 0 ) ) then
!                write(*,*) 'Problem with plan nb_ycell=', this % nb_ycell, dummy_y
!                stop 1
!            end if
!
!            !$OMP ATOMIC
!            this % resid_plan ( dummy_x, dummy_y ) =  this % resid_plan ( dummy_x, dummy_y ) &
!                                                      + real(dtnow,4)
!
!        else if ( this % enum_rt == rt_resid3d ) then
!            !write(*,*) 'will fill resid_3d'
!            ! intermediate pos_prot4, could skip ??
!            pos_prot4 = real ( pos_prot, 4 )
!
!            dummy_x = int ( ( pos_prot4( 1 ) - this % x_origin ) / this % xcell ) !+ 1
!            dummy_y = int ( ( pos_prot4( 2 ) - this % y_origin ) / this % ycell ) !+ 1
!            dummy_z = int ( ( pos_prot4( 3 ) - this % z_origin ) / this % zcell ) !+ 1
!
!
!            !write(*,*) 'dummy_x, y, z ',dummy_x, dummy_y, dummy_z
!           no need modif, now dummy_x can be 0, valid from [1 to nb_cell]
!            if ( ( dummy_x > this % nb_xcell ) .or. ( dummy_x <= 0 ) .or. & ! before <=0
!                 ( dummy_y > this % nb_ycell ) .or. ( dummy_y <= 0 ) .or. &
!                 ( dummy_z > this % nb_zcell ) .or. ( dummy_z <= 0 ) ) then
!                    !write(*,*) 'out box return '
!                    return
!                    !stop 1
!            end if
!
!            !$OMP ATOMIC
!            this % resid_3d ( dummy_x, dummy_y, dummy_z ) =  this % resid_3d ( dummy_x, dummy_y, dummy_z ) &
!                                                             + real(dtnow,4)
!
!        end if
!
!    end subroutine update_residence_time

    !> Update the radial distribution function
    !!
    !! Only used with sdamm
    !! @param this : instance of residence_time
    !! @param dist : distance between the 2 solutes
    !! @param type1, type2 : type of the solutes
    subroutine update_rdf ( this, dist, type1, type2 )
        IMPLICIT NONE
        type ( residence_time ) :: this
        real ( kind=8 ), intent ( in ) :: dist
        integer :: type1, type2

        integer :: dummy_r

        !write(*,*) 'Entry update_rdf ', dist, type1, type2

        if ( this % enum_rt == rt_distance ) then
            !write(*,*) 'will fill resid_dist'

            !! here can trick with dist ?? if dist > threshold, problem no compute dummy_r
            !! certainly faster, and works for box geometry as well

            !! should substract xmin, 0 by default
            dummy_r = int ( ( dist / this % xcell ) - this % x_origin ) + 1

            if ( ( dummy_r <= 0 ) .or. ( dummy_r > this % nb_xcell ) ) return

            !write(*,*) 'dummy_r ', dummy_r, type1, type2

            ! protect if later parallel
            !$OMP  ATOMIC
            this % rdf ( type1, type2, dummy_r ) =  this % rdf ( type1, type2, dummy_r ) + 1.0 !real (1.,4)

            ! make like paolo, if type1 == type2, similar to add 2
            !$OMP ATOMIC
            this % rdf ( type2, type1, dummy_r ) =  this % rdf ( type2, type1, dummy_r ) + 1.0 !real (1.,4)


        else
            write(*,*) 'not implemented'
            stop 1
        end if

    end subroutine update_rdf


    !> Print final residence_time
    !!
    !! Use recursive function : if the 3d residence time has been computed, 2d and 1d will be computed and printed
    !!
    !! @param this : instance residence_time
    !! @param geom_resid : instance of \ref mod_geometry::geometry "geometry"
    !! @param nrun : total number of runs, used for the normalization
    !! @result status : error status
    recursive function print_residence_time ( this, geom_resid, nrun ) result ( status )
        USE mod_geometry

        IMPLICIT NONE
        type ( residence_time ) :: this
        type ( geometry ), intent ( in ) :: geom_resid
        integer :: nrun, status

        !! used if resid3d, can post compute distance and/or plan
        type ( residence_time ) :: tmp_residence

        real ( kind=4 ) :: xyarea, rtmax, factor, tmp_xcell, tmp_xcell_ov2, tmp_sum
        integer :: i, j, k
        character*25 :: format_plan
        character*4 :: tmp_str_int

        !write(*,*) 'Entry print residence time rt', this % enum_rt

        status = 0
        factor = 0.
        xyarea = 0.
        rtmax = 0.
        format_plan = ''

        if ( this % enum_rt == rt_distance ) then

            !factor = 1.0 / ( float( nrun )
            write ( *, * )
            write ( *, * ) '# Mean residence time in 1 A '
            write ( *,* ) 'Check sum of time ', sum ( this % resid_dist )
            if ( geom_resid % type == enum_box ) then
                write ( *, * ) '# Only z distance from the surface ( in fact zmin )'
            !! else sphere
            else
                write ( *, * ) '# Distance between the center of geometry of solutes'
            end if

            !! new version, to test with different setup
            tmp_xcell = this % xcell
            tmp_xcell_ov2 = tmp_xcell / 2.

            !write(*,*) 'rt_distance nb_xcell=', this % nb_xcell, 'with geom_enum ', geom_resid % type
            if ( geom_resid % type == enum_box ) then
                !write(*,*) 'resid3d to distance with box geometry tmp_xcell ', this % xcell, ' tmp_nb_xcell ', this % nb_xcell
                write(*,*) 'sum over x', sum ( this % resid_dist )
            end if

            do i = 1, this % nb_xcell

                !! if sphere correct for geometrical factor
                if ( geom_resid % type == enum_sphere ) then
                    factor = 1.0 / ( ( i * tmp_xcell - tmp_xcell_ov2 ) * ( i * tmp_xcell - tmp_xcell_ov2 ) ) / float( nrun )
                !! else if box, no geometrical factor
                else
                    factor = 1.0 / float ( nrun )
                end if

                !write (*, fmt='(3e12.4,e16.8)' ) ( i * tmp_xcell - tmp_xcell_ov2 ), this % resid_dist ( i ) * factor
                !! need to account for the shift
                write (*, fmt='(3e12.4,e16.8)' ) ( i * tmp_xcell - tmp_xcell_ov2 + this % x_origin ), &
                                                 this % resid_dist ( i ) * factor
            end do

        end if

        if ( this % enum_rt == rt_plan ) then

            write(*,*) 'print rt_plan'
            write(*,*) 'sum of the time ',sum( this % resid_plan )

            write ( tmp_str_int, fmt='(i4)' ) this % nb_xcell
            !write(*,*) 'tmp_str_int!!',trim(tmp_str_int),'!!'

            format_plan = "((" // adjustr(trim ( adjustl( tmp_str_int))) // " (1X,f5.3)))"
            !write(*,*) 'format_plan ', format_plan

            !! transform in normalised probability
            xyarea = this % xcell * this % ycell
            ! resid_tab(i,j) = resid_tab(i,j)/xyarea/ float(nrun)
            ! no point, because want to normalize: just take the maximum !! to check
            ! yes not sure !
            this % resid_plan = this % resid_plan / (xyarea) / float( nrun )
            !! if make the sum, real normalization
            rtmax = MAXVAL ( this % resid_plan )

            !! to avoid Nan if nothing recorded
            if ( rtmax /= 0. ) then
                this % resid_plan = this % resid_plan / rtmax
            else
                this % resid_plan = 0.
            end if

            !!! print the 2D-tabular, need to indicate the shift ??
            write ( *, fmt="('# Mean RT in 1 A**2 area on cells',f8.3,' x',f8.3,'= area=',f8.3, ' fraction to max',e13.4)" ) &
                                this % xcell, this % ycell, xyarea, rtmax
            do j = 1, this % nb_ycell
                write ( *, fmt = format_plan ) ( this % resid_plan(i,j),i= 1, this % nb_xcell )
            enddo
        end if

        ! if resid3d, can easely print residence time distance and/or plan
        ! cannot print, but rather write resid3d if want all data
        ! no problem of shift will be in the grid, what's if call distance, no sense
        if ( this % enum_rt == rt_resid3d ) then

            write ( *, fmt="('# Write 3d grid on disk, normalize to procent time by point * 1000. ')" )
            ! has a sense if geometry is box
            if ( geom_resid % type == enum_box ) then
                !! copy attributes and allocate an empty array
                call copy_resid_time ( this, tmp_residence, rt_distance )
                !write(*,*) 'new shape after copy ', shape ( tmp_residence % resid_dist )
                !write(*,*) 'box average over x,y, shape new', shape ( tmp_residence % resid_dist )
                !write(*,*) 'average nb_zcell ', this % nb_zcell, 'over nb_x, nb_y ', this % nb_xcell, this % nb_ycell

                do k = 1, this % nb_zcell
                    tmp_sum = 0.
                    do i = 1, this % nb_xcell
                        do j = 1, this % nb_ycell
                            tmp_sum = tmp_sum + this % resid_3d ( i, j, k )
                        end do
                    end do
                    !write(*,*) 'tmp_sum ', tmp_sum
                    tmp_residence % resid_dist ( k ) = tmp_sum
                end do

                status = print_residence_time ( tmp_residence, geom_resid, nrun )

                !write(*,*) 'Will delete tmp_residence distance'
                call delete_resid_time ( tmp_residence )

                !write(*,*) 'finish 3d => dist'
            !!just for test apply correct shift
            else if ( geom_resid % type == enum_sphere ) then

                !! copy attributes and allocate an empty array
                call copy_resid_time ( this, tmp_residence, rt_distance )

                do k = 1, this % nb_zcell
                    tmp_sum = 0.
                    do i = 1, this % nb_xcell
                        do j = 1, this % nb_ycell
                            tmp_sum = tmp_sum + this % resid_3d ( i, j, k )
                        end do
                    end do
                    !write(*,*) 'tmp_sum ', tmp_sum
                    tmp_residence % resid_dist ( k ) = tmp_sum
                end do

                status = print_residence_time ( tmp_residence, geom_resid, nrun )

                !write(*,*) 'Will delete tmp_residence distance'
                call delete_resid_time ( tmp_residence )
            end if

        !!!!!!!make for plan
            !! may be usefull if sphere ?? is abvious symetry ??
            !! copy attributes and allocate an empty array
            call copy_resid_time ( this, tmp_residence, rt_plan )

            ! write(*,*) 'box average over z, new shape ', shape ( tmp_residence % resid_plan )
            do i = 1, this % nb_xcell
                do j = 1, this % nb_ycell
                    tmp_sum = 0.
                    do k = 1, this % nb_zcell
                        tmp_sum = tmp_sum + this % resid_3d ( i, j, k )
                    end do
                    tmp_residence % resid_plan ( i, j ) = tmp_sum
                end do
                !write(*,*) 'tmp_sum ', tmp_sum
            !tmp_residence % resid_plan ( i, j ) = tmp_sum
            end do

            !write(*,*) 'will print the plan'
            status = print_residence_time ( tmp_residence, geom_resid, nrun )

            !write(*,*) 'Will delete tmp_residence plan'
            call delete_resid_time ( tmp_residence )

            !write(*,*) 'finish 3d => plan'

        end if

        !write(*,*) 'exit print_residence'
        !! for recursive
        return

    !end subroutine print_residence_time
    end function print_residence_time

    !> Normalize and print radial distribution functions in RDF.log
    !! @param this : instance of residence_time
    !! @param tab_protein : instance \ref mod_array_protein::array_protein_type "array_protein_type"
    !! @param geom_resid : instance of \ref mod_geometry::geometry "geometry"
    !! @param n_rdfs : number of radial distribution functions
    subroutine print_rdf ( this, tab_protein, geom_resid, n_rdfs )
        USE com_sda
        USE mod_geometry
        USE mod_array_protein
        IMPLICIT NONE
        type ( residence_time ), target :: this
        type ( array_protein_type ) :: tab_protein
        type ( geometry ), intent ( in ) :: geom_resid
        integer :: n_rdfs

        ! variable for normalization
        real ( kind=8 ) :: r_low, rho, r_up, const, volume, n_ideal
        integer :: nb_prot, nb_type, i, k, l

        !integer :: io
        !status = 0

        volume = real (geom_resid % pbox % sizebox ( 1 ),8) * real(geom_resid % pbox % sizebox ( 2 ),8) * &
                 real(geom_resid % pbox % sizebox ( 3 ),8)

        write(*,*) 'sizebox ', geom_resid % pbox % sizebox ( : )
        write(*,*) 'sizebox real 8', real ( geom_resid % pbox % sizebox ( 1 ), 8 ), real ( geom_resid % pbox % sizebox ( 2 ), 8 )
        write(*,*) 'volume ', volume
        nb_type = size ( tab_protein % set_of_protein, 1 )
        write(*,*) 'nb_type ', nb_type
        write(*,*) 'this % xcell, this % nb_xcell ', this % xcell, this % nb_xcell

        write(*,*)
        write(*,*) '===================='
        write(*,*)

        open ( io_rdf, file = 'RDF.dat', form='formatted', status ='replace', action ='write' )

        do i = 1, this % nb_xcell
            r_low = ( i-1 ) * this % xcell
            r_up = r_low + this % xcell
            !r_rdf = r_low + ( this % xcell / 2 )

            write ( io_rdf, fmt='(f8.3)', advance ='no' ) r_low + ( this % xcell / 2 )

            do k = 1, nb_type

              ! nuber of protein of type k, should be computed once, in an array
              nb_prot = tab_protein % set_of_protein ( k, 2 ) - tab_protein % set_of_protein ( k, 1 ) + 1
              !write(*,*) 'nb_prot k ', k, nb_prot

              do l = k, nb_type

                rho = real ( nb_prot / volume, 8 )
                const = 4.d0 * pi * rho /3.d0
                n_ideal = const * ( r_up**3 - r_low**3 )
                write(*,*) 'n_ideal ', n_ideal
                write(*,*) 'n_rdfs * nb_prot * n_ideal ', n_rdfs * nb_prot * n_ideal
                write(*,*) 'rdf ', this % rdf ( k, l, i )
                !write(*,*) 'n_rdfs * nb_prot * n_ideal real4', real ( n_rdfs * nb_prot * n_ideal, 4 )

                ! still add one, so correct by dividing by 0.5d0, need to check in book
                this % rdf ( k, l, i ) = this % rdf ( k, l, i ) / real ( n_rdfs * nb_prot * n_ideal, 4 )
                !this % rdf ( k, l, i ) = this % rdf ( k, l, i ) / real ( 0.5d0 * n_rdfs * nb_prot * n_ideal, 4 )

                write ( io_rdf, fmt='(e15.5)', advance ='no' ) this % rdf ( k, l, i )
              end do
            end do

            ! make the line break
            write ( io_rdf, fmt='(a)' )
            ! write to the file
            !write ( io_rdf, * ) r_low + ( this % xcell / 2 ), this % rdf ( :,:, i )
        ! end do nb_xcell
        end do

        close ( io_rdf )
        return

    end subroutine print_rdf

    !> Write the 3d reisdence time grid to UHBD format
    !!
    !! Need to shift to the center of the solute 1 for visualisation
    !! @param this : instance of residence_time
    !! @param shift_center : center of solute 1 ?
    subroutine write_resid_3d ( this, shift_center )
        USE mod_grid

        IMPLICIT NONE
        type ( residence_time ), target :: this
        real ( kind=8 ), dimension(3) :: shift_center

        type ( type_grid ) :: out_grid
        real ( kind=8 ) :: sum_time
        real ( kind=4 ), dimension(: ,: ,:) , pointer :: tmp_pgrid4
        character*128 :: tmp_filename

        tmp_filename='resid3d.grd'
        nullify ( out_grid % p_grid )

        !write(*,*) 'Entry write_resid_3d, create real grid'

        !! try no copy, normalize in %
        sum_time = sum ( this % resid_3d )
        !! to avoid Nnan
        if ( sum_time /= 0. ) then
            !write(*,*) 'sum_time ', sum_time
            !this % resid_3d = real ( this % resid_3d / sum_time * 100. , 4 )
            ! comment from jon, number too small test * 1000
            this % resid_3d = real ( this % resid_3d / sum_time * 100. * 1000. , 4 )

            !write(*,*) 'maxval % ', maxval ( this % resid_3d )
        end if

        !! only one h value in uhbd format
        out_grid % title = 'ResidenceTime 3D'
        out_grid % iform = this % format_resid3d
        out_grid % scale = 1.d0

        ! need correct allocation of the pgrid, for call to write_uhbd_grid
        ! size not import here, rewritten later, need to delte after
        ! BUG write_resid3d
        call allocate_grid ( out_grid, 0, 0, 0, .false. )
        !write(*,*) 'after allocate_grid'

        !! default for UHBD, only one, limitation, could use dx file instead
        out_grid % h = this % xcell
        ! message send before, can be deleted
        if ( ( this % xcell /= this % ycell ) .or.  ( this % xcell /= this % zcell ) ) then
            write(*,*)
            write(*,*) 'WARNING Cell size are not cubic, the UHBD grid will not be readable by VMD'
            write(*,*)
        end if

        out_grid % origin ( 1 ) = this % x_origin - ( this % xcell / 2. ) + shift_center ( 1 )
        out_grid % origin ( 2 ) = this % y_origin - ( this % ycell / 2. ) + shift_center ( 2 )
        out_grid % origin ( 3 ) = this % z_origin - ( this % zcell / 2. ) + shift_center ( 3 )

        out_grid % size_x = this % nb_xcell
        out_grid % size_y = this % nb_ycell
        out_grid % size_z = this % nb_zcell

        !write(*,*) 'new origin', out_grid % origin(:)
        !write(*,*) 'x_origin ', this % x_origin, this % y_origin, this % z_origin
        !write(*,*) 'xcell/2 ', this % xcell / 2.
        !write(*,*) 'shift_center ', shift_center
        !write(*,*) 'origin for calcualtion', this % x_origin + ( this % xcell / 2. )


        !! already real ( 4 ) so direct for writting in uhbd_format, no need copy
        ! write_grid expects a pointer
        tmp_pgrid4 => this % resid_3d

        !! true for optional modif_name, will complete the suffix
        call write_grid ( out_grid, tmp_filename, .true., tmp_pgrid4 )

        !! improtant now to delete the grid
        call delete_grid ( out_grid )

    end subroutine write_resid_3d

end module mod_ResidenceTime



