!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_analytic.f90
!! Define analytical interactions, stored in 1D array

!> Analytical expressions, store energy and force in precomputed array
!!
!! \todo Extend to the case of a surface
MODULE mod_analytic

    ! can be general to all analytic expression, apply cutoff, compute derivative
    ! canot cross reference ! use same mod_grid_type, now can use an helper if wanted !
    use mod_gridtype

    !> Define tyoe analytic interactions (Debye-Huckel/crowders)
    type analytic

        ! for the order in memory, see:
        ! http://www.yolinux.com/TUTORIALS/LinuxTutorialMixingFortranAndC.html
        ! seems better to define ( nbbin, nbtype )

        !> define the arrays to store energy of interaction and force
        !! Could use 2 dimension array, to store for different types / stoke_radius
        !! @{
        real(kind=8), dimension(:,:), pointer :: parray_nrj
        real(kind=8), dimension(:,:), pointer :: parray_force
        !! @}

        !> type of the array
        !! @{
        integer :: type
        !! @}
        !> Array properties, nbbin
        !! @{
        integer :: nbbin
        !! size in Angs. from 0 to max_size
        real(kind=8) :: max_size
        !! spacing
        real(kind=8) :: h
        !! @}
        !> 1 solute representation, already in protein
        !! a bit useless becasue need the sum of the radius
        real(kind=8) :: stoke_radius

        !> Store ki, bjerrum usefull quantity independent of stoke radius
        real ( kind=8 ) :: ki, bjerrum, ionic

        !> Number of different types of solutes
        integer :: nbtypes

    end type analytic

    contains

    !> Allocate the object with default values, arrays filled by param_force_energy:init_analytic
    !! when all data (cutoff,..) are known
    !!
    !! @param this : instance of \ref mod_analytic
    subroutine allocate_analytic( this )
        IMPLICIT NONE
        type ( analytic ) :: this

#ifdef DEBUG
        write(*,*) 'allocate analytic set defaults and nullify all pointers only'
#endif
        this % nbbin = 0
        this % h = 0
        this % max_size = 0
        this % type = 0
        this % nbtypes = 0
        this % stoke_radius = 0.d0
        this % ki = 0.d0
        this % bjerrum = 0.d0
        ! can give a default value
        this % ionic = 0.d0

        nullify ( this % parray_nrj, this % parray_force )

    end subroutine allocate_analytic

    !> Deallocate all analytic arrays.\n
    !! @param this : instance of \ref mod_analytic
    subroutine deallocate_analytic( this )
        IMPLICIT NONE
        type ( analytic ) :: this

        integer :: ierr

        if ( associated ( this % parray_nrj ) ) then
            deallocate( this % parray_nrj, stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating array energy of analytic'
            end if
        end if

        if ( associated ( this % parray_force ) ) then
            deallocate( this % parray_force, stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating array force of analytic'
            end if
        end if

        ! reset the type to 0, unknown
        this % type = 0
        ! general for all analytic
        this % max_size = 0
        this % nbbin = 0
        this % h = 0

        this % nbtypes = 0
        this % stoke_radius = 0.d0

        ! specific to D-H
        this % ki = 0.d0
        this % bjerrum = 0.d0
        this % ionic = 0.d0

        !write(*,*) 'analytic deallocated'
        !call info_analytic( this )

    end subroutine deallocate_analytic

    !> fill the arrays, here need to setup dimension
    !!
    !! @param this : instance of \ref mod_analytic
    !! @param h : bin size of the precomputed arrays
    !! @param max_size : extensiom of DH, half the box for sdamm, c-surface for sda_2proteins
    !! @param stoke_radius : stoke radius of this solute
    subroutine set_analytic_params( this, h, max_size, stoke_radius )
        IMPLICIT NONE
        type ( analytic ) :: this
        real(kind=8), intent(in) :: stoke_radius, h
        real(kind=4) :: max_size

        ! easy to set up the correct value
        this % nbbin = int ( max_size / h )
        !write(*,*) 'this % nbbin ', this % nbbin
        this % h = h
        this % max_size = max_size
        ! assign the stoke_radius of this particular type of solute
        this % stoke_radius = stoke_radius

    end subroutine set_analytic_params

    !> Initialize debye-huckel parameters.
    !!
    !! @param this : instance of \ref mod_analytic
    !! @param opt_ionic : optional, if present set the ionic strength
    subroutine set_dh_params( this, opt_ionic )
        use com_sda
        IMPLICIT NONE
        type ( analytic ) :: this
        real(kind=4), optional :: opt_ionic

        real( kind=8 ) :: ionic

#ifdef DEBUG
        write(*,*) 'Entry set_dh_params'
#endif
        if ( present ( opt_ionic ) ) then
            ionic = real ( opt_ionic, 8 )
        else
            write(*,*) 'ERROR Parameter ionic strehngh is necessary to compute  debye-huckel'
            stop 1
        end if

        ! save as a debye-huckel grid
        this % type = get_nb_analyticaltype( 'debye-huckel' )
        ! phys parameters
        ! bjerrum length [Kg m^3 s^-2], ki = inverse of Debye-Hueckel length [1/m]
        ! Bjerrum legth in m
        this % bjerrum = e_q**2/(4.d0*pi*epsi_o*epssol*kcnst*temper)

        ! ionic strength
        this % ionic = ionic*1000.d0 ! conversion [mol/L] to [mol/m^3]

        ! bjerrum = e_q**2/(4.d0*pi*epsi_o*epssol*kcnst*temper) ! Bjerrum legth in m
        ! inverse debye length
        this % ki = dsqrt( 8.d0* pi * (this % bjerrum) * navo * (this % ionic) ) ! 1/DH length in meters^-1
#ifdef DEBUG
        write(*,*) 'bjerrum ', this % bjerrum
        write(*,*) 'ki ', this % ki
        write(io6,6000) (1.d0/( this % ki )*1e10),( this % bjerrum )*1e10, this % ionic
6000  format( /, 'D-H length : ',f10.3,' [A]',/, &
                 'Bjerrum length : ',f10.3, ' [A]',/, &
                 'Ionic strength : ',f10.3, ' [mol/m^3]' )
#endif

    end subroutine set_dh_params

    !! Precompute values of the debye-huckel array
    !!
    !! @param this : instance of \ref mod_analytic
    !! @param list_stoke_radius : array containing stoke radius of all solutes
    subroutine fill_array_dh ( this, list_stoke_radius )
        IMPLICIT NONE
        type ( analytic ) :: this
        real (kind=8), dimension(:) :: list_stoke_radius

        real (kind=8) :: p_size, dist, dist_m ! distance in meters
        real ( kind=8 ) :: ki, bjerrum ! local copy, to avoid this % ki...
        integer :: ierr, i, type

        ki = this % ki
        bjerrum = this % bjerrum

#ifdef DEBUG
        write(*,*)
        write(*,*) '=== Fill array_dh ==='
#endif

        !write(*,*) 'list_stoke_radius ', list_stoke_radius
        !write(*,*) 'size of list_stoke_radius ', size ( list_stoke_radius )
        this % nbtypes = size ( list_stoke_radius )
        allocate ( this % parray_nrj ( this % nbbin, this % nbtypes ), stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In allocating array energy of analytic'
        end if
        this % parray_nrj = 0.d0

        allocate ( this % parray_force ( this % nbbin, this % nbtypes ), stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'ERROR In allocating array force of analytic'
        end if
        this % parray_force = 0.d0

        ! fill array for the different sizes
        do type = 1, this % nbtypes
            ! paolo  code energy
            ! eij=(netq(igp(ip))*netq(igp(jp))*bjerrum*exp(-ki*(tmp_rij-p_size)))/(tmp_rij*(1.d0+ki*p_size))
            ! paolo force
            !ff_ip_tmp=(qa*qb*bjerrum*exp(-ki*(tmp_rij-p_size))*(ki*tmp_rij+1.d0))/(tmp_rij**2*(1.d0+ki*p_size))
            !> Convert to kT/A
            !ff_ip_tmp=ff_ip_tmp*1.d-10

            ! paolo force latest
            !ff_ip_tmp=(qa*qb*bjerrum*exp(-ki*(tmp_rij-p_size))*(ki*tmp_rij+1.d0))/(tmp_rij**2*(1.d0+ki*p_size))

!            write(*,*) 'For type =', type
!            write(*,*) 'With stoke_radius =', list_stoke_radius ( type )
            ! set correct radius, as the sum of the 2 proteins
            ! transform into meter
            p_size = ( this % stoke_radius + list_stoke_radius ( type ) ) * 1e-10
!            write(*,*) 'set p_size (in meter) ', p_size

            ! loop over bin, will not miss anything
            do i = 1, this % nbbin
                ! distance begins at h
                dist = i * this % h ! dist in Angstrom
                dist_m = dist * 1.0d-10 ! meters

                this % parray_nrj (i, type) = 0.5d0 * ( bjerrum * exp( -ki * (dist_m - p_size) ) ) / &
                                            ( dist_m * (1.d0 + ki * p_size) )

                this % parray_force (i, type) = 0.5d0 * &
                        ( bjerrum * exp( -ki * (dist_m-p_size) ) * (ki * dist_m + 1.d0) ) / &
                            ( dist_m**2*( 1.d0 + ki * p_size ) )
            end do
        end do

    end subroutine fill_array_dh

    !> Specific to Debye-Huckel set zero for  at distance where the grid is available
    !! Allow to call normal electrostatic and debye-huckel without taking care (hopefully)
    !! @param this : instance of \ref mod_analytic
    !! @param cutoff : max distance to which the value is set to 0
    subroutine set_zeroes ( this, cutoff )
        IMPLICIT NONE
        type ( analytic ) :: this
        real ( kind=4 ) :: cutoff

        integer :: i
        real( kind=4 ) :: dist

#ifdef DEBUG
        write(*,*) 'Entry set_zeroes with cutoff ', cutoff
#endif

LOOP_BIN: do i = 1, this % nbbin

            ! use bin number in mod_grid::make_sphere
            ! not a distance in angs.
            dist = i * real ( this % h, 4 )

            if ( dist < cutoff ) then
                ! write(*,*) 'dist < cutoff set to zero ', dist, cutoff
                ! reset to zero all forces and energies inside the cutoff
                this % parray_nrj ( i,: ) = 0.d0
                this % parray_force ( i,: ) = 0.d0
            ! first larger, can exit the loop
            else
                EXIT LOOP_BIN
            end if

        end do LOOP_BIN

    end subroutine set_zeroes

    !> Return precomputed energy values
    !!
    !! @param this : instance of \ref mod_analytic
    !! @param bin : integer bin number
    !! @param type_nb : integer number corresponding to the type of protein
    function get_energy_analytic ( this, bin, type_nb ) result ( energy )
        IMPLICIT NONE
        type ( analytic ) :: this
        integer :: bin, type_nb

        real(kind=8) :: energy

        if ( bin > this % nbbin ) then
#ifdef VERBOSE
            write(*,*) 'Ask bin larger than nbbin for energy ', bin, this % nbbin
            write(*,*) 'dist ', bin * this % h
#endif
            energy = 0.d0
        else
            energy = this % parray_nrj ( bin, type_nb )
        end if

    end function get_energy_analytic

    !> Return precomputed force values
    !!
    !! @param this : instance of \ref mod_analytic
    !! @param bin : integer bin number
    !! @param type_nb : integer number corresponding to the type of protein
    function get_force_analytic ( this, bin, type_nb ) result ( force )
        IMPLICIT NONE
        type ( analytic ) :: this
        integer :: bin, type_nb

        real(kind=8) :: force

        if ( bin > this % nbbin ) then
#ifdef VERBOSE
            write(*,*) 'Ask bin larger than nbbin for force ', bin, this % nbbin
            write(*,*) 'dist ', bin * this % h
#endif
            force = 0.d0
        else
            force = this % parray_force ( bin, type_nb )
        end if

    end function get_force_analytic

    !> Print values of array force and energy.\n
    !! Called by info_analytic
    !! @param this : instance of \ref mod_analytic
    subroutine print_analytic( this )
        IMPLICIT NONE
        type ( analytic ) :: this

        real (kind=8) :: dist
        integer :: i, j

        write(*,*) ' Print array_energy nbbin ', this % nbbin
        do i = 1, this % nbtypes
            write(*,*)
            write(*,*) 'with type ',i
            do j = 1, this % nbbin
                ! transform in distance
                dist = j * this % h
                write(*,*) dist, this % parray_nrj ( j, i )
            end do
            write(*,*)
        end do
        ! print force
        write(*,*) ' Print array_force nbbin ', this % nbbin
        do i = 1, this % nbtypes
            write(*,*)
            write(*,*) 'with type ',i
            do j = 1, this % nbbin
                ! transform in distance
                dist = j * this % h
                write(*,*) dist, this % parray_force ( j, i )
            end do
        end do

    end subroutine print_analytic

    !> Print info about analytical grids.
    !!
    !! Should apply to all, not only D-H
    !! @param this : instance of \ref mod_analytic
    !! @param opt_print : optional, print full arrays
    subroutine info_analytic( this, opt_print )
        IMPLICIT NONE
        type ( analytic ) :: this
        logical, optional :: opt_print

        write(*,*)
        write(*,*) '============='
        write(*,*) 'Info analytic'
        write(*,*)
        write(*,*) 'Type analytic: '
        call print_analyticaltype( this % type )

        write(*,*) 'nbbin ', this % nbbin
        write(*,*) 'h ', this % h
        write(*,*) 'max_size ', this % max_size
        write(*,*) 'is parray energy allocated ', associated( this % parray_nrj )
        write(*,*) 'is parray force allocated ',  associated( this % parray_force )
        write(*,*)

        if ( present ( opt_print ) ) then
            if ( opt_print .eqv. .true. ) then
                call print_analytic( this )
            end if
        end if

    end subroutine info_analytic


END MODULE mod_analytic
