!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_compute_analytic_sdamm.f90
!! Analytic (DH/crowder) Force and energy computation for SDAMM.\n

MODULE mod_compute_analytic_sdamm
     USE mod_grid
     USE mod_array_protein

    contains

!> Compute force with analytic function
!!
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param nb_prot : total number of solutes
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param all_force, all_torque : return all forces and torques
subroutine force_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                     all_force, all_torque )
    USE mod_array_protein
    USE mod_geometry
    USE mod_force_energy
    IMPLICIT NONE

    interface
        subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( geometry ) :: geom
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
        end subroutine compute_dist_pbc
    end interface

    type ( array_protein_type ), target, intent(in) :: tab_prot
    type ( geometry ), intent ( in ) :: geom
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension ( :,: ) :: all_force, all_torque
    integer, intent ( in ) :: nb_prot

    integer :: n1, n2, type_p1, type_p2
    type ( protein ), pointer :: pprot1, pprot2
    real ( kind=8 ), dimension ( 3 ) :: ff1_analytic, ff2_analytic
    real ( kind=8 ), dimension ( 3 ) :: ff1, tt1, ff2, tt2, pos_relat
    real ( kind=4 ) :: dist

    !write(*,*)
    !write(*,*) '=================================='
    !write(*,*) 'Entry force analytic dh sdamm'
    !write(*,*) 'nb_prot =', nb_prot

!! OMP starts here
!$OMP PARALLEL DEFAULT ( NONE ) SHARED ( tab_prot, nb_prot, param_force_energy, geom ) &
!$OMP PRIVATE ( pprot1, pprot2, pos_relat, dist, type_p1, type_p2 ) &
!$OMP PRIVATE ( ff1_analytic, ff1, ff2_analytic, ff2, tt1, tt2 ) &
!$OMP REDUCTION(+:all_force,all_torque) ! could try manual reduction like in force..not necessary for energy in sdamm

!$OMP DO SCHEDULE( DYNAMIC )
PROT1: do n1 = 1, nb_prot

          pprot1 => tab_prot % array ( n1 )
          type_p1 = param_force_energy % array_prot_type ( n1 )

PROT2:    do  n2 = n1 + 1, nb_prot

            pprot2 => tab_prot % array ( n2 )

            call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )
            !write(*,*) 'After compute dist pbc '
            !write (*,*) 'after dist, pos relat ',  dist, pos_relat, dot_product ( pos_relat, pos_relat )

            type_p2 = param_force_energy % array_prot_type ( n2 )

            call pair_force_full_analytic_sdamm( pprot1, pprot2, pos_relat, type_p1, type_p2, dist, &
                    ff2_analytic, ff1_analytic )
            !write(*,*) 'After full analytic'
            !write(0,*) 'Force ff2 DH ', ff2_analytic
            !write(0,*) 'Force ff1 DH ', ff1_analytic

            !write(*,*) 'Compute partial analytic '
            call pair_force_partial_analytic_sdamm( pprot1, pprot2, pos_relat, type_p1, type_p2, dist, &
                    ff2, ff1, tt2, tt1 )
            !write(*,*) 'After partial analytic'
            !write(*,*) 'Force ff1 DH ', ff1
            !write(*,*) 'Force ff2 DH ', ff2
            !write(*,*) 'Torque tt1 DH ', tt1
            !write(*,*) 'Torque tt2 DH', tt2

            ! symetrize the force, factor 2 times larger for full dh compared to SDAMM
            ! like previously, could multiply precomputed array by 0.5 (has been chosen)
            ! it symetrizes the result, seems sorrect
            all_force ( :,n1 ) = all_force ( :,n1 ) + (ff1_analytic + ff1) - (ff2_analytic + ff2)
            all_force ( :,n2 ) = all_force ( :,n2 ) + (ff2_analytic + ff2) - (ff1_analytic + ff1)
            ! torque has been corrected inside the function, not clear
            all_torque ( :,n1 ) = all_torque ( :,n1 ) + tt1
            all_torque ( :,n2 ) = all_torque ( :,n2 ) + tt2

    end do PROT2
end do PROT1
!$OMP END DO
!$OMP END PARALLEL

    !write(*,*) 'End force_analytic_dh'
    !write(*,*) 'f ', f
    !write(*,*) 't ', t
    !write(*,*) 't1 ', t1
    !write(*,*) '######################'

end subroutine force_analytic_dh_sdamm

!> Compute full analytic expression \n
!! To see later if it can be common to sda and sdamm. not far.
!!
!! @param prot1, prot2 : instance of \ref mod_protein::protein "protein"
!! @param pos_relat : position relative
!! @param type_p1, type_p2 : integer for the solute type
!! @param dist : distance between the 2 solutes
!! @param ff2, ff1 : return force on solute 2 and 1, should be exactly the same with full analytic
subroutine pair_force_full_analytic_sdamm( prot1, prot2, pos_relat, type_p1, type_p2, dist, ff2, ff1 )
    USE mod_array_protein
    IMPLICIT NONE
    type ( protein ), intent(in) :: prot1, prot2
    real( kind=4 ), intent ( in ) :: dist
    real ( kind=8 ), dimension ( 3 ), intent(out) :: ff2, ff1
    integer :: type_p1, type_p2

    real ( kind=8 ), dimension ( 3 ) , intent (in) :: pos_relat
    real ( kind=8 ), dimension ( 3 ) :: inv_pos_relat
    ! some need to be extracted only once, dpends 2 prot or sdamm
    real ( kind=8 ) :: ff_tmp, size_epgrid_p1, size_epgrid_p2
    integer :: bin

    ff1 = 0.d0
    ff2 = 0.d0

    !write(*,*) '==============================='
    !write(*,*) 'Entry pair force full analytic 2 proteins'
    !write(*,*) 'type p1 ', type_p1
    !write(*,*) 'type p2 ', type_p2
    !write(*,*) 'tot_charge *tot_charge ', prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total
    !write(*,*) 'dist (after pbc) ', dist

    size_epgrid_p1 = prot1 % p_sogrid % pset_grid ( 1 ) % rp
    size_epgrid_p2 = prot2 % p_sogrid % pset_grid ( 1 ) % rp

    ! compute energy2 , total charge prot2 in analytic grid of prot1
    ! only if full outside of prot1 el.grid
    if ( dist > ( size_epgrid_p1 + prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'outside size_epgrid_p1 ', size_epgrid_p1, ' +dist_max p2 ', prot2 % p_sogrid % distance_max
        !write(*,*) '===Compute ff2==='
        bin = int ( dist / prot1 % p_sogrid % pset_analytic_grid( 1 ) % h )
        ! norm of the force for this interaction
        ff_tmp = get_force_analytic ( prot1 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p2 )
        !> Convert to kT/A
        ff_tmp = ff_tmp * 1.d-10

        ! adjust the correct orientation
!write (0,*) "posrelat ", sqrt(dot_product(pos_relat,pos_relat))
        ff2 = prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total * ff_tmp * ( pos_relat / dist )
!write (0,*) "FF_2 ", ff2
    !else
    !    write(*,*) 'Partialy or full inside ep_grid1, see later '
    end if

    ! compute energy1, no need of change of refererential here
    ! only if full outside if prot2 el.grid
    if ( dist > ( size_epgrid_p2 + prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'outside size_epgrid_p2 ', size_epgrid_p2, ' +dist_max p1 ', prot1 % p_sogrid % distance_max
        !write(*,*) 'Compute ff1, test should be similar if full DH'
        !NJB: Now that pos_relat is passed into routine, can just invert here
        inv_pos_relat = -pos_relat !NJB

        bin = int ( dist / prot2 % p_sogrid % pset_analytic_grid( 1 ) % h ) !NJB
!if ( dist < ( size_epgrid_p1 + prot2 % p_sogrid % distance_max ) ) then
!write (0,*) "BIN" , bin
!write (0,*) prot1 % p_sogrid % pset_analytic_grid( 1 ) % h
!write (0,*) prot2 % p_sogrid % pset_analytic_grid( 1 ) % h
!write (0,*) int ( dist / prot2 % p_sogrid % pset_analytic_grid( 1 ) % h )
!end if
        ! norm of the force for this interaction
        ff_tmp = get_force_analytic ( prot2 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p1 )
        !> Convert to kT/A
        ff_tmp =ff_tmp * 1.d-10
        !write(*,*) 'ff_tmp ', ff_tmp !, 'norm ', sqrt( dot_product(ff_tmp,ff_tmp) )

        ! adjust with the correct orientation
        ff1 =  prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total * ff_tmp * ( inv_pos_relat / dist )
        !write(*,*) 'ff1 ', ff1, 'norm ff1 ', sqrt( dot_product( ff1, ff1 ) )

    !else
    !    write(*,*) 'Partialy or full inside ep_grid2, see later '
    end if

end subroutine pair_force_full_analytic_sdamm

!> Only partial analytic, loop over all atoms in this case.\n
!! Analytic DH is set to zero if inside the grid
!! @param prot1, prot2 : solute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param pos_relat : relative position
!! @param type_p1, type_p2 : integer to get the type of solute
!! @param dist : distance between the solutes
!! @param ff2, ff1, tt2, tt1 : return the torque and force for the 2 solutes
subroutine pair_force_partial_analytic_sdamm( prot1, prot2, pos_relat, type_p1, type_p2, dist, ff2, ff1, tt2, tt1 )
    USE mod_array_protein
    IMPLICIT NONE

    type ( protein ), intent(in) :: prot1, prot2
    real ( kind=8 ), dimension ( 3 ), intent(out) :: ff2, ff1, tt2, tt1
    real ( kind=8 ), dimension(3), intent(in) :: pos_relat
    real ( kind=4 ) :: dist
    integer :: type_p1, type_p2

    interface
          subroutine cross(ab,a,b)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: ab
            real ( kind=8 ), dimension ( 3 ), intent ( in ) :: a, b
          end subroutine cross
    end interface

    real ( kind=8 ), dimension ( 3 ) :: tmp1, tmp2, tt2_copy
    real ( kind=8 ), dimension(3) :: p1, p2, rto, force
    real ( kind=8 ) :: ff_tmp, dist_atom
    real(kind=8) :: size_epgrid_p1, size_epgrid_p2
    integer :: bin, nq1, nq2, j

    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2

    ff1 = (/0.d0, 0.d0, 0.d0/)
    tt1 = (/0.d0, 0.d0, 0.d0/)
    ff2 = (/0.d0, 0.d0, 0.d0/)
    tt2 = (/0.d0, 0.d0, 0.d0/)
    tmp1 = (/0.d0, 0.d0, 0.d0/)
    tmp2 = (/0.d0, 0.d0, 0.d0/)
    size_epgrid_p1 = prot1 % p_sogrid % pset_grid ( 1 ) % rp
    size_epgrid_p2 = prot2 % p_sogrid % pset_grid ( 1 ) % rp

    p_charge1 =>  prot1 % p_sogrid % charge_grid
    p_charge2 =>  prot2 % p_sogrid % charge_grid
    p_charge_position1 => prot1 % p_pos_charge
    p_charge_position2 => prot2 % p_pos_charge
    nq1 = prot1 % nb_charge
    nq2 = prot2 % nb_charge

    !write(*,*) '==Entry partial dh=='
    !write(*,*) 'dist ', dist
    !write(*,*) 'pos_relat ', pos_relat
    !write(*,*) 'total charge prot1 ', prot1 % p_sogrid % charge_total
    !write(*,*) 'total charge prot2 ', prot2 % p_sogrid % charge_total

    ! compute force1, partial charge of prot1  in prot2 analytical grid
    ! will discard case fully inside, already treated
    if ( dist > ( size_epgrid_p2 + prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'dist > size_epgrid_p2 + prot1 % p_sogrid % distance_max =', &
        !    size_epgrid_p2 + prot1 % p_sogrid % distance_max, ' pass, compute f2 '
        ! need torque correction maybe
        goto 100
    end if

    ! only if paritally in prot1.ep.grid
    if ( dist > ( size_epgrid_p2 - prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'Partially inside the grid/analytic of ep2 ', size_epgrid_p1 - prot2 % p_sogrid % distance_max

    FORCE1: do j=1,nq1

               p1 = p_charge_position1(:,j)
               ! relative position charge protein1 to center of protein 2
               rto = - pos_relat + p1 !position charge protein1

               dist_atom = sqrt ( dot_product( rto,rto ) )
               ! only numerical error
               !write(*,*) 'dist atom with rtn', dist_atom
               !write(*,*) 'dist atom with rto', dist_atom != sqrt ( dot_product( rto,rto ) )

               ! from the distance get the correct bin
               bin = int ( dist_atom / prot2 % p_sogrid % pset_analytic_grid( 1 ) % h )

               !ff_tmp = prot2 % p_sogrid % pset_analytic_grid( 1 ) % parray_force ( bin, type_p1 )
               ff_tmp = get_force_analytic ( prot2 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p1 )
               !> Convert to kT/A
               ff_tmp = ff_tmp * 1.d-10

               ! certainly can normalize in original referential, will be always pointing to the center of the sphere
               !write(*,*) 'norm rto ', sqrt( dot_product( rto, rto ) )
               !write(*,*) 'norm rtn ', sqrt( dot_product( rtn, rtn ) )

               !force =  ff_tmp * ( pos_relat / dist_atom ) ! * p_charge1 ( j )
               force = prot2 % p_sogrid % charge_total * p_charge1 ( j ) * ff_tmp * ( rto / dist_atom )

               call cross(tmp1, p1, force)
               ff1 = ff1 + force
               tt1 = tt1 + tmp1

            end do FORCE1

    !else
    !    write(*,*) 'Fully inside epgrid2, already accounted'
    end if

100 continue

    ! compute force2 , partial charge prot2 in analytic grid of prot1
    ! discard case fully outside, already treated
    if ( dist > ( size_epgrid_p1 + prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'dist > size_epgrid_p1 + prot2 % p_sogrid % distance_max =', &
        !    size_epgrid_p1 + prot2 % p_sogrid % distance_max, ' GOTO 200'
        GOTO 200
    end if

    ! only if paritally in prot1 el.grid
    if ( dist > ( size_epgrid_p1 - prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'Compute force2'
        !write(*,*) 'partial analytic/epgrid ', size_epgrid_p1 - prot2 % p_sogrid % distance_max

        ! same as normal force sdamm, 0 inside the epgrid
        FORCE2: do j=1,nq2

                p2 = p_charge_position2 (:,j)
                rto = p2 + pos_relat
                !write(*,*) 'rto ', rto, ' norm=', sqrt( dot_product(rto,rto) )

                dist_atom = sqrt( dot_product(rto, rto) )
                !write(*,*) 'dist  atom', dist_atom

                ! call force_atom_uhbd ( x, force, prot1 % p_sogrid % pset_grid ( grid_number) )
                bin = int ( dist_atom / prot1 % p_sogrid % pset_analytic_grid( 1 ) % h )
                !write(*,*) 'bin ', bin

                ! ff_tmp = prot1 % p_sogrid % pset_analytic_grid( 1 ) % parray_force ( bin, type_p2 )
                ff_tmp = get_force_analytic ( prot1 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p2 )
                ! Convert to kT/A
                ff_tmp = ff_tmp * 1.d-10

                ! assign to the unit vector between charge and center of prot 1
                force =  prot1 % p_sogrid % charge_total* p_charge2( j ) * ff_tmp * ( rto / dist_atom )

                call cross (tmp2, p2, force )
                tt2 = tt2 + tmp2
                ff2 = ff2 + force

          end do FORCE2

    !else
    !    write(*,*) 'Fully inside epgrid1, already accounted'
    end if


200 continue

    ! torque computed on ff1 and ff2, will be asymetric
    ! torque on 2   (-r)x(-ff1), CALL cross(tmp1,r,ff1)
    call cross ( tmp1, pos_relat, ff1 )
    ! torque on 1    (r)x(-ff2), CALL cross(tmp2,r,ff2)
    call cross ( tmp2, pos_relat, ff2 )
    !write(*,*) 'tt1 tmp1 ',tt1, tmp1
    !write(*,*) 'tt2 tmp2 ',tt2, tmp2

    tt2_copy = tt2

    ! should not correct if asymetric, why not ??
    ! was certainly an error, done 2 times ! symetrize here and in calling function
    !ff2 = ff2 - ff1
    tt2 = tt2 - tt1 + tmp1
    tt1 = tt1 - tt2_copy - tmp2
    ! should be something like
    ! tt2 = tt2 + tmp1
    ! tt1 = tt1 - tmp1 ??

    !write(*,*) '== End partial dh =='
    !write(*,*) 'ff2 ', ff2
    !write(*,*) 'ff1 ', ff1
    !write(*,*) 'tt2 ', tt2
    !write(*,*) 'tt1 ', tt1

end subroutine pair_force_partial_analytic_sdamm

!!!!!!!!!!!!!!!!!!!!!  Computation of energy for sdamm analytic !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!> Compute energy for sdamm calculation.\n
!!
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param geom : instance of \ref mod_geometry::geometry "geometry"
!! @param nb_prot : total number of solutes
!! @param param_force_energy : instance of \ref mod_force_energy::type_force_energy "type_force_energy"
!! @param array_energy : return updated array_energy
subroutine energy_analytic_dh_sdamm ( tab_prot, geom, nb_prot, param_force_energy, &
                                      array_energy )
    USE mod_array_protein
    USE mod_geometry
    USE mod_force_energy
    IMPLICIT NONE

    interface

        subroutine compute_dist_pbc ( geom, prot1, prot2, dist, r_corrected )
            USE mod_geometry
            USE mod_protein
            IMPLICIT NONE
            type ( geometry ) :: geom
            type ( protein ), intent ( in ) :: prot1, prot2
            real ( kind=4 ), intent ( out ) :: dist
            real ( kind=8 ), dimension ( 3 ), intent ( out ) :: r_corrected
        end subroutine compute_dist_pbc

    end interface

    type ( array_protein_type ), target, intent(in) :: tab_prot
    type ( geometry ), intent ( in ) :: geom
    type ( type_force_energy ), intent ( in ) :: param_force_energy
    real ( kind=8 ), dimension ( :,: ) :: array_energy
    integer, intent ( in ) :: nb_prot

    integer :: n1, n2, type_p1, type_p2
    type ( protein ), pointer :: pprot1, pprot2
    real ( kind=8 ), dimension ( 3 ) :: pos_relat
    ! energy full_analytic and partial analytic
    real ( kind=8 ) :: energy1_analytic, energy2_analytic, energy1, energy2
    real ( kind=4 ) :: dist

    !write(*,*)
    !write(*,*) '=================================='
    !write(*,*) 'Entry energy analytic dh sdamm'

!$OMP PARALLEL DEFAULT ( NONE ) SHARED ( tab_prot, nb_prot, param_force_energy, geom ) &
!$OMP PRIVATE ( pprot1, pprot2, pos_relat, dist, type_p1, type_p2 ) &
!$OMP PRIVATE ( energy1_analytic, energy1, energy2_analytic, energy2 ) &
!$OMP REDUCTION(+:array_energy) ! could try manual reduction like in force..not necessary for energy in sdamm

!$OMP DO SCHEDULE( DYNAMIC )
PROT1: do n1 = 1, nb_prot
PROT2:    do  n2 = n1 + 1, nb_prot

            pprot1 => tab_prot % array ( n1 )
            pprot2 => tab_prot % array ( n2 )

            call compute_dist_pbc ( geom, pprot1, pprot2, dist, pos_relat )

            ! test surface ? not applicable with analytic dh ! not yet
            ! Julia's case useful for full interaction,
            ! just need to setup a correct analytical expression
            type_p1 = param_force_energy % array_prot_type ( n1 )
            type_p2 = param_force_energy % array_prot_type ( n2 )

            call pair_energy_full_analytic_sdamm( pprot1, pprot2, type_p1, type_p2, dist, &
                    energy1_analytic, energy2_analytic )
            !write(*,*) 'After full analytic'
            !write(*,*) 'Energy e1_analytic DH ', energy1_analytic
            !write(*,*) 'Energy e2_analytic DH ', energy2_analytic

            !write(*,*) 'Compute partial analytic '
            call pair_energy_partial_analytic_sdamm( pprot1, pprot2, pos_relat, type_p1, type_p2, dist, &
                    energy1, energy2 )
            !write(*,*) 'After partial analytic'
            !write(*,*) 'Energy 1 DH ', energy1
            !write(*,*) 'Energy 2 DH ', energy2

            ! nb_types not needed, grid_number == 1 for electrostatic
            array_energy ( 1 + type_p2, n1 )  = &
             array_energy ( 1 + type_p2, n1 ) + energy1_analytic + energy1
            ! total shift, number of full groups type_prot1 subtype of prot1
            array_energy ( 1 + type_p1, n2 )  = &
             array_energy ( 1 + type_p1, n2 ) + energy2_analytic + energy2

    end do PROT2
end do PROT1
!$OMP END DO
!$OMP END PARALLEL

    !write(*,*) 'End energy_analytic_dh_sdamm'
    !write(*,*) 'f ', f
    !write(*,*) 't ', t
    !write(*,*) 't1 ', t1
    !write(*,*) '######################'

end subroutine energy_analytic_dh_sdamm

!> Compute Full DH interactions.\n
!! @todo To see if can be common to sda and sdamm\n
!!
!! Compute p2 solute in DH "grid" of p1
!! @param prot1, prot2 : olute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param type_p1, type_p2 : integer to get the type of solute
!! @param dist : distance between the solutes
!! @param energy1, energy2 : return the energy of solute 1 and 2
subroutine pair_energy_full_analytic_sdamm( prot1, prot2, type_p1, type_p2, dist, energy1, energy2 )
    USE mod_array_protein
    IMPLICIT NONE
    type ( protein ), intent(in) :: prot1, prot2
    real( kind=4 ), intent ( in ) :: dist
    real ( kind=8 ), intent(out) :: energy1, energy2
    integer :: type_p1, type_p2

    ! some need to be extracted only once, dpends 2 prot or sdamm
    real ( kind=8 ) :: en_tmp,  size_epgrid_p1, size_epgrid_p2
    integer :: bin

    energy1 = 0.d0
    energy2 = 0.d0

    !write(*,*) '==============================='
    !write(*,*) 'Entry pair energy full analytic 2 proteins'
    !write(*,*) 'type p1 ', type_p1
    !write(*,*) 'type p2 ', type_p2
    !write(*,*) 'tot_charge * tot_charge ', prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total
    !write(*,*) 'dist (after pbc) ', dist

    size_epgrid_p1 = prot1 % p_sogrid % pset_grid ( 1 ) % rp
    size_epgrid_p2 = prot2 % p_sogrid % pset_grid ( 1 ) % rp

    !write(0,*) 'associated prot1/prot2', associated (prot1), associated ( prot2 )
!    write(0,*) 'associated prot1/prot2', associated ( prot1%p_sogrid ), associated ( prot2%p_sogrid )
!    write(0,*) 'associated prot1/prot2', associated ( prot1%p_sogrid%pset_analytic_grid ), &
!               associated ( prot2%p_sogrid%pset_analytic_grid )

    ! compute energy2 , total charge prot2 in analytic grid of prot1
    ! only if full outside of prot1 el.grid
    if ( dist > ( size_epgrid_p1 + prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'outside size_epgrid_p1 ', size_epgrid_p1, ' +dist_max p2 ', prot2 % p_sogrid % distance_max
        !write(*,*) 'compute energy2'

        bin = int ( dist / prot1 % p_sogrid % pset_analytic_grid( 1 ) % h )
        !write(*,*) 'bin ', bin
        en_tmp = get_energy_analytic ( prot1 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p2 )
        ! Directly in kT
        ! adjust only with total charges
        energy2 = prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total * en_tmp
    !else
    !    write(*,*) 'Partialy or full inside ep_grid1, see later '
    end if

    ! compute energy1, no need of change of refererential here
    ! bin certainly identical for both, because dist is identical
    ! only if full outside if prot2 el.grid
    if ( dist > ( size_epgrid_p2 + prot1 % p_sogrid % distance_max ) ) then

        !write(*,*) 'outside size_epgrid_p2 ', size_epgrid_p2, ' +dist_max p1 ', prot1 % p_sogrid % distance_max
        !write(*,*) 'compute energy 1, test should be similar if full DH'

        bin = int ( dist / prot2 % p_sogrid % pset_analytic_grid( 1 ) % h )
        ! norm of the force for this interaction
        en_tmp = get_energy_analytic ( prot2 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p1 )
        !> Directly in kT

        ! adjust to the correct charge
        energy1 =  prot1 % p_sogrid % charge_total * prot2 % p_sogrid % charge_total * en_tmp
        !write(*,*) 'energy1 ', energy1
    !else
    !    write(*,*) 'Partialy or full inside ep_grid2, see later '
    end if

end subroutine pair_energy_full_analytic_sdamm

!> Only partial analytic, loop over all atoms in this case.\n
!! Analytic has been set to zero if inside the grid
!! @param prot1, prot2 : olute 1 and 2, instance of \ref mod_protein::protein "protein"
!! @param pos_relat : relative position
!! @param type_p1, type_p2 : integer to get the type of solute
!! @param dist : distance between the solutes
!! @param energy1, energy2 : return the energy of solute 1 and 2
subroutine pair_energy_partial_analytic_sdamm( prot1, prot2, pos_relat, type_p1, type_p2, dist, energy1, energy2 )
    USE mod_array_protein
    IMPLICIT NONE

    type ( protein ), intent(in) :: prot1, prot2
    real ( kind=8 ), intent(out) :: energy1, energy2
    real ( kind=8 ), dimension(3), intent(in) :: pos_relat
    real ( kind=4 ) :: dist
    integer :: type_p1, type_p2

    real ( kind=8 ), dimension(3) :: p1, p2, rto
    real ( kind=8 ) :: en_tmp, dist_atom
    real(kind=8) :: size_epgrid_p1, size_epgrid_p2
    !! some need to be extracted only once, dpends 2 prot or sdamm
    integer :: bin, nq1, nq2, j

    real ( kind=8 ), dimension(:), pointer :: p_charge1, p_charge2
    real ( kind=8 ), dimension(:,:), pointer :: p_charge_position1, p_charge_position2

    energy1 = 0.d0
    energy2 = 0.d0

    size_epgrid_p1 = prot1 % p_sogrid % pset_grid ( 1 ) % rp
    size_epgrid_p2 = prot2 % p_sogrid % pset_grid ( 1 ) % rp

    p_charge1 =>  prot1 % p_sogrid % charge_grid
    p_charge2 =>  prot2 % p_sogrid % charge_grid
    p_charge_position1 => prot1 % p_pos_charge
    p_charge_position2 => prot2 % p_pos_charge
    nq1 = prot1 % nb_charge
    nq2 = prot2 % nb_charge

    !write(*,*) '== Entry energy partial dh sdamm =='
    !write(*,*) 'dist ', dist
    !write(*,*) 'pos_relat ', pos_relat
    !write(*,*) 'total charge prot1 ', prot1 % p_sogrid % charge_total
    !write(*,*) 'total charge prot2 ', prot2 % p_sogrid % charge_total

    ! compute force1, partial charge of prot1  in prot2 analytical grid
    ! discard case fully inside, already treated
    if ( dist > ( size_epgrid_p2 + prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'dist > size_epgrid_p2 + prot1 % p_sogrid % distance_max =', &
        !    size_epgrid_p2 + prot1 % p_sogrid % distance_max, ' goto 100 compute energy2 '
        goto 100
    end if

    ! only if partially in prot1 el.grid
    if ( dist > ( size_epgrid_p2 - prot1 % p_sogrid % distance_max ) ) then
        !write(*,*) 'Partially inside the grid/analytic of ep2 ', size_epgrid_p2 - prot1 % p_sogrid % distance_max
        !write(*,*) 'compute energy1 '

    ENERG1: do j = 1, nq1

               p1 = p_charge_position1(:,j)
               ! relative position charge protein1 to center of protein 2
               rto = - pos_relat + p1 !position charge protein1
               !write(*,*) 'rto ', rto, ' norm=', sqrt( dot_product(rto,rto) )
               dist_atom = sqrt ( dot_product( rto,rto ) )

               ! from the distance get the correct bin
               bin = int ( dist_atom / prot2 % p_sogrid % pset_analytic_grid( 1 ) % h )

               en_tmp = get_energy_analytic ( prot2 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p1 )
               !> Directly in kT
               energy1 = energy1 + p_charge1 ( j ) * prot2 % p_sogrid % charge_total * en_tmp
        end do ENERG1

    !else
    !    write(*,*) 'Fully inside epgrid2, already accounted'
    end if

100 continue

    ! compute force2 , partial charge prot2 in analytic grid of prot1
    ! discard case fully outside, already treated
    if ( dist > ( size_epgrid_p1 + prot2 % p_sogrid % distance_max ) ) then
        !write(*,*) 'dist > size_epgrid_p1 + prot2 % p_sogrid % distance_max =', &
        !    size_epgrid_p1 + prot2 % p_sogrid % distance_max, ' return'
        return
    end if

    ! only if paritally in prot1 el.grid
    if ( dist > ( size_epgrid_p1 - prot2 % p_sogrid % distance_max ) ) then
        !write(*,*)
        !write(*,*) 'Compute energy2'
        !write(*,*) 'partial analytic/epgrid ', size_epgrid_p1 - prot2 % p_sogrid % distance_max

        ! same as normal energy sdamm, 0 inside the epgrid
        ENERG2: do j = 1, nq2

                p2 = p_charge_position2 (:,j)
                rto = p2 + pos_relat

                dist_atom = sqrt( dot_product(rto, rto) )
                bin = int ( dist_atom / prot1 % p_sogrid % pset_analytic_grid( 1 ) % h )
                en_tmp = get_energy_analytic ( prot1 % p_sogrid % pset_analytic_grid( 1 ), bin, type_p2 )
                !> Directly to kT
                ! assign to the unit vector between charge and center of prot 1
                energy2 = energy2 + p_charge2 ( j ) * prot1 % p_sogrid % charge_total * en_tmp

          end do ENERG2
    !else
    !    write(*,*) 'Fully inside epgrid1, already accounted'
    end if

end subroutine pair_energy_partial_analytic_sdamm


END MODULE mod_compute_analytic_sdamm
