!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file mod_exclusion_grid.f90
!!
!! Define the exclusion grid

!> Define a specific type for exclusion grid
MODULE mod_exclusion_grid

    USE mod_grid

    !> Specific type for exclusion grid
    !!
    !! Define a grid of type uhbd or dtgrid with logical values\n
    !! Use \ref mod_grid::type_grid "type_grid" as base class
    type type_grid_excl
        !> pointer to a type_grid
        type ( type_grid ), pointer :: pgrid

    end type type_grid_excl

    contains

        !> Allocate memory for exclusion grid
        !!
        !! Called from make_exclusion grid
        !! @param this : instance of type_grid_excl
        !! @param xmax, ymax, zmax : size of the grid ( cell number )
        subroutine allocate_excl_grid ( this, xmax, ymax, zmax )
            IMPLICIT NONE
            type ( type_grid_excl ) :: this
            integer, intent( in ) :: xmax, ymax, zmax

            allocate ( this % pgrid )
            ! exclusion grid, always boolean, be_logical set to true
            call allocate_grid ( this % pgrid, xmax, ymax, zmax, .true. )

        end subroutine allocate_excl_grid

        !> Delete the exclusion grid
        !!
        !! Deallocate only \ref type_grid_excl::pgrid "pgrid", the pointer to the grid.\n
        !! The grid itself must be deleted after
        !!
        !! @param this : instance of type_grid_excl
        subroutine delete_excl_grid ( this )
            IMPLICIT NONE
            type ( type_grid_excl ) :: this

            integer :: ierr

            if ( associated ( this % pgrid ) ) then
                !write(*,*) 'pgrid exclusion grid allocated, delete the grid before'
                call delete_grid ( this % pgrid )
                deallocate ( this % pgrid, stat=ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR In deallocation memory of exclusion grid '
                    stop 1
                endif
            endif

        end subroutine delete_excl_grid

        subroutine print_exclusion_origin ( this )
            IMPLICIT NONE
            type (  type_grid_excl ) :: this

            write(*,fmt='(''grid origin :'',3f10.3 )') this % pgrid % origin
            write (*, fmt='(''shape      :'',3i8)' ) this % pgrid % size_x, this % pgrid % size_y, this % pgrid % size_z
            write (*, fmt='(''spacing h  :'',f8.3)') this % pgrid % h
        end subroutine print_exclusion_origin

        !> Print info about the grid to a log file
        !!
        !! Print a section of the grid to iopoti ( fort.77 )
        !! Maybe more if flex, it is a log can just append, maybe test before and delete if existing
        !! @param this : instance of type_grid_excl
        !! @param ioutput : file descriptor of the log file. By default append the data. If 0 do not write informations
        subroutine print_excl_grid_log ( this, ioutput )
            IMPLICIT NONE
            type (  type_grid_excl ) :: this
            integer :: ioutput

            integer :: nzcnt, i, j, k, io_err
            character*16 :: string_format

            !! trick for false parrallel
            if ( ioutput == 0 ) then
                write(*,*) 'ioutput==0 do not write the log file'
                return
            end if

            open (ioutput, file='exclusion_grids.log', form= 'formatted', position='append', iostat = io_err )
            if ( io_err /= 0 ) then
                write (*,*) 'problem in opening the file'
                stop 1
            end if

            !!do not correspond to center anymore
            !nzcnt = this % size_z / 2
            !! 53 same than sda6_ref
            !nzcnt = 53
            !! print plan at value -0.25 of z
            nzcnt = nint ( ( - this % pgrid % origin ( 3 ) - 0.25 ) / this % pgrid % h )
            !write(*,*) 'without approx ', ( - this % origin ( 3 ) + (-0.25) ) / this % hexcl

            !! b1.0 == print only "1", other print 1 and 0
            write ( string_format, '("(",i4,"b1.0)")' ) this % pgrid % size_x
            !write(*,*) 'format print excl ', string_format

            write ( ioutput, fmt='(a, i4)' ) 'A cross section of excl. grid of pr. #1, xOy plane with k=', nzcnt
! already commented in original version
!CCC      DO K=nzcnt-nzst,nzcnt+nzst,nzst,
            k = nzcnt
            do j = 1, this % pgrid % size_y
                write( ioutput, fmt=string_format ) ( this % pgrid % p_grid % pgrid_logical (i,j,k),i=1,this % pgrid % size_x )
            end do

            close ( ioutput )

        end subroutine print_excl_grid_log

END MODULE mod_exclusion_grid
