!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_flexibility.f90
!! Allow the change in conformation of the solutes
!!
!! Subroutines in \ref change_conformation.f90 implement the different algorithms.


!> Module for flexibility
!!
!! Manage a list of possible conformation ( list of set_of_grid ).
!! define type for flexibility : NodeFlex, list_flex_type and flexible
MODULE mod_flexibility

    USE mod_setofgrid

    !> enumeration of node's state for the list
    integer, parameter :: flex_intern = 0, flex_first = 1, flex_last = 2
    !> enumeration for the algorithm
    integer, parameter :: flex_random= 0, flex_metropolis = 1, flex_minimum = 2, flex_time = 3

    !> Define nodes of list_flex, internal use
    !!
    !! Each p_data points to a \ref mod_setofgrid::sogrid "set_of_grid"
    type NodeFlex
        type ( sogrid ), pointer :: p_data
        integer :: state
        type ( NodeFlex ), pointer :: p_next
        type ( NodeFlex ), pointer :: p_previous
    end type NodeFlex

    !> List of set of grid
    !!
    !! Each group of proteins will point indirectly, by a flexible, to one unique list
    type list_flex_type

        !> in case of dynamical allocation keep copy of list_name_conf and factor grid, may change
        character ( 128 ) :: list_conf_fname

        !> memorize data to load / reload new set of grid
        !! not used
        !! @{
        !!real ( kind=4 ) :: epfct, edfct, hdfct
        !> @}

        !> need to know the maximum size to not reallocate every time
        !! should be allocated when reading input,
        !! \todo need to add max_atom_lj or already included ? to check
        !! @{
        integer :: max_size_movable_charge, max_size_movable_atom
        !> @}

        !! not necessary now for criteria, only one pointer max_size_rcriteria
        !! need for rhit, distance_max
        !> store the maximum over all sogrid
        real ( kind=8 ) :: distance_max

        !> related to the format of list_conformation.dat
        !! should belong to the module, because of reload
        logical :: implicit

        !> pointer on first and last node of the list
        !! @{
        type ( NodeFlex ), pointer :: p_first
        type ( NodeFlex ), pointer :: p_last
        !! @}

        !> keep all data for update, minimal in term of memory
        !! all same type of proteins have the same frequency
        !! because sdamm=> asynchroneous update, add own time_frequence to flexible ( each protein )
        !!         and use std_freq_random
        !! @{
        real ( kind=4 ) :: frequence, std_freq_random
        !> @}
        !> to deal with the different methods of update, add option for nearest_neighbour
        !! @{
        integer :: method_flex
        logical :: nearest_ngb
        !! @}
        !> for setting total number of conformation and the initial nb at the start of the trajectory
        integer :: total_conf, initial_conf


    end type list_flex_type


    !> Main class owned by each protein to allow a change in conformation
    !!
    !! Every flexible protein has a pointer to flexible.\n
    !! it contains the list list_flex_type\n
    !! and additional convenient variable : current_node, current_conf
    !! Manage statistics for each protein : visited_conf, could  extend statistics
    type flexible

        !> pointer to list
        type ( list_flex_type ), pointer :: plist_flex

        !> save value of the current conformation for each protein.
        !! very convenient for sdamm and reading restart file
        integer :: current_conf
        ! when new points to the first node p_data == null  no data
        !> current node
        type ( NodeFlex ), pointer :: p_current_node

        !> for statistics, sum up the number of visited conformations
        !! increments the time spend in each conformation
        real ( kind=4 ), dimension (:), allocatable :: visited_conf

        !> Time resting for the next change to an other conformation\n
        !! sdamm needs asynchrone changes
        !! every protein needs its own time ( kind of gillepsie algorithm )\n
        !! set to time_next_flex (+random_freq) and decrease until < 0
        real ( kind=4 ) :: time_next_flex

    end type flexible

    contains

        !> Constructor flexible\n
        !!
        !! Called by \ref mod_protein::set_flexible \n
        !! The list of conformation should be independently created with \ref new_list_flex \n
        !! Allocate the array for the statistics\n
        !! @param this : instance of a \ref mod_flexibility::flexible "flexible"
        !! @param plist_flex : the list of sogrid to associate to flexible
        subroutine new_flexible ( this , plist_flex )
            IMPLICIT NONE
            type ( flexible ) :: this
            type ( list_flex_type ), target :: plist_flex

            integer :: ierr

            this % current_conf = -1

            !! associate to the list
            this % plist_flex => plist_flex
            !! associate current_node to the head...
            this % p_current_node => plist_flex % p_first

            !! default value for time_flex, will be overwritten by init_time_flex
            this % time_next_flex = 0.
            !this % nearest_ngb = .false.

            !! allocate array for statistics
            !! write(*,*) 'allocate visited conf size total_conf ', plist_flex % total_conf
            allocate ( this % visited_conf ( plist_flex % total_conf ), stat = ierr )
            if ( ierr /= 0 ) then
                write ( *, * ) 'Cannot allocate visited_conf '
            end if
            !! set all to zero explicitely
            this % visited_conf = 0.

        end subroutine new_flexible

        !> Destructor flexible
        !!
        !! Deallocate array for statistics, nullify the pointer ( no deletion of sogrid )\n
        !! @param this : instance of a \ref mod_flexibility::flexible "flexible"
        subroutine delete_flexible ( this )
            IMPLICIT NONE
            type ( flexible ) :: this

            integer :: ierr

            !! delete all sogrid associated to the nodes and the nodes
            !! write(*,*) 'will deallocate visited_conf'
            deallocate ( this % visited_conf, stat = ierr )
            if ( ierr /= 0 ) then
                write ( *,* ) 'ERROR Cannot deallocate visited_conf '
                STOP 1
            end if
            nullify ( this % plist_flex, this % p_current_node )

        end subroutine delete_flexible

        !> Initialize or reinitialze the time for the next change in conformation : \ref flexible::time_next_flex "time_next_flex"
        !!
        !! if \ref list_flex_type::std_freq_random "std_freq_random" > 0, add a random variable
        !! else use exactly \ref list_flex_type::frequence "frequence"
        !! @param this : instance of a \ref mod_flexibility::flexible "flexible"
        subroutine  init_flexible_freq ( this )
            IMPLICIT NONE
            type ( flexible ) :: this

            real ( kind=4 ) :: gaus
            gaus = 0.

            !! if std_freq_random > 0. use an additional random variable
            if ( this % plist_flex % std_freq_random > 0. ) then
                call ggnml4_new ( 1, gaus )
                ! write(*,*) 'new gaus ', gaus
                ! write(*,*) 'transform gaus', gaus * this % plist_flex % std_freq_random
                this % time_next_flex = this % plist_flex % frequence + &
                                           gaus * this % plist_flex % std_freq_random
                !write(*,*) 'new time_next_flex ',this % time_next_flex

            !! else just copy frequence from plist_freq
            else
                this % time_next_flex = this % plist_flex % frequence
            end if

        end subroutine init_flexible_freq

! always warning
!        function get_method_flex_str(this) result(str_method)
!            IMPLICIT NONE
!            type ( list_flex_type ) :: this
!
!            !character :: str_method*(*)
!            character, dimension(15) :: str_method
!
!            !str_method = trim('meTROpolis')
!            str_method = 'meTROpolis'
!
!        end function get_method_flex_str

!!!!!!!!!!!!!!!!!  List_Flex functions   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        ! number of conformation to create at the beginning, may evolve
        ! may add a max number of conf
        ! filename with the list of the conformation, the name of the file

        !> Constructor of list_flexible
        !! @param this : instance of a \ref mod_flexibility::list_flex_type "list_flex_type"
        !! @param flex_method_str : "metropolis", "random", "time" or "minimum"
        !! @param list_conf_fname : name of the input file with all conformations
        !! @param total_conf : total number of conformations
        !! @param initial_conf : specifies if a particular conformation is choosen at the start of a new trajectory
        !! @param nearest_ngb : logical, nearest neigbour only
        !! @param frequence : time for changing conformation
        !! @param std_freq : variance in the frequence
        subroutine new_list_flex ( this, flex_method_str, list_conf_fname, &
                                   total_conf, initial_conf, nearest_ngb, &
                                   frequence, std_freq )

            IMPLICIT NONE
            type ( list_flex_type ) :: this
            character ( 128 ) :: list_conf_fname, flex_method_str
            real ( kind=4 ) :: frequence, std_freq
            integer :: initial_conf, total_conf !max_conf, init_nb_conf, total_conf
            logical :: nearest_ngb

            !! initialze the list
            !this % method_flex = flex_metropolis
            this % list_conf_fname = list_conf_fname
            this % total_conf = total_conf
            this % initial_conf = initial_conf
            this % nearest_ngb = nearest_ngb

            ! not used
            !this % init_nb_conf = init_nb_conf
            !this % max_conf = max_conf

            ! setup frequence of change conf, mean and standart deviation (default 0 ) if needed
            this % frequence = frequence
            this % std_freq_random = std_freq

            ! for the moment setup when reading grid
            this % max_size_movable_charge = 0
            this % max_size_movable_atom = 0
            this % distance_max = 0.

            ! default, to increment when loading grids
            ! not used
            !this % first_conf = 0
            !this % last_conf  = 0

            ! write(*,*) 'flex_method: ', flex_method_str
            if ( flex_method_str == 'metropolis' ) then
                this % method_flex = flex_metropolis
            else if ( flex_method_str == 'random' ) then
                this % method_flex = flex_random
            else if ( flex_method_str == 'time' ) then
                this % method_flex = flex_time
            else if ( flex_method_str == 'minimum' ) then
                this % method_flex = flex_minimum
            else
                write ( *,* ) 'method flex is not known ', flex_method_str
                stop 1
            end if

            ! create first and last node by default
            ! contain no data, sentinelle
            allocate ( this % p_first )
            this % p_first % state = flex_first
            !! set p_data and p_previous to null
            nullify ( this % p_first % p_data, this % p_first % p_previous )

            allocate ( this % p_last )
            this % p_last % state = flex_last
            !! set p_data and p_next to null
            nullify ( this % p_last % p_data, this % p_last % p_next )

            !! initialize relationships, first points to last
            this % p_first % p_next => this % p_last
            this % p_last  % p_previous => this % p_first

        end subroutine new_list_flex

        !> Delete the full list and the associated sogrid
        !!
        !! @param this : instance of a \ref mod_flexibility::list_flex_type "list_flex_type"
        subroutine delete_listflex ( this )
            IMPLICIT NONE
            type ( list_flex_type ) :: this

            integer :: status
            logical :: head_reached

            head_reached = .false.

            !! if already empty
            if ( associated ( this % p_last % p_previous, this  % p_first ) ) &
                 head_reached = .true.

            do while ( head_reached .eqv. .false. )
                call delete_last_dataflex ( this, status )

                if ( associated ( this % p_last % p_previous, this % p_first ) ) &
                     head_reached = .true.
            end do

            deallocate ( this % p_first, this % p_last )
            nullify ( this % p_first, this % p_last )

        end subroutine delete_listflex

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        !> Deleta the last node of the list and the associated sogrid
        !! @param this : instance of \ref list_flex_type
        !! @param status : for error report
        subroutine delete_last_dataflex ( this, status )
            IMPLICIT NONE
            type ( list_flex_type ) :: this
            integer, intent ( out ) :: status

            type ( NodeFlex ), pointer :: tmp_node
            integer :: ierr
            !! error by default
            status = 1

            tmp_node => this % p_last % p_previous % p_previous
            !! should call an explicit delete for sogrid
            call delete_sogrid ( this % p_last % p_previous % p_data )
            !deallocate ( this % p_last % p_previous % p_data
            deallocate ( this % p_last % p_previous % p_data, stat = ierr )
            if ( ierr /= 0 ) then
                write( *, *) 'Cannot deallocate p_data in delete_last_dataflex'
            end if
            !! not necessary to nullify explicitely pointer
            nullify ( this % p_last % p_previous % p_data, this % p_last % p_previous % p_next, &
                      this % p_last % p_previous % p_previous )
            !! deallocate the NodeFlex
            deallocate ( this % p_last % p_previous, stat = ierr )
            if ( ierr /= 0 ) then
                write( *, *) 'Cannot deallocate p_previous in delete_last_dataflex'
            end if
            status = 0

            ! assign new pointer
            tmp_node % p_next => this % p_last
            this % p_last % p_previous => tmp_node

        end subroutine delete_last_dataflex

        !> not used until now, to delete ??
        subroutine delete_first_dataflex ( this, status )
            IMPLICIT NONE
            type ( list_flex_type ) :: this
            integer, intent ( out ) :: status

            type ( NodeFlex ), pointer :: tmp_node

            !! error by default
            status = 1

            tmp_node => this % p_first % p_next % p_next
            !! should call an explicit delete for sogrid
            deallocate ( this % p_first % p_next )
            status = 0
            !! assign new pointer
            tmp_node % p_previous => this % p_first
            this % p_first % p_next => tmp_node

        end subroutine delete_first_dataflex


!!!!!!!!!!!!!!!! insert setofgrid elements

        !> Insert a new sogrid and create a new node
        !!
        !! Could increment the size here
        !! @param this : instance of \ref list_flex_type
        !! @param p_sogrid : pointer to the \ref mod_setofgrid::sogrid "sogrid" to insert
        !! @param head : if true insert first position, if false last position
        !! @param status : for error report, not used
        subroutine insert_sogrid ( this, p_sogrid, head, status )
            IMPLICIT NONE
            type ( list_flex_type ) :: this
            type ( sogrid ), pointer :: p_sogrid
            integer :: status
            logical, intent ( in ) :: head

            type ( NodeFlex ), pointer :: pnew_node, tmp_node
            integer :: ierr

            !write(*,*) 'enter insert_sogrid'
            ! allocate the new node
            allocate ( pnew_node, stat = ierr )
            if ( ierr /= 0 ) then
                write ( *, * ) 'Cannot allocate new node in lflex'
                status = 1
                return
            end if

            !! assign pointer to the correct data
            pnew_node % p_data => p_sogrid
            pnew_node % state = flex_intern

            !! insert to first position
            if ( head .eqv. .true. ) then

                tmp_node => this % p_first % p_next

                this % p_first % p_next => pnew_node
                pnew_node % p_previous => this % p_first
                pnew_node % p_next => tmp_node
                tmp_node % p_previous => pnew_node

            !! insert to last position
            else
                tmp_node => this % p_last % p_previous

                this % p_last % p_previous => pnew_node
                pnew_node % p_previous => tmp_node
                pnew_node % p_next => this % p_last
                tmp_node % p_next => pnew_node

            end if
            !if ( associated ( this % p_last % p_previous % p_data ) ) then
            !    write(*,*) 'last sogrid of last complexe associated'
            !else
            !     write(*,*) 'last sogrid of last complexe NOT associated, problem'
            !end if
            status = 0
            return

        end subroutine insert_sogrid

!!!!!!!!!!!! Module for dealing with change of conformation   !!!!!!!!!!!!!!!

        !! new make sense to use named parameters here
        !! delete counter ?? can get the new conf, imposed, can not ask after the change

        !> Associate pointer pnew_node to the conformation number.
        !!
        !! set pnew_node to nb_conf or shift ( previous or next )
        !! called by change_conf_flex only, which deals with random numbers
        !!
        !! if shift_opt is given moved to a neighbour.
        !! if shift not given and nb_conf /= -1, moved to nb_conf using recursive function, needed to loop throught the list.
        !!
        !! @param this : instance of \ref nodeflex "NodeFlex"
        !! @param pnew_node : pointer to the \ref nodeflex "NodeFlex", associated if status == 0
        !! @param nb_conf : specify the conformation number, used if shift_conf == 0
        !! @param counter : return the current conformation in the case nb_conf is specified explictely
        !! @param shift_conf : if 1 or -1, points to the next or previous node. if 0 use nb_conf
        recursive function  associate_to_conformation ( this, pnew_node, nb_conf, counter, shift_conf ) result ( status )
            IMPLICIT NONE
            type ( NodeFlex ), target :: this
            type ( NodeFlex ), pointer :: pnew_node
            integer, intent ( in ) :: nb_conf, shift_conf
            integer, intent (out) :: counter

            integer :: status
            !! error by default
            status = 1

            !write(*,*) 'entry associate_to_conformation'
            !write(*,*) 'shift_conf ', shift_conf
            !write(*,*) 'nb_conf ', nb_conf
            !write(*,*) 'test p_previous ',associated ( this % p_previous )
            !write(*,*) 'is p_first ', associated ( this % p_first, this )
            !write(*,*) '
            !write(*,*) 'test p_previous % p_data ',  associated ( this % p_previous % p_data )

            if ( shift_conf /= 0 ) then

                ! check if the shift is possible
                if ( ( shift_conf == 1 ) .and. ( associated ( this % p_next % p_data ) ) ) then
                    pnew_node => this % p_next
                    status = 0
                    !write (*,*) 'shifted of +1 state this of next of previous', this % p_next % state, this % p_previous % state
                    !write (*,*) 'shifted of +1 state pnew_node of next of previous', pnew_node % p_next % state, &
                    !               pnew_node % p_previous % state
                else if ( ( shift_conf == -1 ) .and. ( associated ( this % p_previous % p_data) ) ) then
                    pnew_node => this % p_previous
                    !write (*,*) 'shifted of -1'
                    status = 0
                !else
                !    write (*,*) 'shift_conf ', shift_conf, ' not accepted '
                    ! here status to 1, means than the change was rejected
                end if

                ! return a maybe associated pnew_node
                return
            end if


            !! here in case shift_conf == 0, go to conf nb, and return counter ( should be equal to conf_nb )
            !! use recursivity to loop through the list
            if ( this % state == flex_first ) then
                !write (*,*) ' I am the first node ', counter
                status = associate_to_conformation ( this % p_next, pnew_node, nb_conf, counter, shift_conf )
                return
            else if ( this % state == flex_last ) then
                !write (*,*) 'Error: I am the last node, just return counter == ', counter
                !! here status is 1
                return
            !! here intern node, check if correct conformation
            else
                ! increment only if internal node
                counter = counter + 1
                !write(*,*) 'counter ', counter

                !! really simple here, should check limit or get an id
                if ( counter == nb_conf ) then
                    !write(*,*) 'get it, associate pnew_node and return '
                    pnew_node => this
                    !! 0 success
                    status = 0
                    return
                else
                    !! send to the next
                    status = associate_to_conformation ( this % p_next, pnew_node, nb_conf, counter, shift_conf )
                    return
                end if
            end if

        end function associate_to_conformation

        !> Update the p_current_node of flex to conformation nb_conf (== -1 random ) or shift_conf ( +/- 1 )
        !!
        !! if shift_conf_opt present ( +1 / -1) limited to +/- 1
        !!                   not present, do not use shift
        !! both cases if conf_nb == -1 random if shift ( can be the same)
        !!                      full random  (can be the same )
        !! if conf_nb /= -1 and shift_conf is not present : specify the conformation number exactly
        !!
        !! @param this : instance of \ref flexible
        !! @param conf_nb : new conformation number
        !! @param conf_nb : conformation number, if == -1, random
        !! @param shift_conf_opt : optionaly if conf_nb different of -1 (=0) apply the given shift
        subroutine change_conf_flex ( this, conf_nb, shift_conf_opt )
            IMPLICIT NONE
            type ( flexible ) :: this
            integer, intent( in ) :: conf_nb
            integer, optional :: shift_conf_opt

            type ( NodeFlex ), pointer :: p_new_node
            integer :: new_conf, status, counter
            real ( kind=8 ) :: unif
            !! +1 or -1, /  +2 or -2 ?
            integer :: shift_conf

            nullify ( p_new_node )
            shift_conf = 0
            !! int status as error
            status = 1
            ! if a conformation number is asked, need to be 0 by default
            counter = 0

            !write(*,*) 'Entry change_conf nb_conf ', conf_nb
            !write(*,*) 'nearest ', this % plist_flex % nearest_ngb
            !write(*,*) 'current_conf ', this % current_conf

            !if ( present ( shift_conf_opt ) ) then
            !    write(*,*) 'shift_conf present ', shift_conf_opt
            !else
            !    write(*,*) 'shift_conf not present '
            !end if

            !! if random will need a random number unif[0,1]
            if ( conf_nb == -1 ) then
                 call ggubs_thread_safe ( unif )
            end if

            !! now more general, nearest_ngb 0/1 apply in all cases
            if ( this % plist_flex % nearest_ngb ) then

                !! if random and shift_opt want a +1 or -1 uniform
                if ( conf_nb == -1 ) then
                    if ( unif > 0.5 ) then
                        shift_conf = 1
                    else
                        shift_conf = -1
                    end if
                !! if /= -1 (0), not random, apply the shift, shift_opt_conf used by minimum,
                else
                    !just for test, to comment, here really need to check !
                    !if ( present ( shift_conf_opt ) .eqv. .false. ) then
                    !if ( present ( shift_conf_opt ) ) then
                    !   write(*,*) 'mistake !!'
                    !    con
                    !    stop 1
                    !end if
                    !else
                         shift_conf = shift_conf_opt
                    ! else should be zero, but shift_conf is associated
                    !end if
                end if

                !write(*,*) 'Want conf_nb ', conf_nb
                !write ( * , * ) 'Want a shift of ', shift_conf
                status = associate_to_conformation ( this % p_current_node, p_new_node, conf_nb, counter, shift_conf )
                !! test if change has been accepted
                if ( status == 0 ) then

                    ! assign the shift change
                    this % current_conf = this % current_conf + shift_conf
                    this % p_current_node => p_new_node
                    !write(*,*) 'satus == 0, shift has been accepted, new current_conf ', this % current_conf

                ! else p_new_node is still null, current_conf is not modified
                end if

            !! if nearest neigbour is false, no check for opt_shift here....
            !! always call a specific conformation, not a shift
            else
                !! if -1 random over all conformations
                if ( conf_nb == -1 ) then
                    !write (*,*) 'total number of conf', this % plist_flex % total_conf
                    new_conf = int ( unif * ( this % plist_flex % total_conf ) ) + 1
                    !write (*,*) 'full random conf', new_conf
                !! could have a test in opt_shift here...
                !! if set up specify the conformation number explicitely
                else
                    new_conf = conf_nb
                end if
                !write (*,*) 'want a new conf ', new_conf

                !! if equal, nothing to do
                !! else
                if ( new_conf /= this % current_conf ) then
                    ! start always from the first conformation and increments counter to know (to check in fact)
                    status = associate_to_conformation ( this % plist_flex % p_first, p_new_node, new_conf, counter, 0 ) !shift_conf )
                    if ( status == 0 ) then
                      ! counter shoudl be equal to new_conf if function returns correctly
                      this % current_conf = counter
                      ! for test, should not happen, to comment
                      !if ( associated ( p_new_node, this % p_current_node ) ) then
                      !  write(*,*) 'same pointer, SHOULD NOT !!'
                      !end if
                      this % p_current_node => p_new_node
                      !write(*,*) 'new current_conf ', this % current_conf

                      ! just for test
                      !if ( counter /= new_conf ) then
                      !   write(*,*) 'Strange counter/=new_conf ', counter, ' ', new_conf
                      !end if
                    ! for test also, to comment
                    !else
                    !  write(*,*) 'change_conf_flex, status /= 0 error '
                    !  stop 1
                    end if
                ! end if same conformation
                !else
                !    write(*,*) 'new_conf == this % current_conf'
                endif
            endif

            !write(*,*) 'end change conf current ', this % current_conf

        end subroutine change_conf_flex

END MODULE
