!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_force_energy.f90
!! Regroup information useful for computing force and energy with version fast

!> Only one single object of this module should be initialized
MODULE mod_force_energy
    USE mod_grid
    USE mod_gridtype
    USE mod_array_protein

    !> enumeration type for symetry of asymetry of interactions c2g1 : charge of protein2 in grid of solute 1
    !! lennard-jones with surface is c1g2
    integer, parameter :: symetric = 0, c2g1 = -1, c1g2 = 1

    !> Store the grid number to use for computing interaction and theirs types
    !!
    !! Only one single object of this module should be initialized
    !!
    !! It more effective for force and energy calculations with the version fast\n
    !! Add information if electrostatic and desolvation, or apolar and repulsive lennard-jones can be computed together\n
    type type_force_energy

        !> Contains the grid number to use for computation and their \ref mod_grid::type_grid "type_grid".\n
        !! caution, before 2 d, but not used, can add asymetry now add for multiple grids ( lennard-jones only )
        !! could be a separate type for some array set_of_protein
        integer, dimension ( :,: ), allocatable :: array_grid_calc

        !> Make image charge apart, store the group(s) (set_of_proteins ) which need img_charge computation.\n
        !! Typically the surface is excluded, not good for sdamm OpenMP, need to store all proteins
        integer, dimension ( : ), allocatable :: array_img_chg
        !> Same idea, take more time to look at which group it belongs, so save it, used by sdamm
        integer, dimension ( : ),allocatable :: array_prot_type


        !> Total number of grids to use for computation
        integer :: nb_grid
        !> protein id of the surface for image_charge, used only by sdamm ( only 1 for sda_2proteins )
        integer :: surface_id
        !> Speed computation if couple of interactions can be computed together
        !! @{
        logical :: couple_ElDes, couple_ApDesLJrep
        !! @}

        !> Keep specific interaction to compute.\n
        !! img_chg_desolv_input read from input, can be -1, 0 or 1.
        !! metal_desolvation and image charge desolvation (PrometCS)
        !! @{
        logical :: imetaldesolv, iimg_chg, iimg_chg_desolv
        integer :: img_chg_desolv_input
        !! @}

        ! using variable dielectric permit. for the case of protein-surface interaction 1 -use; 0- not use
        ! not use yet, can set up epsilon correction to image-charge or not.
        ! logical is enought for it, should setup correctly and include in the code. maybe latter use a scaling

        !> for PrometCS, analytical correction to image-charge
        real ( kind= 8):: epsilon_scale

        !> State if fast algorithm can be used.\n
        !! new alternative to efast, fast swtiched on if all interactions are symetric can use fast version of force and energy
        !! @{
        logical :: fast_computation
        !! @}

        !> If analytical force/energy are included
        !! Hydrodynamics parameters
        !! @{
        logical :: ianalytic, idebye, ihi
        real(kind=4) :: cutoff_hi
        !real(kind=4) :: h_analytic
        !! @}

    end type type_force_energy

    contains

        ! seems fine for sda_2proteins
        ! sdamm, problems, cannot make the sum over type of loc_int
        ! sdamm would need a grid_calc for every pairs of types, tricky...
        !       rather check association during the computation
        !

        !> Initilaize the oobject
        !! newer version, include asymetry, so need tab_protein to be general ( and sdamm case )
        !! with sdamm need to modify shift, inter-type interactions, need type calculation
        !! @param this : instance of this class
        !! @param tab_protein : instance of \ref mod_array_protein "mod_array_protein"
        !! @param type_calc : instance of type_calc
        !! @param imetal : logical set metal desolvation
        subroutine init_force_energy ( this, tab_protein, type_calc, imetal )
            USE mod_array_protein
            USE mod_type_calc
            IMPLICIT NONE
            type ( type_force_energy ) :: this
            type ( array_protein_type ), intent ( in ) :: tab_protein
            type ( type_calculation ), intent ( in ) :: type_calc
            logical, intent ( in ) :: imetal

            integer :: i, j, k, j2, k2, counter
            logical :: local_interaction, tmp_insert
            !! temporary store only the grid to conserve for grid_calc_energy
            integer, dimension ( enum_grid_force_calc ) :: tmp_grid_calc
            ! now find asymetry
            integer :: loc_int, tmp_nb_subg, shift_grid
            logical :: asym_sdamm

            this % couple_ElDes = .false.
            this % couple_ApDesLJrep = .false.
            ! default true, switch to off if asymetry or lennard-jones is required
            this % fast_computation = .true.
            this % iimg_chg = .false.
            ! set during read_input_file, moved as an input now
            !! this % iimg_chg_desolv = .false.
            ! now need integer to choose
            this % iimg_chg_desolv = .false.
            this % imetaldesolv = .false.

            ! set before during read_input_file, better as an input
            ! set explicitly for sdamm only, after calling this function
            this % ihi = .false.
            ! seems, not set here, certainly not precomputed ! to check
            this % cutoff_hi = 0.
             ! assign metal desolvation
            this % imetaldesolv = imetal

            ! new added, for image charge and sdamm, id of the surface ( better if last )
            this % surface_id = 0

            ! new added, sdamm, must detect asymetry, only one, should make fast_computation false
            ! default false, set up to true if any asymetry
            asym_sdamm = .false.

            ! new added, false by default, set during initialisation in init_analytic
            this % ianalytic = .false.
            this % idebye = .false.

            counter = 0
            tmp_grid_calc = 0
            shift_grid = 0

#ifdef DEBUG
            write (*,*) ' Entry init_force_energy'
#endif

            if ( allocated ( this % array_grid_calc ) ) then
                write (*,*) 'ERROR array_grid_calc already allocated ! This function should be called only once !!'
                stop 1
            end if

            ! difference between sda and sdamm, sdamm need to look interaction inside the same group as well
            ! not sure it is fully correct
            if ( type_calc % enum_type_calc == sdamm ) then
                shift_grid = 0
            else
                shift_grid = 1
            end if

#ifdef DEBUG
            write(*,*) 'shift_grid ', shift_grid
            write(*,*) 'init_force_energy enum_grid_force_calc=', enum_grid_force_calc
#endif

            ! modify with 2 loops
            ! enum_grid_force_calc define mod_gridtype, only the first 5, to extend for analytical
            do i = 1, enum_grid_force_calc

                !write(*,*) 'enum grid i ',i
                !write(*,*) 'tab_protein % nb_grids ', tab_protein % nb_grids
                ! by default false, actif if at least one of the grid  or charge associated
                local_interaction = .false.

                ! need to test different array, array like ljgrid solution ?? certainly
                if ( ( i == 1 ) .or. ( i == 2 ) ) then

                    ! wrong loop here, mix nb_grids and proteins
                    ! not good for sdamm, should check inside the same group
                    ! loop over protein or set_of_grid later
                    do j = 1, tab_protein % nb_grids
                       ! work for sda and sdamm with shift
                       do k = j + shift_grid, tab_protein % nb_grids

                        j2 = tab_protein % set_of_protein ( j, 1 )
                        k2 = tab_protein % set_of_protein ( k, 1 )

                        !write(*,*) 'set j k ', j, k
                        !write(*,*) 'protein j2 k2 ', j2, k2
                        !! will check by couple
                        !write(*,*) 'associated grid j2', &
                        !     is_associated ( tab_protein % array ( j2 ) % p_sogrid % pset_grid ( i ) )
                        !write(*,*) 'nb_charge j2', tab_protein % array ( j2 ) % p_sogrid % nb_charge
                        ! write(*,*) 'associated grid k2', &
                        !     is_associated ( tab_protein % array ( k2 ) % p_sogrid % pset_grid ( i ) )
                        !write(*,*) 'nb_charge k2', tab_protein % array ( k2 ) % p_sogrid % nb_charge

                        ! need to test charge / charge qsq if want to distinguish

                        ! test only the number, SHOULD be consistent
                        if ( ( ( tab_protein % array ( j2 ) % p_sogrid % nb_charge > 0 ) .and. &
                             ( is_associated (tab_protein % array ( k2 ) % p_sogrid % pset_grid ( i )) ) ) .or. &
                             ( ( tab_protein % array ( k2 ) % p_sogrid % nb_charge > 0 ) .and. &
                             ( is_associated (tab_protein % array ( j2 ) % p_sogrid % pset_grid ( i )) ) ) ) then

                             ! at least one asymetric, maybe symetric
                             !write(*,*) 'true interaction for grid ', i
                             local_interaction = .true.
                        !else
                        !    write(*,*) 'wrong interaction for grid ',i
                        end if
                    end do
                  end do

                ! need need nb_atom_surface
                else if ( ( i == 3 ) .or. ( i == 4 ) ) then
                    do j = 1, tab_protein % nb_grids

                      do k = j + shift_grid, tab_protein % nb_grids

                        j2 = tab_protein % set_of_protein ( j, 1 )
                        k2 = tab_protein % set_of_protein ( k, 1 )
                        !write(*,*) 'set j k ', j, k
                        !write(*,*) 'protein j2 k2 ', j2, k2

                        ! test only the number, SHOULD be consistent
                        if ( ( ( tab_protein % array ( j2 ) % p_sogrid % nb_atom_surface > 0 ) .and. &
                             ( is_associated (tab_protein % array ( k2 ) % p_sogrid % pset_grid ( i )) ) ) .or. &
                             ( ( tab_protein % array ( k2 ) % p_sogrid % nb_atom_surface > 0 ) .and. &
                             ( is_associated (tab_protein % array ( j2 ) % p_sogrid % pset_grid ( i )) ) ) ) then

                             local_interaction = .true.
                        end if
                      end do
                    end do

                ! only 5, simplify test here for lennard-jones
                else if ( i==5 ) then

                    ! only test protein 1 p_ljgrid associated if grid OR atom type
                    ! could loop over as well
                    if ( associated ( tab_protein % array( 1 ) % p_sogrid % p_ljgrid ) ) then
                        local_interaction = .true.
                    !else
                    !    write(*,*) 'p_ljgrid not associated'
                    endif

                !! else analytical...

                endif

                ! need separate ? cannot update in test ? no need of lj_present now
                if ( local_interaction ) then
                    counter = counter + 1
                    !write(*,*) 'local_interaction add 1 to counter ',counter
                end if

                !write(*,*) 'enum_type i local interaction ',i, local_interaction
                !write(*,*)

            end do ! end do loop enum_grid

            ! allocate the array
            !write (*,*) 'allocate array_grid_calc with counter ', counter
            this % nb_grid = counter
            allocate ( this % array_grid_calc ( counter, 3 ) )
            this % array_grid_calc = 0

            ! check first for grids in enum
            counter = 1

            !! could use select_charge function !
            !write(*,*) 'nb_types of proteins ', tab_protein % nb_grids
            !write(*,*) 'enum_grid_force_calc ', enum_grid_force_calc

            do i = 1, enum_grid_force_calc
                !write(*,*) 'enum grid i ', i

                ! check for grid_types, could have an other parameter grid, analytical
                loc_int = 0
                ! default 1, overwritten by lennard-jones only
                tmp_nb_subg = 1
                ! just temporary
                tmp_insert = .false.

                if ( ( i == 1 ) .or. ( i == 2 ) ) then
                    ! could loop over tab_prot % set_of_grid, identical always by couple !
                    do j = 1, tab_protein % nb_grids

                      do k = j + shift_grid, tab_protein % nb_grids

                        j2 = tab_protein % set_of_protein ( j, 1 )
                        k2 = tab_protein % set_of_protein ( k, 1 )
                        !write(*,*) 'set j k ', j, k
                        !write(*,*) 'protein j2 k2 ', j2, k2

                        ! test asymetry 1 c1g2
                        if ( ( tab_protein % array ( j2 ) % p_sogrid % nb_charge > 0 ) .and. &
                             ( is_associated (tab_protein % array ( k2 ) % p_sogrid % pset_grid ( i )) ) ) then
                             ! add -1
                             loc_int = loc_int + c1g2
                             !write(*,*) 'new_loc_int for c1g2 ', loc_int
                             tmp_insert = .true.
                        end if
                        ! test asymetry 2 c2g1
                        if ( ( tab_protein % array ( k2 ) % p_sogrid % nb_charge > 0 ) .and. &
                             ( is_associated ( tab_protein % array ( j2 ) % p_sogrid % pset_grid ( i )) ) ) then
                             ! add + 1
                             loc_int = loc_int + c2g1
                             !write(*,*) 'new_loc_int for c2g1 ', loc_int
                             tmp_insert = .true.
                        end if

                      end do
                    end do


                else if ( ( i == 3 ) .or. ( i == 4 ) ) then

                    !! could loop over tab_prot % set_of_grid, identical always by couple !
                    do j = 1, tab_protein % nb_grids

                      do k = j + shift_grid, tab_protein % nb_grids
                        ! test asymetry 1
                        j2 = tab_protein % set_of_protein ( j, 1 )
                        k2 = tab_protein % set_of_protein ( k, 1 )
                        !write(*,*) 'set j k ', j, k
                        !write(*,*) 'protein j2 k2 ', j2, k2

                        ! test asymetry 1 c1g2
                        if ( ( tab_protein % array ( j2 ) % p_sogrid % nb_atom_surface > 0 ) .and. &
                             ( is_associated ( tab_protein % array ( k2 ) % p_sogrid % pset_grid ( i )) ) ) then
                             ! add -1
                             loc_int = loc_int + c1g2
                             !write(*,*) 'new_loc_int ', loc_int
                             tmp_insert = .true.
                        end if
                        ! test asymetry 2 c2g1
                        if ( ( tab_protein % array ( k2 ) % p_sogrid % nb_atom_surface > 0 ) .and. &
                             ( is_associated ( tab_protein % array ( j2 ) % p_sogrid % pset_grid ( i )) ) ) then
                             ! add + 1
                             loc_int = loc_int + c2g1
                             !write(*,*) 'new_loc_int ', loc_int
                             tmp_insert = .true.
                        end if

                      end do
                    end do

                ! else check for 5
                else if ( i==5 ) then

                    ! lennard-jones fixed for the moment assign c1g2
                    if ( associated ( tab_protein % array ( 2 ) % p_sogrid % p_ljgrid ) ) then

                        ! asymetry lennard-jones, fixed to c1g2
                        loc_int = c1g2
                        tmp_nb_subg = tab_protein % array ( 2 ) % p_sogrid % p_ljgrid % nb_ljgrid
                        tmp_insert = .true.
                    end if
                end if

              !write(*,*) 'final loc_int ', loc_int
              !write(*,*) 'final tmp_insert ',tmp_insert
              if ( tmp_insert ) then
                    !write(*,*) 'insert counter ', counter, ' value of i ', i
                    !write(*,*) 'loc_int ', loc_int
                    !write(*,*) 'nb_sub-grids ', tmp_nb_subg
                    this % array_grid_calc ( counter, 1 ) = i
                    ! symetry
                    this % array_grid_calc ( counter, 2 ) = loc_int
                    ! number of sub-grid
                    this % array_grid_calc ( counter, 3 ) = tmp_nb_subg
                    ! increment only if inserted
                    counter = counter + 1
              !else
              !    write(*,*) 'do not insert '
              end if

              ! test for sdamm, must detect any asymetry, 0 symetric or no interaction, fine
              if ( ( type_calc % enum_type_calc == sdamm ) .and. ( loc_int /= 0 ) ) then
                  !write(*,*) 'Detected one asymetry in interactions for sdamm '
                  asym_sdamm = .true.
              end if

            end do

            ! need to check for fast version, if activated can check for couple
            ! fast if all symetric and ( no lennard-jones / subgrids for the moment ), usual case
            ! by default true, if non-symetric or lennard-jones make it false

            !write (*,*) 'size array_grid_calc ', size ( this % array_grid_calc )
            !write(*,*) 'this % nb_grid ', this % nb_grid
            do i = 1, this % nb_grid
                if ( ( this % array_grid_calc ( i, 2 ) /=  symetric ) .or. &
                   ( this % array_grid_calc (i, 1 ) == get_nb_gridtype( 'lennardjones' ) ) ) then
                    !write(*,*) 'set fast_computation to false i ',i
                    write(*,*) 'Asymetry detected, cannot use fast_computation 1'
                    this % fast_computation = .false.
                end if
            end do

            ! test for sdamm, any asymetry makes fast_computation false
            if ( ( type_calc % enum_type_calc == sdamm ) .and. ( asym_sdamm .eqv. .true. ) ) then
                write(*,*) 'Asymetry detected, cannot use fast_computation 2'
                this % fast_computation = .false.
            end if


            ! used by sda_2proteins, sdamm only a global asymetry, no need to compute this
            ! check for fast version, if couple can be computed at the same time
            counter = 1
            !write(*,*) 'array_grid_calc before test Couple size ', this % nb_grid
            !write(*,*) ':,1 ', this % array_grid_calc(:,1)
            !write(*,*) ':,2 ', this % array_grid_calc(:,2)
            !write(*,*) ':,3 ', this % array_grid_calc(:,3)


            ! check for couple, option fast, add symetry condition
            ! checked after array construction
            if ( counter + 1 <= this % nb_grid ) then
                if ( ( this % array_grid_calc ( counter, 1 ) == 1 ) .and. &
                     ( this % array_grid_calc ( counter, 2 ) == symetric ) .and. &
                     ( this % array_grid_calc ( counter + 1, 1 ) == 2 ) .and. &
                     ( this % array_grid_calc ( counter + 1, 2 ) == symetric ) ) then
                    this % couple_ElDes = .true.
                    counter = counter + 2
                else if ( ( this % array_grid_calc ( counter, 1 ) > 0 ) .or. &
                          ( this % array_grid_calc ( counter + 1, 1 ) > 0 ) ) then
                    counter = counter + 1
                end if
            end if

            ! \todo to check, couple_ApDesLJrep not selected in example/HFBI-graphite/run. but run seems correct.
            if ( counter + 1 <= this % nb_grid ) then
                if ( ( this % array_grid_calc ( counter, 1 ) == 3 ) .and. &
                     ( this % array_grid_calc ( counter, 2 ) == symetric ) .and. &
                     ( this % array_grid_calc ( counter + 1, 1 ) == 4 ) .and. &
                     ( this % array_grid_calc ( counter + 1, 2 ) == symetric ) ) then
                    this % couple_ApDesLJrep = .true.
                    counter = counter + 2
                else if ( ( this % array_grid_calc ( counter, 1 ) > 0 ) .or. &
                          ( this % array_grid_calc ( counter + 1, 1 ) > 0 ) ) then
                    counter = counter + 1
                end if
            end if

            ! add new test, if no grids should setup fast to false !
            !write(*,*) 'new test nb_grid ', this % nb_grid
            if ( this % nb_grid == 0 ) then
                this % fast_computation = .false.
            end if

            ! to look if image charge is required, not always
            !write(*,*) 'call init_image_charge'
            call init_image_charge ( this, tab_protein )

            ! problem prometcs sdamm, try to sort easily, but should make better !!
            ! works but really limited
            if ( type_calc % enum_type_calc == sdamm ) then
                if ( ( this % imetaldesolv ) .or. ( this % iimg_chg ) ) then
                    write(*,*)
                    write(*,*) 'Info : sdamm and metaldesolvation or image_charge, force fast_computation to false for safety'
                    write(*,*) 'Info : not necessary if a electrostatic potential grid is associated to the surface'
                    write(*,*) 'Info : code need to be modified to activate it, in mod_force_energy.f90'
                    write(*,*)
                    this % fast_computation = .false.
                end if
            end if

            ! array of nb_proteins which gives directly the type
            ! faster to get than using get_nb_type function

            ! Apply to both sda_2proteins / sdamm, save type only once, get_type
            ! is long (to really test)
            ! write(*,*) 'Create array for protein type '
            allocate ( this % array_prot_type ( size ( tab_protein % array ) ) )
            this % array_prot_type = 0
            do i = 1, size ( tab_protein % array )
                this % array_prot_type ( i ) = get_type( tab_protein, i )
            end do

        end subroutine init_force_energy

        !> Initialize array for image charge.\n
        !! Couyld be used for metal desolv as well.\n
        !! Make separate for the moment, allocate different array, but could be combined to init_force_energy
        !! img_chg can be found by tab_protein, important for sdamm to optimize
        !! to integrate init init_force, to test with sdamm
        !! @param this : instance of this class
        !! @param tab_protein : instance of \ref mod_array_protein "mod_array_protein"
        subroutine init_image_charge( this, tab_protein )
            IMPLICIT NONE
            type ( type_force_energy ) :: this
            type ( array_protein_type ), intent ( in ) :: tab_protein

            integer :: i, count_type_img, ierr

            !write(*,*)
            !write(*,*) 'Entry init_image_charge '

            ! default false, loop over solutes to find if it must be activated
            this %  iimg_chg = .false.
            count_type_img = 0

            ! setup array for array image_charge, use set_of_protein for img_chg apply to a full group
            ! first count needed size
            do i = 1, tab_protein % nb_grids
                !write(*,*) 'protein i ', tab_protein % set_of_protein ( i, 1 )
                !write(*,*) 'img_chg ', tab_protein % array ( tab_protein % set_of_protein ( i, 1 ) ) %  iimg_chg
                if ( tab_protein % array ( tab_protein % set_of_protein ( i, 1 ) ) %  iimg_chg .eqv. .true. )  then
                    !write(*,*) 'This protein wants image_charge calculation '
                    count_type_img = count_type_img + &
                        tab_protein % set_of_protein ( i, 2 ) - tab_protein % set_of_protein ( i, 1 ) + 1
                !else
                !    write(*,*) 'No image charge calculation '
                end if

                ! test presence of surface, should be only one ! used only by sdamm
                if ( tab_protein % array ( tab_protein % set_of_protein ( i, 1 ) ) %  surface .eqv. .true. )  then
                    !write(*,*) 'assign surface_id ', tab_protein % set_of_protein ( i, 1 )
                    this % surface_id = tab_protein % set_of_protein ( i, 1 )
                endif
            end do
            !write(*,*) 'count type_img ', count_type_img

            if ( count_type_img > 0 ) then
                ! assign true, and allocate specific array
                this %  iimg_chg = .true.
                allocate ( this % array_img_chg ( count_type_img ), stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In allocating array_img_chg '
                    stop 1
                end if

                ! by default if image_charge is activated, image_charge_metal_desolv is as well
                ! but can force to not be with -1
                if ( this % img_chg_desolv_input .ge. 0 ) then
                    write(*,*)
                    write(*,*) 'INFO Image charge is activated, by default it also activates'
                    write(*,*) 'INFO the image charge electrostic desolvation.'
                    write(*,*) 'INFO If it is the desired behaviour, be sure to include the desovlation grid with edfct.'
                    write(*,*) 'INFO To force its desactivation you can use desolv_image_charge = -1 '
                    write(*,*)
                    this % iimg_chg_desolv = .true.
                else
                    write(*,*)
                    write(*,*) 'INFO Image charge is activated, but you overwrite the default behaviour '
                    write(*,*) 'INFO in desactivating the image charge for electrostatic desolvation '
                    write(*,*) 'INFO     with desolv_image_charge = -1'
                    write(*,*)
                    this % iimg_chg_desolv = .false.
                end if

            ! no image charge, do not allocate and return
            else
#ifdef DEBUG
                write(*,*) 'No image charge required'
#endif
                return
            end if

            count_type_img = 0

            ! more general than by set_of_protein store for all proteins
            ! one loop only, OpenMP possible for sdamm
            do i = 1, size( tab_protein % array )

                if ( tab_protein % array ( i ) % iimg_chg .eqv. .true. ) then

                    count_type_img = count_type_img + 1
                    this % array_img_chg ( count_type_img ) = i

                    ! to combine with input_parameter now

                    !write(*,*) 'test for desolvation image charge '
                    !! is working, dangerous maybe, could make private
                    !write(*,*) 'enum for type desolv ', desolv
                    !write(*,*) 'number for type desolv ',get_nb_gridtype('desolvation')

                    ! need to use the new input_parameter now
                    !if ( ( is_associated ( tab_protein % array ( &
                    !                     tab_protein % set_of_protein ( i, 1 ) ) % p_sogrid % &
                    !                     pset_grid ( get_nb_gridtype('desolvation' ) ) ) ) .and. &
                    !     ( tab_protein % array ( &
                    !                     tab_protein % set_of_protein ( i, 1 ) ) % nb_charge > 0 ) ) then
                    !     write(*,*) 'Set image charge for desolvation term '
                         ! here could avoid desolv_img_chg from input
                    !     this % iimg_chg_desolv = .true.
                    !end if
                end if
            end do

            !write(*,*) 'Set up array img chg ', this % array_img_chg
        end subroutine init_image_charge

        !> Initialize specificaly analytical interactions.\n
        !! @param this : instance of this class
        !! @param tab_protein : instance of \ref mod_array_protein "mod_array_protein"
        !! @param cut_off : indicates the maximum distance, half the box for sdamm, c-surface for sda_2proteins
        !! @param h_anal : bin size of analytical arrays
        !! @param opt_ionic : optional, used only for the ionic strength at the moment
        subroutine init_analytic( this, tab_protein, cut_off, h_anal, opt_ionic )
            implicit none
            type ( type_force_energy ) :: this
            type ( array_protein_type ), intent ( in ) :: tab_protein
            real (kind=4), intent(in) :: cut_off, h_anal
            real (kind=4), intent(in), optional :: opt_ionic

            type ( sogrid ), pointer :: p_sogrid
            real (kind=8), dimension(:), allocatable :: list_stoke_radius
            real (kind=8) :: stoke !, tmp_max_debye, max_debye use cut_off rahter
            real (kind=4) :: ionic
            integer :: i, count_dh, ierr
#ifdef DEBUG
            integer :: j
#endif
	        ! pointer to flexible node
            type (NodeFlex), pointer :: tmp_node
            type ( protein ), pointer :: tmp_protein

            ionic = 0.
#ifdef DEBUG
            write(*,*)
            write(*,*) '=== Entry mod_force_energy:init_analytic ==='
            write(*,*) 'bin size analytic ', h_anal
            write(*,*) 'cut_off ', cut_off
#endif
            if ( present ( opt_ionic ) ) then
                ionic = opt_ionic
#ifdef DEBUG
                write(*,*) 'set ionic strength ', ionic
#endif
            end if

#ifdef DEBUG
            do i = 1, tab_protein % nb_grids

                write(*,*) 'Protein type = ', i
                p_sogrid => tab_protein % array ( tab_protein % set_of_protein ( i, 1 ) ) % p_sogrid

                write(*,*) 'associated set analytic ', &
                    associated ( p_sogrid % pset_analytic_grid )
                write(*,*) 'idebye associated to grid psgrid idebye ', &
                    p_sogrid % idebye
            end do
#endif
            count_dh = 0

#ifdef DEBUG
            write(*,*)
            write(*,*) '====== Fill array_dh, first loop over types of proteins ====='
#endif
            do i = 1, tab_protein % nb_grids
                ! get set of grid for first protein of the type
                p_sogrid => tab_protein % array ( tab_protein % set_of_protein ( i, 1 ) ) % p_sogrid

                if ( p_sogrid % idebye ) then
#ifdef DEBUG
                    write(*,*) 'type i=', i, ' idebye True'
#endif
                    count_dh = count_dh + 1
                    ! true for both, always use param_force_energy later on
                    this % ianalytic = .true.
                    this % idebye = .true.

!                else
!                    write(*,*) 'type i=', i, ' idebye False'
                end if
            end do
#ifdef DEBUG
            write(*,*) 'count_dh ', count_dh
            write(*,*) '====== Fill array_dh, Second loop ====='
#endif
            ! need the list of all stoke_radius
            allocate ( list_stoke_radius ( count_dh ), stat=ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating list_stoke_radius'
                STOP 1
            end if

            ! fill the list, maybe can be done before ??
            do i = 1, tab_protein % nb_grids
                !write(*,*) 'Fill list_stoke, loop over nb_grids i=', i

                ! Get set_of_grid / flexible from this type
                p_sogrid => tab_protein % array ( tab_protein % set_of_protein ( i, 1 ) ) % p_sogrid

                ! stoke radius for this type of solutes
                stoke = p_sogrid % stoke_radius
                !write(*,*) 'stoke_radius of this protein ', stoke
                if ( stoke == 0.d0 ) then
                    write(*,*) 'ERROR Stoke radius for type i ',i, 'has not been setup'
                    STOP 1
                end if
                list_stoke_radius ( i ) = stoke
            end do
#ifdef DEBUG
            write(*,*) 'list of stoke radius ', list_stoke_radius
#endif
            ! Loop again to fill the array, set up and fill the arrays
            ! need to treat the case of flexibility
            do i = 1, tab_protein % nb_grids

                ! convenient intermediate
                tmp_protein => tab_protein % array ( tab_protein % set_of_protein (i,1) )

                ! case flexible
                !if ( tab_protein % array ( tab_protein % set_of_protein (i, 1 ) ) % iflex ) then
                if ( tmp_protein % iflex ) then

                    ! Point to first allocated (p_first + 1 ) in flexible list
                    tmp_node => tmp_protein % p_flex % plist_flex % p_first % p_next
                    ! First is empty, point to next for first conformation
                    !tmp_node => tmp_node % p_next

                    do while (.not.associated( tmp_node, tmp_protein % p_flex % plist_flex % p_last))
                        ! & !tmp_node, tab_protein % array( tab_protein % set_of_protein (i,1) )% p_flex % plist_flex % p_last))

                        ! Point to sogrid for current conformation
                        p_sogrid => tmp_node % p_data
#ifdef DEBUG
                        write(*,*) 'new conformation flex'
#endif
                        ! stoke radius for this type of solutes, common for all conformations
                        stoke = p_sogrid % stoke_radius

                        ! get the maximum size, 8*ep grid (2 proteins) or half the box ( sdamm )
                        ! case debye-huckel activated
                        if ( p_sogrid % idebye ) then
#ifdef DEBUG
                            write(*,*) 'flexible type i=', i, ' idebye True'
#endif
                            call set_analytic(p_sogrid, 'debye-huckel', list_stoke_radius, real ( h_anal, 8 ), cut_off , ionic)

                        ! else crowder, or play with string'
                        end if
                        ! update the pointer to the next node of the list
                        tmp_node => tmp_node % p_next
                    end do

                ! case not flexible, only one sogrid
                else

                    p_sogrid => tmp_protein % p_sogrid
                    ! stoke radius for this type of solutes
                    stoke = p_sogrid % stoke_radius

                    ! get the maximum size, 8*ep grid (2 proteins) or half the box ( sdamm )
                    ! case debye-huckel activated
                    if ( p_sogrid % idebye ) then
#ifdef DEBUG
                        write(*,*) 'type i=', i, ' idebye True'
#endif
                        ! name as param if so_grid
                        call set_analytic(p_sogrid, 'debye-huckel', list_stoke_radius, real ( h_anal, 8 ), cut_off, ionic)
!                    else
!                           write(*,*) 'Debye false, maybe other analytical...'
                    end if
                end if
            end do

#ifdef DEBUG
            write(*,*)
            write(*,*) '== Analytic after complete setup =='
            !do i = 1, size( tab_protein % array )
            do i = 1, tab_protein % nb_grids
                write(*,*)
                write(*,*) 'Protein type ', i

                !tmp_protein => tab_protein % array ( i )
                tmp_protein =>  tab_protein % array ( tab_protein % set_of_protein (i,1) )

                write(*,*) 'associated set analytic ', &
                    !associated ( tab_protein % array(i) % p_sogrid % pset_analytic_grid )
                    associated ( tmp_protein % p_sogrid % pset_analytic_grid )

                write(*,*) 'idebye associated to grid psgrid idebye ', &
                    !tab_protein % array(i) % p_sogrid % idebye
                    tmp_protein % p_sogrid % idebye

                ! need to loop over flexible
                do j = 1, size ( tmp_protein % p_sogrid % pset_analytic_grid )

                    ! if flexible
                    if ( tmp_protein % iflex ) then

                        ! Point to first allocated (p_first + 1 ) in flexible list
                        tmp_node => tmp_protein % p_flex % plist_flex % p_first % p_next
                        ! First is empty, point to next for first conformation
                        ! tmp_node => tmp_node % p_next

                        do while (.not.associated( tmp_node, tmp_protein % p_flex % plist_flex % p_last))
                            ! Point to sogrid for current conformation
                            p_sogrid => tmp_node % p_data
                            call info_analytic( p_sogrid % pset_analytic_grid (j) )

                            ! update the pointer to the next node of the list
                            tmp_node => tmp_node % p_next
                        end do

                    ! not flexible
                    else
                        call info_analytic ( tmp_protein % p_sogrid % pset_analytic_grid (j) )
                    end if

                end do
            end do
#endif

#ifdef DEBUG
            write(*,*) 'End init_analytic '
            write(*,*)
#endif
        end subroutine init_analytic

        !> Initialization of one analytic of sogrid done in this subfunction, in module so_grid
        !subroutine init_onesogrid_analytic ( p_sogrid, list_stoke ) 'debye-huckel', h_analytic 0.01d0, cut_off , stoke, ionic )
        !    implicit none
        !
        !end subroutine init_onesogrid_analytic

        !> Deallocate all arrays.\n
        !! @param this : instance of this class
        subroutine delete_force_energy ( this )
            implicit none
            type ( type_force_energy ) :: this
            integer :: ierr

            !write(*,*) 'Entry delete_force_energy'
            if ( allocated ( this % array_grid_calc ) ) then
                deallocate ( this % array_grid_calc, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In dealocating param_force_energy % array_grid_calc'
                    stop 1
                endif
            endif

            if ( allocated ( this % array_prot_type ) ) then
                deallocate ( this % array_prot_type, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In dealocating param_force_energy % array_prot_type'
                    stop 1
                endif
            endif

            if ( allocated ( this % array_img_chg ) ) then
                deallocate ( this % array_img_chg, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In dealocating param_force_energy % array_img_chg'
                    stop 1
                endif
            endif

        end subroutine delete_force_energy

        !> Print info about global settings.\n
        !! Would need a check for sdamm, array_grid_calc not used\n
        !! @param this : instance of type_force_energy
        subroutine print_info_param_fe ( this )
            IMPLICIT NONE
            type ( type_force_energy ) :: this
            integer :: i

            write(*,*)
            write(*,*) '==========================='
            write(*,*) 'Summary Param Force Energy '
            write(*,*)  'nb_grid ', this % nb_grid
            write(*,*)
            do i = 1, this % nb_grid
                write(*,*) 'interaction ', i
                call print_gridtype( this % array_grid_calc(i,1) )
                write(*,*) 'symmetry ', this % array_grid_calc(i,2)
                write(*,*) 'number of subgrids : ', this % array_grid_calc(i,3)
            end do

            write ( *,* ) 'fast computation ', this % fast_computation
            write ( *,* ) 'couple_ElDes ', this % couple_ElDes
            write ( *,* ) 'couple_ApDesLJrep ', this % couple_ApDesLJrep
            !write ( *,* ) 'lennard-jones present ', this % lj_present
            write ( *,* )
            write ( *,* ) 'hydrodynamic interaction ', this % ihi
            write ( *,* ) 'image charge flag:  ', this % iimg_chg
#ifdef VERBOSE
            write ( *,* ) 'array array_prot_type ', this % array_prot_type
            if ( allocated ( this % array_img_chg ) ) then
                write ( *,* ) 'Surface id: ', this % surface_id
                write ( *,* ) 'Image charge array: ', this % array_img_chg
            end if
#endif
            write ( *,* ) 'Include desolvation:', this % iimg_chg_desolv
            write ( *,* ) 'Metal desolvation:  ', this % imetaldesolv
            write (*,*)

            if ( this % ianalytic ) then
                write(*,*) 'Analytical interactions is activated'
                !test debye/crowder
            else
                write(*,*) 'No analytical interactions'
            end if

            write ( *,* ) '====================='
            write ( *,* )

        end subroutine print_info_param_fe


  !> Compute the derivative of the energy of position x.\n
  !! Will be used to compute the force on an atom at position x\n
  !! Pass an instance of \ref mod_grid::type_grid "type_grid" as parameter, so it can apply to any grid\n
  !! This version can deal with both uhbd and dtgrid. Overhead estimated to 10 %
  !!
  !! @param x : position of the atom
  !! @param f : return force ( rather derivative of the energy ) computed at x
  !! @param tgrid : instance of the grid used for computation
  subroutine force_atom_uhbd ( x, f, tgrid )
      IMPLICIT NONE
      real ( kind=8 ), dimension(3), intent(in) :: x
      real ( kind=8 ), dimension(3), intent(out) :: f
      type (type_grid ), intent( in ) :: tgrid

      real ( kind=8 ) :: u000, u100, u010, u110, u001, u011, u101, u111, GRIDXJ ,GRIDYK, GRIDZL
      real ( kind=8 ) :: DX, DY, DZ, DXDY, DXDZ, DYDZ, DXDYDZ, DY_DYDZ, DZ_DYDZ, DX_DXDY, DZ_DXDZ, FX, FY, FZ
      real ( kind=8 ) :: h, hpin !, dist2
      integer :: j,k,l

      f = (/0., 0., 0./)

!     overides the global variable
      h = tgrid % h
      hpin = tgrid % hpin

      !dist2 = x(1)**2 + x(2)**2 + x(3)**2
!c    TO CHECK, if lots of entry for nothing, can check before
!! it should be a cut oof here, but done before.. in new version, so no test
      !IF (dist2.GT.tgrid % rpsq) RETURN !in fact miss a lot with surface

      ! GT should work no ?? no because need I+1 !
      j = int(( x(1) - tgrid % origin(1) ) * hpin )
      IF (j.GE.tgrid % size_x .OR. j.LE.0) RETURN
      k = int((x(2) - tgrid % origin(2) ) * hpin )
      IF (k.GE.tgrid%size_y .OR.  k.LE.0 ) RETURN
      l = int((x(3) - tgrid % origin(3) ) * hpin )
      IF (l.GE.tgrid%size_z .OR. l.LE.0) RETURN

      ! no test, for sure real value in force
      u000 = tgrid % p_grid % pgrid_real (J,K,L)
      u100 = tgrid % p_grid % pgrid_real (J+1,K,L)
      u010 = tgrid % p_grid % pgrid_real (J,K+1,L)
      u110 = tgrid % p_grid % pgrid_real (J+1,K+1,L)
      u001 = tgrid % p_grid % pgrid_real (J,K,L+1)
      u011 = tgrid % p_grid % pgrid_real (J,K+1,L+1)
      u101 = tgrid % p_grid % pgrid_real (J+1,K,L+1)
      u111 = tgrid % p_grid % pgrid_real (J+1,K+1,L+1)


!      could save op before
      GRIDXJ = h * dfloat(J) + tgrid % origin(1)
      GRIDYK = h * dfloat(K) + tgrid % origin(2)
      GRIDZL = h * dfloat(L) + tgrid % origin(3)
      ! GRIDJKL = h*dfloat(J,K,L) ??

      DX = (X(1)-GRIDXJ)*hpin
      DY = (X(2)-GRIDYK)*hpin
      DZ = (X(3)-GRIDZL)*hpin
      DXDY = DX*DY
      DXDZ = DX*DZ
      DYDZ = DY*DZ
      DXDYDZ = DXDY*DZ
      DY_DYDZ = DY - DYDZ
      DZ_DYDZ = DZ - DYDZ
      DX_DXDY = DX - DXDY
      DZ_DXDZ = DZ - DXDZ
!
      FX = (U100-U000)* (1.d0-DY_DYDZ-DZ)
      FX = FX + (U110-U010)*DY_DYDZ
      FX = FX + (U101-U001)*DZ_DYDZ
      FX = FX + (U111-U011)*DYDZ
      FX = -FX*hpin
      FY = (U010-U000)* (1.d0-DX-DZ_DXDZ)
      FY = FY + (U011-U001)*DZ_DXDZ
      FY = FY + (U110-U100)* (DX-DXDZ)
      FY = FY + (U111-U101)*DXDZ
      FY = -FY*hpin
      FZ = (U001-U000)* (1.d0-DX_DXDY-DY)
      FZ = FZ + (U011-U010)* (DY-DXDY)
      FZ = FZ + (U101-U100)*DX_DXDY
      FZ = FZ + (U111-U110)*DXDY
      FZ = -FZ*hpin
      f(1) = FX
      f(2) = FY
      f(3) = FZ

      !if (isnan( f(1)) .or. isnan( f(2)) .or. isnan( f(3))) then
      !    write(*,*) 'j, k, l ',j,k,l
      !    write(*,*) 'u000 ', tgrid % p_grid % pgrid_real (J,K,L)
      !    stop '"f1" is a NaN'
      !end if

  end subroutine force_atom_uhbd



  !>  Compute the energy of position x of grid tgrid
  !!
  !! This version can deal with both uhbd and dtgrid
  !! @param x : position of the atom
  !! @param U : return the computed potential
  !! @param tgrid : instance of the grid to use for the computation
  subroutine potential_atom_uhbd ( x, U, tgrid )
      IMPLICIT NONE
      real ( kind=8 ), dimension(3), intent(in) :: x
      real ( kind=8 ), intent(out) :: U
      type (type_grid ), intent(in) :: tgrid

      real ( kind=8 ) :: u000, u100, u010, u110, u001, u011, u101, u111, GRIDXJ ,GRIDYK, GRIDZL !, toto
      real ( kind=8 ) :: DX, DY, DZ, DXDY, DXDZ, DYDZ, DXDYDZ, DY_DYDZ, DZ_DYDZ, DX_DXDY, DZ_DXDZ
      real ( kind=8 ) :: h, hpin !, dist2
      integer :: j,k,l


!     overides the global variable
      h = tgrid % h
      hpin = tgrid % hpin

      !initialize
      U = 0.d0
      u000 = 0.d0
      u100 = 0.d0
      u010 = 0.d0
      u110 = 0.d0
      u001 = 0.d0
      u011 = 0.d0
      u101 = 0.d0
      u111 = 0.d0

      !dist2 = x(1)**2 + x(2)**2 + x(3)**2
      !write(*,*) 'dist2 ', dist2, tgrid%rpsq
!c    TO CHECK, if lots of entry for nothing, can check before

      ! strong approximation here, only half of the larger box !! miss a lot with surface ! even with protein ??
      !IF (dist2.GT.tgrid%rpsq) RETURN

      j = int((x(1) - tgrid%origin(1) ) * hpin )
      IF (j.GE.tgrid%size_x .OR. j.LE.0) RETURN
      k = int((x(2) - tgrid%origin(2) ) * hpin )
      IF (k.GE.tgrid%size_y .OR.  k.LE.0 ) RETURN
      l = int((x(3) -  tgrid%origin(3) ) * hpin )
      IF (l.GE.tgrid%size_z .OR. l.LE.0) RETURN
      !write(*,*) 'did not return '
      !write(*,*) 'u000 ', tgrid % p_grid % pgrid_real (J,K,L)

      u000 = tgrid % p_grid % pgrid_real (J,K,L)
      u100 = tgrid % p_grid % pgrid_real (J+1,K,L)
      u010 = tgrid % p_grid % pgrid_real (J,K+1,L)
      u110 = tgrid % p_grid % pgrid_real (J+1,K+1,L)
      u001 = tgrid % p_grid % pgrid_real (J,K,L+1)
      u011 = tgrid % p_grid % pgrid_real (J,K+1,L+1)
      u101 = tgrid % p_grid % pgrid_real (J+1,K,L+1)
      u111 = tgrid % p_grid % pgrid_real (J+1,K+1,L+1)

!     could save op before
      GRIDXJ = h*dfloat(J) + tgrid%origin(1)
      GRIDYK = h*dfloat(K) + tgrid%origin(2)
      GRIDZL = h*dfloat(L) + tgrid%origin(3)
      ! GRIDJKL = h*dfloat(J,K,L) ??
      DX = (X(1)-GRIDXJ)*hpin
      DY = (X(2)-GRIDYK)*hpin
      DZ = (X(3)-GRIDZL)*hpin
      DXDY = DX*DY
      DXDZ = DX*DZ
      DYDZ = DY*DZ
      DXDYDZ = DXDY*DZ
      DY_DYDZ = DY - DYDZ
      DZ_DYDZ = DZ - DYDZ
      DX_DXDY = DX - DXDY
      DZ_DXDZ = DZ - DXDZ
!!! only change between energy and forces, energy only extrapolation ??
      U = U000* (1.d0-DX_DXDY-DY_DYDZ-DZ_DXDZ-DXDYDZ)
      U = U + U010* (DY_DYDZ-DXDY+DXDYDZ)
      U = U + U001* (DZ_DYDZ-DXDZ+DXDYDZ)
      U = U + U011* (DYDZ-DXDYDZ)
      U = U + U100* (DX_DXDY-DXDZ+DXDYDZ)
      U = U + U110* (DXDY-DXDYDZ)
      U = U + U101* (DXDZ-DXDYDZ)
      U = U + U111*DXDYDZ

      !write(*,*) 'compute energy U ', U
      !if (isnan( U ) ) then
      !    write(*,*) 'j, k, l ',j,k,l
      !    write(*,*) 'u000 ', tgrid % p_grid % pgrid_real (J,K,L)
      !    write(*,*) u000,u001,u010,u011,u100,u110,u101,u111
      !    stop '"U" is a NaN'
      !end if

  end subroutine potential_atom_uhbd

END MODULE mod_force_energy
