!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \defgroup PBC PBC : Periodic Boundary Conditions
!! @{
!!
!!
!! Periodic Boundary Conditions which apply in SDA ( defined in mod_geometry.f90 )
!!
!! Rules used in \ref mod_geometry::update_translation_box "update_translation_box()"
!!
!! With a sphere geometry, no PBC is defined and a surface does not influence the algorithm\n
!! Only used for sda_2proteins, the solute starts at the b surface and the trajectory stops at the c surface\n
!!
!! With a box geometry :
!!  - no PBC :\n
!!    Reflexive boundary conditions along x and y. Nothing is applied along z. Should use escape ??\n
!!    Used with sda_2proteins and association rates calculation with a surface. The solute cannot go below zmin ( exclusion with the surface )\n
!!    It will escape at zmax. The procedure is an approximation, not yet validated\n
!!  - with PBC :\n
!!    -# no surface : PBC in 3D, more usual case with sdamm
!!    -# with a surface : Keep periodic boundary conditions alonz x and y and reflexive on z.\n
!!       Due to the repulsive lennard-jones the solute should not go below the surface\n
!!
!! @}

!> \file mod_geometry.f90
!!
!! Define geometric properties of the simulation box

! try to define a group, add entry only in Modules

! If a surface is present, there is reflective border :
!  - on zmax ( surface should be close to zmin )
!  - along x and y because of the discontinuity of the potential grid
!  - can use as well escape, will apply on zmax ( case of sda_2proteins and association rate calculations with a membrane for instance )
!

!> Define a base class geometry and derivatives : sphere or box
!!
!! Cannot really define a base and derivatives classes, but creates an gemetry object where only the sphere or the box object will be instanciated
!!
!! Different border conditions can be applied to the box\n
!! Normally PBC apply in 3 dimension ( sdamm case ) \n
!! It is modified in the presence of a surface
!!
!! For more details see \ref PBC "group PBC"
MODULE mod_geometry

    !> enumeration type
    integer, parameter :: enum_sphere = 1, enum_box = 2

    !> Derived class sphere
    type sphere
        !> Used for starting and escape distance
        !! @{
        real ( kind=4 ) :: c_surface, b_surface
        !> @}
    end type sphere

    !> Derived class box
    type box
        !> Define the limit of the box position
        !! @{
        real ( kind=4 ) :: x_min, x_max, y_min, y_max, &
                           z_min, z_max
        !> @}
        !> Define the dimension of the box
        real ( kind=4), dimension( 3 ) :: sizebox
    end type box

    !> geometry is a base (common base ) class
    type geometry
        !> type of the contained geometry : box or sphere defined by enumeration
        integer :: type
        !> pointer to a sphere geometry
        type( sphere ), POINTER :: pshere
        !> pointer to a box geometry
        type( box ), POINTER :: pbox

        !> properties of the geometry object
        !> If the solutes can escpape, if there is pbc, if there is a surface
        !! @{
        logical :: escape, pbc, surface
        !> @}
        !> starting and escaping position. redondant wit b and c surface
        !! @{
        real ( kind=4 ) :: start_position, stop_position
        !> @}

    end type geometry

   contains

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!! Base Class

    !> Allocate the geometry
    !!
    !! The type of geometry \ref geometry::type "type" must be setup before the call to this function
    !!
    !! It can be set first with \ref enum_sphere "enum_sphere" or \ref enum_box "enum_box"
    !! @param this : instance of geometry
    !! @param pbc : if periodic boundary conditions should apply
    !! @param escape : if solute(s) can escape
    !! @param surface : if the simulation contains a surface ( modify the computation of distances )
    !! @param c_surf : b and c surface for sphere geometry
    !! @param xmin, xmax, ymin, ymax, zmin, zmax : position of the box
    !! @param start_pos : if a particular position is needed at start up ( should apply both with box and sphere )
    subroutine init_geometry ( this, pbc, escape, surface, c_surf, &
                               xmin, xmax, ymin, ymax, zmin, zmax, start_pos )
        IMPLICIT NONE
        type ( geometry ) :: this
        logical , intent (in) :: pbc, escape, surface
        real ( kind=4 ), intent (in) :: c_surf, xmin, xmax, &
                         ymin, ymax, zmin, zmax, start_pos

        !! set general value
        this % pbc = pbc
        this % escape = escape
        this % surface = surface
        ! could get from box and size
        this % start_position = 0.
        this % stop_position = 0.

        if ( this % type == enum_sphere ) then
            allocate ( this % pshere )
            nullify ( this % pbox )

            ! set up start position (b_surf if sphere)
            this % start_position = start_pos

            call init_geometry_sphere ( this % pshere, this % start_position, c_surf )
            ! always escape in 2 protein sphere
            if ( this % escape ) then
                this % stop_position = c_surf
            end if

        else if ( this %type == enum_box ) then
            allocate ( this % pbox )
            nullify ( this % pshere )
            call init_geometry_box( this % pbox, xmin, xmax, ymin, ymax, &
                                   zmin, zmax )

            this % start_position = start_pos

            if ( this % escape ) then
                this % stop_position = zmax
            else
                this % stop_position = c_surf
            end if

        end if

        ! check was missing
        if ( this % start_position > this % stop_position ) then
            write(*,*) 'start ',this % start_position
            write(*,*) 'stop ',this % stop_position
            write(*,*) 'ERROR Starting position is larger than the stop position !'
            STOP 1
        end if

    end subroutine init_geometry

    !> Delete geometry
    !!
    !! Delete the appropriate derived class
    !! @param this : instance of geometry
    subroutine delete_geometry ( this )
        IMPLICIT NONE
        type ( geometry ) :: this

        if ( associated ( this % pshere ) ) then
            deallocate ( this %  pshere )
            nullify ( this % pshere )
        else if ( associated ( this % pbox ) ) then
            deallocate ( this % pbox )
            nullify ( this % pbox )
        end if

    end subroutine delete_geometry

    !> Get parameters of the geometry object
    !!
    !! Call appropriate derived class
    !! @param this : instance of geometry
    !! @param params_tmp : array to store the results
    subroutine get_parameters ( this, params_tmp )
        IMPLICIT NONE
        type ( geometry ) :: this
        real ( kind=4 ), dimension(6) :: params_tmp

         if ( associated ( this % pshere ) ) then
            call get_parameters_sphere ( this % pshere, params_tmp )
         else
            call  get_parameters_box ( this % pbox, params_tmp )
         end if

    end subroutine get_parameters

    !> Update the translation and apply all border conditions, PBC.\n
    !! Sort of "virtual" function, call the appropiate update of sphere or box.\n
    !! Escape could be checked here. to see later.\n
    !! @param this : instance of geometry
    !! @param translation : new position of the solute before treatment
    !! @param new_pos : new position of the solute after treatment ( PBC or reflective boundaries )
    !! @param cell : box information update, used with PBC (\ref mod_protein::protein::box "box" ) of the solute
    subroutine update_translation ( this, translation, new_pos, cell )
        IMPLICIT NONE
        type ( geometry ) :: this
        real ( kind=8 ), dimension (3), intent ( in ) :: translation
        real ( kind=8 ), dimension (3), intent ( out ) :: new_pos
        integer, dimension (3), intent ( inout ) :: cell

        ! if sphere
        if ( this % type .eq. enum_sphere ) then
             ! keep commented function for later use
             !call update_translation_sphere ( this % pshere, translation, new_pos )
             new_pos = translation

        ! else box
        else
            !write(*,*) 'call update_translation_box'
            call update_translation_box ( this % pbox, translation, new_pos, this % pbc, this % surface, cell )
        end if

    end subroutine update_translation

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!! Derived Class sphere

    ! start position may replace b_surf

    !> Create sphere, should be called by geometry
    !!
    !! Assign b and c surface
    !! @param this : instance of sphere
    !! @param b_surf, c_surf : values of b and c surfaces
    subroutine init_geometry_sphere ( this, b_surf, c_surf )
        IMPLICIT NONE
        type ( sphere ) :: this
        real ( kind=4 ), intent ( in ) :: b_surf, c_surf

        this % b_surface = b_surf
        this % c_surface = c_surf

    end subroutine init_geometry_sphere

    !> Get parameters of the spehere
    !!
    !! Should be called by geometry
    !! @param this : instance of sphere
    !! @param param2s : return array with parameters, only 2 over 6 are filed with sphere
    subroutine get_parameters_sphere ( this, param2s )
        IMPLICIT NONE
        type ( sphere ) :: this
        real ( kind=4 ), dimension(6), intent (out) :: param2s

        param2s(1) = this % b_surface
        param2s(2) = this % c_surface
        param2s(3) = 0.
        param2s(4) = 0.
        param2s(5) = 0.
        param2s(6) = 0.

    end subroutine get_parameters_sphere

    !> Update translation in case of sphere
    !!
    !! Should be called by geometry\n
    !! In case of sphere do nothing
    ! keep function for later, but no real need
!    subroutine update_translation_sphere ( this, translation, new_pos )
!        IMPLICIT NONE
!        type ( sphere ), intent(in) :: this
!        real ( kind=8 ), dimension (3), intent ( in ) :: translation
!        real ( kind=8 ), dimension (3), intent ( out ) :: new_pos
!
!        ! make bd_move translation
!        ! no periodicity in this case
!        ! see later for escape
!        ! now nothing to do
!        new_pos = translation
!
!    end subroutine update_translation_sphere


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
!!!!!!!! Derived Class box

    !> Create box
    !!
    !! Assign the dimension of the box. The \ref box::sizebox "sizebox" is computed
    !! @param this : instance of geometry
    !! @param  xmin, xmax, ymin, ymax, zmin, zmax : the position of the box
    subroutine init_geometry_box ( this, xmin, xmax, ymin, ymax, &
                                   zmin, zmax )
        IMPLICIT NONE
        type ( box ) :: this
        real ( kind=4 ) :: xmin,xmax,ymin,ymax,zmin,zmax

        this % x_min = xmin
        this % x_max = xmax
        this % y_min = ymin
        this % y_max = ymax
        this % z_min = zmin
        this % z_max = zmax

        this % sizebox ( 1 ) = xmax - xmin
        this % sizebox ( 2 ) = ymax - ymin
        this % sizebox ( 3 ) = zmax - zmin

    end subroutine init_geometry_box

    !> Get parameters in case of box
    !!
    !! Should be called by geometry
    !! @param this : instance of box
    !! @param params : return array to store the parameters, use all 6 for sphere
    subroutine get_parameters_box ( this, params )
        IMPLICIT NONE
        type ( box ) :: this
        real ( kind=4 ), dimension(6), intent (out) :: params

        params(1) = this % x_min
        params(2) = this % x_max
        params(3) = this % y_min
        params(4) = this % y_max
        params(5) = this % z_min
        params(6) = this % z_max

    end subroutine get_parameters_box

    !> Apply the translation with box. Take into account periodicity and surface
    !!
    !! Normally PBC apply in 3 dimension for box\n
    !! If a surface is present, there is reflective border on zmax
    !!
    !! Function similar in compute_distance.f90. See if we can merge\n
    !! May add escape ??
    !! @param this : instance of box
    !! @param translation : position of the solute before treatment of periodicity
    !! @param new_pos : return the new position after periodicity and border condition
    !! @param pbc : if pbc should apply
    !! @param surf : if a surface is present
    !! @param cell : box information for solutes
    subroutine update_translation_box ( this, translation, new_pos, pbc, surf, cell )
        IMPLICIT NONE
        type ( box ) :: this
        real ( kind=8 ), dimension (3), intent ( in ) :: translation
        real ( kind=8 ), dimension (3), intent ( out ) :: new_pos
        integer, dimension (3), intent ( inout ) :: cell
        logical, intent ( in ) :: pbc, surf

        !! make bd_move translation
        !write(*,*) 'update_translation_box'
        !write(*,*) 'pbc surface ', pbc, surf
        !write(*,*) 'this % sizebox', this % sizebox

        !! copy by default, modify according to type of box
        new_pos = translation

        if ( pbc ) then
            !new_pos = translation
            if ( translation ( 1 ) > this % x_max ) then
                new_pos ( 1 ) = new_pos ( 1 ) - this % sizebox ( 1 )
                cell ( 1 ) = cell ( 1 ) + 1
                !write(*,*) 'update cell '
            else if ( translation ( 1 ) < this % x_min ) then
                new_pos ( 1 ) = new_pos ( 1 ) + this % sizebox ( 1 )
                cell ( 1 ) = cell ( 1 ) - 1
            end if
            if ( translation ( 2 ) > this % y_max ) then
                new_pos ( 2 ) = new_pos ( 2 ) - this % sizebox ( 2 )
                cell ( 2 ) = cell ( 2 ) + 1
            else if ( translation ( 2 ) < this % y_min ) then
                new_pos ( 2 ) = new_pos ( 2 ) + this % sizebox ( 2 )
                cell ( 2 ) = cell ( 2 ) - 1
            end if

            ! if surface and pbc ( use for sdamm ): z is reflective on top of the box, lj_repulsive avoids going < z_min
            ! would need escape argument here as well here. ?
            if ( surf ) then
            ! if ( surf .and. ( escape .eqv. .false. ) ) then
                ! write(*,*) 'update_translation_box and surf ', new_pos
                if ( translation ( 3 ) > this % z_max ) new_pos( 3 ) = 2 * this % z_max - translation ( 3 )
                !write(*,*) 'update_translation_box and surf z new pos', new_pos
            ! else pbc along z as well
            else
                if ( translation ( 3 ) > this % z_max ) then
                    new_pos ( 3 ) = new_pos ( 3 ) - this % sizebox ( 3 )
                    cell ( 3 ) = cell ( 3 ) + 1
                else if ( translation ( 3 ) < this % z_min ) then
                    new_pos ( 3 ) = new_pos ( 3 ) + this % sizebox ( 3 )
                    cell ( 3 ) = cell ( 3 ) - 1
                end if
            end if
            !new_pos = translation
        ! else reflective boundary condition on x and y, do not modify z or ( can escape ), strange if not ?? should copy by default
        ! defaut do not modify z, like this can escape
        else
            if ( translation ( 1 ) < this % x_min) new_pos( 1 ) = 2 * this % x_min - translation ( 1 )
            if ( translation ( 1 ) > this % x_max) new_pos( 1 ) = 2 * this % x_max - translation ( 1 )
            if ( translation ( 2 ) < this % y_min) new_pos( 2 ) = 2 * this % y_min - translation ( 2 )
            if ( translation ( 2 ) > this % y_max) new_pos( 2 ) = 2 * this % y_max - translation ( 2 )
        end if

    end subroutine update_translation_box

END MODULE mod_geometry
