!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_ljgrid.f90
!! Declare lennard-jones grid (only for ProMetCS force field)

!> Module to define lennard-jones interaction
MODULE mod_ljgrid

    USE mod_grid

    !> independent, used by other modules
    !! could extend with charges
    type type_position_atom

        !> 2d required for atom position(x,y,z : atom_nb )
        !! pointer better ?? can set to null, and allocate later ( with different size ) ??
        real ( kind=8 ), dimension(:,:), pointer :: p_position_atom
        !> size of the array, here, different number for each type
        integer :: nb_pos_atom

    end type type_position_atom

    type type_ljgrid

        !> array for all lj grids, size nb_ljgrid
        type ( type_grid ), dimension(:), pointer :: array_ljgrid

        !> nb lj grid
        integer :: nb_ljgrid, max_pos_atom

        !> array for all position of atoms, introduce new type
        type ( type_position_atom ), dimension(:), allocatable :: array_position_atom_lj

    end type type_ljgrid

    contains

        !> Allocate array for position of the atoms
        !! @param this : instance of \ref mod_ljgrid "mod_ljgrid"
        !! @param nb_atoms : size of the array, always (3 , nb_atoms )
        subroutine allocate_position_atom ( this, nb_atoms )
            IMPLICIT NONE
            type ( type_position_atom ) :: this
            integer, intent ( in ) :: nb_atoms

            integer :: ierr

            this % nb_pos_atom = nb_atoms
            allocate ( this % p_position_atom ( 3, nb_atoms ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocation of p_position_atom '
                stop 1
            end if
            this % p_position_atom = 0.d0

        end subroutine

        !> Copy atom position
        !! @param this : instance of \ref mod_ljgrid::type_position_atom "type_position_atom"
        !! @param pdest : pointer to the destination of type_position
        subroutine copy_position_atom ( this, pdest )
            IMPLICIT NONE
            type ( type_position_atom ), intent ( in ) :: this
            type ( type_position_atom ), pointer :: pdest

            integer :: ierr

            !write(*,*) 'Entry copy_position_atom '

            ! need to test if allocated in the origin
            if ( associated ( this % p_position_atom ) ) then
              !! copy the number of grids
              pdest % nb_pos_atom = this % nb_pos_atom

              ! here allocate the pointer
              allocate ( pdest % p_position_atom ( 3, pdest % nb_pos_atom ), stat = ierr )
              if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocating p_position_atom in copy'
                stop 1
              end if

              ! copy the data of the full array
              pdest % p_position_atom = this % p_position_atom
            !! already done ?
            else
                nullify ( pdest )
                !pdest % nb_pos_atom = 0
            end if

        end subroutine

        !> Deallocate p_pos_atom
        !!
        !! Check if associated done inside the destructor
        !! @param this : instance of \ref type_position_atom
        subroutine deallocate_position_atom ( this )
            IMPLICIT NONE
            type ( type_position_atom ) :: this

            integer :: ierr

            ! test if allocated first
            if ( associated ( this % p_position_atom ) ) then
                deallocate ( this % p_position_atom, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In deallocation of p_pos_atom '
                    STOP 1
                end if
            end if
            nullify ( this % p_position_atom )
            this % nb_pos_atom = 0

        end subroutine

        !> Constructor
        !!
        !! Allocate arrays of pointers, nullified by default and still empty
        !! New do not allocate the array, easier for tests
        !! Will be correctly allocated in function set_lj_X
        !!
        !! First allocate both arrays, but asymetric by default, need to test if null or not
        !! Later, allocate only needed ?
        subroutine allocate_ljgrid ( this, nb_grids )
            IMPLICIT NONE
            type ( type_ljgrid ) :: this
            integer, intent ( in ) :: nb_grids

            !write(*,*) 'Allocate LJ grid, assign size of ', nb_grids

            this % nb_ljgrid = nb_grids
            this % max_pos_atom = 0
            ! nullify pointer by default
            nullify ( this % array_ljgrid )

        end subroutine allocate_ljgrid

        !! need to write copy function
        subroutine copy_ljgrid ( this, pdest )
            IMPLICIT NONE
            type ( type_ljgrid ), intent ( in ) :: this
            type ( type_ljgrid ), pointer :: pdest

            type ( type_position_atom ), pointer :: ppos_atom

            integer :: ierr, i

            !write(*,*)
            !write(*,*) 'Entry copy_ljgrid '
            !write(*,*)

            !! allocate main object
            allocate ( pdest,  stat = ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR In allocating pdest in mod_ljgrid::copy_ljgrid '
                stop 1
            end if
            !! nullify pointer to ljgrid by default
            nullify ( pdest % array_ljgrid )

            !write(*,*) 'pdest associated ', associated ( pdest )

            !! copy the number of grids and the maximum number of atoms
            pdest % nb_ljgrid = this % nb_ljgrid
            pdest % max_pos_atom = this % max_pos_atom
            !write(*,*) 'nb_ljgrid ', this% nb_ljgrid

            ! depends on option, if local copy, allocate its own
            if ( allocated ( this % array_position_atom_lj ) ) then
              !write(*,*) 'allocated array_position_atom_lj in source with size ', this % nb_ljgrid
              allocate ( pdest % array_position_atom_lj ( pdest % nb_ljgrid ), stat = ierr )
              if ( ierr /= 0 ) then
                write(*,*) 'ERROR In allocation of array_position ierr= ', ierr
                stop 1
              end if
               ! clean initialization for atom position
               do i = 1, pdest % nb_ljgrid
                   nullify ( pdest % array_position_atom_lj ( i ) % p_position_atom )
                   pdest % array_position_atom_lj ( i ) % nb_pos_atom = 0
               end do
            !else
            !  write(*,*) 'source array_position is not allocated'
            end if

            ! for grid, assign the pointers only ( no copy of the grids and array needed )
            !write(*,*) 'test original '
            !write(*,*) 'associated array_ljgrid dest ', associated ( pdest % array_ljgrid )
            !write(*,*) 'associated array_ljgrid source ', associated ( this % array_ljgrid )
            !write(*,*) 'assign pointer'
            pdest % array_ljgrid => this % array_ljgrid
            !write(*,*) 'grid ok ?'
            !write(*,*) 'associated array_ljgrid source ', associated ( this % array_ljgrid )
            !write(*,*) 'associated array_ljgrid dest ', associated ( pdest % array_ljgrid )

            ! for atom_position, if full copy is needed
            !write(*,*) 'atom position make a full copy if needed '

            !! can test source or original, same only need a full copy if this function is called
            if ( allocated ( this % array_position_atom_lj ) ) then
              do i = 1, pdest % nb_ljgrid
                !write(*,*) 'i ', i
                !! allocated during tyhe allocation of the array
                !write(*,*) 'allocated pdest % array_position_atom_lj ', allocated ( pdest % array_position_atom_lj )
                !! test is wrong here, undefined
                !write(*,*) 'associated pdest % array_position_atom_lj(i) ', &
                !    associated ( pdest % array_position_atom_lj(i) % p_position_atom )

                ! allocate the new object and makes the copy of data
                ppos_atom => pdest % array_position_atom_lj ( i )
                !call copy_position_atom ( this % array_position_atom_lj ( i ), pdest % array_position_atom_lj ( i ) )
                call copy_position_atom ( this % array_position_atom_lj ( i ), ppos_atom )


                !write(*,*) 'after copy_position_atom is allocated ', allocated ( pdest % array_position_atom_lj )
                !write(*,*) 'associated pdest % array_position_atom_lj(i) % p_position_atom ', &
                !    associated ( pdest % array_position_atom_lj(i) % p_position_atom )
                !associated ( this % array_position_atom_lj ( i ) % p_position_atom
              end do
            end if

            !write(*,*)
            !write(*,*) 'Exit copy_ljgrid '
            !write(*,*)

        end subroutine copy_ljgrid

        !> Delete lj grid
        !! @param this : instance of \ref type_ljgrid
        !! @param opt_copy_omp : optionaly do not delete the associated \ref mod_setofgrid::sogrid "sogrid" and \ref mod_flexibility::flexible "flexible". Default delete all
        subroutine deallocate_ljgrid ( this, opt_copy_omp )
            IMPLICIT NONE
            type ( type_ljgrid ) :: this
            logical, optional, intent ( in ) :: opt_copy_omp

            logical :: copy_omp

            integer :: i, ierr

            !write(*,*) 'Deallocate LJ grid with nb_ljgrids ', this % nb_ljgrid
            if ( present ( opt_copy_omp ) ) then
                copy_omp = opt_copy_omp
            else
                copy_omp = .false.
            end if

            !! deallocate grids and position atom in arrays
            !! if copy_omp, does not delete the uhbd grids, only nullify the pointers
            !write(*,*) 'delete array_ljgrid '
            if ( copy_omp .eqv. .false. ) then

                !write(*,*) 'copy_omp ', copy_omp

                if ( associated ( this % array_ljgrid ) ) then
                  do i = 1, this % nb_ljgrid

                    ! logic, should be associated, but test need for first test
                    if ( is_associated ( this % array_ljgrid( i ) ) .eqv. .true. ) then

                        !write(*,*) 'will delete the uhbd grid i ', i
                        !write(*,*) 'check again is associated ', &
                        !    associated ( this % array_ljgrid( i ) % p_grid % pgrid_real)

                        call delete_grid ( this % array_ljgrid( i ) )
                    else
                        write(*,*) 'p_grid uhbd was not associated'
                    end if

                  end do

                  ! deallocate the arrays themself
                  deallocate ( this % array_ljgrid, stat = ierr )
                  if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In deallocation if array_ljgrid '
                    stop 1
                  end if

                end if ! if array ljgrid
            !else
            !    write(*,*) 'copy omp true, do not delete the array of ljgrids'
            end if

            if ( allocated ( this % array_position_atom_lj ) ) then
              !write(*,*) 'delete array_position_atom_lj nb_ljgrid ', this % nb_ljgrid
              do i = 1, this % nb_ljgrid

                if ( associated ( this % array_position_atom_lj ( i ) % p_position_atom ) ) then

                    !! by default, copy the full array of atom_position
                    !write(*,*) 'will delete the p_position_atom '
                    deallocate ( this % array_position_atom_lj (i ) % p_position_atom, stat = ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'ERROR In dealocation of p_pos_atom X % % p_pso_atom '
                        stop 1
                    end if
                    nullify ( this % array_position_atom_lj ( i ) % p_position_atom )
               else
                    write(*,*) 'p_position_atom was not associated'
               end if

              end do

              ! deallocate the arrays itself
              deallocate ( this % array_position_atom_lj, stat = ierr )
              if ( ierr /= 0 ) then
                write(*,*) 'ERROR In deallocation if array_position_atom_lj '
                stop 1
              end if
            end if ! end if array_position_atom_lj allocated

            ! all case nullify the pointer
            nullify ( this % array_ljgrid )
            ! new added ,s hould be ok
            this % nb_ljgrid = 0
            this % max_pos_atom = 0

        end subroutine deallocate_ljgrid



        !!!!!!!!!!!!!!!!!!!!    methods functions  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !> Set one lennard-jones grid, assign to the nb element of the array
        !!
        !! nb indicates which place in array, filename, only the filename
        subroutine set_ljgrid ( this, filename, nb, fct  )
            IMPLICIT NONE
            type (  type_ljgrid ) :: this
            character, intent ( in ) :: filename*(*)
            integer, intent ( in ) :: nb
            real ( kind=8 ) :: fct

            integer :: i, ierr
            ! mod_grid does not support *(*)
            character*128 :: filename_tmp

            !write(*,*) 'Entry set_lj_grid filename ', filename
            !write(*,*) 'nb ', nb

            filename_tmp = trim ( filename )

            ! test first if first element, need to allocate the array
            if ( associated ( this % array_ljgrid ) .eqv. .false. ) then
                allocate ( this % array_ljgrid ( this % nb_ljgrid ), stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In allocation of array_ljgrid ierr= ', ierr
                    stop 1
                end if

                do i = 1, this % nb_ljgrid
                    nullify ( this % array_ljgrid (i) % p_grid )
                end do
            end if

            ! read the grid
            call read_grid ( this % array_ljgrid ( nb ), filename_tmp, 5, fct )

            ! for test
#ifdef DEBUG
            call info_grid ( this % array_ljgrid ( nb ) )
#endif

        end subroutine set_ljgrid

        !> Set one atom_type, assign to element nb in array
        !!
        !! First version, read the pdb and extract the psotiosion of the atom
        !!
        !! later should accept array from read_pdb, prepare_atom_protein
        subroutine set_atom_lj ( this, filename, nb, atom_type )
            IMPLICIT NONE
            type ( type_ljgrid ) :: this
            character, intent ( in ) :: filename*(*)
            character, intent ( in ) :: atom_type*(*)
            integer, intent ( in ) :: nb

            ! copy from read_pdb, problems with r_criter for sure ! see later
            real ( kind=8 ), dimension(3) :: xv

            character*128 string
            integer :: io, ierr, nat, number_line, i, iusas
            logical :: ex

            xv = 0.d0
            iusas = 93
            !write(*,*) 'Entry set_atom_lj nb ', nb

            if ( allocated ( this % array_position_atom_lj ) .eqv. .false. ) then
                allocate ( this % array_position_atom_lj ( this % nb_ljgrid ), stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In allocation of array_ljgrid ierr= ', ierr
                    stop 1
                end if

                do i = 1, this % nb_ljgrid
                    nullify ( this % array_position_atom_lj (i) % p_position_atom )
                end do
            end if

            inquire ( file=trim(filename), EXIST=ex )
            if ( .NOT.ex ) then
                write (*,*) "pdb file does not exist or file name is missing ", trim(filename)
                stop 1
            end if


            open (iusas,file=filename,form='formatted',status='old')
            nat = 0
            number_line = 0
            !count number of line
            read ( iusas, '(a)', iostat=io ) string
            if ( io < 0 ) then
                write( *,* ) 'file is empty', filename
                return
            endif
            ! fine only 2 characters printed
            !write(*,*) 'atom_type!!',atom_type,'!!'

            do while ( io == 0 )
                if ( (string(1:4).EQ.'ATOM').or.(string(1:6).EQ.'HETATM') ) then
                    ! test atom_type
                    !write(*,*) 'string14:16 !!!',trim(string(14:16)),'!!!'
                    !write(*,*) 'atom_type   !!!',trim(atom_type),'!!!'
                    if ( trim( string(14:16) ) == trim(atom_type) ) then
                        nat = nat+1
                    endif
                endif
                number_line = number_line + 1
                read ( iusas, '(a)', iostat=io ) string
            enddo

            if ( nat == 0 ) then
                write(*,*) 'number of atom is 0 !!'
                return
            endif

            write( *, fmt='("Read file for lennard-jones : ", a)' ) trim(filename)
            write( *, fmt='("Number of atoms in pdb :      ",i8)' ) nat

            allocate ( this % array_position_atom_lj ( nb )  % p_position_atom (3,nat), stat=ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'cannot allocate tmp_atom_position with ', nat, 'atom'
            endif
            this % array_position_atom_lj ( nb )  % p_position_atom = 0.d0

            rewind( iusas )
            io=0
            nat=1
            ! read first line, otherwise make one turn more
            read ( iusas, '(a)', iostat=io ) string

            ! re-read the file and save only position and atom
            do while ( io == 0 )

                if ( (string(1:4).EQ.'ATOM').or.(string(1:6).EQ.'HETATM') ) then
                    ! test atom_type
                    if ( trim( string(14:16) ) .eq. trim(atom_type) ) then
                        !read( string, fmt='(4X,A7,1X,2A1,16X,3F8.3)' ) atom_nb_string, an, an2, xv
                        !write(*,*) 'string(33:60) !!',string(33:60),'!!'
                        read ( string, fmt='(30X,3F8.3)' ) xv
                        !write(*,*) 'xv ', xv
                        this % array_position_atom_lj ( nb )  % p_position_atom(:,nat) = xv
                        nat = nat+1
                    end if

                end if
                read ( iusas, '(a)', iostat=io ) string
            end do

            close ( iusas )

            !write(*,*) 'this % array_position_atom_lj ( nb )  % p_position_atom 1 ', &
            !    this % array_position_atom_lj ( nb )  % p_position_atom(:,1)
            ! nat always one more
            !write(*,*) 'this % array_position_atom_lj ( nb )  % p_position_atom nat-1 ', &
            !    this % array_position_atom_lj ( nb )  % p_position_atom(:,nat-1)

            this % array_position_atom_lj ( nb ) % nb_pos_atom = nat -1
            !write(*,*) 'nb_pos_atom ', nat - 1

            ! try to update the max number of atom
            if ( this % max_pos_atom < (nat - 1) ) then
                this % max_pos_atom = nat - 1
                !write(*,*) 'updtae this max_pos_atom ', this % max_pos_atom
            !else
            !    write(*,*) 'no need to update max_pos_atom ', this % max_pos_atom
            end if
            !write(*,*) 'nb_pos_atom ', nat - 1

        end subroutine set_atom_lj


        subroutine print_info_ljgrid ( this )
            IMPLICIT NONE
            type ( type_ljgrid ) :: this

            integer :: i

            write(*,*)
            write(*,*) 'Info Lennard-jones '
            write(*,*) 'number of atom type and grid ', this % nb_ljgrid
            do i = 1, this % nb_ljgrid
                write(*,*) 'associated p_position_atom ', associated ( this % array_position_atom_lj(i) % p_position_atom )
                write(*,*) 'nb_pos_atom ', this % array_position_atom_lj(i) % nb_pos_atom
            end do

        end subroutine print_info_ljgrid

end module mod_ljgrid
