!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_onecomplexe.f90
!! Define a single "line" for complexes or trajectory file

!> Module to define \ref one_complexe
!! Theses objects are stored in \ref mod_record::record "record"
MODULE mod_onecomplexe

    !> Objects stored in the list ( for complexes ) or in the array ( for trajectory )
    !!
    !! Allow variable number of fields :\n
    !! integer terms ( trajectory number, step number, protein number and conformation) in \ref mod_onecomplexe::one_complexe::int_ocompl "int_compl"\n
    !! energy terms (  electrostatic, el. desolvation, apolar desol. (hydrophobic), repulsive lennard-jones ) in \ref mod_onecomplexe::one_complexe::real_ocompl "real_ocompl". They can be sum up ( el1-2 + el2-1 ) or writen separately\n
    !! other real terms ( occurence, average total energy and standard deviation of the standard energy ) for complexes \n
    !! or ( time ) for trajectory in \ref mod_onecomplexe::one_complexe::other_real_ocompl "other_real_ocompl"
    !!
    !! Print box informations when periodic boundary conditions are used\n
    !! Can write the data in binary or ascii format
    type one_complexe

    !> energy terms, variable number but keep always the order :
    !! electrostatic, el desolvation, apolar desol (hydrophobic), repulsive lennard-jones, lennard-jones, image-charge
    real ( kind=8 ), dimension(:),  allocatable :: real_ocompl
    !> intger to be stored: 1:ntraj, 2:nstep, 3:nprotein, 4: nconf
    integer, dimension(:), allocatable :: int_ocompl

    !> Aditional terms: for trajectory time, for complexes occurence, average energy, deviation of energy terms\n
    !! The standard deviation in only computed when the results are writen to disk, only the sum of the square is updated during the computation
    real ( kind=8 ), dimension(:), allocatable :: other_real_ocompl
    !> box info for sdamm, keep trace when a protein change box with PBC. Used only by tools for computing diffusion.
    integer, dimension(:), allocatable :: box
    !> independent variable used during simulation, used for computing the standart deviation, temporary never printed or read
    !! \todo check if used, in main array now
    real ( kind=8 ) :: sum_square_energy

    !> coordinates and orientation, it has a fixed size and is mandatory
    !! @{
    real ( kind=8 ), dimension( 3 ) :: position
    real ( kind=8 ), dimension( 3 ) :: orient_x
    real ( kind=8 ), dimension( 3 ) :: orient_y
    !> @}

    !> Keep track of the number of activated terms for all arrays
    !! all in record parent now, identical for all the records
    !! @{
    !integer :: nb_int, nb_real, nb_other_real
    !> @}
    !> if true, use the box information with PBC, could move into parent as well
    logical :: ibox

    end type one_complexe
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    !! not used in other modules, but here may clarify the use of the function intented to be used normally
    !! most of the functions can be called from a subroutine in mod_record, see *_wrap functions
    public :: new_onecomplexe

    contains

    ! sufficent to send bit_energy to get the size
    ! problem of summing energy is done completely by one onecomplexe
    ! bit energy and integer stored in record / or complexe and trajectories

    !> Allocate an instance of one_complexe\n
    !! Need only size of variable fields, optional ibox for sdamm
    !! @param this : instance of one_complexe
    !! @param nb_int : size of \ref mod_onecomplexe::one_complexe::int_ocompl "int_ocompl"
    !! @param nb_real : size of  \ref mod_onecomplexe::one_complexe::real_ocompl "real_ocompl"
    !! @param nb_other_real : size of \ref mod_onecomplexe::one_complexe::other_real_ocompl "other_real_ocompl"
    !! @param opt_ibox : optionaly activate the box information
    subroutine new_onecomplexe ( this, nb_int, nb_real, nb_other_real, opt_ibox ) !, nb_type not needed nb_real shoud be correct )
        IMPLICIT NONE
        type ( one_complexe ) :: this
        integer, intent ( in ) :: nb_int, nb_real, nb_other_real
        logical, optional :: opt_ibox

        integer :: ierr

        this % position = 0.d0
        this % orient_x = 0.d0
        this % orient_y = 0.d0
        this % sum_square_energy = 0.d0

        this % ibox = .false.
        if ( present ( opt_ibox ) ) then
           if ( opt_ibox .eqv. .true. ) then
            !write(*,*) 'ibox true allocate array box '
            this % ibox = .true.
            allocate ( this % box ( 3 ), stat = ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'error cannot allocate memory bor ibox'
                stop 1
            end if
            this % box = 0
          endif
        end if

        !this % nb_int = nb_int
        !write(*,* ) 'this nb int', this % nb_int
        allocate( this % int_ocompl ( nb_int ), stat=ierr )
        if ( ierr /= 0 ) then
             write (*,*) 'ERROR Couldn t allocate memory for int_ocompl ', nb_int
             stop 1
        endif
        this % int_ocompl = 0

        !this % nb_real = nb_real
        !write(*,* ) 'this nb real', this % nb_real
        allocate( this % real_ocompl ( nb_real ), stat=ierr )
        if ( ierr /= 0 ) then
             write (*,*) 'ERROR Couldn t allocate memory for real_ocompl ', nb_real
             stop 1
        endif
        this % real_ocompl = 0.d0

        !this % nb_other_real = nb_other_real
        allocate( this % other_real_ocompl ( nb_other_real ), stat=ierr )
        if ( ierr /= 0 ) then
             write (*,*) 'ERROR Couldn t allocate memory for other_real_ocompl ', nb_other_real
             stop 1
        endif
        this % other_real_ocompl = 0.d0

        !write(*,*) 'entry new_onecomplexe'
        !write(*,*) 'nb_int, nb_real, nb_other_real', nb_int, nb_real, nb_other_real
    end subroutine new_onecomplexe

    !> Delete onecomplexe
    !!
    !! Deallocate all arrays
    !! @param this : instance of one_complexe
    subroutine delete_onecomplexe ( this )
            IMPLICIT NONE
            type ( one_complexe ) :: this

            integer :: ierr

            !! reinitialize "static variable"
            this % position =(/0.d0, 0.d0, 0.d0/)
            this % orient_x =(/0.d0, 0.d0, 0.d0/)
            this % orient_y =(/0.d0, 0.d0, 0.d0/)
            this % sum_square_energy = 0.d0

            if ( allocated ( this % box ) ) then
               deallocate ( this % box, stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR Couldn t deallocate memory for box'
                    stop 1
                endif
                this % ibox = .false.
            endif

            if ( allocated( this % int_ocompl ) ) then
                deallocate ( this % int_ocompl, stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR Couldn t deallocate memory for int_ocompl'
                    stop 1
                endif
            endif

            if ( allocated( this % real_ocompl ) ) then
                deallocate ( this % real_ocompl, stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR Couldn t deallocate memory for real_ocompl'
                    stop 1
                endif
            endif

            if ( allocated( this % other_real_ocompl ) ) then
                deallocate ( this % other_real_ocompl, stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR Couldn t deallocate memory for real_ocompl'
                    stop 1
                endif
            endif

        end subroutine delete_onecomplexe

        ! could add sum_square_energy, easier to update only if needed

        !> Set values of an one_complexe instance
        !!
        !! Should be allocated before\n
        !! Input array from record have a fixed size, only the relevent informations ( gieven by the bits ) are stored in internal array of one_complexe
        !! @param this : instance of one_complexe
        !! @param position, box : position of the complex and box information ( default ( 0, 0, 0) )
        !! @param orient_x, orient_y : orientation of the complex
        !! @param array_int_compl : to store in \ref mod_onecomplexe::one_complexe::int_ocompl "int_ocompl"
        !! @param array_real_compl : to store in \ref mod_onecomplexe::one_complexe::real_ocompl "real_ocompl"
        !! @param other_real : to store in \ref mod_onecomplexe::one_complexe::other_real_ocompl "other_real_ocompl"
        !! @param size_other_real : size of array other_real
        !! @param bit_integer, bit_energy : bits information
        !! @param opt_sum : if the sum of energy terms is required
        !! @param nb_to_sum : equal to the number of types in the simulation
        subroutine set_onecomplexe ( this, position, box, orient_x, orient_y, &
                                     array_int_compl, array_real_compl, other_real, size_other_real, &
                                     bit_integer, bit_energy, opt_sum, nb_to_sum )
            IMPLICIT NONE
            type ( one_complexe ) :: this
            integer, dimension ( 3 ) :: box
            real ( kind=8 ), dimension( 3 ), intent ( in ) :: position, orient_x, orient_y
            real ( kind=8 ), dimension( : ), intent( in ) :: array_real_compl
            real ( kind=8 ), dimension( : ), intent( in ) :: other_real
            integer, dimension ( : ), intent ( in ) :: array_int_compl
            logical :: opt_sum
            integer, intent ( in ) :: size_other_real, nb_to_sum

            integer ( kind=2 ) :: bit_energy, bit_integer

            integer :: counter, i, j

            this % position = position
            this % orient_x = orient_x
            this % orient_y = orient_y
            !! should have been allocated here
            if ( this % ibox ) then
                !write(*,*) 'set one complexe this % box', box
                this % box = box
            end if

            !write(*,*)
            !write(*,*) 'entry set onecomplexe nb_to_sum ', nb_to_sum
            !write(*,*) 'opt_sum ', opt_sum
            !write(*,*) 'array real', array_real_compl
            !write(*,*) 'opt_sum ', opt_sum
            !write(*,*) 'size one_complexe: integer, real, other ', size ( this % int_ocompl ), &
            !            size ( this % real_ocompl ), size( this % other_real_ocompl )

            ! here need explicitly the size this % nb_other_real
            ! sender always fixed to 3
            !this % other_real_ocompl = other_real (1 : this % nb_other_real)
            this % other_real_ocompl = other_real ( 1: size_other_real )

            ! need to initialize for trajectory and in case of opt_sum == 1 only increment the data
            this % real_ocompl = 0.d0

            counter = 1
            ! intger to be stored: 1:nrun, 2:nstep, 3:nprotein, 4: nconf ...
            do i = 0, BIT_SIZE( bit_integer ) - 1
               if ( BTEST( bit_integer, i ) ) then
                    this % int_ocompl ( counter ) = array_int_compl ( i + 1 )
                    counter = counter + 1
               end if
            end do
            !write(*,*) 'counter  integer',counter
            ! write(*,*) 'size oreal', size ( this % real_ocompl )

            ! total energy always present, always the first one
            this % real_ocompl ( 1 ) = array_real_compl ( 1 )
            counter = 2
            do i = 0, BIT_SIZE( bit_energy ) - 1
               if ( BTEST( bit_energy, i ) ) then

                    !write(*,*) 'bit_energy set to one for i= ', i
                    if ( opt_sum ) then

                        !! need a loop, with sdamm & opt_sum nb_to_sum == 1 if one type
                        do j = 1, nb_to_sum
                            !write(*,*) 'read data ', array_real_compl( i * nb_to_sum + 1 + j ), i, j
                            this % real_ocompl ( counter ) = this % real_ocompl ( counter ) + &
                                                               array_real_compl( i * nb_to_sum + 1 + j )
                        end do
                        counter = counter + 1

                    else
                        ! same need a loop, sdamm nb_types, sda_2protein 2 ( nb_types also )
                        do j = 1, nb_to_sum
                            !write(*,*) 'counter i * nb_to_sum + 1 + j', counter , i * nb_to_sum + 1 + j
                            !write(*,*) 'value ', array_real_compl ( i * nb_to_sum + 1 + j )
                            this % real_ocompl ( counter ) = array_real_compl ( i * nb_to_sum + 1 + j )
                            counter = counter + 1
                        end do

                    end if

               end if
            end do

        ! set one complexe
        !write(*,*) 'onecomplex set with nb_to_sum=', nb_to_sum
        !write(*,*) 'with energy ', this % real_ocompl
        !write(*,*) 'with array integer ', this % int_ocompl
        !write(*,*) 'with position ', (this % position(m),m=1,3)
        end subroutine set_onecomplexe

        !> Return arrays of data of fixed size
        !!
        !! Sort of inverse of set_onecomplexe, allow to extract easily the data\n
        !! bit_integer and bit_energy will provide the information to select the returned data\n
        !! Extract data from this one_complexe, order and size are fixed ( ntraj, nstep, nprot, nconf...)
        !!   value is null by default, if not present in one_complexe
        !!
        !! Easier to use record % return_compatible_array_wrapper.\n
        !!
        !! \todo add box information
        !!
        !! Need to extend with nb_to_sum, keep same variable name
        !! May add aprameter for size of other_real
        !!
        !! @param this : instance of one_complexe
        !! @param array_int, array_real, other_real : array filled by the function
        !! @param size_other_real : size of array other_real
        !! @param bit_integer, bit_energy : information wanted in return of the function
        !! @param opt_sum : if the energy terms should be added are kept separate
        !! @param nb_to_sum : equal the number of types of solutes
        subroutine return_compatible_array ( this, array_int, array_real, other_real, size_other_real, &
                                             bit_integer, bit_energy, opt_sum, nb_to_sum )
            IMPLICIT NONE
            type ( one_complexe ) :: this
            real ( kind=8 ), dimension( : ), intent ( out ) :: array_real, other_real
            integer, dimension ( : ), intent ( out ) :: array_int
            integer ( kind=2 ), intent ( in ) :: bit_energy, bit_integer
            integer, intent ( in ) :: size_other_real, nb_to_sum
            logical, intent ( in ) :: opt_sum

            call return_array_int ( this, array_int, bit_integer )
            call return_array_real ( this, array_real , bit_energy, opt_sum, nb_to_sum )
            ! just copy other_real, may need a check for the size...
            other_real ( 1: size_other_real ) = this % other_real_ocompl


        end subroutine return_compatible_array

        !> Same as return compatible array, only for \ref mod_onecomplexe::one_complexe::int_ocompl "int_ocompl"
        subroutine return_array_int ( this, array_int , bit_integer )
            IMPLICIT NONE
            type ( one_complexe ) :: this
            integer, dimension ( : ), intent ( out ) :: array_int
            integer ( kind = 2 ), intent ( in ) :: bit_integer

            integer :: i, counter

            array_int = 0
            counter = 1
            do i = 0, size ( array_int ) - 1
                !!can only give dara available in the source
                if (  BTEST( bit_integer, i ) ) then
                     array_int ( i + 1 ) = this % int_ocompl ( counter )
                    counter = counter + 1
                end if
            end do

        end subroutine return_array_int

        !> Same as return_compatible_array, only for \ref mod_onecomplexe::one_complexe::real_ocompl "real_ocompl"
        !!
        !! Extended with nb_types. Convenient if allocate the array inside.\n
        !!
        !! The array should have the correct size, safer if allocated with mod::record::get_nbterms_energy_compatible.
        !! @param this : instance of one_complexe
        !! @param array_real : array to fill, store only valid data
        !! @param bit_energy : indicates which information are stored in one_complexe object
        !! @param opt_sum : indicates if the sum of the terms must be given or all
        !! @param nb_to_sum : number of types of grids
        subroutine return_array_real ( this, array_real , bit_energy, opt_sum, nb_to_sum )
            IMPLICIT NONE
            type ( one_complexe ) :: this
            real ( kind=8), dimension ( : ), intent ( out ) :: array_real
            integer ( kind = 2 ), intent ( in ) :: bit_energy
            logical, intent ( in ) :: opt_sum
            integer, intent ( in ) :: nb_to_sum

            integer :: counter, i, j

            !write(*,*) 'Entry return_array_real nb_to_sum ', nb_to_sum
            !write(*,*) 'Size array_real entry ', size ( array_real )

            ! init at  the base
            array_real = 0.d0

            ! will need 2 counters to be sure of the convertion
            array_real ( 1 ) = this % real_ocompl ( 1 )
            counter = 1

            do i = 0, BIT_SIZE( bit_energy ) - 1

                !write(*,*) 'Bit i ', i, BTEST( bit_energy, i )
                !! i may be larger than 16 ?? no
                if ( BTEST( bit_energy, i ) ) then

                    ! only one bit by interaction, should check if more data to copy
                    if ( opt_sum .eqv. .false. ) then
                        ! copy all the terms
                        do j = 1, nb_to_sum
                            array_real ( nb_to_sum * i + 1 + j ) = this % real_ocompl ( counter + j )
                        end do
                        counter = counter + nb_to_sum
                    else
                        ! only one term for each interaction
                        array_real ( i + 2 ) = this % real_ocompl ( counter )
                        counter = counter + 1
                    end if

                end if

            end do

            !write(*,*) 'array_real out ', array_real
        end subroutine return_array_real

        !> Update statistics for complexes : occurency, average and std deviation
        !!
        !! By default increase occurency by one with the last energy\n
        !! if used from merge_trajectory, take into account correctly the previous data ( sum of the square of the energy )\n
        !! should improve, general occurency, average.,.. and only one loop
        !! @param this : instance of \ref one_complexe
        !! @param previous : pointer to the original one_complexe to be merged in, if not associated update this
        !! @param opt_other_real : optional update the other_real terms.
        subroutine update_statistics ( this, previous, opt_other_real )
            IMPLICIT NONE
            type ( one_complexe ) :: this
            type ( one_complexe ), pointer :: previous
            real ( kind=8 ), dimension ( : ), optional :: opt_other_real
            logical :: opt_update_complexe

            real ( kind=8 ) :: occurency1, occurency2

            ! should be default for normal update
            opt_update_complexe = .false.

            if ( present ( opt_other_real ) ) then
                ! send other_real
                if ( size ( opt_other_real ) == 4 ) then
                    opt_update_complexe = .true.

                ! if send only the total energy, simple update, default value
                else if ( size ( opt_other_real ) == 1 ) then
                    opt_update_complexe = .false.

                else
                    write(*,*) 'ERROR In argument of update statistics'
                    STOP 1
                end if
            end if

            ! a previous one_complexe is explicitely given, comes from merge_complexe or other tool
            if ( associated ( previous ) ) then

                occurency1 = this % other_real_ocompl ( 1 )
                occurency2 = previous % other_real_ocompl ( 1 )
                !write(*,*) 'occurencies case of previous ', occurency1, occurency2
                this % other_real_ocompl ( 1 ) =   occurency1 + occurency2
                this % other_real_ocompl ( 2 ) = ( this % other_real_ocompl ( 2 ) * occurency1 &
                                      + previous % other_real_ocompl ( 2 ) * occurency2 ) / this % other_real_ocompl ( 1 )

                this % sum_square_energy = this % sum_square_energy + previous % sum_square_energy

            ! if previous node is not associated make an update of this onecomplexe
            else
                ! make a simple update, just add the last energy opt_other_real == array_energy ( 1 )
                if ( opt_update_complexe .eqv. .false. ) then
                	!!write(*,*) 'simple update'
                    occurency1 = this % other_real_ocompl ( 1 )
                    !!write(*,*) 'simple update'
                    this % other_real_ocompl ( 1 ) = occurency1 + 1
                    ! update average energy, can be done directly. occurence is minimum 2 at this point
                    this % other_real_ocompl ( 2 ) = ( this % other_real_ocompl ( 2 ) * occurency1 + &
                                                       opt_other_real ( 1 ) ) / this % other_real_ocompl ( 1 )
                    !this % other_real_ocompl ( 3 ) =  this % other_real_ocompl ( 3 ) + &
                    !                                ( opt_other_real ( 1 ) )**2
                    ! in this case opt_other_real ( 1 ) == array_energy ( 1 )
                    this % sum_square_energy = this % sum_square_energy + ( opt_other_real ( 1 ) )**2


                ! makes an update from a previous other_real and previous sum_square
                ! this case is strange from complexe to array, why not use previous ??
                ! opt_other_real ( 4 ) == sum_square energy
                else
                    occurency1 = this % other_real_ocompl ( 1 )
                    occurency2 = opt_other_real ( 1 )
                    !write(*,*) 'occurencies case of opt_other_real ', occurency1, occurency2
                    this % other_real_ocompl ( 1 ) = occurency1 + occurency2
                    this % other_real_ocompl ( 2 ) = ( this % other_real_ocompl ( 2 ) * occurency1 &
                                      + opt_other_real ( 2 ) * occurency2 ) / this % other_real_ocompl ( 1 )
                    !this % other_real_ocompl ( 3 ) =  this % other_real_ocompl ( 3 ) + &
                    !                                ( opt_other_real ( 3 ) )**2

                    this % sum_square_energy = this % sum_square_energy + opt_other_real ( 4 )
                end if

            end if

        end subroutine update_statistics

        !> Test function, convenient for debug
        !! @param this : instance of \ref one_complexe
        subroutine print_onecomplexe ( this )
            IMPLICIT NONE
            type ( one_complexe ) :: this

            character*256 :: string
            character*64 :: formatstr

            formatstr=''
            string=''

            write(*,*) 'print_onecomplexe It is a test, should not be called in runs'

            ! write format, here test, format is done only once
            write(formatstr,'("(", i1, "i8,9f9.3,", i2, "es12.4,", i1, "es12.4)")') &
                             !this % nb_int, this % nb_real, this % nb_other_real
                             size ( this % int_ocompl ), size ( this % real_ocompl ), size ( this % other_real_ocompl )

            !write(*,*) 'format:', formatstr
            write ( string, formatstr ) this % int_ocompl, this % position, &
                    this % orient_x, this % orient_y, this % real_ocompl, &
                    this % other_real_ocompl
            write(*,*) string

        end subroutine print_onecomplexe

        !> Make string, send back a string, only for ascii.
        !!
        !! Give the format string as input, computed only once\n
        !! Binary need an other function
        !! @param this : instance of \ref one_complexe
        !! @param string : return value
        !! @param format_str : format of one line
        subroutine make_ascii_string ( this, string, format_str )
            IMPLICIT NONE
            type ( one_complexe ) :: this
            character, intent( out ) :: string*(*)
            character, intent (in) :: format_str*(*)

            !integer :: i, offset, counter
            !write(*,*) 'Make ascii string, new parameter format_str '
            !write(*,*) 'format_str: ', format_str
            !write(*,*) 'test this % nb_real ', this % nb_real
            !write(*,*) 'test this % nb_int ', this % nb_int
            !write(*,*) 'size this % real_ocompl ', size ( this % real_ocompl )
            !write(*,*) 'size this % other_real_ocompl ', size ( this % other_real_ocompl )
            !write(*,*) 'this % box ', this % box

            ! creates the line
            if ( this % ibox ) then
                write ( string, format_str ) this % int_ocompl, this % position, &
                       this % orient_x, this % orient_y, this % real_ocompl, &
                       this % other_real_ocompl, this % box
            ! normal do not print box info
            else
                write ( string, format_str ) this % int_ocompl, this % position, &
                    this % orient_x, this % orient_y, this % real_ocompl, &
                    this % other_real_ocompl
            end if

            !write(*,*) 'string:',string,'!!!'
        end subroutine make_ascii_string

        !> reading file is done by record or derived, same logic than for writting
        !! for binary should change, for both reading and writting....
        subroutine read_ascii_string ( this, string, format_str )
            IMPLICIT NONE
            type ( one_complexe ) :: this
            character, intent( in ) :: string*(*)
            character, intent( in ) :: format_str*(*)

            !write(*,*) 'read one complexe format_str ', trim ( format_str )
            !write(*,*) 'string:',trim(string)
            !write(*,*) 'format_str', format_str
            !write(*,*) 'size int_ocompl ', size ( this % int_ocompl )
            !write(*,*) 'size int_oreal ', size ( this % real_ocompl )
            !write(*,*) 'size int_oreal ', size ( this % other_real_ocompl )
            !write(*,*) 'size box ', size ( this % box )

            if ( this % ibox ) then
                read ( string, format_str ) this % int_ocompl, &
                   this % position, this % orient_x, this % orient_y, &
                   this % real_ocompl, this % other_real_ocompl, this % box
            else
                read ( string, format_str ) this % int_ocompl, &
                   this % position, this % orient_x, this % orient_y, &
                   this % real_ocompl, this % other_real_ocompl
            end if
            !write(*,*) 'in reading other_real ', this % other_real_ocompl !(3)
        end subroutine read_ascii_string

        !> For reading binary
        subroutine read_binary_string ( this, io_file, box_info, io_err )
            IMPLICIT NONE
            type ( one_complexe ) :: this
            integer, intent ( in ) :: io_file
            integer, intent ( out ) :: io_err
            logical, intent ( in ) :: box_info

            ! here problem, maybe out of the trajectory, this is not defined, use external variable
            if ( box_info ) then
              read ( io_file, iostat = io_err ) this % int_ocompl, &
                   this % position, this % orient_x, this % orient_y, &
                   this % real_ocompl, this % other_real_ocompl, this % box
            else
              read ( io_file, iostat = io_err ) this % int_ocompl, &
                   this % position, this % orient_x, this % orient_y, &
                   this % real_ocompl, this % other_real_ocompl
            end if
            !write(*,* ) 'after reading one line binary io_err box', io_err !, this % box
            !write(*,* ) 'other_real ', this % other_real_ocompl
        end subroutine read_binary_string

        !> For writing binary
        subroutine write_binary_string ( this, io_file, io_err )
            IMPLICIT NONE
            type ( one_complexe ) :: this
            integer :: io_file, io_err

            !write(*,*) 'binary string ibox ', this % ibox, this % box
            if ( this % ibox ) then
               write ( io_file, iostat = io_err ) this % int_ocompl, &
                   this % position, this % orient_x, this % orient_y, &
                   this % real_ocompl, this % other_real_ocompl, this % box
            else
              write ( io_file, iostat = io_err ) this % int_ocompl, &
                   this % position, this % orient_x, this % orient_y, &
                   this % real_ocompl, this % other_real_ocompl
            end if
            !write(*,* ) 'after writing binary io_err ', io_err
            !write(*,* ) 'other_real ', this % other_real_ocompl
        end subroutine write_binary_string

END MODULE mod_onecomplexe
