!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file mod_par_zig.f90
!!
!! Thread-safe random generator with the ziggurat method

!> \defgroup RNG RNG : Random Number Generator
!!
!! Random Number Generator in SDA
!!
!! documentation included in mod_par_zig.f90
!!
!! SDA has two possible implementations for generating random numbers:
!!  - the same method as that was used in SDA 6 : this was kept for comparison with the original code during developement, but it cannot be used within an OpenMP environment
!!  - a Ziggurat thread safe method (see Marsaglia, G. & Tsang, W.W., J. Statist. Software, 5, 8, (2000).) : a method developed for OpenMP, each thread has its own sequence of random numbers ( \ref mod_par_zig.f90 ).
!!
!! The implementation used is chosen at compile time. If OpenMP is activated the Ziggurat method is chosen, otherwise the original SDA 6 routine is used.\n
!! The thread-safe functions are declared in \ref maths_thread_safe.f90 .\n
!! The original implementation is in \ref randgen.f . The critical section is commented in the source code.
!!
!! The Ziggurat method allows full parallelisation of sda_2proteins runs, and parallelisation of BD moves in sdamm simulations ( see \ref make_bd_move_sdamm.f90 ).
!!
!! Both implementations are initialised in \ref read_input_file.f90 ( function \ref initialize_random_generator ).


! Marsaglia & Tsang generator for random normals & random exponentials.
! Translated from C by Alan Miller (amiller@bigpond.net.au)
! Marsaglia, G. & Tsang, W.W. (2000) `The ziggurat method for generating
! random variables', J. Statist. Software, v5(8).
!
! This is an electronic journal which can be downloaded from:
! http://www.jstatsoft.org/v05/i08
! N.B. It is assumed that all integers are 32-bit.
! N.B. The value of M2 has been halved to compensate for the lack of
!      unsigned integers in Fortran.
!
! Latest version - 1 January 2001
! Parallel version - October 2006
!
! This version has been customised for parallel processing use,
! specifically with OpenMP.  Each thread uses its own pseudo-random 
! sequence. (Gib Bogle)
!--------------------------------------------------------------------------

!> For use with OpenMP
!!
!! Each thread has its own random generator sequence
!! Use for bd_move, both sda_2proteins and sdamm\n
!! see \ref RNG "Random Generator Number"
MODULE Par_Zig_mod

   use omp_lib

   IMPLICIT NONE

   PRIVATE

   INTEGER,  PARAMETER  ::  DP=SELECTED_REAL_KIND( 12, 60 )
   REAL(DP), PARAMETER  ::  m1=2147483648.0_DP,   m2=2147483648.0_DP,      &
                            half=0.5_DP
   REAL(DP), parameter  ::  dn0=3.442619855899_DP, tn0=3.442619855899_DP,    &
                            vn=0.00991256303526217_DP,                     &
                                                de0=7.697117470131487_DP, & !q,
                            te0=7.697117470131487_DP,                       &
                            ve=0.003949659822581572_DP

    integer, save :: par_n = 0, par_step
    integer, allocatable, save ::  par_jsr(:), par_kn(:,:), par_ke(:,:)
    real(DP), allocatable, save :: par_wn(:,:), par_fn(:,:), par_we(:,:), par_fe(:,:)

   PUBLIC  :: par_zigset, par_shr3, par_uni, par_rnor, par_rexp, delete_par_zig, generate_seed_from_clock


CONTAINS


SUBROUTINE par_zigset( npar, par_jsrseed, grainsize)

   !! one difference, but still working, keep oginal
   INTEGER, INTENT(IN)  :: npar, grainsize, par_jsrseed(*)
   !!integer, dimension (:) :: par_jsrseed

   INTEGER  :: i, kpar
    REAL(DP) dn, tn, de, te, q

    par_n = npar
    par_step = grainsize
    ! First we need to allocate all the non-volatile arrays with the size npar
    allocate(par_jsr(0:npar*par_step))
    allocate(par_kn(0:127,0:npar-1))
    allocate(par_ke(0:255,0:npar-1))
    allocate(par_wn(0:127,0:npar-1))
    allocate(par_fn(0:127,0:npar-1))
    allocate(par_we(0:255,0:npar-1))
    allocate(par_fe(0:255,0:npar-1))

    !! initialize array
    par_jsr = 0
    par_kn = 0
    par_ke = 0
    par_wn = 0
    par_fn = 0
    par_we = 0
    par_wn = 0
    par_fe = 0

! Now treat each instance separately
do kpar = 0,npar-1
   !  Set the seed
   par_jsr(kpar*par_step) = par_jsrseed(kpar+1)

   !  Tables for RNOR
   dn = dn0
   tn = tn0
   q = vn*EXP(half*dn*dn)
   par_kn(0,kpar) = int ( (dn/q)*m1 )
   par_kn(1,kpar) = 0
   par_wn(0,kpar) = q/m1
   par_wn(127,kpar) = dn/m1
   par_fn(0,kpar) = 1.0_DP
   par_fn(127,kpar) = EXP( -half*dn*dn )
   DO  i = 126, 1, -1
      dn = SQRT( -2.0_DP * LOG( vn/dn + EXP( -half*dn*dn ) ) )        ! dn
      par_kn(i+1,kpar) = int ( (dn/tn)*m1 )
      tn = dn                                                        ! tn
      par_fn(i,kpar) = EXP(-half*dn*dn)
      par_wn(i,kpar) = dn/m1
   END DO

   !  Tables for REXP
   de = de0
   te = te0
   q = ve*EXP( de )
   par_ke(0,kpar) = int ( (de/q)*m2 )
   par_ke(1,kpar) = 0
   par_we(0,kpar) = q/m2
   par_we(255,kpar) = de/m2
   par_fe(0,kpar) = 1.0_DP
   par_fe(255,kpar) = EXP( -de )
   DO  i = 254, 1, -1
      de = -LOG( ve/de + EXP( -de ) )                                ! de
      par_ke(i+1,kpar) = int ( m2 * (de/te) )
      te = de                                                        ! te
      par_fe(i,kpar) = EXP( -de )
      par_we(i,kpar) = de/m2
   END DO
enddo
   RETURN
END SUBROUTINE par_zigset

!! must deallocate array, valgrinds complains
subroutine delete_par_zig ( dummy )
    IMPLICIT NONE
    integer :: dummy

    ! dummy variable, necessary with subroutines ?
    dummy = 0

    if ( .not.allocated ( par_jsr ) ) then
        return
    endif

    deallocate ( par_jsr )
    deallocate ( par_kn )
    deallocate ( par_ke )
    deallocate ( par_wn )
    deallocate ( par_fn )
    deallocate ( par_we )
    deallocate ( par_fe )
end subroutine delete_par_zig

!  Generate random 32-bit integers
FUNCTION par_shr3(kpar) RESULT( ival )
   INTEGER  ::  ival, kpar
   integer :: jz, jsr

   jsr = par_jsr(kpar*par_step)
   jz = jsr
   jsr = IEOR( jsr, ISHFT( jsr,  13 ) )
   jsr = IEOR( jsr, ISHFT( jsr, -17 ) )
   jsr = IEOR( jsr, ISHFT( jsr,   5 ) )
   par_jsr(kpar*par_step) = jsr
   ival = jz + jsr
   RETURN
END FUNCTION par_shr3

!  Generate uniformly distributed random numbers, sequence kpar
FUNCTION par_uni(kpar) RESULT( fn_val )
    integer :: kpar
   REAL(DP)  ::  fn_val

    if (kpar >= par_n) then
        write(*,*) 'thread number exceeds initialized max: ',kpar,par_n-1
        stop 1
    endif
   fn_val = half + 0.2328306e-9_DP * par_shr3(kpar)
   RETURN
END FUNCTION par_uni

!  Generate random normals, sequence kpar
FUNCTION par_rnor(kpar) RESULT( fn_val )
   REAL(DP)             ::  fn_val
   integer :: kpar

   REAL(DP), PARAMETER  ::  r = 3.442620_DP
   REAL(DP)             ::  x, y
   integer :: iz, hz

!   IF( .NOT. initialized ) CALL zigset( jsr )
    if (kpar >= par_n) then
        write(*,*) 'thread number exceeds initialized max: ',kpar,par_n
        stop 1
    endif
   hz = par_shr3(kpar)
   iz = IAND( hz, 127 )
   IF( ABS( hz ) < par_kn(iz,kpar) ) THEN
      fn_val = hz * par_wn(iz,kpar)
   ELSE
      DO
         IF( iz == 0 ) THEN
            DO
               x = -0.2904764_DP* LOG( par_uni(kpar) )
               y = -LOG( par_uni(kpar) )
               IF( y+y >= x*x ) EXIT
            END DO
            fn_val = r+x
            IF( hz <= 0 ) fn_val = -fn_val
            RETURN
         END IF
         x = hz * par_wn(iz,kpar)
         IF( par_fn(iz,kpar) + par_uni(kpar)*(par_fn(iz-1,kpar)-par_fn(iz,kpar)) < EXP(-half*x*x) ) THEN
            fn_val = x
            RETURN
         END IF
         hz = par_shr3(kpar)
         iz = IAND( hz, 127 )
         IF( ABS( hz ) < par_kn(iz,kpar) ) THEN
            fn_val = hz * par_wn(iz,kpar)
            RETURN
         END IF
      END DO
   END IF
   RETURN
END FUNCTION par_rnor

!  Generate random exponentials, sequence kpar
FUNCTION par_rexp(kpar) RESULT( fn_val )
   REAL(DP)  ::  fn_val
   integer :: kpar

   REAL(DP)  ::  x
   integer :: iz, jz

!   IF( .NOT. initialized ) CALL Zigset( jsr )
    if (kpar >= par_n) then
        write(*,*) 'thread number exceeds initialized max: ',kpar,par_n-1
        stop 1
    endif
   jz = par_shr3(kpar)
   iz = IAND( jz, 255 )
   IF( ABS( jz ) < par_ke(iz,kpar) ) THEN
      fn_val = ABS(jz) * par_we(iz,kpar)
      RETURN
   END IF
   DO
      IF( iz == 0 ) THEN
         fn_val = 7.69711 - LOG(par_uni(kpar) )
         RETURN
      END IF
      x = ABS( jz ) * par_we(iz,kpar)
      IF( par_fe(iz,kpar) + par_uni(kpar)*(par_fe(iz-1,kpar) - par_fe(iz,kpar)) < EXP( -x ) ) THEN
         fn_val = x
         RETURN
      END IF
      jz = par_shr3(kpar)
      iz = IAND( jz, 255 )
      IF( ABS( jz ) < par_ke(iz,kpar) ) THEN
         fn_val = ABS( jz ) * par_we(iz,kpar)
         RETURN
      END IF
   END DO
   RETURN
END FUNCTION par_rexp

!> suffer overflow, not sure correct in original case
subroutine generate_seed_from_clock ( seed )
    IMPLICIT NONE
    integer, intent ( out ) :: seed
    ! important to avoid overflow
    integer(kind=8) :: ss, mm, hh, dd
    real ( kind=8) :: dss,dmm,dhh,ddd
    integer :: io_unit !ierr
    real (kind=8), parameter :: b = 4.294967291E9

    io_unit = 110

    call SYSTEM(' date +%S%M%H%j > zbql1234.tmp')
    open( io_unit, file='zbql1234.tmp')
    read ( io_unit, fmt='(3(I2),I3)' ) SS,MM,HH,DD
    !write(*,*) 'ss,mm,hh,dd ',ss,mm,hh,dd

    close( io_unit )
    call SYSTEM('rm zbql1234.tmp')
    DSS = DINT( (real(SS,8)/6.0E1) * B)
    DMM = DINT( (real(MM,8)/6.0E1) * B)
    DHH = DINT( (real(HH,8)/2.4E1) * B)
    DDD = DINT( (real(DD,8)/3.65E2)* B)

    seed = INT ( DMOD(real(SS+MM+HH+DD,8),B) )

end subroutine generate_seed_from_clock

END MODULE par_zig_mod



