!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_setofgrid.f90
!! store all grids and position of atoms


!> Module for storing data needed for a specific conformation.\n
!! store all grids and position of atoms
MODULE mod_setofgrid
    USE mod_gridtype
    USE mod_grid
    USE mod_analytic
    USE mod_exclusion_grid
    USE mod_ljgrid

    !> Store all grids ( UHBD type), and "charges": charges, accessibility and their position
    !!
    !! Pointer to exclusion grid as well, see \ref mod_exclusion_grid.f90\n
    !! Many proteins may refer to the same instance of sogrid ( case of sdamm )
    type sogrid

        !> value of the effective charge, and square value for desolvation
        !! @{
        real ( kind = 8 ), dimension(:), allocatable :: charge_grid
        real ( kind = 8 ), dimension(:), allocatable :: charge_square_grid
        !> @}
        !> position of the effective charge
        real ( kind = 8 ), dimension(:,:), allocatable :: position_charge

        !> reaction criteria associated to this set of grid
        !! make sense for sda flex, for sdamm ??
        !! by default associated to protein2
        !! change associate to all proteins we want to consider possible criteria
        !! seems easier to store only array of position for sda_flex and sdamm, limitation if HOMODIMER !!
        real ( kind=4 ), dimension (:,:), allocatable :: reaction_criteria_pos

        !> accessibility position of the atoms at the surface, computed by \ref solva::solva "solva()"
        !! @{
        real ( kind = 8 ), dimension(:,:), allocatable :: list_atom_surface
        real ( kind = 8 ), dimension(:), allocatable :: accessibility
        !> array listing which atoms from the entire PDB list are the surface atoms (i.e., accessible) - used only in tools
        integer, dimension(:), allocatable :: which_atom_surface
        !> should this array be allocated? default: false
        logical :: alloc_which_at_surf
        !> @}



        !> fixed size array which contains pointers to grid objects ( uhbd type): electrostatic, desolvation, hydrophobic, lennard-jones repulsive ( sdamm ), lennar-jones
        type ( type_grid ), dimension(:), pointer :: pset_grid
        !> fixed size array which contains pointers to analytical grid
        type ( analytic ), dimension(:), pointer :: pset_analytic_grid

        !> pointer to exclusion grid
        type ( type_grid_excl ), pointer :: p_excl_grid

        !> pointer to module mod_ljgrid, include all grids, all atoms
        type ( type_ljgrid ), pointer :: p_ljgrid

        !> parameter for exclusion, put it here even if not implemented\n
        !! keep old format because old subroutine
        !! @{
        real ( kind=8 ) :: hexcl
        real ( kind=8 ) :: distance_max
        !> @}

        real ( kind=8 ) :: charge_total
        integer :: nb_charge, nb_atom_surface, nb_reaction_criteria

        !> rotation and translation diffusion coefficients, as read in the input file\n
        real ( kind=8 ) :: diffusion_trans, diffusion_rotat

        !> In case of hydrodynamic interaction stoke_radius must be > 0.
        !! Used by Debye-Huckel also now, need to correct / test for HI,
        !! explict hi variable needed
        real(kind=8) :: stoke_radius
        !> Save if debye-huckel / analytical interaction is activated
        !! Needed because full load done only later by param_force_energy
        logical :: idebye

        !> Array to store cut off for each grid in sogrid with equivalent in all other sogrids
        real(kind=4) , dimension (:,:), allocatable :: cut2_array !NJB
 
        !> Largest cut off in cut2_array
        real(kind=4) :: max_cut2

    end type sogrid

    contains

        !> Initialze sogrid with the total number of grid declared in \ref mod_gridtype.f90
        !!
        !! The array of grid has a fixed sized
        !! @param this : instance of \ref sogrid
        subroutine allocate_sogrid( this )
            IMPLICIT NONE
            type ( sogrid ) :: this

            integer :: ierr=1,i
            integer :: nb_grid

            !initialize all to null
            nullify( this % pset_grid )
            nullify( this % pset_analytic_grid )
            nullify ( this % p_ljgrid )

            ! number is fixed here
            nb_grid = get_sizeof_sgrid()

            !to check if already existing, but dangerous before initialisation, need history variable
            allocate( this % pset_grid ( nb_grid ) , stat=ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR Couldn t allocate memory for set of grid ',nb_grid
                stop 1
            endif

            !dangerous pointer ?? not initiliased in mod_grid, try
            ! now initialized if called read_grid, but better to be sure
            do i = 1, nb_grid !get_sizeof_sgrid() !nb_grid
                nullify ( this % pset_grid ( i ) % p_grid )
                this % pset_grid ( i ) % grid_type = i
                !! initaliaze data of grid
                this % pset_grid ( i ) % origin = (/0., 0., 0./)

                ! default real_values
                this % pset_grid ( i ) % is_logical = .false.

                !write(*,*) 'test associated ', associated ( this % pset_grid (1) % p_grid % pgrid_real )
            end do

            ! number is fixed here, 2 = debyeh and "crowders"
            nb_grid = get_sizeof_analytic_grid()
            allocate( this % pset_analytic_grid ( nb_grid ) , stat=ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR Couldn t allocate memory for set of analytical grid ',nb_grid
                stop 1
            endif

            do i = 1, nb_grid
                ! call the default/empty constructor, nullify all
                ! will be allocated in MainX by param_force_energy:init_analytical
                call allocate_analytic( this % pset_analytic_grid ( i ) )
            end do
!            write(*,*) 'after default allocation of analytic grid'

            ! init variable
            this % charge_total = 0
            this % nb_charge = 0
            this % nb_atom_surface = 0
            this % nb_reaction_criteria = 0
            this % diffusion_trans = 0.d0
            this % diffusion_rotat = 0.d0
            this % stoke_radius = 0.d0
            !this % idebye = .true.
            this % idebye = .false.

            nullify ( this % p_excl_grid )
            this % alloc_which_at_surf = .false.

        end subroutine allocate_sogrid

        !> Delete sogrid
        !! @param this : instance of \ref sogrid
        subroutine delete_sogrid ( this )
            IMPLICIT NONE
            type ( sogrid ) :: this
            integer :: ierr=0

            integer :: i, nb_grid

            !write(*,*)
            !write(*,*) 'entry delete_sogrid'

            ! need to split position and charge. position may be deallocated before to avoid multiple copies ( sdamm not flex )
            if ( allocated ( this % charge_grid ) ) then
                deallocate( this % charge_grid, this % charge_square_grid, stat=ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR In deallocation memory of charge_grid '
                    stop 1
                endif
                if ( allocated ( this % position_charge ) ) then
                    deallocate ( this % position_charge, stat=ierr )
                    if ( ierr /= 0 ) then
                        write (*,*) 'ERROR In deallocation memory of position_charge'
                        stop 1
                    endif
                end if
            endif

            ! delete all grid ( uhbd or dtgrid )
            nb_grid = size( this % pset_grid )
            do i=1,nb_grid
                call delete_grid ( this % pset_grid( i ) )
            enddo
            !!deallocate the array
            deallocate ( this % pset_grid , stat=ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR In deallocation array of pset_grid'
                stop 1
            endif
            nullify ( this % pset_grid )

            ! delete all analytical
            nb_grid = size( this % pset_analytic_grid )
            do i=1,nb_grid
                call deallocate_analytic ( this % pset_analytic_grid( i ) )
            enddo
            !!deallocate the array
            deallocate ( this % pset_analytic_grid , stat=ierr )
            if ( ierr /= 0 ) then
                write (*,*) 'ERROR In deallocating array of analytical grid '
                stop 1
            endif
            nullify ( this % pset_analytic_grid )


            ! delete accessibility grid and list_atom_surface
            if ( allocated( this % accessibility ) ) then
                ! write ( *,* ) 'deallocate accessibility and list_atom_surface'
                ! deallocate( this % accessibility, this % list_atom_surface, stat = ierr )
                deallocate( this % accessibility, stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR In deallocating memory of accessibility'
                 endif
            endif

            if ( allocated ( this % list_atom_surface ) ) then
                deallocate ( this % list_atom_surface, stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR In deallocating memory of list atom surface'
                endif
            end if

            if ( allocated ( this % which_atom_surface ) ) then
                deallocate ( this % which_atom_surface, stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR In deallocating memory of which_atom_surface'
                endif
            end if

            ! cancel pointer to reaction, not responsible of the object destruction
            ! nullify ( this % p_reaction_criteria )
            if ( allocated ( this % reaction_criteria_pos ) ) then
                deallocate ( this % reaction_criteria_pos, stat = ierr )
                if ( ierr /=0 ) then
                    write (*,*) 'ERROR In deallocating memory of reaction_criteria_pos'
                endif
            endif

            ! delete exclusion_grid, need to delete the grid explicetely before deallocating the object
            ! grid have been allocated in make_exclusion_grid
            if ( associated ( this % p_excl_grid ) ) then
                ! write ( *,* ) 'deallocate excl grid '
                call delete_excl_grid ( this % p_excl_grid )
                deallocate ( this % p_excl_grid, stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR In deallocating pointer on exclusion grid '
                 endif
                 ! and nullify the pointer
                 nullify ( this % p_excl_grid )
            end if

            ! test if p_ljgrid associated
            if ( associated ( this % p_ljgrid ) ) then
                !write(*,*) 'Delete p_ljgrid with option copy_omp is false, really delete grids and atom_position '
                call deallocate_ljgrid ( this % p_ljgrid, .false. )
                deallocate ( this % p_ljgrid, stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR In deallocating pointer on exclusion grid '
                 endif
                 ! and nullify the pointer
                 nullify ( this % p_ljgrid )

            end if

        end subroutine delete_sogrid

        !> Load one grid or one "charge", can read binary or ascii format
        !! Add analytical interactions ( negative grid number )
        !! @param this : instance of \ref sogrid
        !! @param filename : name of the file
        !! @param t_grid : type of the grid from 1 to 9, see \ref mod_gridtype.f90
        !! @param scfct_opt : optionaly the multiplicative factor to apply ( default 0.)
        !! @param iform_opt : optionaly, ascii or binary
        subroutine load_grid ( this, filename, t_grid, scfct_opt, iform_opt )
            IMPLICIT NONE
            type ( sogrid ), intent( inout ) :: this
            character*128 filename
            integer, intent (in) :: t_grid
            integer, intent(in), optional :: iform_opt
            real ( kind=8 ), optional :: scfct_opt
            real ( kind=8 ) :: scfct
            integer :: iform=0

#ifdef DEBUG
            write(*,*) 'Entry mod_setofgrid::load_grid t_grid=', t_grid
#endif

            if ( present(iform_opt) ) then
                iform = iform_opt
            else
                iform = 1
            endif

            if ( present(scfct_opt) ) then
                scfct = scfct_opt
            else
                scfct = 0.
            endif

            ! real, 3D grid + charge
            if ( t_grid > 0 ) then
                ! load uhbd type grid, call allocation
                if ( (t_grid >= 1) .and. (t_grid <= 4) ) then
                    ! write(*,*) 'scfct ', scfct
                    ! flush ( 6 )
                    !write(*,*) 'Read grid ', trim(filename)
                    call read_grid ( this % pset_grid ( t_grid ), filename, t_grid, scfct ) !, .false., iform )
                ! load charges, need to be modified
                ! this not used normally, old numbering before lennard-jones
                else if ( ( t_grid >= 5 ) .and. ( t_grid <= 7 ) ) then
                    ! done in prepare_protein and make_exclusion_grid, could be added to module BUT order is important
                    ! problem with exclusion or not for calling
                    ! write(*,*) ' t_grid = ', t_grid ,' can be done, only by prep_protein '
                    ! read_uhbd_logical_grid
                ! should be better to test name, charge, charge_square
                !else if ( (t_grid == 11) .or. (t_grid== 12) ) then
                else if ( ( t_grid == get_nb_gridtype('charge') ) .or. ( t_grid == get_nb_gridtype('charge_square' ) ) ) then
                    ! write(*,*) 'will read the charge in file t_grid=', filename, t_grid
                    call read_charge ( this, filename )
                else
                    write(*,*) 'WARNING Wrong parameters in mod_set_ofgrid::load_grid, should be an error and stop ??'
                endif
            endif

        end subroutine load_grid

        !> allocate and load lj_grid, should be done only once
        !! need more option, only one of lj / atom should be allocated
        !! keep one function, need type_to_load : 1 or 2, both are array of characters
        subroutine load_ljgrid ( this, type_to_load, lj_filename, pdb_filename, opt_fct )
            IMPLICIT NONE
            type ( sogrid ) :: this
            integer, intent ( in ) :: type_to_load !, nb_lj
            character, dimension(:) :: lj_filename*(*)
            character :: pdb_filename*(*)
            real ( kind=8 ), optional :: opt_fct

            real ( kind=8 ) :: fct
            integer :: nb_lj, ierr, i

            if ( present ( opt_fct ) ) then
                fct = opt_fct
                !write(*,*) 'set fct ', fct
            else
                fct = 1.0
            end if

#ifdef DEBUG
            write(*,*)
            write(*,*) 'Entry load_ljgrid '
            write(*,*) 'lj_filename ', lj_filename
            write(*,*) 'pdb_filename ',trim(pdb_filename)
#endif
            nb_lj = size(lj_filename)

#ifdef DEBUG
            write(*,*) 'Allocate ', nb_lj, ' lenard-jones grids and atom types'
#endif

            if ( ( associated( this % p_ljgrid ) .eqv. .false. ) .and. ( nb_lj > 0 ) ) then
                !write(*,*) 'allocate this % p_ljgrid with size ', nb_lj
                allocate ( this % p_ljgrid, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In allocation of p_ljgrid'
                    STOP 1
                endif
                call allocate_ljgrid ( this % p_ljgrid, nb_lj )
            else
                write(*,*) 'this % p_ljgrid already allocated with size '
                !stop 1
            end if

            !write(*,*) 'ok, no error associated p_ljgrid ? ', associated ( this % p_ljgrid )

            ! here need to loop over the nb_lj, use entries in lj_filename
            ! load the grids
            if ( type_to_load == 1 ) then
                do i = 1, size ( lj_filename )
                    call set_ljgrid ( this  % p_ljgrid, lj_filename(i), i, fct )
                end do
            ! load the atom_type
            else if ( type_to_load == 2 ) then
                do i = 1, size ( lj_filename )
                    call set_atom_lj ( this % p_ljgrid, pdb_filename, i, lj_filename(i) )
                end do
            end if

        end subroutine

        !> Initialize one analytic array
        !!
        !! Can be done only after cut_off has been computed and other stoke_radius known
        !!
        !! @param this : instance of \ref sogrid
        !! @param name_type : string name debye-huckel or crowder
        !! @param list_stoke : list all stoke radius by type of proteins
        !! @param bin_size : of the precomputed array
        !! @param max_size : size of cut-off, already corrected in case of box
        !! @param opt_extra_param : optional, ionic stength at the moment
        subroutine set_analytic ( this, name_type, list_stoke, bin_size, max_size, opt_extra_param )
            implicit none
            type ( sogrid ) :: this
            character, intent(in) :: name_type*(*)
            real (kind=8), dimension(:) :: list_stoke
            real(kind=8) :: bin_size
            real(kind=4) :: max_size
            real(kind=4), optional :: opt_extra_param

            integer :: code, grid_code_nb

#ifdef DEBUG
            write(*,*) 'Entry mod_setofgrid::set_analytic'
            write(*,*) 'stoke_radius', this % stoke_radius
#endif

            code = get_nb_analyticaltype( name_type )
            ! convention: use - code
            grid_code_nb = -code

#ifdef DEBUG
            write(*,*) 'grid_code_nb ', grid_code_nb
#endif

            ! general, whatever the type
            call set_analytic_params( this % pset_analytic_grid ( grid_code_nb ), bin_size, max_size, this % stoke_radius )

#ifdef DEBUG
            write(*,*) 'Entry set_analytic '
            write(*,*) 'code ', code, ' stands for ', name_type
            write(*,*) 'bin_size ', bin_size
#endif
            select case ( code )
                case ( -1 )
#ifdef DEBUG
                    write(*,*) 'choose debye-huckel'
#endif
                    ! compute kappa, bjerrum, need ionic strength
                    call set_dh_params( this % pset_analytic_grid ( 1 ), opt_extra_param )
                    ! fill the arrays
                    call fill_array_dh ( this % pset_analytic_grid ( 1 ), list_stoke )
                    ! specific to Debye-H, set zeroes if the electrostatic grid is present
                    call set_zeroes ( this % pset_analytic_grid ( 1 ), this % pset_grid ( 1 ) % rp )

                case ( -2 )
#ifdef DEBUG
                    write(*,*) 'choose other, not implemented'
#endif
                    write(*,*) 'choose -2 not implemented'
                    STOP 1

                case DEFAULT
                    write(*,*) 'ERROR code not valid in set_analytic'
                    STOP 1

            end select

        end subroutine set_analytic

        !> Allocate charge array and square of the charges for electrostatic desolvation
        !! @param this : instance of \ref sogrid
        !! @param nat : number of charges
        subroutine allocate_charge( this, nat )
            IMPLICIT NONE
            type ( sogrid ) :: this
            integer, intent ( in ) :: nat
            integer :: ierr

            ierr = 1

            allocate( this % position_charge (3 , nat ), this % charge_grid ( nat ), this % charge_square_grid ( nat ), stat=ierr )
            if ( ierr /=0 ) then
              write(*,*) 'ERROR In memory allocation'
              stop 1
            endif

            this % position_charge = 0.d0

        end subroutine allocate_charge

        !> Allocate \ref sogrid::accessibility "accessibility" and  \ref sogrid::list_atom_surface "list_atom_surface" array
        !! @param this : pointer to an instance of \ref sogrid ( why pointer ?)
        !! @param nat2 : number of accessibility atoms
        subroutine allocate_acc_surf( this, nat2 )
            IMPLICIT NONE
            type ( sogrid ), intent(inout), pointer :: this
            integer, intent(in) :: nat2
            integer :: ierr=0

!            write(*,*) 'is allocated ',  ALLOCATED( this % accessibility )
!            write(*,*)
            allocate( this % accessibility ( nat2 ), this % list_atom_surface ( 3, nat2 ), stat=ierr )
            if ( ierr /=0 ) then
              write(*,*) 'ERROR In memory allocation of accessibility'
              STOP 1
            endif
            this % accessibility = 0.d0
            this % list_atom_surface = 0.d0
            
            if ( this % alloc_which_at_surf ) then
                allocate( this % which_atom_surface( nat2 ), stat=ierr )
                if ( ierr /=0 ) then
                  write(*,*) 'ERROR In memory allocation of accessibility'
                  STOP 1
                endif
                this % which_atom_surface(:) = 0
            end if

        end subroutine allocate_acc_surf

        !> Function to read charges, and make charge_square
        !! @param this : instance of \ref sogrid
        !! @param echa_filename  : effective charge (*.echa) filename
        subroutine read_charge ( this, echa_filename )
           IMPLICIT NONE
           type ( sogrid ) :: this
           character *128 echa_filename
           real ( kind=8 ) :: qtot
           integer :: i

           ! call external \ref read_eff_charges.f90
           call read_eff_charges ( this, echa_filename, qtot )
           !write(*,*) 'after read_eff_charges qtot ', qtot

           ! test this array, all allocated in allocate charge
           if ( allocated ( this % charge_grid ) ) then
               !write(*,*) 'size charge_grid ', size ( this % charge_grid )
               ! make square
               do i=1, size ( this % charge_grid )
                   this % charge_square_grid(i) =  this % charge_grid(i) * this % charge_grid(i)
               enddo
               this % charge_total = qtot
               this % nb_charge = size ( this % charge_grid )
           ! initialize to zero, should be done in constructor already
           else
               this % charge_total = 0.d0
               this % nb_charge = 0
           endif


           !write(*,*) 'charge total', qtot
           !write(*,*) 'charge 1 ', this%charge_grid(1), ( this%position_charge(i,1),i=1,3 )
           !write(*,*) 'charge 3 ', this%charge_grid(3), ( this%position_charge(i,3),i=1,3 )
           !write(*,*) 'charge 2 square ', this%charge_square_grid(2)
        end subroutine read_charge

        !> Used ??
        function get_max_debye( this ) result ( dist )
            IMPLICIT NONE
            type ( sogrid ) :: this
            real ( kind=8 ) :: dist

            write(*,*) 'Get max debye, default 8 times the radius '

            dist = 0.d0
            if ( is_associated ( this % pset_grid ( 1 ) ) ) then
                write(*,*) 'rp ', this % pset_grid ( 1 ) % rp
                dist = 8.d0 * this % pset_grid ( 1 ) % rp
                write(*,*) 'max_debye ', dist
            else
                write(*,*) 'ERROR p_grid electrostatic not allocated '
                STOP 1
            end if

        end function get_max_debye

        !> Write acceessibility into a file
        !!
        !! It allows to be computed only once, and only read for the next simulations\n
        !! Store a maximum of information, the pdb does not even need to be read\n
        !! It is called by \ref prepare_atom_protein::prepare_atom_protein "prepare_atom_protein()"
        !! @param this : instance of \ref sogrid
        !! @param access_filename : name of the file, p1.access
        !! @param center_prot : center of geometry of the solute
        !! @param vdw : list of the Van der Walls radius
        !! @param max_vdw : maximum value of the vdw array
        !! @param max_dist : maximum radius of the protein \ref sogrid::distance_max "distance_max"
        !! @param param_probe : the value of the probe used to compute the accessibility
        subroutine write_accessibility ( this, access_filename, center_prot, vdw, max_vdw, max_dist, param_probe )
            USE com_sda
            IMPLICIT NONE
            type ( sogrid ) :: this
            character*128 :: access_filename
            real ( kind=8 ), dimension ( 3 ) :: center_prot
            real ( kind=8 ), dimension ( : ) :: vdw
            real ( kind=8 ) :: max_vdw, max_dist
            type ( probe_type ) :: param_probe

            integer :: io_access = 200, i
            logical :: ex

#ifdef DEBUG
            write(*,*) 'Entry write_accessibility  file ', trim ( access_filename )
            write(*,*) 'max_vdw ', max_vdw
            write(*,*) 'max_dist ', max_dist
            !write(*,*) 'size ', size ( this % list_atom_surface )
#endif

            inquire ( file=access_filename, EXIST=ex )
            if ( ex ) then
                write (*,*) "WARNING Accessibility already exists !", access_filename
                !stop 1
            end if

            open ( io_access, file= access_filename, form='formatted', status = 'replace' ) !, iostat = io_err )

            write ( io_access, fmt='("## center",3f11.5)' ) center_prot
            write ( io_access, fmt='("## max van der waals",f11.5)' ) max_vdw
            write ( io_access, fmt='("## max distance     ",f11.5)' ) max_dist
            write ( io_access, fmt='("## probe water      ",f11.5)' ) param_probe % probew
            write ( io_access, fmt='("## threshld         ",f11.5)' ) param_probe % thresold
            write ( io_access, fmt='("## Accessibility, x, y, z, access, vdw, nb atoms:",i10 )' ) this % nb_atom_surface

            if ( this % alloc_which_at_surf ) then
                do i = 1, this % nb_atom_surface
                    write( io_access, fmt = '("ATOM",2X,i6,18X,3f8.3,X,f7.4,f7.4)' ) &
                           this % which_atom_surface(i), this % list_atom_surface(:,i), this % accessibility(i), vdw ( i )
                end do
            else
                do i = 1, this % nb_atom_surface
                    !write( io_access, fmt='("ATOM",26X,3f8.3,X,f8.3)' ) this % list_atom_surface(:,i), this % accessibility(i)
                    write( io_access, fmt = '("ATOM",26X,3f8.3,X,f7.4,f7.4)' ) &
                           this % list_atom_surface(:,i), this % accessibility(i), vdw ( i )
                end do
            end if

            close ( io_access )

        end subroutine write_accessibility

        !> Read accessibility file
        !!
        !! Optional but save lots of time to not recompute solva for large system\n
        !! Read as well vwd, max_vdw and max_dist computed previously\n
        !! @param this : input, instance of \ref sogrid
        !! @param access_filename : input, name of the file, p1.access
        !! @param center_prot : output, center of geometry of the solute
        !! @param vdw : output, list of the Van der Walls radius
        !! @param max_vdw : output, maximum value of the vdw array
        !! @param max_dist : output, maximum radius of the protein \ref sogrid::distance_max "distance_max"
        !! @param param_probe : input, instance of \ref com_sda::probe_type "probe_type", for checking whether the re-calculation is needed
        !! @param opt_recalculate : optional, input, if param_probe values are different than in the file, stop reading (default: .true.)
        subroutine read_accessibility ( this, access_filename, center_prot, vdw, max_vdw, max_dist, param_probe, opt_recalculate )
            USE com_sda
            IMPLICIT NONE
            type ( sogrid ) :: this
            character*128 :: access_filename
            real ( kind=8 ), dimension ( 3 ) :: center_prot
            real ( kind=8 ), dimension ( : ), pointer :: vdw
            real ( kind=8 ) :: max_vdw, max_dist
            type ( probe_type ) :: param_probe
            logical, optional :: opt_recalculate

            real ( kind=4 ) tmp_probew, tmp_thres
            integer :: io_access = 200, i, ierr
            logical :: ex

            center_prot = 0.d0
            tmp_probew = 0.
            tmp_thres = 0.
            !status = .false.

            write (*,*)
            write (*, fmt='("Read accessibility file : ",(a))' ) trim( access_filename )

            inquire ( file=access_filename, EXIST=ex )
            if ( .NOT.ex ) then
                write (*,*) "accessibility file does not exist !! ", access_filename
                stop 1
            end if
            open ( io_access, file= access_filename, form='formatted', action = 'read', status = 'old' )
            !! could read center now
            read  ( io_access, fmt='(9X,3f11.5)' ) center_prot
            !write (*,*) 'read access center:', center_prot

            read  ( io_access, fmt='(20X,f11.5)' ) max_vdw
            !write (*,*) 'read access max vdw:', max_vdw

            read  ( io_access, fmt='(20X,f11.5)' ) max_dist
            !write (*,*) 'read access max_dist:', max_dist

            read  ( io_access, fmt='(20X,f11.5)' ) tmp_probew
            read  ( io_access, fmt='(20X,f11.5)' ) tmp_thres

            if ( ( tmp_probew /= param_probe % probew ) .or. ( tmp_thres /= param_probe % thresold ) ) then
                if ( present( opt_recalculate ) .and. .not.opt_recalculate ) then
                    param_probe % probew = tmp_probew
                    param_probe % thresold = tmp_thres
                else
                    write (*,* ) 'Parameters in accessibility have changed, will recompute the accessibility '
                    close ( io_access )
                    !status = .true.
                    return
                end if
            endif

            ! read number of atoms present
            read  ( io_access, fmt='(49X,i10)' ) this % nb_atom_surface
            !write (*,*) 'read access nb_atom_surface:', this % nb_atom_surface

            ! allocate array, vdw
            allocate ( vdw ( this % nb_atom_surface ), stat = ierr )
            if ( ierr /= 0 ) then
                write ( *,* ) 'ERROR Cannot allocate vdw in read_accessibility'
                stop 1
            end if
            vdw = 0.d0

            !! accessibility
            if ( .not.allocated( this % accessibility ) ) then
                allocate ( this % accessibility ( this % nb_atom_surface ), stat=ierr )
                if ( ierr/= 0 ) then
                    write (*,*) 'ERROR Problem in the allocation of grid accessibility in read accessibility', &
                                 this % nb_atom_surface, '; err: ', ierr
                   stop 1
                endif
            endif
            this % accessibility = 0.d0
            if ( .not.allocated( this % list_atom_surface ) ) then
                allocate ( this % list_atom_surface ( 3, this % nb_atom_surface ), stat=ierr )
                if ( ierr/= 0 ) then
                    write (*,*) 'ERROR Problem in the allocation of grid list_atom_surface in read accessibility', &
                                 this % nb_atom_surface, '; err: ', ierr
                   stop 1
                endif
            endif

            if ( this % alloc_which_at_surf ) then
                if ( .not.allocated( this % which_atom_surface ) ) then
                    allocate ( this % which_atom_surface( this % nb_atom_surface ), stat=ierr )
                    if ( ierr/= 0 ) then
                        write (*,*) 'ERROR Problem in the allocation of which_atom_surface in read accessibility', &
                                     this % nb_atom_surface, '; err: ', ierr
                       stop 1
                    endif
                end if

                do i = 1, this % nb_atom_surface
                    read ( io_access, fmt='(6X,i6,18X,3f8.3,X,f7.4,f7.4)' ) &
                           this % which_atom_surface(i), this % list_atom_surface(:,i), this % accessibility(i), vdw(i)
                end do
            else
                do i = 1, this % nb_atom_surface
                    !read ( io_access, fmt='(30X,3f8.3,X,f8.3)' ) this % list_atom_surface(:,i), this % accessibility(i)
                    read ( io_access, fmt='(30X,3f8.3,X,f7.4,f7.4)' ) &
                           this % list_atom_surface(:,i), this % accessibility(i), vdw(i)
                end do
            end if

            close ( io_access )

        end subroutine read_accessibility

        !> Print info of this set_of_grid
        subroutine set_ofgrid_info( this )
           IMPLICIT NONE
           type ( sogrid ) :: this
           integer :: i

           write(*,*)
           write(*,*) '=== Info Set of grids ==='
           if ( associated ( this % pset_grid )) &
                     write (*,*) ' pset_grid is allocated '

           write(*,*) '  grid stored automatically: '
           do i= 1 , size ( this % pset_grid )
               !if ( associated ( this % pset_grid ) ) &
               !     write (*,*) ' pset_grid is allocated '
               if ( associated ( this % pset_grid ( i ) % p_grid ) ) then
                    write( *, * ) '    grid_type: '
                    call print_gridtype( this % pset_grid( i ) % grid_type )
                    call info_grid( this % pset_grid ( i ) )
                !else
                !    write(*,*) 'no grid associated'
                endif
           enddo
           !write (*,*) '   grid filled manually: '
           !write (*,*), '  charge '

           if ( allocated( this % charge_grid )) then
              write(*,*) 'charges have been read '
              if ( allocated( this % charge_square_grid ) .eqv. .false. ) &
                  write(*,*) ' Problem with charges square !!!!!'
              if ( allocated( this % position_charge ) .eqv. .false. ) &
                  write(*,*) 'Problem with position charge !!!!! '

              write (*,fmt='("total charge:", f8.3," nb charges ", i6)' ) this % charge_total, this % nb_charge
           else
              write(*,*) 'charges have not been read '
           endif

           if ( allocated( this % list_atom_surface )) then
              write(*,*) ' atom position and accessibility have been read number atom surface ', &
                           this % nb_atom_surface
              if ( allocated( this % accessibility ) .eqv. .false. ) &
                  write(*,*) '  Problem with accessibility !!!!!'
           endif

           if ( associated ( this % p_excl_grid ) ) then
               write(*,*) 'exclusion grid has been computed '
           !else
           !    write(*,*) ' exclusion grid has not been computed '
           endif

           if ( associated ( this % pset_grid )) &
                     write (*,*) ' pset_grid is associated '

           write(*,*) 'analytical grid stored automatically: size ', size ( this % pset_analytic_grid )
           do i= 1 , size ( this % pset_analytic_grid )
               !if ( associated ( this % pset_grid ) ) &
               !     write (*,*) ' pset_grid is allocated '
               if ( associated ( this % pset_analytic_grid ( i ) % parray_nrj ) ) then
                    write( *, * ) '    grid_type: '
                    !call print_gridtype( this % pset_analytic_grid( i ) % grid_type )
                    call info_analytic( this % pset_analytic_grid ( i ) )
                else
                    write(*,*) 'no grid associated'
                endif
           enddo

       end subroutine set_ofgrid_info

END MODULE mod_setofgrid
