!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file restart.f90
!!
!! To read and write restart file

!> Read all restart files and associate to p_record
!!
!! Read all restart files for all type of calculations ( complexes or trajectories )\n
!! In ascii or binary format
!!
!! @param p_record : pointer ( of type \ref mod_record::record "record ) to which associate the restart
!! @param tab_prot : instance of \ref mod_protein::protein "protein"
!! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
!! @param filename : name of the restart file
!! @param io_restart : file descriptor of the file
!! @param account_occur : if occurence must be taken into account ( used by sda_koff )
!! @param opt_nb_types : optional indicates the number of different types of solutes
subroutine read_restart ( p_record, tab_prot, type_calc, filename, io_restart, account_occur, &
                          opt_nb_types )
    USE mod_record
    USE mod_type_calc
    IMPLICIT NONE
    type ( record ), pointer :: p_record
    type ( type_calculation ) :: type_calc
    type ( array_protein_type ), target :: tab_prot
    character (128) :: filename
    ! need to send an unique integer for file identifier, defined in com_sda
    integer :: io_restart
    logical :: account_occur
    integer, optional, intent (in) :: opt_nb_types

    integer :: ierr, i
    integer :: nb_record, nb_types
    real ( kind = 8 ), dimension ( 2, 3 ) :: cm

    nb_record = 0
    cm = 0.d0

    write(*,*)
    write(*,*) '=========================================================================================='
    write(*,*) 'Read restart complexe file ', trim ( filename ), ' account_occurency ', account_occur
    write(*,*)
    
    nb_types = 0
    allocate ( p_record, stat = ierr )
    if ( ierr /= 0 ) then
        write ( *, * ) 'Problem with allocate p_restart'
        stop 1
    end if
    nullify ( p_record % pcomplexe, p_record % ptrajectory )

    if ( present( opt_nb_types ) ) then
        nb_types = opt_nb_types
    end if

    ! to see later, add initialize_record ? need simple allocator
    p_record % header_str = ''

    p_record % io_compl = io_restart
    !!need to assign ptab_prot before reading header
    p_record % ptab_prot => tab_prot
    p_record % nb_types = nb_types

    ! read the header, and setup most variables
    ! setup ascii or binary, bit's...all data available in header
    call read_header_record ( p_record, filename, cm, nb_record )

    ! can simplify with initialize_record...

    ! if sdamm, expect to read a trajectory with a default array of nb_protein,  only 1 restart conformation
    ! create p_restart only temporarly, then delete it.
    if ( type_calc % enum_type_calc == sdamm ) then

        ! create trajectory
        allocate ( p_record % ptrajectory, stat = ierr )
        if ( ierr /= 0 ) then
            write ( *, * ) 'Problem with allocate ptrajectory in restart '
        end if

        !write(*,*) 'nb_max_complexe ', p_record % nb_max_complexe
        !write(*,*) 'test nb_types ', p_record % nb_types
        p_record % ptrajectory % size_array = p_record % nb_protein

        !! should call initalize_record, only problem because write header file in initialze_record
        call initialize_trajectory ( p_record % ptrajectory, p_record )

        !! load data in trajectory
        call read_file_record ( p_record )

        !! sdamm restart to initialize the trajectory
        !! here copy position of all atoms, could use ptab_prot
        do i = 1, size ( tab_prot % array )

            !! always present
            tab_prot % array ( i ) %  position = 0.d0
            ! default value
            tab_prot % array ( i ) %  box = 0
            tab_prot % array ( i ) %  position = p_record % ptrajectory % array_trajectory ( i ) %  position
            !write(*,*) 'copy to protein position ', p_record % ptrajectory % array_trajectory ( i ) %  position
            tab_prot % array ( i ) %  orientation_x = p_record % ptrajectory % array_trajectory ( i ) %  orient_x
            tab_prot % array ( i ) %  orientation_y = p_record % ptrajectory % array_trajectory ( i ) %  orient_y

            ! best way to normalize ?? original way, norm before
            ! DEBUG, to recopy sda_flex method
            ! sdamm way, norm only z
            ! test normalisation of axis : test_module/test_froce_force_energy2
            call norm ( tab_prot % array ( i ) % orientation_x )
            call norm ( tab_prot % array ( i ) % orientation_y )
            call cross ( tab_prot % array ( i ) % orientation_z, &
                         tab_prot % array ( i ) %  orientation_x, tab_prot % array ( i ) %  orientation_y )
            ! only to reproduce sdamm, not norm on x,y
            !call norm ( tab_prot % array ( i ) % orientation_z )

            ! could read nprot, but should be well ordered
            if ( get_indice_nconf ( p_record ) > 0 ) then
                !write(*,*) 'Restart read a conformation number ', &
                !            p_record % ptrajectory % array_trajectory ( i ) % int_ocompl ( p_record % indice_nconf )
                call initialize_conformation( tab_prot % array ( i ), &
                                     p_record % ptrajectory % array_trajectory ( i ) % int_ocompl ( p_record % indice_nconf ) )
            ! test if restart does not contain flexible input, but protein is felxible (examples 128_flex)
            else if ( tab_prot % array ( i ) % iflex ) then
                !write(*,*) 'assign conf numb 1'
                call initialize_conformation( tab_prot % array ( i ), 1 )
            ! else not flexible, ntohing to do
            end if

            ! test if cell info present, default 0
            if ( p_record % box_info .eqv. .true. ) then
                tab_prot % array ( i ) % box = p_record % ptrajectory % array_trajectory ( i ) % box
            !! new added, should be in constructor anyway
            else
                tab_prot % array ( i ) % box = 0
            end if

        end do

        ! need to close the file if just restart sdamm
        close ( p_record % io_compl )
        nullify ( p_record % ptab_prot )

        ! can delete p_record if just sdamm. donot_check_array == true
        call delete_record ( p_record, .true. )
        deallocate ( p_record, stat = ierr )
        if ( ierr /= 0 ) then
            write ( *, * ) 'Problem with allocate p_restart'
            stop 1
        end if
        nullify ( p_record )

#ifdef VERBOSE
        ! test value in memory
        !write(*,'(a,/)') 'Restart initialized'
        write(*,*)
        write(*,*) '=== Restart initialized ==='
        do i = 1 , size ( tab_prot % array )
            write(*,*) 'protein ', i
            write(*,*) 'pos ', tab_prot % array ( i ) %  position
            write(*,*) 'orient_x ', tab_prot % array ( i ) %  orientation_x
            write(*,*) 'orient_y ', tab_prot % array ( i ) %  orientation_y
            write(*,*) 'orient_z ', tab_prot % array ( i ) %  orientation_z
        end do
#endif


    ! else can read all file, then send back pointer as argument, need to keep record
    ! or read on the fly, need restart variable defined as mod_record ??
    else if ( ( type_calc % enum_type_calc == sda_energy ) .or. &
              ( type_calc % enum_type_calc == sda_koff   ) .or. & !! ) then
              !! jon extension of a restart for complexe
              ( type_calc % enum_type_calc == sda_2proteins ) ) then

        ! should check if complexe or trajectory, consider complexe by default at this point
        if ( p_record % type_record == complexe_enum ) then

            allocate ( p_record % pcomplexe )

            !! should have better way now, with initialize record for both traj and enum
            call new_listcomplexe ( p_record % pcomplexe, 0.d0, .false. , .false. )

        ! else trajectory
        else

            allocate ( p_record % ptrajectory, stat = ierr )
            if ( ierr /= 0 ) then
                write ( *, * ) 'Problem with allocate ptrajectory in restart '
            end if

            if ( account_occur ) then
                write(*,*)
                write(*,*) 'INFO Input parameter account_occurency has no effect with a trajectory as restart '
                write(*,*)
            end if
            ! write(*,*) 'nb_max_complexe ', p_record % nb_max_complexe
            ! should call initalize_record, only problem because write header file in initialze_record
            call initialize_trajectory ( p_record % ptrajectory, p_record )

        end if

        ! common complexe trajectory
        ! for traj can proceed by steps or multiple of steps
        call read_file_record ( p_record )

        ! trick for account_occurency, modify here, do not care anymore later  in sda_koff, sda_energy
        if ( ( p_record % type_record == trajectory_enum ) .or. ( account_occur .eqv. .false. ) .or. &
             ( type_calc % enum_type_calc == sda_energy ) ) then

            !write(*,*) 'Set all occurencies to 1.'
            !Write(*,*)
            call reset_position ( p_record, ierr )
            if ( ierr /= 0 ) then
                write(*,*) 'ERROR Restart file is empty '
                stop 1
            end if

            ! set all occurencies to 1.
            do while ( ierr == 0 )
                p_record % p_onecompl % other_real_ocompl ( 1 ) = 1.
                call getNext ( p_record, ierr )
            end do

        end if

    end if ! end sda-koff/sda_energy

end subroutine read_restart

!> Case when a restart is not given, generate one from the pbd files
!!
!! Used only by sda_koff and sda_energy at the moment
!! @param p_record : pointer ( of type \ref mod_record::record "record" ) to which associate the restart
!! @param tab_prot : instance of \ref mod_protein::protein "protein"
subroutine restart_origin_conf ( p_record, tab_prot )
    USE mod_record
    USE mod_type_calc
    IMPLICIT NONE
    type ( record ), pointer :: p_record
    type ( array_protein_type ), target :: tab_prot

    type ( type_calculation ) :: type_calc
    integer, dimension( 3 ) :: box
    real ( kind=8 ), dimension( 3 ) :: position, orient_x, orient_y
    type ( one_complexe ), pointer :: p_tmp_onecomplexe
    integer, dimension ( 2 ) :: tmp_array_integer
    real ( kind=8 ), dimension ( 3 ) :: tmp_array_energy, tmp_other_real
    integer :: ierr

    !write(*,*) 'Entry restart_origin_conf'

    nullify ( p_tmp_onecomplexe )
    nullify ( p_record )
    tmp_array_integer = 1
    tmp_array_energy = 0.
    tmp_other_real = 1.

    position = (/0.d0, 0.d0, 0.d0/)
    orient_x = (/1.d0, 0.d0, 0.d0/)
    orient_y = (/0.d0, 1.d0, 0.d0/)
    box = 0

    type_calc % enum_type_calc = sda_2proteins
    !! compiler indicates not initialisation
    type_calc % type_calc_str = get_string ( type_calc )

    !! create artificial p_record
    allocate ( p_record, stat = ierr )
    if ( ierr /= 0 ) then
        write ( *, * ) 'Problem with allocate p_restart'
        stop 1
    end if

    nullify ( p_record % ptab_prot )
    nullify ( p_record % pcomplexe, p_record % ptrajectory )


    p_record % type_record = complexe_enum
    p_record % nb_other_real = 3
    p_record % bit_integer = O'0'
    p_record % bit_energy = O'0'
    p_record % io_compl = 0

    allocate ( p_record % pcomplexe, stat = ierr )
    if ( ierr /= 0 ) then
        write ( *, * ) 'Problem in allocating pcomplexe'
        stop 1
    end if

    !!need to assign ptab_prot before reading header
    p_record % ptab_prot => tab_prot
    !position = real ( tab_prot % array ( 2 ) % center_protein - tab_prot % array ( 1 ) % center_protein, 4 )
    position = tab_prot % array ( 2 ) % center_protein - tab_prot % array ( 1 ) % center_protein

    !! create the list complexe
    !call new_listcomplexe ( this % pcomplexe, rmsd_min, one_run, one_conf )
    call new_listcomplexe ( p_record % pcomplexe, 0.d0, .false., .false. )


    !! default complexe
    p_record % record_type_calc = type_calc
    p_record % opt_sum = .false.
    p_record % nb_max_complexe = 1


    !1! equivalent to ntraj, nstep
    call set_bit_integer( p_record, 0 )
    call set_bit_integer( p_record, 1 )
    ! only electrostatic
    call set_bit_energy ( p_record, 0 )
    !write(*,*) 'bit_integer ', p_record % bit_integer
    !write(*,*) 'bit_energy ', p_record % bit_energy

    call set_variable_field ( p_record )

    !! cretae only one onecomplexe, with default value
    allocate ( p_tmp_onecomplexe )
    call new_onecomplexe ( p_tmp_onecomplexe, p_record % nb_int, p_record % nb_real, p_record % nb_other_real )

    !call set_onecomplexe ( p_tmp_onecomplexe, position, box, orient_x, orient_y, &
    !                       tmp_array_integer, tmp_array_energy, tmp_other_real, p_record % nb_other_real, &
    !                       p_record % bit_integer, p_record % bit_energy, p_record % opt_sum, p_record % nb_types )
                           !p_record % record_type_calc )
    call set_onecomplexe_wrap ( p_record, position, box, orient_x, orient_y, &
                                tmp_array_integer, tmp_array_energy, tmp_other_real, &
                                opt_p_onecomplexe = p_tmp_onecomplexe )
    !call print_onecomplexe ( p_tmp_onecomplexe )

    call insert_new_node ( p_record % pcomplexe % p_first % p_next, p_tmp_onecomplexe )
    p_record % pcomplexe % nb_complexe = 1

    !call print_listcomplexe( p_record % pcomplexe, .false. )

end subroutine restart_origin_conf

!> Write restart
!!
!! Final step for sdamm
!! @param p_record : pointer on the \ref mod_record::record "record" to write into file
!! @param io_restart : file descriptor of the output file
subroutine write_restart ( p_record, io_restart )
    USE mod_record
    IMPLICIT NONE
    type ( record ) :: p_record
    integer :: io_restart

    integer :: io_err, memo_io

#ifdef VERBOSE
    write(*,*) 'Entry write restart'
#endif
    ! keep memo of the initial output file
    memo_io = p_record % io_compl

    if ( p_record % format_ascii .eqv. .true. ) then
        open ( io_restart, file='restart.ascii', form='formatted', action='write', iostat = io_err )
        if ( io_err /= 0 ) then
            write(*,*) 'problem in opening file restart.ascii '
        end if

    else
        open ( io_restart, file='restart.bin', form='unformatted', action='write', iostat = io_err )
        if ( io_err /= 0 ) then
            write(*,*) 'problem in opening file restart.bin '
        end if
    end if

    p_record % io_compl = io_restart
#ifdef VERBOSE
    write(*,*) 'write restart file io_restart ', io_restart
    write(*,*) 'p_record % opt_sum ', p_record % opt_sum
    write(*,*) 'p_record % nb_types ', p_record % nb_types
#endif

    ! should check sdamm, ptrajec or complexe,.. to see later
    call write_header_record ( p_record )
    ! optional true for last_record
    call write_array ( p_record % ptrajectory, p_record, .true. )

    ! close the restart file
    close ( io_restart )

    !! assign normal input/ouptut
    p_record % io_compl = memo_io

end subroutine write_restart
