!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file solva.f90
!!
!! Define an approximate calculation of the solvant accessibility

!> Compute solvent accessibility
!!
!! Compute the solvant accessibility, used to determine which atoms will be keep in sda ( with a solvent accessibility > threshold )\n
!! This version is free of use\n
!! This version use OpenMP\n
!!
!! @param na : number ofatoms
!! @param x : array with the position of atoms
!! @param a : array with the Van der Walls radius of each atom
!! @param acc : retrun the solvent accessibility for all atoms
!! @param probe : water probe
subroutine solva ( na,x,a,acc,probe )
!c       subroutine solva(nats,xyz,rads,accs,probe,zslice)

!c       na = number atom, x = xyz, is an array of atom position
!         a = array atom_vdw, accs = array of accessibility for all atoms
!c       probe and zslice parameter ( only probe can be modifed,
!c       zslice fixed at 0.05 !!
!.       zslice not used !! certainly old version
!**, not used with naccess
!*************************************************************
!**  SOLVA - SIMPLE METHOD TO CALCULATE ACCESSBILITIES **
!*************************************************************
!**
!. points on accessible surface are generates for every atom
!. surface element is d(phi)*d(cos(theta))*a
!. phi and cos(theta) divide equally: 2pi/Nphi=dphi~2/Ntheta
!. i.e. Nphi~pi*Ntheta, 4pi*a**2*dens=Nphi*Ntheta=
!. pi*Nphi**2 => Nphi=2*a*sqrt(dens), Ntheta=Nphi/pi
!. Typically, a=3-3.5A, dens=5-40, Nphi=13-44,Ntheta=4-14.
!. Dot value in A**2 as 4pi*a**2/Nphi*Ntheta should be
!. recomputed for each atom !!
!.. dots are precomputed for the sake of speed:
!.. the same number of dots for any atom
      IMPLICIT NONE

      integer, intent ( in ) :: na
      real ( kind=8 ), dimension(:,:), intent ( in ) :: x ! atom_pos(3,nat)
      real ( kind=8 ), dimension(:), intent ( in ) :: a ! atom_vdw
      real ( kind=8 ), dimension(:), intent ( out ) :: acc !  atom access
      real ( kind=4 ), intent ( in ) :: probe

      !! like for force, just intermediate tabular
      real ( kind=4 ), dimension( 99,32 ) :: xfa, yfa, zfa
      real ( kind=4 ), dimension( na ) :: a2
      integer ( kind=4 ), dimension( na ) :: iex

      real ( kind=4 ), dimension(3) :: d
      real ( kind=4 ) :: dens, armax, ddmax, pi, twopi, fopi,srf, dval
      real ( kind=4 ) :: htheta, hphi, adots, phi, cphi, sphi
      real ( kind=4 ) :: ctheta, stheta, dd, arave
      integer :: Nphi, Ntheta, iphi, itheta, ia, i, nneib, npoi

#ifdef DEBUG
    write(*,*)
    write(*,*) 'Entry solva method'
#endif

!     need to play with it to reduce the time
! this parameter tunes accuracy
      dens = 80.

      armax = 2.5
      arave = 1.7 + probe
      ddmax = ( 2. * probe + 2.0 * armax )**2

      pi = atan( 1. ) * 4. !datan( 1.d0 ) * 4.
      twopi = 2. * pi
      fopi = 4. * pi
      srf = fopi*arave**2
      Nphi = int ( 2 * arave * sqrt( dens ) )
      Ntheta = int ( Nphi / pi )
      htheta = 2./ float(Ntheta)
      hphi = twopi / float(Nphi)
      adots = float( Nphi*Ntheta )

      DO 301 iphi = 1,Nphi
          phi = ( iphi - 0.5 ) * hphi
          cphi = cos( phi )
          sphi = sin( phi )
          DO 201 itheta = 1,Ntheta
              ctheta = -1.0 + ( itheta - 0.5 ) * htheta
              stheta = sqrt ( 1.0 - ctheta**2 )
              xfa(iphi,itheta) = stheta*cphi
              yfa(iphi,itheta) = stheta*sphi
              zfa(iphi,itheta) = ctheta
  201     CONTINUE
  301 CONTINUE

      DO ia = 1,na
          a2(ia) = real ( (a(ia)+probe)**2, 4 )
      END DO

!$OMP PARALLEL DEFAULT ( NONE ) SHARED ( acc, na, ddmax, fopi, a2, Nphi, Ntheta, x, a, probe, adots, xfa, yfa, zfa ) &
!$OMP PRIVATE ( nneib, iex, dd, d, srf, dval, iphi, itheta, npoi )

!$OMP DO SCHEDULE( DYNAMIC )
AT_LOOP : &
      do ia = 1, na !100
          nneib = 0
          DO 101 i = 1,na
              iex(i) = 0
              dd = real ( (x(1,i)-x(1,ia))**2 + (x(2,i)-x(2,ia))**2 + &
                   (x(3,i)-x(3,ia))**2, 4 )
              IF (dd.GT.ddmax) iex(i) = 1
              IF (iex(i).NE.1) nneib = nneib + 1
  101     CONTINUE
!
          srf = fopi*a2(ia)
          dval = srf/adots
!       write(6,*) ia,nneib,srf
!
!.. now make the points
          npoi = 0
          DO 300 iphi = 1,Nphi
              DO 200 itheta = 1,Ntheta
                  d(1) = real ( x(1,ia) + (a(ia)+probe)*xfa(iphi,itheta), 4 )
                  d(2) = real ( x(2,ia) + (a(ia)+probe)*yfa(iphi,itheta), 4 )
                  d(3) = real ( x(3,ia) + (a(ia)+probe)*zfa(iphi,itheta), 4 )
!
!   check distances
                  DO 400 i = 1,na
                      IF (i.EQ.ia) GO TO 400
                      IF (iex(i).EQ.1) GO TO 400
                      dd = real ( (x(1,i)-d(1))**2 + (x(2,i)-d(2))**2 + &
                                  (x(3,i)-d(3))**2, 4 )
                      IF (dd.LT.a2(i)) GO TO 401
  400             CONTINUE
                  npoi = npoi + 1
  401             CONTINUE
  200         CONTINUE
  300     CONTINUE
          ! no need to protect update with OpenMP
          acc(ia) = dval*npoi
      end do AT_LOOP
!$OMP END DO
!$OMP END PARALLEL

 end subroutine solva
