!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file solva_naccess.f90
!!
!! Compute solvent accessibility with naccess algorithm

! \bug version naccess, crash on 64 bits with gfortran and OpenMP (not in debug mode or with gdb). Only at home (AMDx2) !
! zslice fixed at zslice = 0.050

!> Compute solvent accessibility
!!
!! Use naccess algorithm, can not redistribute\n
!! Need to modify \ref prepare_atom_protein and recompile in order to use this version
!!
!!
!! @param nats : number ofatoms
!! @param xyz : array with the position of atoms
!! @param rads : array with the Van der Walls radius of each atom
!! @param accs : return solvent accessibility for all atoms
!! @param probew : water probe
subroutine solva ( nats, xyz, rads, accs, probew )
!**
!*************************************************************
!**  SOLVA - LEE & RICHARDS TYPE ACCESSIBLITY CALCULATIONS **
!*************************************************************
!**
!** Calculate accessible surface area for a group of atoms.
!** The accessible area for a given atom is calculated by the
!** formula:
!**     (arcsum) x (atom radius+probe radius) x (deltaz)
!** Numerical integration is carried out over z. in each z-
!** section, the arcsum for a given atom is the arclength of
!** the circle (intersection of the atom sphere with the z-
!** section) that is not interior to any other atom circles
!** in the same z-section.
!**
!*************************************************************
!**
!**  error parameter  - this gives accuracy of calculation
!**                   - suitable values are 0.01 (high
!**                     accuracy) to 0.1 (low accuracy)
!**                   - in detail the z sections are spaced
!**                     at about error*diameter of atom
!**
!**  probe size       - radius of probe in angstroms
!**                   - suitable value for water = 1.4
!**
!*************************************************************
    IMPLICIT NONE
    !integer, parameter :: ncube = 200000, nac = 300, nint = 2000
    ! sdamm latest parameters
    integer, parameter :: ncube = 90000, nac = 400, nint = 2000
    real ( kind = 8 ), parameter :: zslice = 0.05
!. maxs = max number of atoms in PDB file
!. ncube = maximum number of cubes allowed for placing of atoms
!. nac   = maximum number of atoms per cube
!. nint  = maximum number of sphere intersections
!. normally they are taken from naccess parameter file:
!
!     the following are dimensioned to the max no of atoms (maxs)
!
    interface
        subroutine sortag(a,n,tag)
            IMPLICIT NONE
            real ( kind=8 ), dimension ( n ) :: a
            integer, dimension ( n ) :: tag
            integer, intent ( in ) :: n
        end subroutine sortag
    end interface

    integer, intent ( in ) :: nats
    real ( kind = 8 ), dimension ( :, : ), intent ( in ) :: xyz ! change my version xyz(3,nat), important (;,;) otherwise error
    real ( kind = 8 ), dimension ( : ), intent ( in ) :: rads
    real ( kind = 8 ), dimension ( : ), intent ( out ) :: accs
    real ( kind = 4 ), intent ( in ) :: probew


    integer, dimension ( nats ) :: cube
    real ( kind = 8 ), dimension ( nats ) :: rad, radsq

!   the following are dimensioned to the max no of intersections
!   of neighbouring spheres (nint)
    integer, dimension ( nint ) :: inov, tag
! defined as real*8 in sda6
    real ( kind = 8 ), dimension ( nint ) :: arci, arcf, dx, dy, d, dsq

    integer, dimension ( nac, ncube ) :: natm
    integer, dimension ( ncube ) :: itab
! defined explicitely as real*8
    real ( kind = 8 ) :: b, trig_test
    real ( kind = 8 ) :: xmin, ymin, zmin, xmax, ymax, zmax

    !! integer not declared
    integer :: ict, karc, idim, jidim, kjidim, mkji, nzp
    integer :: l, i, j, k, m, kji, n, nm, in, io, ir
    real ( kind = 8 ) :: pi, pix2, rmax, area, xr, yr, zr, rr, rrx2, rrsq, zres, zgrid, rsec2r, rsecr, rsecn, rsec2n
    real ( kind = 8 ) :: alpha, beta, ti, arcsum, t, tt, tf, parea

#ifdef DEBUG
    write(*,*)
    write(*,*) 'Entry solva naccess method'
#endif

!   initialise variables, constants
    xmin = 9999.d0
    ymin = 9999.d0
    zmin = 9999.d0
    xmax = -9999.d0
    ymax = -9999.d0
    zmax = -9999.d0
    !zslice = 0.05d0

    ict = nint
    pi = acos(-1.0)
    pix2 = 2.d0 * pi
    !write(*,*) 'nats ', nats
    !write(*,*) 'probew ', probew
    !write(*,*) 'size xyz ', size(xyz)
    !write(*,*) 'xyz(1,1)', xyz(1,1)
    !write(*,*) 'rads(1)', rads(1)

!
!     -- Radius of an atom sphere = atom radius + probe radius
!     -- Find maxima and minima
!
      rmax = 0.d0
      karc = ict
      DO i = 1,nats
          rad(i) = rads(i) + probew
          radsq(i) = rad(i)**2
          IF (rad(i).GT.rmax) rmax = rad(i)
          IF (xmin.GT.xyz(1,i)) xmin = xyz(1,i)
          IF (ymin.GT.xyz(2,i)) ymin = xyz(2,i)
          IF (zmin.GT.xyz(3,i)) zmin = xyz(3,i)
          IF (xmax.LT.xyz(1,i)) xmax = xyz(1,i)
          IF (ymax.LT.xyz(2,i)) ymax = xyz(2,i)
          IF (zmax.LT.xyz(3,i)) zmax = xyz(3,i)
      END DO

!     rmax = max diameter
      rmax = rmax*2.
!
!     -- Cubicals containing the atoms are setup.
!     -- The dimension of an edge equals the largest atom sphere radius
!     -- The cubes have a single index
!     -- Minimum of 3 by 3 cubic grid
!     -- EXIT if max cubes exceeded
!
      idim = int ( (xmax-xmin)/rmax ) + 1

      IF (idim.LT.3) idim = 3
      jidim = int ( (ymax-ymin)/rmax ) + 1
      IF (jidim.LT.3) jidim = 3
      jidim = idim*jidim
      kjidim = int ( (zmax-zmin)/rmax ) + 1
      IF (kjidim.LT.3) kjidim = 3
      kjidim = jidim*kjidim
      IF (kjidim.GT.ncube) THEN
          WRITE (6,*) 'rmax=',rmax
          write(6,*)'ERROR In solva_naccess: max cubes exceeded'
          stop 1
      END IF
      !write(*,*) 'idim ',idim
      !write(*,*) 'jidim ',jidim
      !write(*,*) 'kjidim ',kjidim

!
!     -- Prepare upto ncube cubes each containing upto nac atoms. The cube index
!     -- is kji. The atom index for each cube is in itab

      itab = 0
      DO l = 1,nats
          i = int ( (xyz(1,l)-xmin)/rmax ) + 1
          j = int ( (xyz(2,l)-ymin)/rmax )
          k = int ( (xyz(3,l)-zmin)/rmax )
          kji = k*jidim + j*idim + i
          n = itab(kji) + 1
          IF (n.GT.nac) then
            write(*,*) 'l, n nac rmax. xmin, xmax', l, n, nac, rmax, xmin, xmax
            write(*,*) 'ERROR In solva_naccess: max atoms per cube exceeded'
            stop 1
          end if
          itab(kji) = n
          natm(n,kji) = l
          cube(l) = kji
      END DO

      !write(*,*) 'itab(1:500) ', itab(1:500) !itab !itab(1)
      !write(*,*) 'natm(1,500) ',natm(1,500) !natm(1,1)
      !write(*,*) 'cube(1) ',cube(1)
      !write(*,*) 'n ',n
!
!     -- Process each atom in turn
!
      nzp = int ( 1./zslice + 0.5 )
!     write(*,*) 'nzp ', nzp
      DO ir = 1,nats
          kji = cube(ir)
          io = 0
          area = 0.
          xr = xyz(1,ir)
          yr = xyz(2,ir)
          zr = xyz(3,ir)
          rr = rad(ir)
          rrx2 = rr*2.
          rrsq = radsq(ir)
!
!     -- Find the 'mkji' cubes neighboring the kji cube
!
          DO k = -1,1,1
              DO j = -1,1,1
                  DO i = -1,1,1
                      mkji = kji + k*jidim + j*idim + i
                      IF (mkji.GE.1) THEN
                          IF (mkji.GT.kjidim) GO TO 14
                          nm = itab(mkji)
                          IF (nm.GE.1) THEN
!
!     -- record the atoms in inov that neighbor atom ir
!
                              DO m = 1,nm
                                  in = natm(m,mkji)
                                  IF (in.NE.ir) THEN
                                      io = io + 1
                                      IF (io.GT.ict) THEN
                                        write(*,*) 'ERROR in solva_naccess: intrsctns > max'
                                        STOP 1
                                      END IF
                                      dx(io) = xr - xyz(1,in)
                                      dy(io) = yr - xyz(2,in)
                                      dsq(io) = dx(io)**2 + dy(io)**2
                                      d(io) = sqrt(dsq(io))
                                      inov(io) = in
                                  END IF
                              END DO
                          END IF
                      END IF
                  END DO
              END DO
          END DO
   !write(*,*) 'dsq(1:10) ', dsq(1:10)
   14     IF (io.GE.1) THEN
!
!     z resolution determined
!
              zres = rrx2/nzp
              zgrid = xyz(3,ir) - rr - zres/2.
          ELSE
              area = pix2*rrx2
              GO TO 18
          END IF
!
!     section atom spheres perpendicular to the z axis
!
          !write(*,*) 'radsq ',radsq
          DO i = 1,nzp
              zgrid = zgrid + zres
!
!     find the radius of the circle of intersection of
!     the ir sphere on the current z-plane
!
              rsec2r = rrsq - (zgrid-zr)**2
              rsecr = sqrt(rsec2r)
              DO k = 1,karc
                  arci(k) = 0.0
              END DO
              karc = 0
              DO j = 1,io
                  in = inov(j)
!
!     find radius of circle locus
!
                  rsec2n = radsq(in) - (zgrid-xyz(3,in))**2
                  IF (rsec2n.LE.0.0) GO TO 10
                  rsecn = sqrt(rsec2n)
!
!     find intersections of n.circles with ir circles in section
!
                  IF (d(j).GE.rsecr+rsecn) GO TO 10
!
!     do the circles intersect, or is one circle completely inside the other?
!
                  b = rsecr - rsecn
                  IF (d(j).GT.abs(b)) GO TO 20
                  IF (b.LE.0.0) GO TO 9
                  GO TO 10
!
!     if the circles intersect, find the points of intersection
!
   20             karc = karc + 1
                  IF (karc.GE.ict) THEN
                      write(*,*) 'ERROR in solva_naccess : max intersections exceeded2'
                      STOP 1
                  END IF
!
!     Initial and final arc endpoints are found for the ir circle intersected
!     by a neighboring circle contained in the same plane. The initial endpoint
!     of the enclosed arc is stored in arci, and the final arc in arcf
!     law of cosines
!
                  trig_test = (dsq(j)+rsec2r-rsec2n)/ (2.*d(j)*rsecr)
                  IF (trig_test.GE.1.0) trig_test = 0.99999
                  IF (trig_test.LE.-1.0) trig_test = -0.99999
                  alpha = acos(trig_test)
!
!     alpha is the angle between a line containing a point of intersection and
!     the reference circle center and the line containing both circle centers
!
                  beta = atan2(dy(j),dx(j)) + pi
!
!     beta is the angle between the line containing both circle centers and the x-axis
!
                  ti = beta - alpha
                  tf = beta + alpha
                  IF (ti.LT.0.0) ti = ti + pix2
                  IF (tf.GT.pix2) tf = tf - pix2
                  arci(karc) = ti
                  IF (tf.GE.ti) GO TO 3
!
!     if the arc crosses zero, then it is broken into two segments.
!     the first ends at pix2 and the second begins at zero
!
                  arcf(karc) = pix2
                  karc = karc + 1
    3             arcf(karc) = tf
   10         END DO

              !write(*,*) 'arcf ', arcf
              !write(*,*) 'arci ', arci
!
!     find the accessible surface area for the sphere ir on this section
!
              IF (karc.NE.0) GO TO 19
              arcsum = pix2
              GO TO 25
!
!     The arc endpoints are sorted on the value of the initial arc endpoint
!
   19         CALL sortag(arci(1),karc,tag)
!
!***************************************
!     calculate the accessible area
!***************************************
!

              arcsum = arci(1)
              t = arcf(tag(1))
              IF (karc.EQ.1) GO TO 11
              DO k = 2,karc
                  IF (t.LT.arci(k)) arcsum = arcsum + arci(k) - t
                  tt = arcf(tag(k))
                  IF (tt.GT.t) t = tt
              END DO
   11         arcsum = arcsum + pix2 - t
!
!     The area/radius is equal to the accessible arc length x the section thickness.
!
   25         parea = arcsum*zres
!
!     Add the accessible area for this atom in this section to the area for this
!     atom for all the section encountered thus far
!
              area = area + parea
    9     END DO
!
!     scale area to vdw shell
!
          !write(*,*) 'area ',area
   18     b = area*rr
          accs(ir) = b
!------------------------------------------------------------------
! The following line converts from accessible to contact surface
!         c=(b*(rad(ir)-probe)**2)/(rad(ir)**2)
!------------------------------------------------------------------
      END DO
!    write(*,*) 'accs(1) ', accs(1)
!    write(*,*)
!    write(*,*) 'End solva_naccess method'
!    write(*,*)

end subroutine solva

subroutine sortag(a,n,tag)
      IMPLICIT NONE
      real ( kind=8 ), dimension ( n ) :: a
      integer, dimension ( n ) :: tag
      integer, intent ( in ) :: n

      integer, dimension ( 16 ) :: iu, il
      integer :: tg, i, j, k, l, m, ij
      real ( kind=8 ) :: t, tt

      do i = 1,n
          tag(i) = i
      end do

      !write(*,*) 'entry sortarg'
      !write(*,*) 'a ',a
      !write(*,*) 'n', n

      m = 1
      i = 1
      j = n
    5 IF (i.GE.j) GO TO 70
   10 k = i
      ij = (j+i)/2
      t = a(ij)
      IF (a(i).LE.t) GO TO 20
      a(ij) = a(i)
      a(i) = t
      t = a(ij)
      tg = tag(ij)
      tag(ij) = tag(i)
      tag(i) = tg
   20 l = j
      IF (a(j).GE.t) GO TO 40
      a(ij) = a(j)
      a(j) = t
      t = a(ij)
      tg = tag(ij)
      tag(ij) = tag(j)
      tag(j) = tg
      IF (a(i).LE.t) GO TO 40
      a(ij) = a(i)
      a(i) = t
      t = a(ij)
      tg = tag(ij)
      tag(ij) = tag(i)
      tag(i) = tg
      GO TO 40
   30 a(l) = a(k)
      a(k) = tt
      tg = tag(l)
      tag(l) = tag(k)
      tag(k) = tg
   40 l = l - 1
      IF (a(l).GT.t) GO TO 40
      tt = a(l)
   50 k = k + 1
      IF (a(k).LT.t) GO TO 50
      IF (k.LE.l) GO TO 30
      IF (l-i.LE.j-k) GO TO 60
      il(m) = i
      iu(m) = l
      i = k
      m = m + 1
      GO TO 80
   60 il(m) = k
      iu(m) = j
      j = l
      m = m + 1
      GO TO 80
   70 m = m - 1
      IF (m.EQ.0) RETURN
      i = il(m)
      j = iu(m)
   80 IF (j-i.GE.1) GO TO 10
      IF (i.EQ.1) GO TO 5
      i = i - 1
   90 i = i + 1
      IF (i.EQ.j) GO TO 70
      t = a(i+1)
      IF (a(i).LE.t) GO TO 90
      tg = tag(i+1)
      k = i
  100 a(k+1) = a(k)
      tag(k+1) = tag(k)
      k = k - 1
      IF (t.LT.a(k)) GO TO 100
      a(k+1) = t
      tag(k+1) = tg
      GO TO 90
end subroutine sortag
