!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> @file contacts_sdamm.f90
!! compute, and print contacts from sdamm trajectory


!> Program contacts_sdamm
!! \ingroup tools
!! Finding contacts between protein pairs or protein and a surface.
!! Used with SDAMM-type of trajectory.
program contacts_sdamm
    USE mod_record
    USE mod_sda_input
    USE mod_input_parameter
    USE genbox_subroutines
    USE analyze_subroutines
    USE com_sda
    IMPLICIT NONE
    
    interface
        subroutine create_filename ( pdb_filename, suffix, out_filename, opt_shift )
            IMPLICIT NONE
            character*128 :: pdb_filename, out_filename
            character, intent( in ) :: suffix*(*)
            integer, optional :: opt_shift
        end subroutine create_filename
        
        subroutine write_pdb_bfact( prot, pdb_file, out_filep, bfact, occup, map_2CA )
            use mod_protein
            IMPLICIT NONE

            type( protein ), intent(in) :: prot
            real( kind = 8 ), dimension(:), intent(in), optional :: bfact, occup
            character(128), intent(in) :: pdb_file
            integer, intent(in) :: out_filep
            logical, optional :: map_2CA
        end subroutine write_pdb_bfact
        
        function get_atoms_min_dist( prot1, prot2, max_contact_dist, min_contact_dist ) result ( contact_atoms )
            use mod_protein
            IMPLICIT NONE

            type ( protein ), intent (in) :: prot1, prot2
            real ( kind = 8 ), intent (in) :: max_contact_dist, min_contact_dist
            integer, dimension ( :,: ), allocatable :: contact_atoms
            ! 'helper' array
            integer, dimension ( :,: ), allocatable :: contact_test
            ! counters:
            integer :: i, j, k
            integer :: nb_contacts, max_nb_contacts, err
            real ( kind = 8 ) :: cur_dist
            real ( kind = 8 ), dimension (3) :: cur_atom1, cur_atom2, prev_atom
            logical :: new_contact
        end function
    end interface
    
    character ( 128 ) :: input_file, traj_file, filename_access
    character ( 128 ) :: output_matrix_file, output_pdb_file, output_time_file, output_stats_file
    character ( 128 ) :: tmp_arg
    character*64, dimension(:,:), allocatable :: list_pdb_f
    
    integer :: nargs, tot_nlines
    ! file pointers and status
    integer :: trajf_p, outputf_p, outputf_time_p, outputf_stats_p, ierr
    ! number of protein types; total number of molecules; number of accessible atoms; nb of frames in the trajectory; step for reading the trajectory
    integer :: nb_prot_types, nb_mol, nb_surf_atoms, nb_all_atoms, nb_frames, step
    ! if there is a surface, which molecule and which type is it
    integer :: surf_mol_nb, surf_type_nb
    ! array of dimensions nb_surf_atoms_prot1 x nb_surf_atoms_prot2, with counts of contacts - reusable!
    integer, dimension (:,:), allocatable :: contacting_atoms
    ! counters
    integer :: i, j, k, l, f, c
    integer :: int_size, position1, position2, offset1, offset2
    integer :: prot_set1, prot_set2, type1, type2, cur_nb_contacts, cur_nb_complexes
    ! temporary information, from the current comparison
    ! the 'contact_matrix' is a binary bit matrix recording which molecules contact which for each trajectory frame
    ! 'contact_atoms_tmp' is a matrix of size 2 x cur_nb_contacts - keeping info about which atoms are in contact
    integer, dimension (:,:), allocatable :: contact_matrix, contact_atoms_tmp
    ! longest_complex_per_mol - array of dim( nb_mol, 2 ) with values per each molecule: how long was the longest-lasting contact and with which molecule
    ! cur_complex_per_mol - array of dim( nb_mol, nb_mol ) with marking which molecules contact which for how long
    integer, dimension (:,:), allocatable :: longest_complex_per_mol, cur_complex_per_mol
    integer, dimension (:), allocatable :: sum_contacts1, sum_contacts2, nb_mol_per_type, sum_complex_per_mol
    
    real( kind = 8 ) :: max_contact_inter, min_contact_intra, cur_cntr_dist, max_vdw, cur_min_cntr_dist, tmp_delta
    real( kind = 4 ) :: tmp_real4
    real( kind = 8 ), dimension (:), pointer :: tmp_real
    real( kind = 8 ), dimension (:,:), pointer :: tmp_list_atom
    real( kind = 8 ), dimension(:,:), allocatable :: tmp_list_atom_surface, tmp_list_atom_surface2, avg_contacts_per_mol
    real( kind = 8 ), dimension (:), allocatable :: max_radius
    
    logical :: iflex, one_run, one_conf, isurf, map_bfact, any_contact

    type ( record ) :: record_traj
    type ( array_protein_type ), target :: tab_prot
    ! pointer to the molecule that is a surface
    type ( protein ), pointer :: surf_prot, tmp_prot, tmp_prot2
    type ( protein ) :: cur_prot1, cur_prot2
    type ( sogrid ), pointer :: tmp_sogrid
    ! pointer to the trajectory
    type ( Trajectory ), pointer :: traj
    type ( param_list ) :: list_param
    type ( one_complexe ), pointer :: cur_complexe
    type ( probe_type ) :: param_probe

    one_run = .true.
    one_conf = .true.
    map_bfact = .false.
    nb_surf_atoms = 0
    nb_all_atoms = 0
    int_size = bit_size( nb_surf_atoms )
!    write(*,*) 'int size: ', int_size
    trajf_p = 111

    nargs = 5
    if ( iargc() .lt. nargs )  then
        ! wrong number of arguments given - print info and exit
        write (*,*) 'contacts_sdamm   -   find contacts between pairs of molecules from SDAMM-type trajectory'
        write (*,*) ''
        write (*,*) 'USAGE:'
        write (*,*) '        ./contacts_sdamm   sda_input_file   max_inter_dist   min_intra_dist   type1   type2  '
        write (*,*) '                       [-traj traj]  [-step step]'
        write (*,*) ''
        write (*,*) 'max_inter_dist - maximum distance between two atoms of different molecules at which'
        write (*,*) '                 contact is assumed;'
        write (*,*) 'min_intra_dist - minimum distance between two atoms of the same molecule, which'
        write (*,*) '                 defines that these atoms make two separate contacts'
        write (*,*) 'type1, type2   - type numbers for which to perform the analysis; the order is the same'
        write (*,*) '                 as given in sda_input_file'
        write (*,*) 'traj           - trajectory; if not given, the file specified in sda_input_file will be used'
        write (*,*) 'step           - use only frames every given step (default: use all frames)'
        write (*,*) ''
        STOP
    end if
    
    call getarg( 1, input_file )
    write(*,*) 'input argument - input file: ', input_file

    call read_input_sda( tab_prot, input_file, list_param, list_pdb_f, iflex, .true., .false. )
    nb_prot_types = size( tab_prot % set_of_protein (:,1) )
    write(*,*) 'detected number of protein types: ', nb_prot_types
    allocate( nb_mol_per_type( nb_prot_types ), stat = ierr )
    if ( ierr .ne. 0 ) then
        write(*,*) 'ERROR Could not allocate nb_mol_per_type of size: ', nb_prot_types
        STOP 1
    end if

    call getarg( 2, tmp_arg )
    read (tmp_arg, '(f8.3)') max_contact_inter
    write(*,*) 'input argument - max inter-molecular distance btwn two atoms: ', max_contact_inter

    call getarg( 3, tmp_arg )
    read (tmp_arg, '(f8.3)') min_contact_intra
    write(*,*) 'input argument - min intra-molecular distance btwn two atoms: ', min_contact_intra
    
    call getarg( 4, tmp_arg )
    read (tmp_arg, '(i3)') type1
    call getarg( 5, tmp_arg )
    read (tmp_arg, '(i3)') type2
    write(*,*) 'input argument - molecule types: ', type1, ' and ', type2
    
    ! read other arguments, if given
    traj_file = list_param % ftrajectories
    step = 1

    if( iargc() .gt. 5 ) then
        i = 6 ! initial number of argument
        do while ( i .lt. iargc() )
            call getarg( i, tmp_arg )
            if ( trim( tmp_arg ) == "-traj" ) then
                call getarg( i + 1, traj_file )
                write(*,*) 'input argument - trajectory file: ', traj_file
                i = i + 2
            else if ( trim( tmp_arg ) == "-step" ) then
                call getarg( i + 1, tmp_arg )
                read (tmp_arg, '(i5)') step
                write(*,*) 'input argument - step: ', step
                i = i + 2
            else
                write(*,*) 'ERROR Wrong argument given: ', tmp_arg, '; check your input and try again'
                STOP 1
            endif
        end do
    end if
    
    flush(6)

    allocate( max_radius( nb_prot_types ), stat = ierr )
    if ( ierr .ne. 0 ) then
        write(*,*) 'ERROR Could not allocate max_radius ',nb_prot_types,'; ierr: ', ierr
        STOP 1
    end if
    
    !---------------------------------------
    ! read the accessibility
    write(*,*) ' INFO: reading accessibility...'
    do i = type1, type2
        write(*,*) '     type ',i,'...'
        nb_mol_per_type( i ) = tab_prot % set_of_protein( i, 2 ) - tab_prot % set_of_protein( i, 1 ) + 1
        if ( nb_mol_per_type( i ) .le. 0 ) then
            write(*,*) 'ERROR No molecules of type ', i, ' detected!'
            STOP 1
        end if
        
        nullify( tmp_list_atom )
        filename_access = ''
        tmp_prot => tab_prot % array( tab_prot % set_of_protein( i,1 ) )
        nb_all_atoms = size( tmp_prot % p_sogrid % list_atom_surface(1,:) )
        allocate( tmp_list_atom_surface( 3,size( tmp_prot % p_sogrid % list_atom_surface(1,:) ) ), stat = ierr )
        if ( ierr .ne. 0 ) then
            write(*,*) 'ERROR could not allocate tmp_list_atom_surface of size: ',&
                        size( tmp_prot % p_sogrid % list_atom_surface(1,:) )
            STOP 1
        end if
        tmp_list_atom_surface(:,:) = tmp_prot % p_sogrid % list_atom_surface(:,:)
        
!        write(*,*) '  tmp_prot % p_sogrid % list_atom_surface size: ', size(tmp_prot % p_sogrid % list_atom_surface(1,:))
        param_probe % probew = 0.0
        param_probe % thresold = 0.0
        
        ! problem with conversion: the 'read_input_sda' takes character*64 array, while 'create_filename' takes character*128 !
        tmp_arg = ''
        write ( tmp_arg, '(a)' ) list_pdb_f( i,1 )
!        write(*,*) ' creating access file name from: ',list_pdb_f( i,1 ), ' (tmp_arg: ',tmp_arg,')'
        call create_filename ( tmp_arg, "access", filename_access )
!        write(*,*) ' filename_access: ',filename_access
        
        if( allocated( tmp_prot % p_sogrid % which_atom_surface ) ) then
            deallocate( tmp_prot % p_sogrid % which_atom_surface )
            nullify( tmp_prot % p_which_atom_surface )
        end if
        if( allocated( tmp_prot % p_sogrid % accessibility ) ) then
            deallocate( tmp_prot % p_sogrid % accessibility )
        end if
        if( allocated( tmp_prot % p_sogrid % list_atom_surface ) ) then
            deallocate( tmp_prot % p_sogrid % list_atom_surface )
            nullify( tmp_prot % p_list_atom_surface )
        end if
        
        tmp_prot % p_sogrid % alloc_which_at_surf = .true.
        call read_accessibility ( tmp_prot % p_sogrid, filename_access, tmp_prot % center_protein, tmp_real, &
                                  max_vdw, max_radius( i ), param_probe, opt_recalculate = .false. )
        tmp_prot % p_which_atom_surface => tmp_prot % p_sogrid % which_atom_surface
        tmp_prot % p_list_atom_surface => tmp_prot % p_sogrid % list_atom_surface
        nb_surf_atoms = size( tmp_prot % p_which_atom_surface )
        
!        write(*,*) ' accessibility read; nb of atom surface: ', nb_surf_atoms
!        flush(6)
        
        if ( nb_surf_atoms .eq. 0 ) then
            write(*,*) 'ERROR No surface atoms - check whether the file is correct: ', filename_access
            STOP 1
        end if
        if ( ( tmp_prot % p_which_atom_surface( 1 ) .lt. 0 ) .or. &
             ( tmp_prot % p_which_atom_surface( 1 ) .gt. nb_all_atoms ) ) then
            write(*,*) 'ERROR surface atom numbering is wrong - check whether the accessibility file contains'
            write(*,*) 'ERROR a column with integers next to "ATOM" keywords!'
            STOP 1
        end if
        
        allocate( tmp_list_atom( 3,nb_surf_atoms ), stat = ierr )
        if ( ierr .ne. 0 ) then
            write(*,*) 'ERROR Could not allocate tmp_list_atom of size: ', nb_surf_atoms
            STOP 1
        end if
        tmp_list_atom(:,:) = 0
        do j = 1, nb_surf_atoms
            tmp_list_atom( :,j ) = tmp_list_atom_surface( :,tmp_prot % p_which_atom_surface( j ) )
        end do
        deallocate( tmp_list_atom_surface )
        tmp_prot % p_list_atom_surface => tmp_list_atom
        tmp_prot % nb_atom_surface = nb_surf_atoms
        
        ! associate the updated info with other proteins of the same type
        if ( tab_prot % set_of_protein( i,2 ) .gt. tab_prot % set_of_protein( i,1 ) ) then
            do j = tab_prot % set_of_protein( i,1 ) + 1, tab_prot % set_of_protein( i,2 )
                tmp_prot2 => tab_prot % array( j )
                
                allocate( tmp_sogrid, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'error in allocation; err: ', ierr
                    stop 1
                end if
                call allocate_sogrid( tmp_sogrid )
                tmp_sogrid % alloc_which_at_surf = .true.
                allocate ( tmp_sogrid % list_atom_surface( 3, tmp_prot % nb_atom_surface ), &
                           tmp_sogrid % which_atom_surface( tmp_prot % nb_atom_surface ), stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'error in allocation (tmp_grid); err: ', ierr
                    stop 1
                end if
                
                tmp_sogrid % which_atom_surface( : ) = tmp_prot % p_sogrid % which_atom_surface( : )
                tmp_sogrid % list_atom_surface( :,: ) = tmp_list_atom(:,:)
                tmp_prot2 % p_sogrid => tmp_sogrid
                tmp_prot2 % p_list_atom_surface => tmp_sogrid % list_atom_surface
                tmp_prot2 % p_which_atom_surface => tmp_sogrid % which_atom_surface
                tmp_prot2 % nb_atom_surface = tmp_prot % nb_atom_surface
                nullify( tmp_prot2 )
                nullify( tmp_sogrid )
            end do
        end if
        write(*,*) '     ...DONE!'
    end do ! read accessibility for each molecule type
    !---------------------------------------
    
    !---------------------------------------
    ! prepare the trajectory
    call prepare_read_traj( record_traj, trajf_p, traj_file, tot_nlines )

    if ( record_traj % record_type_calc % enum_type_calc .ne. sdamm ) then
        write(*,*) 'ERROR The file ', traj_file, ' does not contain trajectory, check the input and try again'
        stop 1
    end if
    
    call read_file_record( record_traj )
!    call print_bits( record_traj )
    call set_variable_field ( record_traj )
    traj => record_traj % ptrajectory
    
    call reset_position( record_traj )
    ierr = 0
    !---------------------------------------
    
    !---------------------------------------
    ! find the surface (if exists)
    nullify ( surf_prot )
    isurf = .false.
    nb_mol = size( tab_prot % array )
    surf_mol_nb = -1
    surf_type_nb = -1
    do i = 1, nb_mol
        if ( tab_prot % array (i) % surface ) then
            surf_type_nb = get_type( tab_prot, i )
            if ( ( surf_type_nb .eq. type1 ) .or. ( surf_type_nb .eq. type2 ) ) then
                isurf = .true.
                surf_mol_nb = i
                surf_prot => tab_prot % array (i)

                write(*,*) ' INFO: Surface found! Surface is mol.nb. ',surf_mol_nb,'; of type: ',surf_type_nb
            else
                surf_type_nb = -1
            end if
            EXIT
        end if
    end do

    nb_frames = tot_nlines / nb_mol
    write (*,*) '  INFO: Expected number of frames: ', nb_frames
    if ( ( step .gt. nb_frames ) .or. ( step .lt. 1 ) ) then
        write(*,*) 'ERROR The given step value is wrong: ', step, '; please check your input and try again!'
        call delete_array_protein ( tab_prot )
        call delete_record( record_traj, .true. )
        deallocate( tmp_real, max_radius )
        nullify( surf_prot, tmp_prot, tmp_prot2, tmp_sogrid, cur_complexe )
        
        STOP 1
    end if

    flush(6)
    !---------------------------------------
    ! the output files:
    outputf_p = trajf_p + 1
    outputf_time_p = trajf_p + 2
    outputf_stats_p = trajf_p + 3
    
    output_pdb_file = "molecule_pair"
    output_matrix_file = "contact_matrix_pair"
    output_time_file = "contacts_vs_time"
    write( output_stats_file, '(a,i2.2,a,i2.2,a)') 'contacts_stats', type1 ,'-', type2, '.dat'
    !---------------------------------------
    
    cur_prot1 = tab_prot % array ( tab_prot % set_of_protein( type1, 1 ) )
    cur_prot2 = tab_prot % array ( tab_prot % set_of_protein( type2, 1 ) )
    allocate( sum_contacts1( cur_prot1 % nb_atom_surface ), sum_contacts2( cur_prot2 % nb_atom_surface ), stat = ierr )
    if ( ierr .ne. 0 ) then
        write (*,*) "ERROR (contacts_sdamm): cannot allocate arrays ", cur_prot1 % nb_atom_surface, &
                    " and ", cur_prot2 % nb_atom_surface
        STOP
    end if
    ! can be super big matrix!
    allocate( contacting_atoms( cur_prot1 % nb_atom_surface, cur_prot2 % nb_atom_surface ), stat = ierr )
    if ( ierr .ne. 0 ) then
        write (*,*) "ERROR (contacts_sdamm): cannot allocate matrix ", cur_prot1 % nb_atom_surface, &
                    " x ", cur_prot2 % nb_atom_surface
        STOP
    end if
    contacting_atoms(:,:) = 0
    ! minimal distance between centers of the two molecules of different types
    cur_min_cntr_dist = max_radius( type1 ) / 2.0 + max_radius( type2 ) / 2.0
    call open_contact_time_output( outputf_time_p, output_time_file, type1, type2 )

    allocate( avg_contacts_per_mol( nb_mol, 2 ), stat = ierr )
    if ( ierr .ne. 0 ) then
        write (*,*) "ERROR (contacts_sdamm): cannot allocate avg_contacts_per_mol, size: ", nb_mol, "; err: ", ierr
        STOP
    end if
    avg_contacts_per_mol( :,: ) = 0.0
    
    ! the 'contact_matrix' is a binary bit matrix recording which molecules contact which for each trajectory frame
    allocate( contact_matrix( nb_mol, int( nb_mol/int_size ) + 1 ), stat = ierr )
    if ( ierr .ne. 0 ) then
        write (*,*) "ERROR (contacts_sdamm): cannot allocate contact_matrix ", nb_mol, &
                   " x ", int( nb_mol/int_size ) + 1, "; err: ", ierr
        STOP
    end if
!    write(*,*) ' contact_matrix size: ', size( contact_matrix( :,1 ) ), ' x ', size( contact_matrix( 1,: ) )
    call open_contact_matrix_output( outputf_p, output_matrix_file, type1, type2 )
    write( outputf_p ) nb_mol, int( nb_frames/step ), type1, type2
!    write ( tmp_arg, '("(",i4,"i4)")' ) nb_mol

    allocate( longest_complex_per_mol( nb_mol,2 ), cur_complex_per_mol( nb_mol,nb_mol ), stat = ierr )
    if ( ierr .ne. 0 ) then
        write (*,*) "ERROR (contacts_sdamm): cannot allocate longest_complex_per_mol and cur_complex_per_mol; size: ", nb_mol, "^2"
        STOP
    end if
    longest_complex_per_mol(:,:) = 0
    cur_complex_per_mol(:,:) = 0
    
    allocate( sum_complex_per_mol( nb_mol ), stat = ierr )
    if ( ierr .ne. 0 ) then
        write (*,*) "ERROR (contacts_sdamm): cannot allocate sum_complex_per_mol; size: ", nb_mol
        STOP
    end if
    sum_complex_per_mol(:) = 0

    
    prot_set1 = tab_prot % set_of_protein( type2, 1 )
    prot_set2 = tab_prot % set_of_protein( type1, 2 )

    ! if these are the same types, then I don't want to calculate the contacts twice for the same molecule pair
    if ( type1 .eq. type2 ) then
        prot_set2 = prot_set2 - 1
    end if
    
    cur_prot1 = tab_prot % array( tab_prot % set_of_protein( type1, 1 ) )
    allocate( tmp_list_atom_surface( 3, cur_prot1 % nb_atom_surface ), stat = ierr )
    if ( ierr .ne. 0 ) then
        write (*,*) "ERROR (contacts_sdamm): cannot allocate tmp_list_atom_surface; size: ", cur_prot1 % nb_atom_surface," x 3"
        STOP
    end if
    ! set here the original, PDB coordinates:
    tmp_list_atom_surface( :,: ) = cur_prot1 % p_list_atom_surface( :,: )
    
    if ( type1 .eq. type2 ) then
        tmp_list_atom_surface2 = tmp_list_atom_surface
    else
        cur_prot1 = tab_prot % array( tab_prot % set_of_protein( type2, 1 ) )
        allocate( tmp_list_atom_surface2( 3, cur_prot1 % nb_atom_surface ), stat = ierr )
        if ( ierr .ne. 0 ) then
          write (*,*) "ERROR (contacts_sdamm): cannot allocate tmp_list_atom_surface2; size: ", cur_prot1 % nb_atom_surface," x 3"
          STOP
        end if
        tmp_list_atom_surface2( :,: ) = cur_prot1 % p_list_atom_surface( :,: )
    end if

!------------------------------------------------------------------------------
    !MAIN LOOPS
    write(*,*) ''
    write(*,*) 'Preparations OK'
    write(*,*) 'Start of the main analysis...'
    do f = 1, nb_frames, step ! for each frame
        write( outputf_time_p, fmt = '(i5)', advance = "no" ) f
!        write(*,*) "      frame: ", f
        contact_matrix( :,: ) = 0
        cur_nb_complexes = 0
        do k = tab_prot % set_of_protein( type1, 1 ), prot_set2 !take each protein
!            write(*,*) '      k = ',k

            ! update the coordinates of surface atoms of the first protein
            cur_prot1 = tab_prot % array( k )
            cur_prot1 % p_list_atom_surface( :,: ) = tmp_list_atom_surface( :,: )
            cur_complexe => traj % array_trajectory( (f - 1)*nb_mol + k )
            call transform_surf_atom_coords( cur_prot1, cur_complexe )

            position1 = int( ( k - 1 ) / int_size ) + 1
            offset1 = modulo( k - 1, int_size )

            if ( type1 .eq. type2 ) then
                prot_set1 = k + 1
            end if
            ! get the other proteins
            do l = prot_set1, tab_prot % set_of_protein( type2, 2 )
                cur_prot2 = tab_prot % array( l )
                cur_prot2 % p_list_atom_surface( :,: ) = tmp_list_atom_surface2( :,: )
                cur_complexe => traj % array_trajectory( (f - 1)*nb_mol + l )
                cur_prot2 % position = cur_complexe % position
                
                ! get the distance between the COMs of the molecules
                call compute_distance( cur_prot1, cur_prot2, tmp_real4 )
                cur_cntr_dist = real( tmp_real4, kind = 8 )

                ! if the molecules' centers are close enough - try to find the contacting atoms
                if ( cur_cntr_dist .le. ( cur_min_cntr_dist + max_contact_inter ) ) then
!                    write(*,*) '        l = ',l
!                    write(*,*) '          cur_dist = ', cur_cntr_dist , ' --- CLOSE!'
                    flush(6)
                    call transform_surf_atom_coords( cur_prot2, cur_complexe )

                    ! compute the distances between each pair of accessible atoms and get the contacting pairs
                    contact_atoms_tmp = get_atoms_min_dist( cur_prot1, cur_prot2, max_contact_inter, min_contact_intra )

                    ! if there were found contacts...
                    if ( contact_atoms_tmp(1,1) .ne. 0 ) then
!                        write(*,*) 
!                        write(*,*) ' CONTACTS FOUND: ', contact_atoms_tmp( :,: )
!                        write(*,*) 

                        cur_nb_complexes = cur_nb_complexes + 1
                        cur_nb_contacts = size( contact_atoms_tmp( :,1 ) )
                        
                        ! update the info in 'contact_matrix'
                        position2 = int( ( l - 1 ) / int_size ) + 1
                        offset2 = modulo( l - 1, int_size )
                        contact_matrix( k, position2 ) = ibset( contact_matrix( k, position2 ), offset2 )
                        contact_matrix( l, position1 ) = ibset( contact_matrix( l, position1 ), offset1 )
                        
                        ! update the average and variance - 'online' algorithm
                        tmp_delta = real( cur_nb_contacts - avg_contacts_per_mol( k, 1 ), kind = 8 )
                        avg_contacts_per_mol( k, 1 ) = avg_contacts_per_mol( k, 1 ) + tmp_delta / ( int( ( f - 1 )/step ) + 1 )
                        avg_contacts_per_mol( k, 2 ) = avg_contacts_per_mol( k, 2 ) + &
                                                       tmp_delta*( cur_nb_contacts - avg_contacts_per_mol( k, 1 ) )
                        
                        do c = 1, cur_nb_contacts
                            contacting_atoms( contact_atoms_tmp( c,1 ), contact_atoms_tmp( c,2 ) ) = &
                                contacting_atoms( contact_atoms_tmp( c,1 ), contact_atoms_tmp( c,2 ) ) + 1
                        end do ! contacts_tmp
                    end if ! contact_atoms_tmp
                    deallocate( contact_atoms_tmp )
                end if ! cntr-2-cntr distance
                flush(6)
            end do ! l
        end do ! k
        
        ! update the longest complex matrix:
        ! j in ( 1, nb_mol ); i in ( 1, 10 )
!        write(*,*) "---FIND LONGEST COMPLEX---"
        do j = 1, size( contact_matrix( :,1 ) )
            any_contact = .false.
            do i = 1, size( contact_matrix( 1,: ) )
                do k = 0, int_size - 1
                    ! if there was a contact between the two molecules in the current frame...
                    position1 = ( i-1 )*int_size + k + 1
                    if( position1 .gt. nb_mol ) then
                       exit
                    end if
                    
                    if ( btest( contact_matrix( j,i ), k ) ) then
!           write(*,*) 'j:',j,'; position1: ', position1, '; longest: ',longest_complex_per_mol( j,2 ), &
!                      '; current: ',cur_complex_per_mol( j,position1 )
                        ! add to the counter
                        cur_complex_per_mol( j,position1 ) = cur_complex_per_mol( j,position1 ) + 1
!                         cur_complex_per_mol( position1,j ) = cur_complex_per_mol( position1,j ) + 1
                        if ( f.eq.nb_frames ) then
                            if ( longest_complex_per_mol( j,2 ) .lt. cur_complex_per_mol( j,position1 ) ) then
!                                write(*,*) '  !new longest complex! i',i,'; j: ',j, '; k: ',k,'; position1: ',position1
                                longest_complex_per_mol( j,1 ) = position1
                                longest_complex_per_mol( j,2 ) = cur_complex_per_mol( j,position1 )
                            end if
                        end if
                        
                        !registered complex
                        any_contact = .true.

                    else
                        ! if there was no contact - check whether this was the longest recorded complex...
                        if ( longest_complex_per_mol( j,2 ) .lt. cur_complex_per_mol( j,position1 ) ) then
!                            write(*,*) '  !new longest complex! i',i,'; j: ',j, '; k: ',k,'; position1: ',position1
                            longest_complex_per_mol( j,1 ) = position1
                            longest_complex_per_mol( j,2 ) = cur_complex_per_mol( j,position1 )
                        end if
                        !... and reset the counter
                        cur_complex_per_mol( j,position1 ) = 0
                        cur_complex_per_mol( position1,j ) = 0
                    end if
                end do !k
            end do !i
!            write( outputf_p, tmp_arg ) contact_matrix( j,: )
            write( outputf_p ) contact_matrix( j,: )
        
            ! update the information about which molecules were in complexes in this frame
            if( any_contact ) then
              sum_complex_per_mol( j ) = sum_complex_per_mol( j ) + 1
!             sum_complex_per_mol( l ) = sum_complex_per_mol( l ) + 1
            end if
        end do !j
        
        write( outputf_time_p, fmt = '(i7)' ) cur_nb_complexes
        flush( outputf_time_p )
        flush( outputf_p )
        flush(6)
    end do ! frames

    close( outputf_p )
    close( outputf_time_p )
    
    write(*,*) ' Wrote time statistics and binary contact map to files. '
    write(*,*) ' main analysis DONE!'
    write(*,*) ''
    write(*,*) ' Writing PDB files...'
    
    ! create pdb output for each molecule pair
    call open_contact_output( outputf_p, output_pdb_file, type1, type2 )

    sum_contacts1( : ) = sum( contacting_atoms, dim = 2 )
    sum_contacts2( : ) = sum( contacting_atoms, dim = 1 )
!    write(*,*) 'sum_contacts1: ', sum_contacts1
!    write(*,*) 'sum_contacts2: ', sum_contacts2
    ! WRITE TO THE FILE !
    ! problem with conversion: the 'read_input_sda' takes character*64 array, while 'create_filename' takes character*128 !
    tmp_arg = ''
    write ( tmp_arg, '(a)' ) list_pdb_f( type1,1 )
    call write_pdb_bfact( tab_prot % array ( tab_prot % set_of_protein( type1, 1 ) ), tmp_arg, outputf_p, &
               bfact = real( sum_contacts1, kind = 8 ) / ( nb_mol_per_type( type1 ) * ( nb_frames / step + 1 ) ) * 100., &
               map_2CA = map_bfact )
    tmp_arg = ''
    write ( tmp_arg, '(a)' ) list_pdb_f( type2,1 )
    call write_pdb_bfact( tab_prot % array ( tab_prot % set_of_protein( type2, 1 ) ), tmp_arg, outputf_p, &
              bfact = real( sum_contacts2, kind = 8 ) / ( nb_mol_per_type( type2 ) * ( nb_frames / step + 1 ) ) * 100., &
              map_2CA = map_bfact )
    close( outputf_p )
    write(*,*) '  ... DONE!'
    
    ! write out stats per molecule
    write(*,*) ' Writing statistics per molecule...'
    open( outputf_stats_p, file = output_stats_file, form = 'formatted', status = 'replace', iostat = ierr )
    if ( ierr .ne. 0 ) then
        write(*,*) " ERROR (open_contact_output): could not open the file ", output_stats_file
        STOP
    end if

    write( outputf_stats_p, '(a)' ) "# File generated by SDA ver 7.0, program contacts_sdamm.f90"
    write( outputf_stats_p, '(a,i2,a,i2)' ) "# Statistics of complex formation for types ",type1, " and ",type2
    write( outputf_stats_p, '(a)' ) "# Column 1: molecule number; column2: % time forming complexes;"
    write( outputf_stats_p, '(a)' ) "# column 3-4: avg +/- stdev of number of contacts per complex;"
    write( outputf_stats_p, '(a)' ) "# column 5-6: duration of the longest complex and with which mol."
    
    write( tmp_arg, fmt='(a)' ) '(i5,X,f6.2,X,f6.2,X,f6.2,X,i6,X,i5)'
    tmp_arg = trim( tmp_arg )
!    write(*,*) 'sum_complex_per_mol: ', sum_complex_per_mol
    
    do i = 1, nb_mol
        write( outputf_stats_p, fmt=tmp_arg ) i, &
           real( sum_complex_per_mol( i ), kind = 8 )/( real( nb_frames, kind = 8 ) / real( step, kind = 8 ) ) * 100., &
           avg_contacts_per_mol( i,1 ), sqrt( avg_contacts_per_mol( i,2 )/( nb_frames / step ) ), &
           longest_complex_per_mol( i,2 ), longest_complex_per_mol( i,1 )
    end do
    
    ! close all the files and cleanup
    close( outputf_stats_p )
    close( trajf_p )
    write(*,*) '  ... DONE!'
    flush( 6 )

    call delete_array_protein ( tab_prot )
    call delete_record( record_traj, .true. )

    deallocate( contacting_atoms, sum_contacts1, sum_contacts2 )
    deallocate( tmp_real, max_radius )
    deallocate( longest_complex_per_mol, cur_complex_per_mol )
    deallocate( sum_complex_per_mol, avg_contacts_per_mol )

    nullify( surf_prot, tmp_prot, tmp_prot2, tmp_sogrid, cur_complexe )
    
end program contacts_sdamm

subroutine transform_surf_atom_coords( prot, cur_prot )
    USE mod_protein
    USE mod_onecomplexe
    IMPLICIT NONE
    
    integer :: s
    real ( kind=8 ), dimension(3,3) :: matrix_rot
    real( kind = 8 ), dimension (3) :: cur_rotZ, cur_rotX, cur_rotY

    type ( protein ), intent (inout) :: prot
    type ( one_complexe ), intent (in) :: cur_prot
    
    cur_rotX(:) = cur_prot % orient_x(:)
    cur_rotY(:) = cur_prot % orient_y(:)
    call norm( cur_rotX )
    call norm( cur_rotY )
    call cross( cur_rotZ, cur_rotX, cur_rotY )
    matrix_rot (:,1) = cur_rotX
    matrix_rot (:,2) = cur_rotY
    matrix_rot (:,3) = cur_rotZ

    ! rotation according to the axes
    call update_rotation_protein( prot, matrix_rot )

    ! translation of the protein
    do s = 1, prot % nb_atom_surface
        prot % p_list_atom_surface( :,s ) = prot % p_list_atom_surface( :,s ) + cur_prot % position
    end do
    prot % position(:) = cur_prot % position(:)
    
end subroutine transform_surf_atom_coords

! prepare the PDB file for output
subroutine open_contact_output( file_p, file_core_name, type1, type2 )
    IMPLICIT NONE

    character (128), intent (in) :: file_core_name
    integer, intent (in) :: file_p, type1, type2
    integer :: err
    character (128) :: output_file

    write( output_file, '(a,i2.2,a,i2.2,a)' ) trim(file_core_name), type1, '-', type2, '.pdb'
    open( file_p, file = output_file, form = 'formatted', status = 'replace', iostat = err )
    if ( err .ne. 0 ) then
        write(*,*) " ERROR (open_contact_output): could not open the file ", output_file
        STOP
    end if

    write( file_p, '(a)' ) "REMARK    File generated by SDA ver 7.0, program contacts_sdamm.f90"
    write( file_p, '(a)' ) "REMARK    Molecules have default positions; the numbers in the B-factor column"
    write( file_p, '(a)' ) "REMARK    show the percentage of the simulation time the given residue was in contact"
    write( file_p, '(a)' ) "REMARK    with a residue of another molecule, averaged over all molecules of the same type"
end subroutine open_contact_output

! Prepare output file with info about number of contacts per each frame
subroutine open_contact_time_output( file_p, file_core_name, pair1, pair2 )
    IMPLICIT NONE

    character (128), intent (in) :: file_core_name
    integer, intent (in) :: file_p, pair1, pair2
    integer :: err
    character (128) :: output_file

    write( output_file, '(a,i2.2,a,i2.2,a)' ) trim(file_core_name), pair1, '-', pair2, '.dat'
    open( file_p, file = output_file, form = 'formatted', status = 'replace', iostat = err )
    if ( err .ne. 0 ) then
        write(*,*) " ERROR (open_contact_output): could not open the file ", output_file
        STOP
    end if

    write( file_p, '(a)' ) "# File generated by SDA ver 7.0, program contacts_sdamm.f90"
    write( file_p, '(a,i2,a,i2)' ) "# Number of complexes between molecules of type ", pair1, " and ", pair2
    write( file_p, '(a)' ) "# in each trajectory frame (column 1)"
end subroutine open_contact_time_output

! Prepare output file with detailed binary contact matrix per each timestep (binary file)
subroutine open_contact_matrix_output( file_p, file_core_name, pair1, pair2 )
    IMPLICIT NONE

    character (128), intent (in) :: file_core_name
    integer, intent (in) :: file_p, pair1, pair2
    integer :: err
    character (128) :: output_file

    write( output_file, '(a,i2.2,a,i2.2,a)' ) trim(file_core_name), pair1, '-', pair2, '.bin'
    open( file_p, file = output_file, form = 'unformatted', status = 'replace', iostat = err )
    if ( err .ne. 0 ) then
        write(*,*) " ERROR (open_contact_output): could not open the file ", output_file
    end if
    
!    write( file_p, '(a)' ) "# File generated by SDA ver 7.0, program contacts_sdamm.f90"
!    write( file_p, '(a,i2,a,i2)' ) "# Number of contacts between molecules of type ", pair1, " and ", pair2
!    write( file_p, '(a)' ) "# in each trajectory frame"

end subroutine open_contact_matrix_output

!> Write a pdb to the file, with specific values put in the b-factor and/or occupancy columns
!! @param prot : input, the protein
!! @param pdb_file : input, filename of the original PDB file with all the atoms of the protein
!! @param out_filep : input, pointer to the open output file
!! @param bfact : input, optional, values to put in the b-factor column (or null)
!! @param occup : input, optional, values to put in the occupancy column (or null)
!! @param map_2CA : input, optional, logical value whether to map the bfactor values to CAlpha atoms (default: .true.)
subroutine write_pdb_bfact( prot, pdb_file, out_filep, bfact, occup, map_2CA )
    use mod_protein
    IMPLICIT NONE
    
    type( protein ), intent(in) :: prot
    real( kind = 8 ), dimension(:), intent(in), optional :: bfact, occup
    character(128), intent(in) :: pdb_file
    integer, intent(in) :: out_filep
    logical, optional :: map_2CA
    
    character(80) :: pdb_line, out_str
    character(6) :: occup_str, bfact_str, zero_str
    integer :: in_filep, err, i_at, acc_at, cur_resno, last_resno, total_resno
    logical :: is_bfact, is_occup, is_map
    real( kind = 8 ) :: cur_sum
    real( kind = 8 ), dimension(:), allocatable :: out_bfact
    
    is_bfact = .false.
    is_occup = .false.
    if ( present( occup ) ) then
        is_occup = .true.
    end if
    if ( present( bfact ) ) then
        is_bfact = .true.
    end if
    
    is_map = .true.
    if ( present( map_2CA ) ) then
        is_map = map_2CA
    end if
    
    in_filep = 126
    open( in_filep, file = pdb_file, form = 'formatted', status = 'old', iostat = err )
    if ( err .ne. 0 ) then
        write(*,*) " ERROR (write_pdb_bfact): could not open the file ", pdb_file
        STOP
    end if
    
    read ( in_filep, '(a)', iostat = err ) pdb_line
    if ( err < 0 ) then
        write( *,* ) 'file is empty', pdb_file
        stop 1
    endif

    zero_str = ''
    write( zero_str, fmt = '(f6.2)' ) 0.0
    i_at = 0
    acc_at = 1

    cur_resno = -1
    last_resno = -1
    total_resno = 0
    
    if ( is_map ) then
        if ( .not.is_bfact ) then
            write(*,*) "ERROR (write_pdb_bfact): bfactor mapping set but no bfactor values given!"
            stop
        end if
        
        allocate( out_bfact( size( bfact ) ), stat = err )
        if ( err .ne. 0 ) then
            write (*,*) "ERROR (write_pdb_bfact): cannot allocate out_bfact, size: ", size( bfact )
            STOP
        end if
        out_bfact(:) = 0.0
        cur_sum = 0.0

        do while ( err .eq. 0 ) ! first, map the bfactor values to CA
            if ( pdb_line(1:4).EQ.'ATOM') then
                last_resno = cur_resno
                out_str = ''
                out_str( 1:54 ) = pdb_line( 1:54 )
                i_at = i_at + 1
                read( pdb_line( 23:26 ), '(i4)' ) cur_resno
!                write(*,*) 'cur_resno: ', cur_resno
                if ( cur_resno .ne. last_resno ) then
                    if ( total_resno .gt. 0 ) then ! this is not the first residue
!                    write(*,*) '   total_resno: ', total_resno, '; cur_sum: ', cur_sum
                        out_bfact( total_resno ) = cur_sum
                    end if
                    total_resno = total_resno + 1
                    cur_sum = 0.0
                end if
                if( prot % p_which_atom_surface( acc_at ) .eq. i_at ) then
                    cur_sum = cur_sum + bfact( acc_at )
                    acc_at = acc_at + 1
                end if

            end if ! read ATOM record
            read ( in_filep, '(a)', iostat = err ) pdb_line
        end do
        close( in_filep )
!        write(*,*) 'total_resno: ', total_resno, '; new bfactors: ', out_bfact

        open( in_filep, file = pdb_file, form = 'formatted', status = 'old', iostat = err )
        if ( err .ne. 0 ) then
            write(*,*) " ERROR (write_pdb_bfact): could not open the file ", pdb_file
            STOP
        end if

        read ( in_filep, '(a)', iostat = err ) pdb_line
        if ( err < 0 ) then
            write( *,* ) 'file is empty', pdb_file
            stop 1
        endif

        cur_resno = 1
        i_at = 0
        acc_at = 1
        do while ( err .eq. 0 )
            if ( pdb_line(1:4).EQ.'ATOM') then
                out_str = ''
                out_str( 1:54 ) = pdb_line( 1:54 )
                i_at = i_at + 1

                bfact_str = zero_str
                occup_str = zero_str
                if( prot % p_which_atom_surface( acc_at ) .eq. i_at ) then
                    if ( is_occup ) then
                        write( occup_str, fmt = '(f6.2)' ) occup( acc_at )
                    end if
                    acc_at = acc_at + 1
                end if

!                write(*,*) 'pdb_line(13:16): ',trim( pdb_line(13:16) )
                if( trim( pdb_line(13:16) ) .eq. ' CA' ) then
                    write( bfact_str, fmt = '(f6.2)' ) out_bfact( cur_resno )
                    cur_resno = cur_resno + 1
                end if
                
                out_str = trim( out_str ) // occup_str // bfact_str
                write( out_filep, fmt = '(a)' ) out_str
            end if
            read ( in_filep, '(a)', iostat = err ) pdb_line
        end do
        write( out_filep, '(a)' ) "TER"
        
        if( cur_resno .eq. 1 ) then
            write(*,*) "WARNING (write_pdb_bfact): no CAlpha atoms found!"
        end if

        close( in_filep )
        deallocate( out_bfact )

    else ! not mapping bfactor to CA:
        do while ( err .eq. 0 )
            if ( pdb_line(1:4).EQ.'ATOM') then
                out_str = ''
                out_str( 1:54 ) = pdb_line( 1:54 )
                i_at = i_at + 1

                bfact_str = zero_str
                occup_str = zero_str
                if( prot % p_which_atom_surface( acc_at ) .eq. i_at ) then
                    if ( is_occup ) then
                        write( occup_str, fmt = '(f6.2)' ) occup( acc_at )
                    end if
                    if ( is_bfact ) then
                        write( bfact_str, fmt = '(f6.2)' ) bfact( acc_at )
                    end if

                    acc_at = acc_at + 1
                end if

                out_str = trim( out_str ) // occup_str // bfact_str
                write( out_filep, fmt = '(a)' ) out_str
            end if
            read ( in_filep, '(a)', iostat = err ) pdb_line
        end do
        write( out_filep, '(a)' ) "TER"

        close( in_filep )
    end if ! for mapping the bfactor values to CA
        
end subroutine write_pdb_bfact
