!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file energy_analyze.f90
!! extract energy from sdamm trajectory

!> Extract energy from sdamm trajectory
!! \ingroup tools
program energy_analyze
      use mod_record
      use genbox_subroutines
      use analyze_subroutines
      use mod_input_parameter
      use mod_sda_input
      IMPLICIT NONE

      ! filename; string storing the current argument
      character ( 128 ) :: trajf, sda_inputf
      character ( 128 ) :: format_string
      character ( 64 ), dimension (:,:), allocatable :: list_pdbf

      ! matrices for results:
      ! mean_step -> mean energy values per each step
      ! mean_mol -> mean energy values per each molecule
      ! (analogous: var_step and var_mol)
      real ( kind = 8 ), dimension( :,: ),  allocatable :: mean_step, var_step, mean_mol, var_mol
      real ( kind = 8 ) :: delta
      ! vector with energies from 'onecomplexe'
      real ( kind = 8 ), dimension( : ), allocatable :: real_array, other_real_array
      ! easier access to an energy value
      real ( kind = 8 ) :: cur_energy

      ! pointers to open files:
      integer :: trajf_p, enerf_p, enertf_p
      ! total number of lines in the output file, number of steps, number of molecules, nb of molecule types
      integer :: tot_nlines, nsteps, nmols, nb_types
      ! iterators
      integer :: i, j, tmp
      ! number of energy terms in output; how often was the output written; number of energy types in the output
      ! n_pairs - number of combinations with repetitions
      integer :: n_eterms, nprint, n_etypes, n_pairs
      ! result of an operation
      integer :: ierr
      ! iterators for molecules, timesteps, types and energy terms (es, es-desolv, ap-desolv, lj...)
      integer :: cur_mol, cur_step, cur_type, cur_eterm, cur_eterm_out, step_shift, cur_interact
      integer :: nargs
      integer :: cur_pair
      integer, dimension( :,: ), allocatable :: pairs
      ! vector with step, mol_no etc., from 'onecomplexe'
      integer, dimension( 4 ) :: int_array

      type ( record ) :: traj
      type ( param_list ) :: input_params
      type ( array_protein_type ) :: tab_prot
      
      ! is the first molecule a surface? are the molecules flexible?
      logical :: surf, iflex
      
      iflex = .false.
      nargs = 2
      
! read the arguments
      !if ( iargc() .ne. nargs )  then
      if ( command_argument_count() /= nargs ) then
         ! wrong number of arguments given - print info and exit
         write (*,*) 'energy_analyze - compute energies and averages from SDAMM trajectory'
         write (*,*)
         write (*,*) 'USAGE:'
         write (*,*) '    ./energy_analyze   traj   sda_input_file'
         write (*,*) 'traj           - trajectory filename'
         write (*,*) 'sda_input_file - the input file used for this simulation'
         write (*,*)
         STOP
      end if

      call getarg(1, trajf)
      call getarg(2, sda_inputf)

      trajf_p = 111
      
!! init record
      call prepare_read_traj ( traj, trajf_p, trajf, tot_nlines )

      ! number of energy terms:
      n_eterms = traj % nb_real
      write(*,*) 'number of energy terms: ', n_eterms
      
      nmols = traj % nb_protein
      nsteps = int ( tot_nlines / nmols )
      write (*,*) ' expected number of molecules: ', nmols
      write (*,*) ' expected number of lines: ', tot_nlines
      write (*,*) ' expected number of steps: ', nsteps
      write (*,*) ' sum the energy terms: ', traj % opt_sum
      
      ! allocate all matrices
      allocate( mean_mol ( n_eterms, nmols ), stat=ierr )
      if ( ierr /= 0 ) then
        write (*,*) 'error: could not allocate memory for mean_mol ', n_eterms, '*', nmols
        STOP
      endif
      mean_mol (:,:) = 0.0
   
      allocate( var_mol ( n_eterms, nmols ), stat=ierr )
      if ( ierr /= 0 ) then
        write (*,*) 'error: could not allocate memory for var_mol ', n_eterms, '*', nmols
        STOP
      endif
      var_mol (:,:) = 0.0

      allocate( real_array ( n_eterms ), stat=ierr )
      if ( ierr /= 0 ) then
        write (*,*) 'error: could not allocate memory for real_array ', n_eterms
        STOP
      endif

      allocate( other_real_array ( traj % nb_other_real ), stat=ierr )
      if ( ierr /= 0 ) then
        write (*,*) 'error: could not allocate memory for other_real_array ', traj % nb_other_real
        STOP
      endif

      ! read the trajectory
      call read_file_record ( traj )
      call reset_position ( traj )
      
      ! get other information (don't read the pdbs, don't extend array)
      call read_input_sda ( tab_prot, sda_inputf, input_params, list_pdbf, iflex, .false., .false. )
      surf = input_params % isurf
      nb_types = size( tab_prot % set_of_protein (:,1) )
     ! combinations with repetitions: find all possible pairs from set of nb_types
      n_pairs = int( (nb_types + 1)*nb_types / 2 )
      
      nprint = input_params % freq_print
            
      n_etypes = ( n_eterms - 1 ) / nb_types
      
      allocate( pairs( nb_types, nb_types ), stat = ierr )
      if ( ierr /= 0 ) then
        write (*,*) 'error: could not allocate memory for pairs ', nb_types, '*', nb_types
        STOP
      endif
      
      tmp = 1
      do i = 1, nb_types
          do j = i, nb_types
              pairs( i,j ) = tmp
              pairs( j,i ) = tmp
              tmp = tmp + 1
          end do
      end do
      
      if ( traj % opt_sum ) then
        allocate( mean_step ( n_eterms, nsteps ), stat=ierr )
      else
        allocate( mean_step ( n_etypes*n_pairs + 1, nsteps ), stat=ierr )
      end if
      if ( ierr /= 0 ) then
        write (*,*) 'error: could not allocate memory for mean_step ', n_eterms, '*', nsteps
        STOP
      endif
      mean_step (:,:) = 0.0
      write(*,*) 'allocated ',size(mean_step(:,1)),' x ',size(mean_step(1,:)),' mean_step matrix'
    
      if ( traj % opt_sum ) then
        allocate( var_step ( n_eterms, nsteps ), stat=ierr )
      else
        allocate( var_step ( (n_eterms - 1)/nb_types*n_pairs + 1, nsteps ), stat=ierr )
      end if
      if ( ierr /= 0 ) then
        write (*,*) 'error: could not allocate memory for var_step ', n_eterms, '*', nsteps
        STOP
      endif
      var_step (:,:) = 0.0
    
      write (*,*) 'nprint: ', nprint, '; nb_types: ', nb_types, '; n_etypes: ', n_etypes
      
      ierr = 0
      cur_mol = 0
      cur_step = 0
      cur_type = 0
      step_shift = 0

!      write(*,*) '=== (reading energies) ==='
      if (surf) then
         nmols = nmols - 1
      end if
      
      ! read each line and get energy values:
      do i = 1, tot_nlines
        if ( ierr /= 0 ) then
            write (*,*) 'error: could not update pointer! current pos: ', traj % ptrajectory % current_pos, &
                          ', size: ', traj % ptrajectory % size_array
            STOP
        endif

        int_array (:) = 0
        real_array (:) = 0.d0
        other_real_array (:) = 0.d0
        !call return_compatible_array ( traj % p_onecompl, int_array, real_array, other_real_array, &
        !                               traj % bit_integer, traj % bit_energy, traj % opt_sum, traj % nb_types )
        call return_compatible_array_wrap ( traj, int_array, real_array, other_real_array ) !traj % p_onecompl,

        cur_mol = int_array ( 3 )
        cur_type = get_type ( tab_prot, cur_mol )
        cur_step = int_array ( 2 )
        cur_step = int ( cur_step / nprint )
        if ( cur_step .eq. 0 ) then
            step_shift = 1
        end if
        if ( i.eq.1 .and. cur_step.gt.1 ) then
            step_shift = -1 * cur_step + 1
        end if
        cur_step = cur_step + step_shift
        
!        write(*,*) ' line_no: ',i, '; cur_mol: ',cur_mol, '; cur_step: ',cur_step, '; cur_type: ', cur_type
      
        ! read each energy
        do j = 1, n_eterms
          cur_energy = real_array( j )
          cur_eterm = ceiling( real( ( j - 2 + nb_types ) / nb_types ) + 1 )
          cur_interact = modulo( j - 2 + nb_types, nb_types ) + 1
          cur_pair = pairs( cur_type, cur_interact )
          
!          write(*,*) '   cur_energy: ',cur_energy,'; energy term no.: ',j,'; cur_eterm: ',cur_eterm, &
!                     '; cur_interact: ',cur_interact, '; cur_pair: ', cur_pair
!          flush(6)
          mean_mol( j, cur_mol ) = mean_mol( j, cur_mol ) + dble( cur_energy / nsteps )
          
          ! if the energy terms per molecule type were summed up, we just read one by one
          if ( j.eq.1 .or. traj % opt_sum .or. nb_types.eq.1 ) then
            mean_step( j, cur_step ) = mean_step( j, cur_step ) + dble( cur_energy / nmols )
          ! else, if there are more than 1 type of molecules...
          else if ( j.ne.1 ) then
              cur_eterm_out = int( ( cur_eterm - 2 )*n_pairs + 1 + cur_pair )
!              write(*,*) '   writing to mean_step( ',cur_eterm_out,' )'
!              flush(6)
              mean_step( cur_eterm_out, cur_step ) = mean_step( cur_eterm_out, cur_step ) + dble( cur_energy / nmols )
          end if
        enddo
      
        call getNext ( traj, ierr ) !traj % p_onecompl,
      enddo
    
      ! and now - variances
      ierr = 0
      step_shift = 0
      call reset_position ( traj )
      do i = 1, tot_nlines
        if ( ierr /= 0 ) then
            write (*,*) 'error: could not update pointer! current pos: ', traj % ptrajectory % current_pos, &
                          ', size: ', traj % ptrajectory % size_array
            STOP
        endif

        int_array (:) = 0
        real_array (:) = 0.d0
        other_real_array (:) = 0.d0
        !call return_compatible_array ( traj % p_onecompl, int_array, real_array, other_real_array, &
        !                               traj % bit_integer, traj % bit_energy, traj % opt_sum, traj % nb_types )
        call return_compatible_array_wrap ( traj, int_array, real_array, other_real_array ) !traj % p_onecompl,

        cur_mol = int_array ( 3 )
        cur_type = get_type ( tab_prot, cur_mol )
        cur_step = int_array ( 2 )
        cur_step = int ( cur_step / nprint )
        if ( cur_step .eq. 0 ) then
            step_shift = 1
        end if
        if ( i.eq.1 .and. cur_step.gt.1 ) then
            step_shift = -1 * cur_step + 1
        end if
        cur_step = cur_step + step_shift
      
!        write(*,*) ' line_no: ',i, '; cur_mol: ',cur_mol, '; cur_step: ',cur_step, '; cur_type: ', cur_type
        do j = 1, n_eterms
          cur_energy = real_array( j )
          cur_eterm = ceiling( real( ( j - 2 + nb_types ) / nb_types ) + 1 )
          cur_interact = modulo( j - 2 + nb_types, nb_types ) + 1
          cur_pair = pairs( cur_type, cur_interact )

!          write(*,*) '   cur_energy: ',cur_energy,'; energy term no.: ',j,'; cur_eterm: ',cur_eterm, &
!                     '; cur_interact: ',cur_interact, '; cur_pair: ', cur_pair
!          flush(6)
          
          delta = cur_energy - mean_mol(j, cur_mol)
          var_mol(j, cur_mol) = var_mol(j, cur_mol) + dble( delta*delta / nsteps )

          if ( ( j .eq. 1 ) .or. traj % opt_sum .or. ( nb_types.eq.1 ) ) then
            delta = cur_energy - mean_step(j, cur_step)
            var_step(j, cur_step) = var_step(j, cur_step) + dble( delta*delta / nmols)
          else
            cur_eterm_out = int( ( cur_eterm - 2 )*n_pairs + 1 + cur_pair )
!            write(*, *) '   writing to mean_step( ', cur_eterm_out, ' )'
!            flush(6)
            delta = cur_energy - mean_step( cur_eterm_out, cur_step )
            var_step( cur_eterm_out, cur_step ) = var_step( cur_eterm_out, cur_step ) + dble( delta*delta / nmols )
          end if
        enddo
      
        call getNext ( traj, ierr )
      enddo

      ! open output files
      enerf_p = 2
      open(unit = enerf_p, file = 'ene_prot.dat', status = 'replace')

      enertf_p = 3
      open(unit = enertf_p, file = 'ene_time.dat', status = 'replace')
     
! pre-formatted strings for output
      write ( format_string, '("(i8,", i2, "es12.3)")') ( size( mean_step(:, 1) ) + size( var_step(:, 1) ) )

      ! write the results
      traj % description_column_str = ' '
      call make_header_int_pairs_ascii ( traj, .true. )
!      write (*,*) 'header: ',traj % description_column_str

      write(enertf_p, *) '# Averages for the whole system per each step'
!      if ( surf ) then
!          write (enertf_p, *) '# (excluding the first molecule, i.e., the surface)'
!      end if
      write(enertf_p, *) '# STEP    ', trim( traj % description_column_str )
      do i = 1, nsteps
        write (enertf_p, format_string) i*nprint, mean_step (:, i), var_step (:, i)
      enddo
      flush( enertf_p )

      write ( format_string, '("(i8,", i2, "es12.3)")') ( size( mean_mol(:, 1) ) + size( var_mol(:, 1) ) )

      traj % description_column_str = ' '
      call make_header_energy_ascii ( traj, .true. )
      write(enerf_p,*) '# Averages over ', nsteps, ' steps'
      write(enerf_p,*) '# MOLECULE   ', trim( traj % description_column_str )
      ! restate the original number of molecules
      if ( surf ) then
          nmols = nmols + 1
      end if
      do i = 1, nmols
        write (enerf_p, format_string) i, mean_mol (:, i), var_mol (:, i)
      enddo
      flush( enerf_p )
    
      ! clean memory
      deallocate ( mean_step, var_step, mean_mol, var_mol, real_array, other_real_array, list_pdbf )
      call delete_record ( traj, .true. )
      call delete_array_protein ( tab_prot, .false. )
      
      call exit
end program energy_analyze
