!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file genbox_subroutines.f90
!! general functions used for the gen_init_box.f90
!! \ingroup tools

!> Module declaration
MODULE genbox_subroutines
    !> list of variables that have to be included in the input file
    type genbox_list
        !! total number of molecules (prot + surf)
        integer :: nb_mol
        !! number of different types of molecules (currently, max.2)
        integer :: nb_types
        !! is molecule no.1 a surface?
        logical :: surf
        !! molecular weight of each molecule type - if only one type, molw1 must be given
        real ( kind = 8 ), dimension (:), allocatable :: molw
        !! density of each molecule type [g/L] - if only one type, dens1 must be given
        real ( kind = 8 ), dimension (:), allocatable :: dens
        !! list of pdb files with molecules' structures
        character ( 1750 ), dimension (:), allocatable :: pdbs
        !! seed for random number generator
        integer :: seed
        !! dimensions of surface [AA] and surface origin (has to aggree with PDB coordinates)
        !! if surf==.true., this has to be given
        real ( kind = 8 ), dimension ( 3 ) :: surf_dim, surf_beg
        !! Real between 0.0 and 1.0 that defines conservative should genbox be in avoiding overlaps:
        !! 1.0 = guarantee no overlaps (default) ; 0.0 = no check for overlaps
        real ( kind = 8 ) :: fit_factor
        !! Maximum number of iterations in which genbox tries to place a solute
        !! Used to prevent infinite loop at high concentrations
        integer :: max_iter
    end type genbox_list

    contains
    
    function test_all ( vect1, vect2 ) result ( res )
        IMPLICIT NONE
        real ( kind = 8 ), dimension (:), intent ( in ) :: vect1, vect2
        logical :: res

        res = all ( vect1 .eq. vect2 )
    end function test_all

    subroutine set_def_list_values ( ll )
        IMPLICIT NONE
        type ( genbox_list ), intent ( inout ) :: ll

        ll%nb_mol = 0
        ll%nb_types = 0
        ll%seed = 0
        ll%surf = .false.
        ll%surf_dim = (/ 0.0 , 0.0 , 0.0 /)
        ll%surf_beg = (/ 0.0 , 0.0 , 0.0 /)
        !NJB: new optional parameters to prevent infinite loops of high concentrations
        ll%fit_factor = 1.0
        ll%max_iter = 5000

    end subroutine set_def_list_values

    subroutine read_genbox_input( inputf, ll )
        IMPLICIT NONE

        integer, intent ( in ) :: inputf
        type ( genbox_list ), intent ( inout ) :: ll
        integer :: io_stat, i

        character ( 25 ) :: key
        character ( 1750 ) :: val
        character ( 2000 ) :: Line_Buffer

        Line_Buffer = ''
        key = ''
        val = ''
        io_stat = 0

!        write (*,*) '---starting reading---'
        do while ( io_stat .eq. 0 )
            Line_Buffer = ''
            key = ''
            val = ''
            read ( inputf, '(a)', iostat = io_stat ) Line_Buffer
            if ( ( Line_Buffer(1:1) .eq. '#' ) .or. ( Line_Buffer(1:1) .eq. '!' ) .or. &
                 ( len_trim ( Line_Buffer ) .eq. 0)  ) then
                CYCLE
            end if

            call extract_key_value ( key, val, Line_Buffer )

            select case ( key )
              case ( 'nmols' )
                call get_int_val ( val, ll%nb_mol )
                write (*,*) '---nb_mols: ', ll%nb_mol
                
              case ( 'ntype' )
                call get_int_val ( val, ll%nb_types )
                if ( ll%nb_types .gt. 0 ) then
                    allocate( ll%molw( ll%nb_types ), ll%dens( ll%nb_types ), stat = io_stat )
                    if ( io_stat /= 0 ) then
                       write (*,*) 'ERROR could not allocate memory for arrays (molw and dens) ', ll%nb_types
                       STOP 1
                    endif
                    allocate( ll%pdbs( ll%nb_types ), stat = io_stat )
                    if ( io_stat /= 0 ) then
                       write (*,*) 'ERROR could not allocate memory for array pdbs ', ll%nb_types
                       STOP 1
                    endif
                else
                    write (*,*) 'ERROR read nb_types = ', ll%nb_types
                    STOP 1
                end if
                write (*,*) '---nb_types: ', ll%nb_types
                
              case ( 'surf' )
                call get_logic_val ( val, ll%surf )
                write (*,*) '---surface: ', ll%surf
                
              case ( 'mw' )
                if ( ll%nb_types .eq. 1 ) then
                    call get_real_val( val, ll%molw( 1 ) )
                else
                    call get_real_arr_val ( val, ll%molw )
                end if
                write (*,*) '---molw: ', ll%molw
                
              case ( 'dens' )
                if ( ll%nb_types .eq. 1 ) then
                    call get_real_val( val, ll%dens( 1 ) )
                else
                    call get_real_arr_val ( val, ll%dens )
                end if
                write (*,*) '---dens: ', ll%dens
                
              case ( 'pdbfiles' )
                if ( ll%nb_types .gt. 1 ) then
                    call get_char_arr_val( val, ll%pdbs )
                else
                    ll%pdbs( 1 ) = val
                end if
                write (*,*) '---pdbfiles: ', ll%pdbs
                
              case ( 'seed' )
                call get_int_val ( val, ll%seed )
                write (*,*) '---seed: ', ll%seed
                
              case ( 'surfdim' )
                call get_real_arr_val ( val, ll%surf_dim )
                write (*,*) '---surf dim: ', ( ll%surf_dim ( i ), i = 1,3 )
                
              case ( 'surfbeg' )
                call get_real_arr_val ( val, ll%surf_beg )
                write (*,*) '---surf beg: ', ( ll%surf_beg ( i ), i = 1,3 )
                
              case ( 'fit_factor' )
                call get_real_val( val, ll%fit_factor )
                write (*,*) '---fit_factor: ', ll%fit_factor

              case ( 'max_iter' )
                call get_int_val( val, ll%max_iter )
                write (*,*) '---max_iter: ', ll%max_iter

              case default
                write (*,*) 'unrecognized keyword: ', key
            end select
        end do ! end input file loop

    end subroutine read_genbox_input

    subroutine extract_key_value ( key, val, string )
        implicit none
        character( 25 ), intent ( out ) :: key
        character( 1750 ), intent ( out ) :: val
        character( 2000 ), intent ( in ) :: string

        integer :: pos1, pos2

        pos1 = 1
        pos2 = INDEX ( string ( pos1: ), "=" )
        key = trim ( adjustl ( string ( pos1:pos1+pos2-2 ) ) )
        val = trim ( adjustl ( string ( pos2+1: ) ) )

    end subroutine extract_key_value

    subroutine get_int_val ( string, val )
        IMPLICIT NONE
        character ( 1750 ), intent ( in ) :: string
        integer, intent ( out ) :: val

        read (string, *) val
    end subroutine get_int_val

    subroutine get_real_val ( string, val )
        IMPLICIT NONE
        character ( 1750 ), intent ( in ) :: string
        real ( kind = 8 ), intent ( out ) :: val

        read ( string, * ) val
    end subroutine get_real_val

    subroutine get_logic_val ( string, lval )
        IMPLICIT NONE
        character ( 1750 ), intent ( in ) :: string
        integer :: val
        logical, intent ( out ) :: lval

        read (string, *) val

        if ( val .eq. 1 ) then
            lval = .true.
        else if ( val .eq. 0 ) then
            lval = .false.
        else
            write(*,*) 'problems reading boolean value in input file'
            STOP
        end if
    end subroutine get_logic_val

    subroutine get_char_arr_val ( string, val )
        IMPLICIT NONE
        character ( 1750 ), intent ( in ) :: string
        character ( 1750 ), dimension (:) :: val
        
        character ( 1750 ) :: substr
        integer :: i
        integer :: pos1, pos2, arr_size

        arr_size = size( val )

!        write (*,*) '---reading array: ', string

        pos1 = 1
        do i = 1, arr_size - 1
            pos2 = INDEX ( string ( pos1: ), ',' )
!            write (*,*) '---pos2: ', pos2

            substr = trim ( adjustl ( string ( pos1:pos1+pos2-2 ) ) )
            val( i ) = substr
!            write (*,*) '---substr: ', substr, '; remaining: ', string ( pos1+pos2-2: )
            pos1 = pos1 + pos2
        end do

        !last value
        substr = trim ( adjustl ( string ( pos1 + 1: ) ) )
!        write (*,*) '---substr: ', substr
        val ( arr_size ) = substr
    end subroutine get_char_arr_val
    
    subroutine get_real_arr_val ( string, val )
        IMPLICIT NONE
        character ( 1750 ), intent ( in ) :: string
        real ( kind = 8 ), dimension (:), intent ( inout ) :: val

        character ( 1750 ), dimension (:), allocatable :: substr
        integer :: i, ierr, arr_size
        
        arr_size = size( val )
        allocate( substr( arr_size ), stat = ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR could not allocate memory for array ', arr_size
            STOP 1
        endif
        
        call get_char_arr_val( string, substr )
        do i = 1, size( val )
           call get_real_val ( substr( i ), val( i ) )
        end do
        
        deallocate( substr )
    end subroutine get_real_arr_val

    subroutine wrong_input ( msg, sval )
        IMPLICIT NONE
        character ( * ), intent ( in ) :: msg, sval

        write (*,*) 'ERROR reading input file: ', trim( msg ), ' = ', trim( sval )
    end subroutine wrong_input

    subroutine vect_dist_real8 ( v1, v2, dist )
        IMPLICIT NONE
        real ( kind = 8 ), dimension (:), intent (in) :: v1, v2
        real ( kind = 8 ), intent (out) :: dist
        real ( kind = 8 ) :: tmp
        integer :: i

        if ( size ( v1 ) /= size ( v2 ) ) then
            write (*,*) 'ERROR attempt to calculate distance btwn vectors of different sizes'
            STOP 1
        end if

        tmp = 0.0
        dist = 0.0
        do i = 1, size ( v1 )
            tmp = v1 ( i ) - v2 ( i )
            dist = dist + tmp * tmp
        end do
        dist = sqrt ( dist )
    end subroutine vect_dist_real8

    subroutine rotate ( rnd_vec, new_rot, opt_short )
        IMPLICIT NONE
        ! vector with random numbers for 3 rotation angles
        real ( kind = 8 ), dimension (:), intent (in) :: rnd_vec
        ! input vector, which will contain the rotated version
        real ( kind = 8 ), dimension (:), intent (inout) :: new_rot
        logical, optional :: opt_short
        
        ! PI constant
        real ( kind = 8 ) :: PI
        ! trigonometric functions of the random angles
        real ( kind = 8 ) :: cos1, sin1, cos2, sin2, cos3, sin3
        ! rotation matrix
        real ( kind = 8 ), dimension ( 3,3 ) :: rot_mat
        ! old copy of the input vector
        real ( kind = 8 ), dimension ( 6 ) :: old_vec
        integer :: i, last
        logical :: is_short
        
        is_short = .false.
        if ( present( opt_short ) ) then
            is_short = opt_short
        end if

        if ( size ( rnd_vec ) .lt. 3 ) then
            write (*,*) 'ERROR subroutine rotate, less than 3 random numbers given!'
            stop 1
        else if ( size ( rnd_vec ) .gt. 3 ) then
            write (*,*) 'WARNING more than 3 random numbers given, only first 3 will be taken'
        end if

        PI = acos( dble ( -1.0 ) )
        cos1 = cos( rnd_vec ( 1 )*2*PI )
        sin1 = sin( rnd_vec ( 1 )*2*PI )
        cos2 = cos( rnd_vec ( 2 )*2*PI )
        sin2 = sin( rnd_vec ( 2 )*2*PI )
        cos3 = cos( rnd_vec ( 3 )*2*PI )
        sin3 = sin( rnd_vec ( 3 )*2*PI )

        rot_mat ( 1,1 ) = cos1 * cos3 - sin1 * cos2 * sin3
        rot_mat ( 1,2 ) = cos1 * sin3 + sin1 * cos2 * cos3
        rot_mat ( 1,3 ) = sin1 * sin2
        rot_mat ( 2,1 ) = - sin1 * cos3 - cos1 * cos2 * sin3
        rot_mat ( 2,2 ) = - sin1 * sin3 + cos1 * cos2 * cos3
        rot_mat ( 2,3 ) = sin2 * cos1
        rot_mat ( 3,1 ) = sin2 * sin3
        rot_mat ( 3,2 ) = - sin2 * cos3
        rot_mat ( 3,3 ) = cos2

        if ( is_short ) then
            old_vec(1:3) = new_rot
            last = 0
        else
            old_vec (:) = new_rot (:)
            last = 1
        end if
        do i = 0,last
            new_rot ( i*3 + 1 ) = rot_mat ( 1,1 )*old_vec( i*3 + 1 ) + rot_mat ( 2,1 )*old_vec ( i*3 + 2 ) + &
                              rot_mat ( 3,1 )*old_vec ( i*3 + 3 )
            new_rot ( i*3 + 2 ) = rot_mat ( 1,2 )*old_vec( i*3 + 1 ) + rot_mat ( 2,2 )*old_vec ( i*3 + 2 ) + &
                              rot_mat ( 3,2 )*old_vec ( i*3 + 3 )
            new_rot ( i*3 + 3 ) = rot_mat ( 1,3 )*old_vec( i*3 + 1 ) + rot_mat ( 2,3 )*old_vec ( i*3 + 2 ) + &
                              rot_mat ( 3,3 )*old_vec ( i*3 + 3 )
        end do
    end subroutine rotate

    subroutine write_init_pos ( nb_mol, nb_types, centers, rot )
        IMPLICIT NONE
        real ( kind = 8 ), dimension (:,:), intent ( in ) :: centers, rot
        integer, intent ( in ) :: nb_mol, nb_types
        integer :: file_out, i

        file_out = 11
        ! check the sizes of 
        if ( ( size ( centers (:, 1) ) .ne. size ( rot (:, 1) ) ) .or. &
             ( size ( centers (:, 1) ) .ne. nb_mol ) .or. ( size ( rot (:, 1) ) .ne. nb_mol ) ) then
            write (*,*) 'ERROR subroutine write_init_pos: sizes of given matrices differ: centers (:,1) = ', &
                         size ( centers (:, 1) ), '; rot (:,1) = ',size ( rot (:, 1) ),'; given nb of mol = ',nb_mol
            stop 1
        end if

        open ( file_out, file='initial_pos', form='formatted', status='replace')
        ! write the header
        write ( file_out, '(a)', advance='no' ) '##StepNb SolutNb   TransX   TransY   TransZ    Rot1X    Rot1Y    Rot1Z    Rot2X'
        write ( file_out, '(a)' ) '    Rot2Y    Rot2Z   TotEnergy        Elec        Time '
        write ( file_out, '(a,i4,i4,a)' ) '##version0.2 sdamm           trajectory', nb_mol, nb_types,'  6 15  1  1'

501    format ( 2i8, 9f9.3, 3e12.4 )

        do i = 1, nb_mol
            write ( file_out, 501 ) 0, i, centers ( i,: ), rot ( i,: ), 0., 0., 0.
        end do

        close ( file_out )
    end subroutine
    
    function get_mol_type( mols_per_type, max_mol_types, cur_mol ) result ( cur_type )
      IMPLICIT NONE
      
      integer, intent( in ) :: cur_mol, max_mol_types
      integer, dimension( max_mol_types ), intent( in ) :: mols_per_type
      integer :: cur_type
      integer :: cur_min, cur_max, max_mols
    
!      max_mol_types = size( mols_per_type )
      max_mols = sum( mols_per_type )
      if ( cur_mol.gt.max_mols ) then
        write(*,*) 'ERROR: function get_mol_type; cur_mol: ',cur_mol,' is larger than', &
                   ' total number of molecules(', max_mols,')!'
        cur_type = -1
        stop
      end if
      if ( cur_mol.eq.max_mols ) then
        cur_type = max_mol_types
        return
      end if
      
      if( max_mol_types.eq.1 ) then
        cur_type = 1
        return
      end if
      
      cur_type = 1
      cur_min = 1
      cur_max = mols_per_type( cur_type )
      do while ( cur_type.lt.max_mol_types )
        if ( cur_mol.le.cur_max ) then
          return
        end if
        cur_type = cur_type + 1
        cur_min = cur_max + 1
        cur_max = cur_max + mols_per_type( cur_type )
      end do
    end function get_mol_type
    
    ! takes the array with radii and returns indexes to this array in a descending order
    subroutine sort_radius( radius_per_type, nb_types,  sorted )
        IMPLICIT NONE
        real( kind = 8 ), dimension(*), intent( in ) :: radius_per_type
        integer, intent( in ) :: nb_types
        integer, dimension(:), allocatable , intent(inout) :: sorted
        integer, dimension(:), allocatable :: tmp
        integer :: ierr, cur_idx, insert_idx
        
        allocate( sorted( nb_types ), tmp( nb_types ), stat = ierr )
        if ( ierr.ne.0 ) then
            write(*,*) 'ERROR (sort_radius): cannot allocate sorted, size: ', nb_types
            stop
        end if
        
        ! first - trivial cases: 1 or two radii only
        if ( nb_types.lt.3 ) then
            if ( nb_types.eq.1 ) then
                sorted = 1
            else
                if ( radius_per_type( 1 ).ge.radius_per_type( 2 ) ) then
                    sorted( 1 ) = 1
                    sorted( 2 ) = 2
                else
                    sorted( 1 ) = 2
                    sorted( 2 ) = 1
                end if
            end if
            return
        end if
        
        ! real sorting:
        sorted = 0
        tmp = 0
        sorted( 1 ) = 1
!        write(*,*) '--- sorting ---'
        
        do cur_idx = 2, nb_types
!            write(*,*) 'cur_idx: ', cur_idx
            
            insert_idx = cur_idx - 1
            do while( insert_idx.gt.0 )
                if( radius_per_type( sorted( insert_idx ) ).gt.radius_per_type( cur_idx ) ) then
                    exit
                end if
                insert_idx = insert_idx - 1
            end do
            
!            write(*,*) 'insert_idx: ', insert_idx
            if ( insert_idx.eq.( cur_idx - 1 ) ) then
                sorted( cur_idx ) = cur_idx
                cycle
            end if
            ! inserting new number
            tmp( insert_idx + 1 ) = cur_idx
            
            if( insert_idx.ne.0 ) then
                tmp( 1:insert_idx ) = sorted( 1:insert_idx )
            end if
            tmp( ( insert_idx + 2 ):nb_types ) = sorted( ( insert_idx + 1 ):( nb_types - 1 ) )
            sorted = tmp
            tmp = 0
        end do
        
    end subroutine sort_radius

end module
