!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> @file mod_sda_input_tools.f90
!! \ingroup tools
!!
!! Define module mod_sda_input, for reading sda input file by the tools
!!
!! Used only by tools to retrieve the data, not by sda. Maybe common later, but here is simpler
!! Read main parameters and can load pdbs ( used by trajectory2dcd.f90 )
!!
!! Special case where arrays have a limit size. May need to modify the values and recompile
module mod_sda_input

    integer, parameter :: MAXCONF=20, MAXGRID=5

    contains

!> Function for reading the sda input file
!!
!! Can read any parameters ( just need to extend ).
!! Always read Type_Calculation, and create the array_protein.
!! Can load pdb files if load_pdb is True
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param filename : name of the sda input file
!! @param tmp_p_list : define all parameters in \ref mod_input_parameter::param_list "param_list"
!! @param all_filename_pdb : to store all pdb filenames
!! @param iflex : if some solutes are flexible
!! @param iload_pdb : if the pdb need to be loaded
!! @param extend_array : if the array for the pdb need to be extended ( to a maximum size, case of flexibility )
!! @param opt_all_prot : boolean, whether to read all molecule types or not in sda_2protein mode - default: false
subroutine read_input_sda ( tab_prot, filename, tmp_p_list, all_filename_pdb, iflex, iload_pdb, extend_array, opt_all_prot )
    USE mod_array_protein
    USE mod_input_parameter
    USE mod_interface
    USE mod_type_calc
    IMPLICIT NONE

    interface
        subroutine read_single_parameter ( key, string, inp_p_list, param_probe, been_read )
            USE com_sda
            USE mod_interface
            USE mod_input_parameter
            implicit none
            character( 25 ), intent(in) :: key
            character( 200 ), intent(in) :: string
            type ( param_list ) :: inp_p_list
            type ( probe_type ) :: param_probe
            logical :: been_read
        end subroutine read_single_parameter
    end interface

    type ( array_protein_type ) :: tab_prot
    character*64 :: filename
    type ( param_list ) :: tmp_p_list
    type ( parameter_metaldesolv ) :: param_metaldesolv
    character*64, dimension(:,:), allocatable :: all_filename_pdb
    logical, optional :: opt_all_prot
    logical :: iflex, iload_pdb, extend_array, read_all_prot

    type ( type_calculation ) :: type_calc
    ! need only for interface
    type ( probe_type ) :: param_probe
    integer :: io_unit = 2223, ierr, io
    logical :: ex, test_prot1, been_read
    !real ( kind=8 ) :: dvalue
    !real ( kind=4 ) :: fvalue
    !integer :: ivalue
    !logical :: lvalue
    character ( 200 ) :: string
    character ( 128 ) :: cvalue
    character ( 25 ) :: key
    integer :: i, count_grid, count_prot

    write(*,*)
    write(*,*) 'Entry read_input_sda ', trim( filename )

    ex = .false.
    cvalue=''
    string=''
    key=''
    count_grid = 0
    count_prot = 0
    test_prot1 = .true.
    type_calc % enum_type_calc = 0 !sda_2proteins
    param_metaldesolv % imetaldesolv = .false.
    
    read_all_prot = .false.
    if ( present( opt_all_prot ) ) then
        read_all_prot = opt_all_prot
    end if

    !> use general parameters, same as sda
    call set_default_values ( tmp_p_list )
    write(*,*) 'tmp_p_list % iflex ', tmp_p_list % iflex

    open (io_unit, file=filename, form= 'formatted', status='old', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open ', trim ( filename )
    end if
    read ( io_unit, '(a)', iostat=io ) string
    if ( io /= 0 ) then
        write( *,* ) 'file is empty ', trim ( filename )
        return
    end if

    !> allocate array
    allocate ( all_filename_pdb ( MAXGRID, MAXCONF ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot allocate all_filename_pdb ( MAXGRID, MAXCONF )'
        stop 1
    end if

BEGIN: &
    do while ( io == 0 )

        if ( ( string(1:1) == '#' ) .or. ( string(1:1) == '!' ) .or. &
             ( len_trim ( string ) == 0)  ) then
              !write(*,*) 'string not take into account:', trim(string)
              read ( io_unit, '(a)', iostat=io ) string
              CYCLE BEGIN
        end if

        !write(*,*) 'string accepted:', trim(string)
        !! to check that a keyword has been effectively read
        been_read = .false.
        call extract_key ( key, string )
        !write(*,*) 'after extract key'
        !write(*,*) 'extracted key ', key

        if ( key == 'GROUP' ) then
            call read_value ( string, cvalue )

            !! read total_solute and nb_grids from type_calculation group
            if ( cvalue == "Type_Calculation" ) then
                !write(*,*) 'Type_Calculation'
                call read_group ( tmp_p_list, cvalue, io_unit, been_read )
                type_calc = tmp_p_list % t_calc
                !write(*,*) 'type calculation', type_calc % enum_type_calc
                !write(*,*) 'type calculation ', get_string ( type_calc )
                !write(*,*) 'total_number of protein ', tmp_p_list % total_solutes
                !write(*,*) 'total_number of grids ', tmp_p_list % total_grids

                !! if sdamm sure to have nb_proteins
                if ( type_calc % enum_type_calc == sdamm ) then
                    !write(*,*) 'allocate array sdamm ', tmp_p_list % total_solutes
                    call new_array_protein ( tab_prot, tmp_p_list % total_solutes, tmp_p_list % total_grids )
                end if

            else if ( cvalue == 'Solute_Grid' ) then
                !write(*,*) 'Solute_Grid'
                call read_group ( tmp_p_list, cvalue, io_unit, been_read )

                !if ( ( count_grid == 0 ) .and. ( type_calc % enum_type_calc /= sdamm ) ) then
                if ( ( .not.read_all_prot ) .and. test_prot1 .and. ( type_calc % enum_type_calc /= sdamm ) ) then
                    test_prot1 = .false.
                    if ( tmp_p_list  % iflex ) then
                        write(*,*) 'sda type, first protein is flexible, need to load it, array 2 protein'
                        call new_array_protein ( tab_prot, 2, 2 )
                    else
                        write(*,*) 'sda type, first protein is not flexible, discard array 1 protein'
                        call new_array_protein ( tab_prot, 1, 1 )
                        read ( io_unit, '(a)', iostat=io ) string
                        cycle BEGIN
                    end if
                end if

                !write(*,*) 'number of solute ', tmp_p_list % nb_solute
                !! update grid counter
                count_grid = count_grid + 1
                !write(*,*) 'update count_grid ', count_grid

                !! update iflex, not counted here
                if ( tmp_p_list % iflex ) iflex = .true.

                if ( count_grid == 1 ) then
                  tab_prot % set_of_protein ( count_grid , 1 ) = 1
                else
                  tab_prot % set_of_protein ( count_grid , 1 ) = tab_prot % set_of_protein ( count_grid - 1 , 2 ) + 1
                end if

                tab_prot % set_of_protein ( count_grid , 2 ) = &
                                  tab_prot % set_of_protein ( count_grid , 1 ) + tmp_p_list % nb_solute - 1
                !write(*,*) 'update set_of_protein count_grid ', count_grid, &
                !           tab_prot % set_of_protein ( count_grid, 1 ), tab_prot % set_of_protein ( count_grid, 2 )

                !! create protein
                do i = tab_prot % set_of_protein ( count_grid, 1 ), tab_prot % set_of_protein ( count_grid, 2 )
                    !! subroutine new_protein ( this, irot, isurf, iflex, iimg_chg, id )
                    call new_protein ( tab_prot % array ( i ), tmp_p_list % irot, tmp_p_list % isurf, &
                         tmp_p_list % iflex, .false., i )
                end do

                ! here no need to load pdb a priori
                if ( iload_pdb ) then
                    if ( tab_prot % array ( count_prot + 1 ) % iflex ) then
                        !write(*,*) 'read_input_sda will load flex ', trim ( tmp_p_list % list_conf_name )
                        !write(*,*) 'total_conf ', tmp_p_list % total_conf
                        !write(*,*) 'count_grid ', count_grid

                        ! no need of loading pdb here
                        call load_all_pdbs ( tab_prot, tmp_p_list % list_conf_name, count_grid, &
                                             all_filename_pdb ( count_grid, : ), tmp_p_list % total_conf, extend_array )
                    else
                        !write(*,*) 'will load not flex ', trim ( tmp_p_list % pdb_filename )
                        !write(*,*) 'count_grid ', count_grid
                        call load_all_pdbs ( tab_prot , tmp_p_list % pdb_filename, count_grid, &
                                             all_filename_pdb ( count_grid, : ), tmp_p_list % total_conf, extend_array )
                    end if
                end if

                ! update count_prot
                count_prot = count_prot +  tmp_p_list % nb_solute
                !write(*,*) 'update count_prot ', count_prot

                !! reset flex tag, but general iflex is still true
                if ( tmp_p_list % iflex ) tmp_p_list % iflex = .false.

            else if ( cvalue == 'Complexes' ) then
                call read_group ( tmp_p_list, cvalue, io_unit, been_read )

            else if ( cvalue == 'Timestep' ) then
                call read_group ( tmp_p_list, cvalue, io_unit, been_read )
                !write(*,*) 'read timestep dt1', tmp_p_list % dt1

            else if ( cvalue == 'Geometry' ) then
                !write(*,*) 'read geometry'
                call read_group ( tmp_p_list, cvalue, io_unit, been_read )

             !else if ( cvalue == 'Geometry' ) then
                !write(*,*) 'read geometry'
             !   call read_group ( tmp_p_list, cvalue, io_unit, been_read )

             ! new added
             else if ( cvalue == 'RateCalculation' ) then
                !been_read = .true.
                call read_group ( tmp_p_list, cvalue, io_unit, been_read )

             else if ( cvalue == 'ReactionCriteria' ) then
                !been_read = .true.
                call read_group ( tmp_p_list, cvalue, io_unit, been_read )

             else if ( cvalue == 'ResidenceTime' ) then
                !been_read = .true.
                call read_group ( tmp_p_list, cvalue, io_unit, been_read )

             else if ( cvalue == 'MetalDesolvation' ) then
                param_metaldesolv % imetaldesolv = .true.
                !param_metaldesolv %  Zadw =  tmp_p_list % Zadw
                !param_metaldesolv %  Eadw =  tmp_p_list % Eadw
                !param_metaldesolv %  Radw =  tmp_p_list % Radw
                !param_metaldesolv %  Aadw =  tmp_p_list % Aadw

                call read_group ( tmp_p_list, cvalue, io_unit, been_read )

            end if

        end if ! GROUP

        ! all single parameters should be read here
        if ( been_read .eqv. .false. ) then
          !write(*,*) 'call read_single key ',trim(key), ' string ', trim(string)
          call read_single_parameter( key, string, tmp_p_list, param_probe, been_read )
        end if
        ! add other to avoid warning, need to be similar with read_input file
        ! what need to be independent : key, Inp_Par_List, type_calc, bool sda/tool ( init of random_generator )
        !                               been_read, problem with param_probe...
        ! ***    Random number seed, and simulation parameters

        ! check if something has been read
        ! except for GROUP because warning in read group, a possible mistake in group is checked before
        if ( ( been_read .eqv. .false. ) .and. ( key /= "GROUP" ) ) then
            write(*,*)
            !write(*,*) 'Warning: line ', Line_Read_Count,' was not read, check your spelling'
            write(*,*) 'WARNING Line not read '
            write(*,*) 'WARNING Line in question ', trim(string)
            write(*,*)
        end if

        ! outside the function
        read ( io_unit, '(a)', iostat=io ) string

    end do BEGIN
    close ( io_unit )

end subroutine read_input_sda

subroutine load_pdb ( psgrid, filename, opt_center_prot, opt_van, opt_return_center )
    USE mod_ReactionCriteria
    USE mod_setofgrid
    USE mod_van
    IMPLICIT NONE

    interface
        subroutine read_pdb ( pdb_filename, rcriter, tmp_atom_letter, tmp_atom_position, tmp_rcriter_pos, counter_prot )
            USE mod_ReactionCriteria
            CHARACTER*128 pdb_filename
            type ( react_criter ) :: rcriter
            real ( kind=8 ), dimension(:,:), intent (inout), pointer :: tmp_atom_position, tmp_rcriter_pos
            character*2, dimension(:), intent(inout), pointer :: tmp_atom_letter
            integer, intent ( in ) :: counter_prot
        end subroutine read_pdb
    end interface

    type ( sogrid ), pointer :: psgrid
    character*64 :: filename
    logical, optional :: opt_center_prot
    real ( kind= 8 ), dimension ( : ), allocatable, optional :: opt_van
    !real ( kind= 8 ), dimension ( 3 ), intent (out), optional :: opt_return_center
    real ( kind= 8 ), dimension ( 3 ), optional :: opt_return_center
    !real ( kind= 8 ), dimension ( : ), allocatable, optional :: opt_return_center

    type ( type_van ) :: tvan
    logical :: bool_center_prot, bool_opt_van

    character*128 :: filename128
    !! necessary for read_pdb from module
    type ( react_criter ) :: rcriter
    real ( kind=8 ), dimension(:,:), pointer :: tmp_atom_position, tmp_rcriter_pos
    real ( kind = 8 ), dimension ( 3 ) :: center
    character*2, dimension(:), pointer :: tmp_atom_letter
    logical :: assign_center


    integer :: i, ierr, nat !counter_prot
    real( kind = 8 ) :: max_dist, tmp_dist

    !write(*,*)
    !write(*,*) 'Entry load_pdb, file ', trim ( filename )

    !! init, rcriter needed by read_pdb, not used here
    rcriter % type_reaction = off_rc
    nullify ( rcriter % p_assoc_rate )
    nullify ( tmp_rcriter_pos )
    nullify ( tmp_atom_letter, tmp_atom_position )
    center = 0.d0
    max_dist = 0d0
    tmp_dist = 0d0

    !! deal with option
    if ( present ( opt_center_prot ) ) then
        bool_center_prot = opt_center_prot
    !! default will shift protein, such as center is at zero
    else
        bool_center_prot = .true.
    end if

    !! will create  and fill array of van radius
    if ( present ( opt_van ) ) then
        bool_opt_van = .true.
        ! initialize the new module, will check if the user provides additional atom types
        ! first implementation, allocate and deallocate every time, later should pass an argument
        if ( allocated ( tvan % array_atom ) .eqv. .false. ) &
            call init_van( tvan )
    !! default does not return full array of van
    else
        bool_opt_van = .false.
    end if

    assign_center = .FALSE.
    if( present( opt_return_center ) ) then
        assign_center = .TRUE.
    end if

    filename128 = filename

    ! subroutine read_pdb ( pdb_filename, rcriter, tmp_atom_letter, tmp_atom_position, tmp_rcriter_pos, counter_prot )
    call read_pdb ( filename128, rcriter, tmp_atom_letter, tmp_atom_position, tmp_rcriter_pos, 0 )

    ! allocate tmp_atom_position
    ! write(*,*) 'allocate list_atom_surface ', size ( tmp_atom_position(1,:) )
    nat = size ( tmp_atom_position(1,:) )
    !write(*,*) 'nat ', nat

    allocate ( psgrid, stat = ierr )
    if ( ierr /=0 ) then
        write(*,*) 'cannot allocate psgrid'
        stop 1
    end if

    call allocate_sogrid( psgrid ) !, get_sizeof_sgrid()  )
    !write(*,*) 'ok allocate sogrid'

    allocate ( psgrid % list_atom_surface ( 3, nat ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'ERROR allocation'
        stop 1
    end if

    !! need to allocate array
    if ( bool_opt_van ) then
        allocate ( opt_van ( nat ) )
    end if

    do i = 1, nat
        center = center + tmp_atom_position(:,i)
        if ( bool_opt_van ) then
            !call van1 ( tmp_atom_letter ( i ), opt_van ( i ) )
            call get_VdW ( tvan, tmp_atom_letter ( i ), opt_van ( i ) )
        end if
    end do
    center = center / nat
    if( assign_center ) then
        opt_return_center = center
!        write(*,*) 'center of this protein: ',center
    end if

    ! should make the shift of center, which center ??? from traj or from read_pdb ???
    if ( bool_center_prot ) then
        do i = 1, nat
            psgrid % list_atom_surface(:,i) = tmp_atom_position(:,i) - center
            tmp_dist = psgrid % list_atom_surface(1,i)**2 + psgrid % list_atom_surface(2,i)**2 + &
                       psgrid % list_atom_surface(3,i)**2
            if ( tmp_dist > max_dist**2 ) then
                max_dist = sqrt( tmp_dist )
            end if
        end do
    else
        do i = 1, nat
            psgrid % list_atom_surface(:,i) = tmp_atom_position(:,i)
            tmp_dist = psgrid % list_atom_surface(1,i)**2 + psgrid % list_atom_surface(2,i)**2 + &
                       psgrid % list_atom_surface(3,i)**2
            if ( tmp_dist > max_dist**2 ) then
                max_dist = sqrt( tmp_dist )
            end if
        end do
    end if

    !! update psgrid % nb_atom_surface
    psgrid % distance_max = max_dist
    psgrid % nb_atom_surface = nat
    !write(*,*) 'psogrid nat ', psgrid % nb_atom_surface

    !write(*,*) 'position at1 ', psgrid % list_atom_surface ( :,1 )
    !write(*,*) ' distance from center ', sqrt ( dot_product ( psgrid % list_atom_surface ( :,1 ), &
    !                                            psgrid % list_atom_surface ( :,1 ) ) )

    !! deallocate array not used
    deallocate ( tmp_atom_letter )
    deallocate ( tmp_atom_position )

    if ( allocated ( tvan % array_atom ) .eqv. .true. ) &
        call delete_van( tvan )

end subroutine load_pdb

!> Laod all pdbs in case of flexibility, save with the intermediate of tab_prot and flexibility
!!
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param filename : name of the listconformation file
!! @param nb_prot_group : number of group of solutes
!! @param onedim_filename_pdb :
!! @param total_conf : number of conformation
!! @param extend_array : if the array for the pdb need to be extended ( to a maximum size, case of flexibility )
subroutine load_all_pdbs ( tab_prot, filename, nb_prot_group, onedim_filename_pdb, total_conf, extend_array )
    USE mod_array_protein
    USE mod_flexibility
    IMPLICIT NONE

! defined in the same scope ? not needed ? even give error at compilation
!    interface
!        subroutine load_pdb ( psgrid, filename, opt_center_prot, opt_van, opt_return_center )
!            USE mod_ReactionCriteria
!            USE mod_setofgrid
!            USE mod_van
!            IMPLICIT NONE
!            type ( sogrid ), pointer :: psgrid
!            character*64 :: filename
!            logical, optional :: opt_center_prot
!            real ( kind= 8 ), dimension ( : ), allocatable, optional :: opt_van
            !real ( kind = 8 ), dimension ( 3 ), intent (out), optional :: opt_return_center
            !real ( kind = 8 ), dimension ( 3 ), optional :: opt_return_center
!            real ( kind= 8 ), dimension ( : ), allocatable, optional :: opt_return_center
!        end subroutine load_pdb

!        subroutine read_listconf_pdb ( prot, filename, onedim_filename_pdb, total_conf, bool_extend )
!            USE mod_protein
!            USE mod_flexibility
!            USE mod_input_parameter
!            !USE param
!            IMPLICIT NONE
!            type ( protein ) :: prot
!            character*64 :: filename
!            character*64, dimension ( MAXCONF ) :: onedim_filename_pdb
!            integer :: total_conf
!            logical :: bool_extend
!        end subroutine read_listconf_pdb
!    end interface

    type ( array_protein_type ) :: tab_prot
    character*64 :: filename
    integer :: nb_prot_group
    character*64, dimension ( MAXCONF ) :: onedim_filename_pdb
    integer :: total_conf
    logical :: extend_array

    type ( sogrid ), pointer :: p_tmp_sogrid
    type ( list_flex_type ), pointer :: p_listflex
    type ( flexible ), pointer :: p_tmp_flex
    character*128 :: filename128, method128
    integer :: prot1, i !, max_natoms

    !write(*,*) 'Entry load_all_pdbs', trim ( filename )
    !write(*,*) 'total conf ', total_conf
    if ( total_conf > MAXCONF ) then
        write(*,*) 'Number of conformations > MAXCONF=',MAXCONF, 'recompile trajectory2dcd '
        stop 1
    end if
    !write(*,*) 'filename ', filename
    !write(*,*) 'onedim filename', onedim_filename_pdb


    nullify ( p_tmp_sogrid, p_listflex, p_tmp_flex )
    filename128=''
    method128='random'

    prot1 = tab_prot % set_of_protein ( nb_prot_group, 1 )
    ! write(*,*) 'prot1 ', prot1

    ! if prot2 flexible ??
    if ( tab_prot % array ( prot1 ) % iflex ) then
        !write(*,*) 'flex'
        filename128 = filename
        !! allocate new_flexible
        allocate ( p_listflex )
        !! 2 and 2 correpsond to total_conf and max_conf, to update
        call new_list_flex ( p_listflex, method128, filename128, 0, 0, .false., 0., 0. )!, 0., 1 )

        !! set new flexible, allocated in the function
        call set_flexible ( tab_prot % array ( prot1 ), p_listflex )

        call read_listconf_pdb ( tab_prot % array ( prot1 ), filename, onedim_filename_pdb, total_conf, extend_array )

        !! associate so_grid
        tab_prot % array ( prot1 ) % p_sogrid => p_listflex % p_first % p_next % p_data
        if ( associated ( tab_prot % array ( prot1 ) % p_sogrid ) .eqv. .false. ) then
            write(*,*) 'load_all_pdbs p_sogrid is NOT associated, problem'
            stop 1
        end if
        !! just recopy to be sure, done extra before
        tab_prot % array ( prot1 ) % p_flex % p_current_node => p_listflex % p_first % p_next

    ! not flexible
    else
        !write(*,*) 'load simple pdb'
        call load_pdb ( p_tmp_sogrid, filename, opt_return_center = tab_prot % array( prot1 ) % center_protein )
        tab_prot % array ( prot1 ) % p_sogrid => p_tmp_sogrid
        tab_prot % array ( prot1 ) % p_list_atom_surface => p_tmp_sogrid % list_atom_surface
        if ( p_tmp_sogrid % alloc_which_at_surf ) then
            tab_prot % array ( prot1 ) % p_which_atom_surface => p_tmp_sogrid % which_atom_surface
        end if
        onedim_filename_pdb ( 1 ) = filename
        !write(*,*) 'onedim filename ', onedim_filename_pdb
    end if

    ! if other protein in group should associate, sdamm
    ! assign other prot sogrid to the new created
    ! write(*,*) 'nb group ', tab_prot % set_of_protein ( nb_prot_group, 2 ), tab_prot % set_of_protein ( nb_prot_group, 1 )
    if ( tab_prot % set_of_protein ( nb_prot_group, 2 ) > tab_prot % set_of_protein ( nb_prot_group, 1 ) ) then
        ! write(*,*) 'apply new grid to all group of set of protein'

        do i =  tab_prot % set_of_protein ( nb_prot_group, 1 ) + 1, tab_prot % set_of_protein ( nb_prot_group, 2 )
            if ( tab_prot % array ( prot1 ) % iflex ) then
                !write(*,*) 'all flexible'
                call set_flexible ( tab_prot % array ( i ), p_listflex )
                tab_prot % array ( i ) % p_flex % p_current_node => p_listflex % p_first % p_next
                tab_prot % array ( i ) % p_sogrid => tab_prot % array ( i ) % p_flex % p_current_node % p_data

            else
                !write(*,*) 'all not flexible'
                tab_prot % array ( i ) % p_sogrid => p_tmp_sogrid
                tab_prot % array ( i ) % p_list_atom_surface => p_tmp_sogrid % list_atom_surface
                if ( p_tmp_sogrid % alloc_which_at_surf ) then
                    tab_prot % array ( i ) % p_which_atom_surface => p_tmp_sogrid % which_atom_surface
                end if
            end if
        end do
    !else
    !    write(*,*) 'unique protein in this set of protein'
    end if

end subroutine load_all_pdbs


!> Read the list of conformations if flexible
!!
!! particular version for this tool. may merged to common module
!!
!! @param prot : instance of \ref mod_protein::protein "protein"
!! @param filename : name of the list of conformations
!! @param onedim_filename_pdb : will store the pdb names
!! @param total_conf : will store the number of conformations
!! @param bool_extend : if the arrays must be extended to all have the same size ( dcd format )
subroutine read_listconf_pdb ( prot, filename, onedim_filename_pdb, total_conf, bool_extend )
    USE mod_protein
    USE mod_flexibility
    USE mod_input_parameter
    USE mod_interface
    IMPLICIT NONE

    interface
!        subroutine load_pdb ( psgrid, filename, opt_center_prot, opt_van )
!            USE mod_ReactionCriteria
!            USE mod_setofgrid
!            IMPLICIT NONE
!            type ( sogrid ), pointer :: psgrid
!            character*64 :: filename
!            logical, optional :: opt_center_prot
!            real ( kind= 8 ), dimension ( : ), allocatable, optional :: opt_van
!        end subroutine load_pdb

        subroutine extend_array_double ( array_in, new_size )
            IMPLICIT NONE
            real ( kind = 8 ), dimension ( :, : ), allocatable :: array_in
            integer :: new_size
        end subroutine extend_array_double
    end interface

    type ( protein ) :: prot
    character*64 :: filename
    character*64, dimension ( MAXCONF ) :: onedim_filename_pdb
    integer :: total_conf
    logical :: bool_extend

    type ( param_list ) :: tmp_p_list
    type ( sogrid ), pointer :: p_tmp_sogrid
    character ( 200 ) :: string
    character ( 128 ) :: cvalue
    character ( 25 ) :: key
    character*64 :: filename64
    character ( 128 ) :: copy_original_pdbname
    integer :: io_unit = 2222, ierr, io, status
    integer :: i, nb_max_atoms, first_mode, current_conf
    logical :: ex, bool_change_size, explicit, been_read

    real ( kind = 8 ), dimension ( :, : ), pointer :: p_tmp

    write(*,*) 'Entry read_listconf ', trim(filename)
    !write(*,*) 'filename ', trim(filename)

    nullify ( p_tmp_sogrid )
    nullify ( p_tmp )
    nb_max_atoms = 0
    ex = .false.
    bool_change_size = .false.
    filename64 = ''
    first_mode = 1
    current_conf = 0
    explicit = .true.
    copy_original_pdbname = ''
    !total_conf = 1

    !! simple version of read_listconf in read_input file
    inquire ( file=filename, exist = ex )
    !write(*,*) ' check filename ', trim ( filename ), ' ex:',ex
    if ( .not.ex ) then
        write( io6,* ) 'ERROR Input file is missing:', trim( filename )
        stop 1
    endif

    call set_default_values ( tmp_p_list )
    key=''
    !string =''
    io = 0
    status = 1
    ierr = 1

    !write(*,*) 'read_list_conf total_conf ', total_conf

    !! trick to use the same group than Solute_Grid when reading, could make a group Conf
    cvalue='Solute_Grid'

    open (io_unit, file=filename, form= 'formatted', status='old', iostat = ierr )
    read ( io_unit, '(a)', iostat=io ) string
    if ( io < 0 ) then
        write( *,* ) 'file is empty', trim ( filename )
        return
    endif

BEGIN: &
    do while ( io == 0 )

        if ( ( string(1:1) == '#' ) .or. ( string(1:1) == '!' ) .or. &
             ( len_trim ( string ) == 0)  ) then
              !write(*,*) 'string not take into account:', trim(string)
              read ( io_unit, '(a)', iostat=io ) string
              CYCLE BEGIN
        end if

        if ( string == 'EXPLICIT' ) then
            !write(*,*) 'EXPLICIT reading of list conf'
            explicit = .true.
            read ( io_unit, '(a)', iostat=io ) string
            CYCLE BEGIN
        else if  ( string == 'IMPLICIT' ) then
            !write(*,*) 'IMPLICIT reading of list conf'
            explicit = .false.
            read ( io_unit, '(a)', iostat=io ) string
            CYCLE BEGIN
        endif
!
        been_read = .false.
        call extract_key ( key, string )

        if ( ( .not.explicit ) .and. ( key == 'first_mode' ) ) then
            !write (*,*) 'first_mode string', trim ( string )
            call read_value ( string, first_mode )
            !write ( string, fmt='(i4)' ) first_mode
            !write (*,*) 'first mode integer ', first_mode
            first_mode = first_mode - 1

            read ( io_unit, '(a)', iostat=io ) string
            cycle begin
        end if

        if ( key == 'GROUP' ) then

            !write (*,*) 'read GROUP explict=', explicit

            first_mode = first_mode + 1
            current_conf = current_conf + 1

            !! with the trick above on cvalue read the same group Solute_Grid than normal
            if ( ( explicit ) .or. &
                 ( ( explicit .eqv. .false. ) .and. ( current_conf == 1 ) ) ) then
                call read_group ( tmp_p_list, cvalue, io_unit, been_read )
            end if
            !write(*,*) 'pdb file ', trim ( tmp_p_list % pdb_filename )

            if  ( explicit .eqv. .false. ) then
                !write(*,*) 'explicit is false make a loop total_conf ', total_conf
                copy_original_pdbname = tmp_p_list % pdb_filename
                !write(*,*) 'original pdb_name:',trim(copy_original_pdbname),'!!'
             else
                !write(*,*) 'explicit is true set total_conf to 1'
                total_conf = 1
             end if

            !! if explicit total_conf is one
            do i = 1, total_conf

                if ( explicit .eqv. .false. ) then
                    tmp_p_list % pdb_filename = copy_original_pdbname
                    !write(*,*) 'before modif of name',tmp_p_list % pdb_filename
                    call modify_fname_implicit ( tmp_p_list, first_mode )
                    first_mode = first_mode + 1
                end if

                !write(*,*) 'will load pdb',trim (tmp_p_list % pdb_filename ),'!!'
                !! should use load_all_pdbs
                call load_pdb ( p_tmp_sogrid, tmp_p_list % pdb_filename )

                !! insert at the last position
                call insert_sogrid ( prot % p_flex % plist_flex, p_tmp_sogrid, .false., status )
                !write(*,*) 'status after insert_sogrid ', status
                !! increment max_conf, could be in insert_sogrid
                prot % p_flex % plist_flex % total_conf = prot % p_flex % plist_flex % total_conf + 1


                !! copy name of pdbs
                write(*,*) 'copy name pdb to array at position ', prot % p_flex % plist_flex % total_conf
                filename64 = trim( tmp_p_list % pdb_filename)
                onedim_filename_pdb ( prot % p_flex % plist_flex % total_conf ) = filename64
                !write(*,*) 'ok'

                !! update max_atoms, so_grid % nb_atoms update done in load_pdb
                !! not really working, to check later, not a bug but uncorrectly updated
                if ( p_tmp_sogrid % nb_atom_surface > nb_max_atoms ) then
                    bool_change_size = .true.
                    nb_max_atoms =  p_tmp_sogrid % nb_atom_surface
                    !write(*,*) 'update nb_max_atoms in read_list_conf ', nb_max_atoms
                !else
                !    write(*,*) 'not increment nb_max_atoms'
                end if
            end do

        end if ! GROUP

        if ( ( been_read .eqv. .false. ) .and. ( key /= "GROUP" ) ) then
            write(*,*)
            write(*,*) 'WARNING A line was not read, check your spelling'
            write(*,*) 'WARNING Line in question ', trim(string)
            write(*,*)
        end if

        read ( io_unit, '(a)', iostat=io ) string

    end do BEGIN

    !! can close list_conformation
    close ( io_unit )

    write(*,*) 'total conf', prot % p_flex % plist_flex % total_conf
    write(*,*) 'nb_max_atoms', nb_max_atoms

    prot % p_flex % plist_flex % max_size_movable_atom = nb_max_atoms
    prot % p_flex % p_current_node => prot % p_flex % plist_flex % p_first % p_next

    !! Extend array, all will have the same size
    if ( (bool_change_size).and.(bool_extend) ) then
        write (*,*)
        write (*,*) 'extend list_atom to write to size', nb_max_atoms
        do i = 1, prot % p_flex % plist_flex % total_conf

            !subroutine extend_array_double ( array_in, new_size )
            p_tmp => prot % p_flex % p_current_node % p_data % list_atom_surface

            !! could test if associated
            if ( size ( prot % p_flex % p_current_node % p_data % list_atom_surface )/3 < nb_max_atoms ) then
                call extend_array_double ( prot % p_flex % p_current_node % p_data % list_atom_surface, nb_max_atoms )
                !call extend_array_double ( p_tmp, nb_max_atoms )

                ! update size the variable
                prot % p_flex % p_current_node % p_data % nb_atom_surface = nb_max_atoms
            end if

            prot % p_flex % p_current_node => prot % p_flex % p_current_node % p_next
        end do
        !! come back to default
        prot % p_flex % p_current_node => prot % p_flex % plist_flex % p_first % p_next
    !else
    !    write(*,*) 'not extend, strange or only one'
    !    stop 1
    endif

end subroutine read_listconf_pdb

end module mod_sda_input

