!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> @file read_record.f90
!!
!! tools for reading / modifying records ( complexes file or trajectory )

!> \defgroup tools

!> Program read_record
!!
!! \ingroup tools
!! General program for reading \ref mod_record::record "record"\n
!! Multiple options :\n
!!   - convert between binary and ascii\n
!!   - extract energy terms : compute partial or total energy for sdamm\n
!!   - split conformation : split a complex file with a flexible solute into n files\n
!!     to use with -format_cluster, it makes the format compatible with the clustering program
program read_record
    USE mod_record
    IMPLICIT NONE

    interface
        subroutine concatenate_traj ( o_record, traj_concat_f )
            USE mod_record
            type ( record ), intent ( inout ) :: o_record
            character ( 256 ), dimension (:), intent ( in ) :: traj_concat_f
        end subroutine concatenate_traj
    end interface

    type ( record ) :: o_record

    ! needed for center_protein for sda_2proteins, sda_koff,..
    ! may be for sdamm, reading position and orientation from trajectory
    ! allocate 2 proteins by default for reading sda_2proteins, if sdamm, desallocate and reallocate a new one with correct size
    ! uses also to divide by steps, allow to compute total energy ??
    type ( array_protein_type ), target :: tab_prot

    integer :: nb_argument, count_arg, i, io_record
    integer :: nb_concat_traj
    integer :: extr_fr_beg, extr_fr_end
    integer :: param_int1
    integer :: nb_max_complexe, nb_line
    
    character (64) :: str_arg
    character (64) :: filename, filename_no_comp ! , filename_react_crit
    character ( 256 ), dimension (:), allocatable :: traj_concat_f
    
    ! options of the program
    logical :: header_only, extract, convert, split, format_cluster_bool, convertToMM, convertFromMM
    logical :: only_no, large_traj_nb
    logical :: concat, extract_frames_opt, extract_dist_opt !, split_success

    real ( kind= 8 ), dimension ( 2, 3) :: cm
    real( kind = 8 ) :: max_c2c_dist

    io_record = 155
    count_arg = 1
    nb_line = 0
    param_int1 = 0
    header_only = .false.
    extract = .false.
    convert = .false.
    split = .false.
    format_cluster_bool = .false.
    only_no = .false.
    concat = .false.
    large_traj_nb = .false.
    convertToMM = .false.
    convertFromMM = .false.
    cm = 0.d0
    
    extract_frames_opt = .false.
    extr_fr_beg = 0
    extr_fr_end = -1
    
!    split_success = .false.
!    filename_react_crit = ''
    
    extract_dist_opt = .false.
    max_c2c_dist = 0.0
    
    nb_concat_traj = 0
    filename_no_comp = ''
    filename = ''

    ! minimum for initialising record and tab_prot
    nullify ( o_record % pcomplexe, o_record % ptrajectory, o_record % ptab_prot )

!!!!!!!!!! Read arguments and options
    nb_argument = command_argument_count()

    !! no arguement print help
    if ( nb_argument == 0 ) then
        write (*,*) ' read_record   -  give information or transform complexes or trajectory files'
        write (*,*) ''
        write (*,*) ' USAGE:'
        write (*,*) './read_record filename [-header_only] [-extract] [-convert] [-convertToMM] [-split nb_conformation]'
        write (*,*) '                       [-large_traj_nb] [-only_no file_no] [-concat traj2 traj3 ...] [-extract_frames beg end]'
        write (*,*) '                       [-extract_max_dist c2c_dist] [-format_cluster]'
        write (*,*)
        write (*,*) ' filename   - the file with complexes of trajectory (binary or ascii); the output filenames are fixed!'
        write (*,*)
        write (*,*) ' !NB! use only one option at a time! '
        write (*,*)
        write (*,*) ' -header_only  : print only information about file header (type of file and summary of content)'
        write (*,*)
        write (*,*) ' -extract      : extract energy terms from a trajectory, and compute total energy;'
        write (*,*) '                 works only with trajectory ( sdamm or sda );'
        write (*,*) '                 only one choice of the energy term is possible, if several terms needed,'
        write (*,*) '                 run several times, each time changing the output extracted_energy;'
        write (*,*) '                 (see also ./energy_analyze tool)'
        write (*,*)
        write (*,*) ' -convert      : convert ascii to binary or binary to ascii'
        write (*,*)
        write (*,*) ' -convertToMM  : convert ascii trajectories to SDAMM format'
        write (*,*)
        !write (*,*) ' -convertFromMM: convert ascii trajectories from SDAMM format'
        !write (*,*) 
        write (*,*) ' -split  nb_conformation  : split a complexes file generated with flexible molecules,'
        write (*,*) '                into nb_conformation complexes file: complex1, 2, 3...'
        write (*,*)
        write (*,*) ' -format_cluster : transform a complexes file into a format needed by the cluster program'
        write (*,*) '             !NB! if molecules are flexible, apply -split BEFORE, otherwise the conformation number is lost'
        write (*,*)
        write (*,*) ' -large_traj_nb : apply with -format_cluster to extend the width of the StepNb column '
        write (*,*) '                 (needs to be used only if there are errors in the clustering program);'
        write (*,*) '                 if applied with -extract_frames or -extract_max_dist,'
        write (*,*) '                 this will deal better with very large files'
        write (*,*)
        write (*,*) ' -only_no file_no : (for flexible molecules) extract specific conformations from a complexes file;'
        write (*,*) '                 read the conformation from file_no;'
        write (*,*) '                 can be cluster_details_1 (just read the first column of this file)'
        write (*,*)
        write (*,*) ' -concat traj2 traj3 ... : concatenate output trajectories from restarts of sdamm (if necessary,'
        write (*,*) '                 it deletes the first frames and incomplete frames)'
        write (*,*)
        write (*,*) ' -extract_frames beg end :  extract only frames between "beg" and "end" (including those);'
        write (*,*) '                 output file is "extracted_frames"; '
        write (*,*) '                 "beg" = 0 -> extract from the first frame;'
        write (*,*) '                 "end" = -1 -> extract to the last frame;'
        write (*,*) '                 use negative numbers to get only last k frames, e.g.:'
        write (*,*) '                 "beg" = -30 and "end" = -1 -> extract last 30 frames'
        write (*,*) '                 (see also -large_traj_nb option)'
        write (*,*)
        ! working on it
!        write (*,*) ' -split_assoc_traj criteria : split the trajectories into successful and unsuccessful ones;'
!        write (*,*) '                         "criteria" is the file with reaction criteria to define successful trajectories'
!        write (*,*)
        write (*,*) ' -extract_max_dist c2c_dist : extract from the sda_2proteins trajectories only the frames where the'
        write (*,*) '                  two solutes were maximum c2c_dist apart (distance between the COMs)'
        write (*,*) '                  (see also -large_traj_nb option)'
        
        stop
    end if

    do while ( count_arg <= nb_argument )

      call getarg( count_arg, str_arg )
      write (*,*) 'argument ', count_arg, trim( str_arg )

      if ( count_arg == 1 ) then
           write ( *, * ) 'read file ', trim ( str_arg )
           filename = str_arg
      else if ( str_arg == '-header_only' ) then
             header_only = .true.
      else if ( str_arg == '-extract' ) then
            extract = .true.
      else if ( str_arg == '-convert' ) then
            convert = .true.
      else if ( str_arg == '-convertToMM' ) then
            convertToMM = .true.
      else if ( str_arg == '-convertFromMM' ) then
            convertFromMM = .true.
      else if ( str_arg == '-split' ) then
            split = .true.
      else if ( ( count_arg == 3 ) .and. split ) then
            !read ( param_int1, fmt='(a)' ) trim ( str_arg )
            !write ( param_int1, fmt='(i)' ) trim ( str_arg )
            read ( str_arg, fmt='(i3)' ) param_int1
            write(*,*) 'read param ', param_int1

      else if ( str_arg == '-format_cluster' ) then
            format_cluster_bool = .true.

      else if ( str_arg == '-large_traj_nb' ) then
            large_traj_nb = .true.

      else if ( str_arg == '-only_no' ) then
            only_no = .true.
            call getarg( count_arg +1, str_arg )
            filename_no_comp = trim ( str_arg )
            count_arg = count_arg + 1
      else if ( str_arg == '-concat' ) then
          concat = .true.
          nb_concat_traj = nb_argument - 2
          if ( nb_concat_traj > 0 ) then
              allocate ( traj_concat_f ( nb_concat_traj ) )
          end if
!      else if ( str_arg == '-split_assoc_traj' ) then
!          split_success = .true.
!          if ( nb_argument .lt. count_arg + 1 ) then
!              write(*,*) 'ERROR split_assoc_traj option chosen but no file with reaction criteria given - will quit!'
!              stop 1
!          end if
!          call getarg( count_arg + 1, str_arg )
!          filename_react_crit = trim ( str_arg )
!          write(*,*) 'SPLIT ASSOC TRAJ; react crit: ', trim(filename_react_crit)
!          count_arg = count_arg + 1
          
      else if ( str_arg == '-extract_max_dist' ) then
          extract_dist_opt = .true.
          if ( nb_argument .lt. count_arg + 1 ) then
              write(*,*) 'ERROR -extract_max_dist option chosen but no maximal distance given - will quit!'
              STOP 1
          end if
          call getarg( count_arg + 1, str_arg )
          read ( str_arg, fmt='(f5.2)' ) max_c2c_dist
          count_arg = count_arg + 1
          write(*,*) 'EXTRACT_MAX_DIST; max_c2c_dist: ', max_c2c_dist
          if ( max_c2c_dist .le. 0.0 ) then
              write(*,*) 'ERROR -extract_max_dist option chosen with maximal distance less than 0 - will quit!'
              STOP 1
          end if
          if ( max_c2c_dist .le. 5.0 ) then
              write(*,*) 'WARNING -extract_max_dist option chosen with maximal distance less than 5.0 - very strict limit!'
          end if


!          write (*,*) 'CONCAT: count_arg = ', count_arg, '; nb_concat_traj = ', nb_concat_traj, 'nb_argument = ', nb_argument
      else if ( concat ) then ! read each trajectory filename
          traj_concat_f ( count_arg - 2 ) = str_arg
          write (*,*) 'count_arg: ', count_arg, '; curr filename: ', traj_concat_f ( count_arg - 2 )
      else if ( str_arg == '-extract_frames' ) then
          extract_frames_opt = .true.
          if ( ( nb_argument - 2 ) .eq. 0 ) then
              write(*,*) ' ERROR extract_frame option chosen but no limits given - will quit'
              STOP 1
          end if
      else if ( extract_frames_opt ) then
          param_int1 = 2
          if ( large_traj_nb ) then
              param_int1 = 3
          end if
          if ( ( nb_argument - param_int1 ) .eq. 2 ) then
              read ( str_arg, fmt = '(i9)' ) extr_fr_beg
              count_arg = count_arg + 1
              call getarg( count_arg, str_arg )
              read ( str_arg, fmt = '(i9)' ) extr_fr_end
          else if ( ( nb_argument - param_int1 ) .eq. 1 ) then
              read ( str_arg, fmt = '(i9)' ) extr_fr_beg
              extr_fr_end = -1
              write(*,*) ' WARNING only one argument given, will extract from: ', extr_fr_beg, ' to the end'
          else
              write(*,*) ' ERROR extract_frame option chosen but wrong limits given - will quit'
              STOP 1
          end if
      else
            write ( *, * ) 'option ', trim (str_arg), ' not defined'
            stop 1
      end if
      count_arg = count_arg + 1
    end do
!!!!!!!!!!  Read header

    write(*,*) 'filename ', filename
    !write(*,*) 'only_no ', only_no

    !write(*,*) 'len(filename_no_comp) ', len(filename_no_comp)
    if ( filename_no_comp /= "" ) then
        write(*,*) 'filename_no_comp ', filename_no_comp
    end if

    ! read number of all infos from header and nb_protein
    ! test if file is binary or ascii, setup format_ascii necessary, and read the header
    ! tab_prot has not been created and associated, so cannot read center of mass

    ! simple initialization, mainly nullify pointer, job easier for a general initialize_record
    call init_record ( o_record, opt_iocmx = io_record )
    call read_header_record ( o_record, filename, cm, nb_line )
    ! could be done in read_header_record
    !nb_line = o_record % nb_max_complexe

    write(*,*) 'cm 1', cm ( 1, : )
    write(*,*) 'cm 2', cm ( 2, : )
    write (*,*) 'Number of records ', nb_line

    !! if header_only exit now, nothing has been allocated, could use a return value in read_header_record ( nb_protein )
    if ( header_only ) stop

    ! create array protein, before finalizing the record
    !call new_array_protein ( tab_prot, o_record % nb_protein, 1 )
    call new_array_protein ( tab_prot, o_record % nb_protein, 2 )

    do i = 1, o_record % nb_protein
        !call new_protein ( tab_prot % array ( i ), .true., .false., .false., i )
        call new_protein ( tab_prot % array ( i ), .true., .false., .false., .false., i )
        tab_prot % array ( i ) % center_protein = cm ( i, : )
    end do

    ! need correct setup of set_of_protein for deleting, make a stupid one
    ! make most simple one, now important for correct cm
    !write(*,*) 'o_record % nb_protein ', o_record % nb_protein
    tab_prot % set_of_protein ( 1, 1 ) = 1
    tab_prot % set_of_protein ( 1, 2 ) = 1
    tab_prot % set_of_protein ( 2, 1 ) = o_record % nb_protein
    tab_prot % set_of_protein ( 2, 2 ) = o_record % nb_protein

!    if ( ( extract_frames_opt .or. split_success .or. extract_dist_opt ) .and. large_traj_nb ) then
    if ( ( extract_frames_opt .or. extract_dist_opt ) .and. large_traj_nb ) then
        call initialize_record ( o_record, tab_prot, opt_read = .true., opt_large_traj = .true. )
    else
    ! new version, otpional .true. do not write the header
    !write(*,*) 'new call intialize record '
        call initialize_record ( o_record, tab_prot, opt_read = .true. )
!!!!!!!!!!!! Read data
        ! for traj can proceed by steps or multiple of steps
        call read_file_record ( o_record )
        ! call print_record ( o_record, .false. )
    end if

!!!!! If Converts function
!!!!! File is loaded in memory, all traj or complexes
!!!!! Now print output
    if ( convert ) then
       call convert_format ( filename, o_record )
    else if ( convertToMM ) then
       call convert_tomm (filename, o_record )
    else if ( extract ) then
        call extract_terms ( o_record, nb_max_complexe )

    else if ( split ) then
        call split_conformation ( o_record, param_int1 )

    else if ( format_cluster_bool ) then
        call format_cluster ( o_record, filename, large_traj_nb )

    else if ( only_no ) then
        !write(*,*) 'got it'
        call extract_no_from_file ( o_record, filename_no_comp, filename )
        
    else if ( concat ) then
        if ( o_record % type_record /= trajectory_enum ) then
            write (*,*) 'ERROR this is not a trajectory! "concat" option concatenates only trajectories!'
            stop 1
        end if
        
        write (*,*) 'will concatenate trajectories: ', traj_concat_f(:), ' (size: ', size(traj_concat_f), ')'
        call concatenate_traj ( o_record, traj_concat_f )

    else if ( extract_frames_opt ) then
        if ( o_record % type_record /= trajectory_enum ) then
            write (*,*) 'ERROR this is not a trajectory! "extract_frames" option works only with trajectories!'
            stop 1
        end if
        
        write (*,*) ' will extract frames from: ', extr_fr_beg, ' to: ', extr_fr_end
        call extract_frames ( o_record, extr_fr_beg, extr_fr_end, large_traj_nb )
        
!    else if ( split_success ) then
!        write(*,*) 'will split the association trajectories into successful and unsuccessful'
!        call split_success_traj( o_record, filename_react_crit, large_traj_nb )
!        
    else if ( extract_dist_opt ) then
        write(*,*) 'will extract the frames where the two solutes are max. ', max_c2c_dist, ' apart'
        call extract_dist ( o_record, max_c2c_dist, large_traj_nb )
    end if

    write(*,*) 'delete array_protein '
    !! if something allocated, need to deallocate
    call delete_array_protein ( tab_prot )
    !write(*,*) 'array_prot deleted'
    !! set true for donot_check_array, only for tools, default false for sda program
    write(*,*) 'delete record'
    if ( large_traj_nb ) then
        o_record % nb_max_complexe = 1
        o_record % ptrajectory % size_array = 1
    end if
    call delete_record ( o_record, .true. )
    write(*,*) 'END'

end program read_record


!>
!!
!!
subroutine convert_tomm ( filename, o_record )
    USE mod_record
    IMPLICIT NONE
    character *64 filename
    type ( record ) :: o_record
    character*64 :: filename_out
    integer :: i,j,io_sdamm,io_err

    io_sdamm = 55
    

    filename_out = trim( filename ) // '.fort55'
    write(*,*) 'opening file for old trajectory format ', filename_out
    open ( io_sdamm, file=filename_out, form='formatted', action='write', iostat = io_err )
    write(io_sdamm,fmt='("#",i5)') o_record % nb_protein
    write (*,*) 'number of proteins',o_record % nb_protein
    !write (io_sdamm, fmt='("#",3f9.3,i5)' ) o_record % ptab_prot % array ( 1 ) % center_protein,1
    do i=1,  o_record % nb_protein
        write (*,*) 'working on ',i
        write ( io_sdamm, fmt='("#",3f9.3,i5)' ) &
                o_record % ptab_prot % array ( o_record % ptab_prot % set_of_protein ( 1 ,1 ) ) % center_protein,&
                o_record % ptab_prot % set_of_protein ( 1 ,1 ) 
      ! else only one type, use same cm
    end do
    write (*,*), 'size of trajectory', size (o_record % ptrajectory % array_trajectory )
    write (*,*), 'size of array', o_record % ptrajectory % size_array
    do j=1, o_record % nb_max_complexe
       write (io_sdamm,fmt='(2i8,9f9.3,4e12.4,3i3,5e12.4)') &
          o_record % ptrajectory % array_trajectory (j) % int_ocompl(1), &
          o_record % ptrajectory % array_trajectory (j) % int_ocompl(2), &
          o_record % ptrajectory % array_trajectory (j) % position, &
          o_record % ptrajectory % array_trajectory (j) % orient_x, &
          o_record % ptrajectory % array_trajectory (j) % orient_y, &
          o_record % ptrajectory % array_trajectory (j) % real_ocompl(1), &
          o_record % ptrajectory % array_trajectory (j) % real_ocompl(2), &
          o_record % ptrajectory % array_trajectory (j) % real_ocompl(3), &
          o_record % ptrajectory % array_trajectory (j) % real_ocompl(4), &
          o_record % ptrajectory % array_trajectory (j) % box, &  
          0.0,0.0,0.0,0.0,0.0       
    end do
    
    close ( io_sdamm )

end subroutine convert_tomm

!> Convert from ascii to binary or inverse
!!
!! The output filename is automaticaly created ( append .bin or .ascii )
!! @param filename : name of the input file
!! @param o_record : instance of \ref mod_record::record "record"
subroutine convert_format ( filename, o_record )
    USE mod_record
    IMPLICIT NONE
    character *64 filename
    type ( record ) :: o_record

    character*64 :: filename_out
    integer :: io_output, io_record, io_err

    io_output = 156
    io_record = o_record % io_compl

    write(*,*)
    write(*,*) 'Entry convert_format'

    !! invert the format
    o_record % format_ascii =  .not.( o_record % format_ascii )
    o_record % io_compl = io_output

    ! get box info
    write(*,*) 'o_record % box_info ', o_record % box_info

    !write(*,*) 'current_pos for traj ', o_record % ptrajectory % current_pos

    !! should delete suffix first, if present
    if ( o_record % format_ascii .eqv. .true. ) then
        write (*,*) 'will write ascii format '
        filename_out = trim( filename ) // '.ascii'
        open ( o_record % io_compl, file=filename_out, form='formatted', action='write', iostat = io_err )
        if ( io_err /= 0 ) then
             write(*,*) 'problem in opening file ', filename
        end if
    !! write binary format
    else
        filename_out = trim( filename ) // '.bin'
        write (*,*) 'will write binary format '
        open ( o_record % io_compl, file=filename_out, form='unformatted', action='write', iostat = io_err )
        if ( io_err /= 0 ) then
             write(*,*) 'problem in opening file ', filename
        end if
    end if

    write (*,*) 'filename_out ', trim(filename_out)
    !! write in output file
    !write(*,*) 'format ascii ', o_record % format_ascii

    call write_header_record ( o_record )
    call write_record ( o_record, .false. )
    close ( io_output )

    !! before exit and deletion comes back to initial data
    !! could be a problem when deleting and closing file
    o_record % format_ascii = .not.( o_record % format_ascii )
    o_record % io_compl = io_record

end subroutine convert_format

!> Extract and sum up energies
!!
!! Used for sdamm, can extract specific terms or the total energy\n
!! Note the electrostatic energy is correct\n
!! @param o_record : instance of \ref mod_record::record "record"
!! @param nb_record : total number of lines ( the number of steps will be computed inside the function
subroutine extract_terms ( o_record, nb_record )
    USE mod_record
    IMPLICIT NONE
    type ( record ) :: o_record
    integer :: nb_record

    integer :: number, i, j, io_unit, ierr
    integer :: nb_prot, nb_step
    character*16 :: filename
    real ( kind=8 ) :: sum_energy, time

    filename= 'extracted_energy'
    io_unit = 1
    number = -1
    ierr = 1

    write ( *, fmt = '(''term :            0  total energy'')' )
    !! optional true, output for menu choice
    call print_bits ( o_record, .true. )
    write(*,*)

    write ( *, fmt = '(''Extract energy terms :'')', advance = "no" )
    read ( *, * ) number

    write(*,*) 'your entry:', number
    open (io_unit, file=filename , form= 'formatted', status='new', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open file, one already existing ?? '
    end if

    !! compute sum of total energy
    if ( o_record % record_type_calc % enum_type_calc == sdamm ) then
        nb_prot = size ( o_record % ptab_prot % array )
    !! sda_2proteins
    else
        nb_prot = 1
    end if
    !write ( *, * ) 'size array protein ', size ( o_record % ptab_prot % array )
    nb_record = o_record % nb_max_complexe
    !write ( *, * ) 'nb_record ', nb_record
    !write ( *, * ) 'nb_prot ', nb_prot
    !write ( *, * ) 'o_record % nb_protein ', o_record % nb_protein
    !write ( *, * ) ' nb_step = nb_record / nb_prot ', nb_record / nb_prot
    !! nb_prot incorrect, need to change order
    nb_step = nb_record / nb_prot

    !write(*,*) 'Entry loop'
    !write(*,*) 'nb_prot ', nb_prot
    !write ( *, * ) ' nb_step ', nb_step
    !write(*,*) 'size traj_array', size ( o_record % ptrajectory % array_trajectory )

    !! should check ?? division is ok
    if ( o_record % type_record == trajectory_enum ) then
        do i = 1, nb_step
            sum_energy = 0.
            time = o_record % ptrajectory % array_trajectory ( (i-1) * nb_prot + 1 ) % other_real_ocompl ( 1 )
            !write (*,*) 'time'
            !write ( *, * ) 'write nb_step ', nb_step
            !write ( *, * ) 'array indic time ', i * nb_prot + 1, time
            do j = 1, nb_prot
                !write(*,*) 'energy ', o_record % ptrajectory % &
                !            array_trajectory ( (i-1) * nb_prot + j ) % real_ocompl ( 1 )

                sum_energy = sum_energy + o_record % ptrajectory % &
                             array_trajectory ( (i-1) * nb_prot + j ) % real_ocompl ( number + 1 )
            end do
            !write ( *, * ) 'sum_energy ', time, sum_energy
            write ( io_unit, fmt= '( f10.3, f10.3 )' ) time / 1000., sum_energy
        end do
    else
        write(*,*)
        write(*,*) 'NOT YET implemented for complexe'
    end if

    close ( io_unit )

end subroutine extract_terms

!> Split multiple conformation complexes
!!
!! Always complex by default, should not be too much work for traj, but why to do it ??\n
!! Will create nb_conf output files\n
!! @param o_record : instance of \ref mod_record::record "record"
!! @param nb_conf : the total number of conformations
subroutine split_conformation ( o_record, nb_conf )
    USE mod_record
    USE mod_array_protein
    USE com_sda
    IMPLICIT NONE
    type ( record ) :: o_record
    integer :: nb_conf

    character*64 fnameout_tmp
    character*5 str_numb_tmp
    type ( record ), dimension( nb_conf ), target :: array_record
    !type ( record ), dimension (:), allocatable, target :: array_record
    !type ( record ), dimension( nb_conf ), pointer :: array_record
    type ( NodeComplexe ), pointer :: tmp_node_complexe
    type ( one_complexe ), pointer :: p_new_onecomplexe, p_old_onecomplexe

    !! max size come from com_sda, global parameter, how to get this value ??
    !! function of one_complexe would be the best
    integer, dimension ( max_nb_integer_bit ) :: array_int
    !real ( kind=8), dimension ( max_nb_real_bit ) :: array_real
    real ( kind=8 ), dimension(:), allocatable :: array_real
    integer :: size_real

    !! add for new copy_record, expects a pointer
    type ( record ), pointer :: p_record

    integer :: i, tmp_conf, ierr
    integer :: indice_nconf

    nullify ( tmp_node_complexe, p_new_onecomplexe, p_old_onecomplexe )
    nullify ( p_record )
    array_int = 0

    write(*,*) 'Entry split_conformation param ', nb_conf
    !write(*,*) 'size of array_record ', size ( array_record )

    ! called 4 times !! need intermediate value
    size_real = get_nbterms_energy_compatible ( o_record )
    !allocate ( array_real ( get_nbterms_energy_compatible ( o_record ) ), stat = ierr )
    allocate ( array_real ( size_real ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot allocate array_real with size: ', &
            get_nbterms_energy_compatible ( o_record )
    end if
    array_real = 0.d0
    !write(*,*) 'After allocation array_real '

!!!!! initialize array_record, always complexe, size will be extended on the fly, just check max and rmsd_min
    do i = 1, nb_conf
        ! nullify pointer, safer
        nullify ( array_record ( i ) % pcomplexe, array_record ( i ) % ptrajectory )
        p_record => array_record ( i )

        ! copy record, recently make full copy with an emplty default list_complexe
        ! ( for complexe not yet for trajectory )
        ! default header, string format not copied. Better if we change the format
        call copy_record ( o_record, p_record, .true. )
        array_record ( i ) % io_compl = o_record % io_compl + i

        !!setup output file, after init of a new io_compl !
        write ( str_numb_tmp, fmt='(i4)' ) i
        str_numb_tmp = adjustl(str_numb_tmp)
        fnameout_tmp = 'complex'//trim(str_numb_tmp)
        open ( array_record ( i ) % io_compl, file= fnameout_tmp, form= 'formatted', status='new', iostat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'WARNING Cannot open file, already existing ?? '
            write(*,*) 'WARNING Will produce default output fort.*'
            write(*,*)
        end if

        ! set bit for conformation to zero, data output will be lost
        ! all the other stay by default like in o_record
        array_record ( i ) % bit_integer = ibclr ( o_record % bit_integer, 3 )

        ! set the rest of important initialization for bits
        ! should be identical for energy, just modified for integer
        call set_variable_field ( array_record ( i ) )
    end do

    nullify ( p_record )

    !! to do improve get_indice_nconf/ntraj
    !indice_nconf = get_indice_nconf ( o_record )
    indice_nconf = o_record % indice_nconf
    write(*,*) 'indice_nconf ', indice_nconf
    if ( indice_nconf == 0 ) then
        write(*,*) "indice nconf is zero, are u sure it's present ??"
        stop 1
    end if

!!!!!!!!!!!! Make loop for filling splitted complexes
    !! init pointer to the first in loop
    tmp_node_complexe => o_record % pcomplexe % p_first
    !write(*,*) ' o_record nb complexe ', o_record % pcomplexe % nb_complexe

    do i = 1, o_record % pcomplexe % nb_complexe

        tmp_node_complexe => tmp_node_complexe % p_next
        p_old_onecomplexe => tmp_node_complexe % p_data

        ! extract current conf
        tmp_conf = p_old_onecomplexe % int_ocompl ( indice_nconf )
        !write(*,*) 'tmp_conf', tmp_conf

        !! initialize onecomplexe, keep same format after, just need to update
        allocate ( p_new_onecomplexe )
        call new_onecomplexe ( p_new_onecomplexe, array_record ( tmp_conf ) % nb_int, &
                               array_record ( tmp_conf ) % nb_real, array_record ( tmp_conf ) % nb_other_real )

        call return_array_int ( p_old_onecomplexe, array_int, o_record % bit_integer )
        !write(*,*) 'returned array_int:', array_int
        !write(*,*) 'size array_int ', size(array_int)
        !write(*,*) 'size array real old_onecomplexe ', size( p_old_onecomplexe % real_ocompl )
        !write(*,*) 'array real of old_onecomplexe', p_old_onecomplexe % real_ocompl
        !write(*,*) 'test nb_types ', array_record ( tmp_conf ) % nb_types

        ! error, because input array real is not compatible, works only if size is correct
        ! imprtant to allocate with good size, see above. Difficult to set a maximum, 20 in sda
        call return_array_real ( p_old_onecomplexe, array_real, o_record % bit_energy, &
                                o_record % opt_sum, o_record % nb_types )
        ! will return a size of ?
        !write(*,*) 'retrurned array_real ', array_real

        !call set_onecomplexe ( p_new_onecomplexe, &
        !       p_old_onecomplexe % position, p_old_onecomplexe % box, p_old_onecomplexe % orient_x, p_old_onecomplexe % orient_y, &
        !       array_int, array_real, p_old_onecomplexe % other_real_ocompl, array_record ( tmp_conf ) % nb_other_real, &
        !       array_record ( tmp_conf ) % bit_integer, array_record ( tmp_conf ) % bit_energy, &
        !       array_record ( tmp_conf ) % opt_sum ,  array_record ( tmp_conf ) % nb_types )

        call set_onecomplexe_wrap ( array_record ( tmp_conf ), p_old_onecomplexe % position, p_old_onecomplexe % box, &
                                    p_old_onecomplexe % orient_x, p_old_onecomplexe % orient_y, &
                                    array_int, array_real, p_old_onecomplexe % other_real_ocompl, &
                                    opt_p_onecomplexe = p_new_onecomplexe )

        ! only for test, print to the console
        !call print_onecomplexe ( p_new_onecomplexe )


       ! always insert onecomplexe at the end of the list
       call insert_new_node ( array_record ( tmp_conf ) % pcomplexe % p_last, p_new_onecomplexe )
       ! should not be necessary in fact, but not implemented in insert_new_node
       array_record ( tmp_conf ) % pcomplexe % nb_complexe = array_record ( tmp_conf ) % pcomplexe % nb_complexe + 1

    end do ! loop nb_complexes

    ! not necessary to do, deleted at function exit
    deallocate ( array_real )

!!!!!!! print and delete all record for each conformation, strange still complain form valgrind
    do i = 1, nb_conf

        !!opt false do not recompute dispersion
        ! write(*,*) 'write_record i', i

        ! due to recent change need to force each record to setup the correct format before writing
        call make_format_string_ascii( array_record ( i ) )

        call write_record ( array_record ( i ), .false. )
        call delete_record ( array_record ( i ), .true. )

    end do
    !write(*,*) 'end split conformation'

end subroutine split_conformation

!> Make the format compatible to the cluster program
!!
!! Simply delete additional data ( nprotein, nconf ) data will be lost
!! and 0 to the missing data
!! cannot fully follow the orignal version, but can limit changes
!! 2 integer : ntraj, nstep, 9 coordinates, fixe the rest at : total_energy, ep, ed, hd, lj, occu, average, std_dev
!! general function, copy o_record parameters, modify accordingly to the out
!! @param o_record : instance of \ref mod_record::record "record"
!! @param filename : input file name
!! @param large_traj_nb : bool, indicates if the integer field should be printed larger (in case of large step number)
subroutine format_cluster ( o_record, filename, large_traj_nb )
    USE mod_record
    USE com_sda
    IMPLICIT NONE

    interface
        subroutine initialize_record_cluster_format ( record_comp, o_record, large_traj_nb ) !, need_sum_terms )
        USE mod_record
        type ( record ), target :: record_comp
        type ( record ) :: o_record
        logical, intent (in) :: large_traj_nb
        !logical, intent (out) :: need_sum_terms
        end subroutine initialize_record_cluster_format

        subroutine sum_energy_terms( o_record, array_real )
            USE mod_record
            IMPLICIT NONE
            type ( record ), target :: o_record
            real ( kind=8 ), dimension (:) :: array_real
        end subroutine sum_energy_terms
    end interface

    character *64 filename
    type ( record ) :: o_record
    logical :: large_traj_nb

    type ( record ), target :: record_comp
    type ( NodeComplexe ), pointer :: tmp_node_complexe
    type ( one_complexe ), pointer :: p_old_onecomplexe, p_new_onecomplexe
    real ( kind=8 ), dimension (:), allocatable :: array_real

    real ( kind=8 ), dimension( 3 ) :: other_real
    integer, dimension (:), allocatable :: array_int
    integer :: size_energy_comp, size_int_comp
    integer :: ierr
    !logical :: need_sum_terms
    ! test if error valgrind due to box
    integer, dimension(3) :: empty_box
    logical :: need_sum_energy_terms

    write(*,*)
    write(*,*) 'Entry format_cluster '
    write(*,*)

    ! nullfy to be sure
    nullify ( record_comp % pcomplexe, record_comp % ptrajectory )
    nullify ( p_old_onecomplexe, p_new_onecomplexe )
    nullify ( tmp_node_complexe )

    other_real = 0.d0
    empty_box = 0
    need_sum_energy_terms = .false.

    ! create new record comp_record from o_record with parameter for comp_cluster
    call initialize_record_cluster_format ( record_comp, o_record, large_traj_nb ) !, need_sum_terms )

    ! allocate correctly array_real, use o_record ?? for reading, but maybe more stuffs to ouput !
    !write(*,*) 'MAX size energy ', max( get_nbterms_energy_compatible ( record_comp, .true. ), &
    !                                    get_nbterms_energy_compatible ( o_record, .true. ) )
    size_energy_comp = max( get_nbterms_energy_compatible ( record_comp, .true. ), &
                            get_nbterms_energy_compatible ( o_record, .true. ) )
    ! allocate ( array_real ( get_nbterms_energy_compatible ( o_record ) ), stat = ierr )
    allocate ( array_real ( size_energy_comp ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot allocate array_real with size: ', &
            get_nbterms_energy_compatible ( record_comp )
    end if
    array_real = 0.d0

    ! here problem input maybe larger ( conf number for instance ), need a max ??
    !write(*,*) 'MAX size_int ', max( get_nbterms_integer_compatible ( record_comp ), get_nbterms_integer_compatible ( o_record ) )
    !write(*,*) 'allocata array_int with size ',  get_nbterms_integer_compatible ( record_comp )

    size_int_comp = max ( get_nbterms_integer_compatible ( record_comp ), &
                          get_nbterms_integer_compatible ( o_record ) )
    allocate ( array_int ( size_int_comp ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot allocate array_real with size: ', size_int_comp
    end if
    array_int = 0

    !!setup outpout file
    filename = trim(filename)//'.comp_cluster'
    write(*,*) 'filename: ', trim(filename)
    !!need to be done manually, or add  an option .. before opening
    record_comp % io_compl = o_record % io_compl + 1
    open ( record_comp % io_compl, file= filename, form= 'formatted', status='new', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open file, one already existing ?? '
        stop 1
    end if

    ! set size, identical
    !record_comp % nb_max_complexe = o_record % pcomplexe % nb_complexe
    !write(*,*) 'max size', record_comp % nb_max_complexe
    !write(*,*) 'o_record nb_complexe', o_record % pcomplexe % nb_complexe
    !write(*,*) 'tests bits get_max_bits ', get_max_bit_energy (o_record)

    ! test if more than 4 energy_terms
    if ( get_max_bit_energy( o_record) > 4 ) then
        need_sum_energy_terms = .true.
        write(*,*)
        write(*,*) 'Info: There are some additional energy terms in the input,'
        write(*,*) 'Info: The clust format supports only 4 fields for the energy terms. The following rules will apply:'
        write(*,*) 'Info: 1. Lennard-Jones(PrometCS) will be added to Soft-core LJ'
        write(*,*) 'Info: 2. Image-charge will be added to Electrostatic energy'
        write(*,*) 'Info: 3. Metal Desolvation will be added to Soft-core LJ'
        write(*,*)
    end if

    !tmp_node_complexe => o_record % pcomplexe % p_first
    ierr = 0
    call reset_position ( o_record, ierr )
    if ( ierr /=0 ) then
        write(*,*) 'Record contains no data'
        stop 1
    end if
    ! may need an argument in reset_position, assign old into the first argument, then updated by getNext
    ! bug, was missing, p_old_onecomplexe is a convenient shortcut, will be updated by getNext option
    p_old_onecomplexe => o_record % p_onecompl
!!!!!!! loop over complexes
    !do i = 1, o_record % pcomplexe % nb_complexe
    do while ( ierr == 0 )

        ! initialize onecomplexe, keep same format after, just need to update
        allocate ( p_new_onecomplexe ) ! stat = ierr )
        call new_onecomplexe ( p_new_onecomplexe, record_comp % nb_int, record_comp % nb_real, record_comp % nb_other_real )

        call return_compatible_array_wrap ( o_record, array_int, array_real, other_real )

        ! here need to modify in case of image-charge,  desolvation image-charge,, lennard-jones/rep lennard-jones
        ! only if one these terms are present
        if ( need_sum_energy_terms ) then
            ! intern function, comp_cluster has a maximum of 4 fields for energy
            call sum_energy_terms( o_record, array_real )
        end if

        call set_onecomplexe_wrap ( record_comp, p_old_onecomplexe % position, p_old_onecomplexe % box, &
                                    p_old_onecomplexe % orient_x, p_old_onecomplexe % orient_y, &
                                    array_int, array_real, other_real, opt_p_onecomplexe = p_new_onecomplexe )
!        call print_onecomplexe ( p_new_onecomplexe )

        ! always insert onecomplexe at the end
        call insert_new_node ( record_comp % pcomplexe % p_last, p_new_onecomplexe )
        ! not necessary in fact
        record_comp % pcomplexe % nb_complexe = record_comp % pcomplexe % nb_complexe + 1

        call getNext( o_record, ierr, opt_p_onecomplexe = p_old_onecomplexe )
    end do

    ! write output and delete
    ! error format, box bad initialization
    call write_record ( record_comp, .false. )
    call delete_record ( record_comp, .true. )

    deallocate ( array_real, array_int )

end subroutine format_cluster

!> Read a list of complexes belonging to the smae cluster
!!
!! Used for recursive clustering. The format is very simple : one integer by line ( compatible with cluster_details )
!! @param o_record : instance of  \ref mod_record::record "record"
!! @param filename_no_comp : file name from where to extract the list of number
!! @param filename : name of the initial complexe file
subroutine extract_no_from_file ( o_record, filename_no_comp, filename )
    USE mod_record
    USE com_sda
    IMPLICIT NONE

    interface
        subroutine read_array_no ( filename, array_no )
        character *64 filename
        integer, dimension(:), allocatable :: array_no
        end subroutine read_array_no

        subroutine initialize_record_cluster_format ( record_comp, o_record, large_traj_nb ) !, need_sum_terms )
        USE mod_record
        type ( record ), target :: record_comp
        type ( record ) :: o_record
        logical, intent (in) :: large_traj_nb
        !logical, intent (out) :: need_sum_terms
        end subroutine initialize_record_cluster_format
    end interface

    ! need to declare the function ?
    logical :: is_in_array

    character *64 filename, filename_no_comp
    type ( record ) :: o_record
    ! local defined for output
    !type ( record ), target :: p_record
    type ( record ) :: record_comp
    type ( NodeComplexe ), pointer :: tmp_node_complexe
    type ( one_complexe ), pointer :: p_old_onecomplexe, p_new_onecomplexe

    real ( kind=8 ), dimension (  max_nb_real_bit ) :: array_real
    real ( kind=8 ), dimension( 3 ) :: other_real
    integer, dimension ( max_nb_integer_bit ) :: array_int

    ! read filename_no_comp, fill array_no
    integer, dimension(:), allocatable :: array_no
    integer :: ierr, size_array, count_in, count_out
    logical :: b_inarray

    nullify ( p_old_onecomplexe, p_new_onecomplexe )
    nullify ( tmp_node_complexe ) !, p_record )

    write(*,*) 'Entry extract_no_from_file'

    array_int = 0
    array_real= 0.
    other_real = 0.
    b_inarray = .false.

    ! allocate and fill array_no, no order in clustering file
    call read_array_no ( filename_no_comp, array_no )
    size_array = size ( array_no )
    write(*,*) 'size_array ', size_array

    write(*,*) 'initialze record'
    ! create new record comp_record from o_record with parameter for comp_cluster
    ! not remembering why need_sum_terms
    call initialize_record_cluster_format ( record_comp, o_record, .false. )
    !!setup output file, can be done by calling function
    filename = trim(filename)//'.only_no.comp_cluster'
    write(*,*) 'filename', filename

    !!need to be done manually, or add  an option .. before opening
    record_comp % io_compl = o_record % io_compl + 1
    open ( record_comp % io_compl, file= filename, form= 'formatted', status='new', iostat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'Cannot open file, one already existing ?? '
        stop 1
    end if

    count_in = 1
    count_out = 1
    !tmp_node_complexe => o_record % pcomplexe % p_first
    call reset_position( o_record, ierr )
    if ( ierr /=0 ) then
        write(*,*) 'Record contains no data'
        stop 1
    end if
    ! assign shortcut first time, later done with getNext
    p_old_onecomplexe => o_record % p_onecompl

!!!!!!! loop over complexes
    !do i = 1, o_record % pcomplexe % nb_complexe
    do while ( ierr == 0 )
        !write(*,*) 'i ',i
        !! associate next node and next onecomplexe
        !tmp_node_complexe => tmp_node_complexe % p_next
        !p_old_onecomplexe => tmp_node_complexe % p_data

        b_inarray = is_in_array ( array_no, size_array, count_in )
        write(*,*) 'count_in ', count_in, b_inarray


        ! match, make a copy in record_comp
        if ( b_inarray ) then

            !! initialize onecomplexe, keep same format after, just need to update
            allocate ( p_new_onecomplexe ) ! stat = ierr )

            call new_onecomplexe ( p_new_onecomplexe, record_comp % nb_int, record_comp % nb_real, record_comp % nb_other_real )
            ! get data from the original
            call return_compatible_array_wrap ( o_record, array_int, array_real, other_real )
            ! set for the new record
            !call set_onecomplexe ( p_new_onecomplexe, &
            !   p_old_onecomplexe % position, p_old_onecomplexe % box, p_old_onecomplexe % orient_x, p_old_onecomplexe % orient_y, &
            !   array_int, array_real, other_real, record_comp % nb_other_real, &
            !   record_comp % bit_integer, record_comp % bit_energy, record_comp % opt_sum , record_comp % nb_types )
               !record_comp % record_type_calc )

            call set_onecomplexe_wrap ( record_comp, p_old_onecomplexe % position, p_old_onecomplexe % box, &
                                        p_old_onecomplexe % orient_x, p_old_onecomplexe % orient_y, &
                                        array_int, array_real, other_real, opt_p_onecomplexe =  p_new_onecomplexe )
            !call print_onecomplexe ( p_new_onecomplexe )

            !! always insert onecomplexe at the end
            call insert_new_node ( record_comp % pcomplexe % p_last, p_new_onecomplexe )
            ! not necessary in fact
            record_comp % pcomplexe % nb_complexe = record_comp % pcomplexe % nb_complexe + 1


            count_out = count_out + 1
        end if

        count_in = count_in + 1
        call getNext( o_record, ierr, opt_p_onecomplexe = p_old_onecomplexe )

    end do

    write(*,*) 'count_out ', count_out


    deallocate ( array_no )
    !!write output and delete
    call write_record ( record_comp, .false. )
    call delete_record ( record_comp, .true. )

end subroutine extract_no_from_file

!> Utility function
!! @param array_no : input array
!! @param size_array : size of input array
!! @param count_in : value to check if present in array
!! @result b_in : boolean true or false
function is_in_array ( array_no, size_array, count_in ) result ( b_in )
    IMPLICIT NONE
    integer, intent ( in ) :: size_array, count_in
    integer, dimension( size_array ), intent ( in ) :: array_no

    integer :: i
    logical :: b_in

    b_in = .false.
    do i =1, size_array
        if ( array_no( i ) == count_in ) then
            b_in = .true.
            !return b_in
            exit
        end if
    end do

end function is_in_array

!> Utility function
subroutine read_array_no ( filename, array_no )
    IMPLICIT NONE
    character *64 filename
    integer, dimension(:), allocatable :: array_no

    logical :: ex
    integer :: io, number_line
    integer :: iusas = 23
    character *128 string

    inquire ( file=filename, EXIST=ex )
    if ( .NOT.ex ) then
        write (*,*) "file does not exist or file name is missing ", trim ( filename )
        stop 1
    end if

    open (iusas,file=filename,form='formatted',status='old')
    !nat = 0
    number_line = 0
    !count number of line
    read ( iusas, '(a)', iostat=io ) string
    if ( io < 0 ) then
      write( *,* ) 'file is empty', trim ( filename )
      return
    endif

    do while ( string(1:1).NE.'' )

      if ( string(1:1).EQ.'#' ) then
          !number_line = number_line
          write(*,*) 'read #'
      !endif
      else
          number_line = number_line + 1
      endif
      read ( iusas, '(a)', iostat=io ) string
      !write(*,*) 'string:',string,'!!'
    enddo

    write(*,*) 'number line ', number_line
    allocate ( array_no ( number_line ) )
    array_no = 0

    rewind( iusas )

    ! read again and record data
    number_line = 0
    !count number of line
    read ( iusas, '(a)', iostat=io ) string
    do while ( string(1:1).NE.'' )

      if ( string(1:1).EQ.'#' ) then
          !number_line = number_line
          write(*,*) 'read #'
      !endif
      else
          number_line = number_line + 1
          !write(*,*) 'string:',trim(string),'!!'
          read ( string, fmt='(I4)' ) array_no( number_line )
      end if
      read ( iusas, '(a)', iostat=io ) string
      !write(*,*) 'string:',string,'!!'
    enddo

    write(*,*) 'number line2 ', number_line

end subroutine read_array_no

!>Initialize destination array for cluster_format
!!could be used in more than 1 function
!! general subroutine, initialize a empty record with default values for cluster format
!! Used internally to read_record
subroutine initialize_record_cluster_format ( record_comp, o_record, large_traj_nb ) !, need_sum_terms )

    USE mod_record
    IMPLICIT NONE
    type ( record ), target :: record_comp
    type ( record ) :: o_record
    logical, intent (in) :: large_traj_nb
    !logical, intent (out) :: need_sum_terms


    type ( record ), pointer :: p_record
    ! nullfy to be sure
    nullify ( record_comp % pcomplexe, record_comp % ptrajectory )

    ! other_real = 0.
    ! not really copy, but copy main attribute, added the listcomplexe
    p_record => record_comp
    call copy_record ( o_record, p_record )

    !need_sum_terms = .false.

    ! only ntraj and nstep
    record_comp % bit_integer = 3
    ! all 4 implemented energy
    record_comp % bit_energy = 15
    ! should already be fine, if complexe in output
    record_comp % nb_other_real = 3
    ! modify attributes accordingly to new format
    ! always opt_sum true, always ascii format
    record_comp % opt_sum = .true.

    record_comp % format_ascii = .true.
    ! add because of error, badly initialised ( problem in copy )
    record_comp % box_info = .false.

    ! set up all other related data to variable field
    call set_variable_field ( record_comp )
    !write(*,*) 'force the format creation '
    call make_format_string_ascii( record_comp )

    if ( large_traj_nb ) then
        write( record_comp % format_string_ascii, '("(", i1, "i10,9f9.3,", i2, "es12.4,", i1, "es12.4)")') &
                                         record_comp % nb_int, record_comp % nb_real,record_comp % nb_other_real
        write(*,*) 'new format str ', record_comp % format_string_ascii
    end if

    ! set size, identical, may changed after
    record_comp % nb_max_complexe = o_record % pcomplexe % nb_complexe
    write(*,*) 'max size', record_comp % nb_max_complexe
    write(*,*) 'o_record nb_complexe', o_record % pcomplexe % nb_complexe

    !write(*,*) 'end initialisation'
    !write(*,*)
end subroutine initialize_record_cluster_format

! quick fix
! new added, make internal to read_record, but could be in mod_record also
! \todo check if working with opt_sum = 1
subroutine sum_energy_terms( o_record, array_real )
    USE mod_record
    IMPLICIT NONE
    type ( record ), target :: o_record
    real ( kind=8 ), dimension (:) :: array_real

    integer :: i

    !write(*,*) 'Entry sum_energy_terms '
    !write(*,*) 'size array_real ', size (array_real)
    !write(*,*) 'array_real ', array_real

    do i = 4, BIT_SIZE( o_record % bit_energy ) - 1
    !do i = 0, BIT_SIZE( o_record % bit_energy ) - 1

        if ( BTEST( o_record % bit_energy, i ) ) then
            !write(*,*) 'bit ',i,' is set'

            ! Lennard-Jones PrometCS, add to column 3
            if ( i == 4 ) then
                !write(*,*) 'Lennard-Jones PrometCS found'
                if ( o_record % opt_sum ) then
                    array_real(5) = array_real( 2 + i) + array_real(3 + i)
                else
                    array_real(8) = array_real(8) + array_real(2 + 2*i)
                    array_real(9) = array_real(9) + array_real(3 + 2*i)
                    !write(*,*) 'Copy LJprometCS ',array_real(8) , array_real(9)
                end if
                !array_real(3) =
            end if

            ! Image-charge, add to electrostatic
            if ( i == 5 ) then
                !write(*,*) 'Image charge found'
                if ( o_record % opt_sum ) then
                    array_real(2) = array_real(2) + array_real(3 + 2*i)
                else
                    array_real(2) = array_real(2) + array_real(2 + 2*i)
                    array_real(3) = array_real(3) + array_real(3 + 2*i)
                    !write(*,*) 'Copy Image charge ',array_real(2) , array_real(3)
                end if

            end if

            !Metal desolvation, add to LJ
            if ( i == 6 ) then
                !write(*,*) 'Metal desolvation found'
                if ( o_record % opt_sum ) then
                    array_real(5) = array_real(5) + array_real(3 + 2*i)
                else
                    array_real(8) = array_real(8) + array_real(2 + 2*i)
                    array_real(9) = array_real(9) + array_real(3 + 2*i)
                    !write(*,*) 'Copy MetalDesolv ',array_real(4) , array_real(5)
                end if
            end if

        !else
        !    write(*,*) 'bit ',i,' is not set'
        end if
    end do

    !write(*,*) 'array_real ', array_real

end subroutine sum_energy_terms

subroutine concatenate_traj ( o_record, traj_concat_f )
     USE mod_record
     IMPLICIT NONE
     
     type ( record ), intent ( inout ) :: o_record
     character ( 256 ), dimension (:), intent ( in ) :: traj_concat_f

     character ( 15 ) :: format_file
     integer :: nb_mols, last_step, last_line, last_mol
     integer :: ioerr, curr_file_i, out_file_i
     integer :: nb_files, curr_file, curr_frame, curr_step, tot_nb_lines
     
     real ( kind = 8 ) :: curr_time, last_time

     nb_files = size ( traj_concat_f ) + 1
     out_file_i = 111
     curr_file_i = 112
     nb_mols = o_record % nb_protein
     curr_step = 0
     curr_time = 0.d0
     last_time = 0
     
     ! open the output file
     if ( o_record % format_ascii ) then
         format_file = 'formatted'
     else
         format_file = 'unformatted'
     end if

     open ( out_file_i, file = 'concatenated_traj', status = 'NEW', iostat = ioerr, form = format_file )
     if ( ioerr /= 0 ) then
       write (*,*) 'ERROR Cannot create new file "concatenated_traj"!'
       STOP 1
     end if
     

     ! write the header only once
     o_record % io_compl = out_file_i
     call write_header_record ( o_record )
     
     do curr_file = 1, nb_files ! read each file
!         write (*,*) 'curr file: ', curr_file
        last_line = o_record % ptrajectory % size_array
        last_mol = o_record % ptrajectory % array_trajectory ( last_line ) % int_ocompl ( 2 )

        ! check whether there is an incomplete frame
        if ( last_mol /= nb_mols ) then
            write (*,*) 'INFO: incomplete frame detected, will fix...'
            last_line = o_record % ptrajectory % size_array - last_mol
        end if
        last_step = o_record % ptrajectory % array_trajectory ( last_line ) % int_ocompl ( 1 )
        last_time = o_record % ptrajectory % array_trajectory ( last_line ) % other_real_ocompl (1)
!        write(*,*) 'updating last_step: ', last_step, ', last_mol: ',last_mol
         
         if ( curr_file /= 1 ) then ! first trajectory has been already read
!            write(*,*) 'deleting old traj...'
!            write(*,*) 'delete a ptrajectory'
            call delete_trajectory ( o_record % ptrajectory, o_record, .true. )

            deallocate ( o_record % ptrajectory )
            nullify ( o_record % pcomplexe, o_record % ptrajectory )
            o_record % header_str = ''
            o_record % io_compl = curr_file_i
            
            call read_header_record ( o_record, traj_concat_f ( curr_file - 1 ), opt_nb_record = tot_nb_lines )
            
            if ( o_record % type_record /= trajectory_enum ) then
                write (*,*) 'ERROR this is not a trajectory! "concat" option concatenates only trajectories!'
                stop 1
            end if
            
            allocate ( o_record % ptrajectory, stat = ioerr )
            if ( ioerr /= 0 ) then
                write (*,*) 'Problem in allocating ptrajectory '
                stop 1
            end if
            call initialize_trajectory ( o_record % ptrajectory, o_record )

!            write(*,*) 'reading new traj...'
            call read_file_record ( o_record )
            last_line = o_record % ptrajectory % size_array
            last_mol = o_record % ptrajectory % array_trajectory ( last_line ) % int_ocompl ( 2 )

            ! check whether there is an incomplete frame
            if ( last_mol /= nb_mols ) then
                last_line = o_record % ptrajectory % size_array - last_mol
            end if
         end if
             
     ! first frame has to be discarded, because the molecules can be too close to each other or
     ! it can be a duplicate of the last frame of the previous trajectory
         do curr_frame = nb_mols + 1, last_line
             if ( curr_file /= 1 ) then
                ! read the step number from the current trajectory file
                curr_step = o_record % ptrajectory % array_trajectory ( curr_frame ) % int_ocompl ( 1 )
                curr_time = o_record % ptrajectory % array_trajectory ( curr_frame ) % other_real_ocompl (1)
                ! correct the numbering, if this is not the first file
                o_record % ptrajectory % array_trajectory ( curr_frame ) % int_ocompl ( 1 ) = curr_step + last_step
                o_record % ptrajectory % array_trajectory ( curr_frame ) % other_real_ocompl (1) = curr_time + last_time
             end if
         end do
         ! write the trajectory to the output file
         o_record % io_compl = out_file_i
         call write_part_array( o_record % ptrajectory, o_record, first = nb_mols + 1, last = last_line )
         
         close ( curr_file_i )

     end do
     
     ! close the file
     close  ( out_file_i )

end subroutine concatenate_traj

subroutine extract_frames ( i_record, extr_fr_beg, extr_fr_end, large_traj_nb )
    use mod_record
    IMPLICIT NONE
    
    type ( record ), intent ( in ) :: i_record
    integer, intent ( inout ) :: extr_fr_beg, extr_fr_end
    logical, intent( in ) :: large_traj_nb

    character ( 15 ) :: format_file
    character ( 512 ) :: string
    integer :: out_file_i, ioerr, max_nb_frames, i
    type ( record ), pointer :: o_record
    
    if ( i_record % format_ascii ) then
       format_file = 'formatted'
    else
       format_file = 'unformatted'
    end if

    max_nb_frames = int ( i_record % nb_max_complexe / i_record % nb_protein )
    if ( extr_fr_end .lt. 0 ) then
        if ( extr_fr_end .eq. -1 ) then
            extr_fr_end = max_nb_frames
        else
            extr_fr_end = max_nb_frames + extr_fr_end
        end if
    end if
    if ( extr_fr_beg .lt. 0 ) then
        if ( extr_fr_beg .eq. -1 ) then
            extr_fr_beg = max_nb_frames - 1
        else
            extr_fr_beg = max_nb_frames - 1 + extr_fr_beg
        end if
    end if
    
    if ( extr_fr_end .lt. extr_fr_beg ) then
        write(*,*) ' ERROR given value for "end" is less than "beg"!'
        stop 1
    end if
    
    if ( extr_fr_end .gt. max_nb_frames ) then
        write (*,*) 'WARNING Given value for "end" is larger than maximal number of frames (',max_nb_frames,')'
        extr_fr_end = max_nb_frames
        
        if ( extr_fr_beg .eq. 0 ) then
            write(*,*) '       and value for "beg" is 0 - will not extract the entire trajectory'
            write(*,*) 'ERROR Will not extract the entire trajectory'
            STOP 1
        else
            write(*,*) '       will extract frames from "beg" to the end'
        end if
    end if
    
    out_file_i = 111
    open ( out_file_i, file = 'extracted_frames', status = 'NEW', iostat = ioerr, form = format_file )
    if ( ioerr /= 0 ) then
      write (*,*) 'ERROR Cannot create new file "extracted_frames"!'
      STOP 1
    end if

    nullify( o_record )
    call copy_record( i_record, o_record, opt_header = .true. )
    o_record % io_compl = out_file_i
    call write_header_record ( o_record )
    
    if ( large_traj_nb ) then
        ! the file was rewinded, so I have to get rid of the first two lines:
        read ( i_record % io_compl , fmt = '(a)', iostat = ioerr ) string
        read ( i_record % io_compl , fmt = '(a)', iostat = ioerr ) string
        i_record % ptrajectory % current_pos = 1
        do i = 1, ( extr_fr_beg - 1 ) * i_record % nb_protein
            i_record % ptrajectory % current_pos = 1
            call read_string_trajectory ( i_record % ptrajectory, i_record, ioerr )
        end do
        do i = 1, ( extr_fr_end - extr_fr_beg + 1 ) * i_record % nb_protein
            i_record % ptrajectory % current_pos = 1
            call read_string_trajectory ( i_record % ptrajectory, i_record, ioerr )
            call write_part_array ( i_record % ptrajectory, o_record, 1, 1 )
        end do
    else
        call write_part_array ( i_record % ptrajectory, o_record, extr_fr_beg * i_record % nb_protein + 1, &
                           extr_fr_end * i_record % nb_protein )
    end if

    close  ( out_file_i )
    nullify( o_record )
  
end subroutine extract_frames

!subroutine split_success_traj( o_record, filename_react_crit, large_traj_nb )
!    use mod_record
!    use mod_ReactionCriteria
!    IMPLICIT NONE
!    
!    type( record ), intent( inout ) :: o_record
!    character( 64 ), intent( in ) :: filename_react_crit
!    type ( react_criter ), target :: rcriteria
!    logical, intent( in ) :: large_traj_nb
!    
!    ! we can assume that number of successful trajectories will be quite small
!    integer, dimension( : ), allocatable :: success_traj
!    integer :: nsuccess_traj, max_nsuccess_traj, ioerr, max_nsteps, i, out_f
!    integer, dimension( 4 ) :: tmp_int_array
!    
!    character( 128 ) :: output_trajs
!    character*512 :: string
!    
!    type( one_complexe ), pointer :: cur_step
!    
!    max_nsteps = o_record % nb_max_complexe
!    if ( max_nsteps .eq. 0 ) then
!        write(*,*) 'ERROR trajectory has size of 0!'
!        stop 1
!    end if
!    
!    max_nsuccess_traj = 1000
!    allocate ( success_traj( max_nsuccess_traj ), stat = ioerr )
!    if ( ioerr /= 0 ) then
!        write (*,*) 'Problem in allocating success_traj'
!        stop 1
!    end if
!    success_traj = 0
!    nsuccess_traj = 0
!    
!    !open the output file
!    output_trajs = "successful_trajectories"
!    out_f = 116
!    if ( o_record % format_ascii ) then
!      open ( unit = out_f, file = output_trajs, form = 'formatted', status = 'new', iostat = ioerr )
!      if ( ioerr .ne. 0 ) then
!        write( *,* ) 'ERROR could not open the output file!'
!        STOP 1
!      endif
!    else
!      open ( unit = out_f, file = output_trajs, form = 'unformatted', status = 'new', iostat = ioerr )
!      if ( ioerr .ne. 0 ) then
!        write( *,* ) 'ERROR could not open the output file!'
!        STOP 1
!      endif
!    endif
!    o_record % io_compl = out_f
!
!!    call allocate_react_criter ( rcriteria, 'association', filename_react_crit, '', '', 2, 1 )
!!    
!!    !first, find the successful trajs - maybe find all the frames where the ligand is close to the b site
!!    do i = 1, max_nsteps
!!        cur_step => o_record % ptrajectory % array_trajectory( i )
!!        call return_array_int( cur_step, tmp_int_array, o_record % bit_integer )
!!        if ( get_index( tmp_int_array( 1 ), success_traj ).ne. -1 ) then
!!            ! we have to check the current step
!!            
!!            
!!            ! if the reaction criteria are satisfied (with a margin), mark the trajectory as successful
!!            nsuccess_traj = nsuccess_traj + 1
!!            success_traj( nsuccess_traj ) = tmp_int_array( 1 )
!!        else
!!            ! this trajectory is already marked as successful
!!            if ( o_record % format_ascii ) then
!!                call make_ascii_string( cur_step, string, o_record % format_string_ascii )
!!                write ( o_record % io_compl, fmt='(a)' ) trim( string )
!!            else
!!                call write_binary_string ( cur_step, o_record % io_compl, ioerr )
!!            end if
!!            flush( o_record % io_compl )
!!        end if
!!    end do
!!    
!!    !then, extract only those traj numbers
!!    !write to a separate file
!!    write(*,*) 'Found ', nsuccess_traj, ': ', success_traj( 1:nsuccess_traj )
!    
!end subroutine split_success_traj


subroutine extract_dist ( i_record, max_c2c_dist, large_traj_nb )
    use mod_record
    IMPLICIT NONE

    type( record ), intent( in ) :: i_record
    logical, intent( in ) :: large_traj_nb

    type( record ), pointer :: o_record
    type( one_complexe ), pointer :: cur_step
    real( kind = 8 ) :: max_c2c_dist, cur_dist
    integer :: max_nsteps, i, out_f, ioerr, nsteps_out
    character( 128 ) :: output_trajs
    character( 512 ) :: string
    character( 15 ) :: format_str
    
    max_nsteps = i_record % nb_max_complexe
    if ( max_nsteps .eq. 0 ) then
        write(*,*) 'ERROR trajectory has size of 0!'
        stop 1
    end if
    
    if ( i_record % format_ascii ) then
        format_str = "formatted"
    else
        format_str = "unformatted"
    end if
    
    output_trajs = "traj_close_c2c"
    out_f = 116
    open ( unit = out_f, file = output_trajs, form = format_str, status = 'new', iostat = ioerr )
    if ( ioerr .ne. 0 ) then
      write( *,* ) 'ERROR could not open the output file!'
      STOP 1
    endif
    nullify( o_record )
    call copy_record( i_record, o_record, opt_header = .true. )
    o_record % io_compl = out_f
    call write_header_record ( o_record )
    if ( large_traj_nb ) then
        i_record % ptrajectory % current_pos = 1
        call read_string_trajectory ( i_record % ptrajectory, i_record, ioerr )
    end if
    
    nsteps_out = 0
    do i = 1, max_nsteps
        if ( large_traj_nb ) then
            i_record % ptrajectory % current_pos = 1
            call read_string_trajectory ( i_record % ptrajectory, i_record, ioerr )
            cur_step => i_record % ptrajectory % array_trajectory( 1 )
        else
            cur_step => i_record % ptrajectory % array_trajectory( i )
        end if
        cur_dist = sqrt ( dot_product ( cur_step % position, cur_step % position ) )
        if ( cur_dist .le. max_c2c_dist ) then
            ! write it out to a file
            if ( i_record % format_ascii ) then
                call make_ascii_string( cur_step, string, i_record % format_string_ascii )
                write ( o_record % io_compl, fmt='(a)' ) trim( string )
            else
                call write_binary_string ( cur_step, o_record % io_compl, ioerr )
            end if
            flush( o_record % io_compl )
            nsteps_out = nsteps_out + 1
        end if
    end do
    
    write(*,*) 'Found ', nsteps_out, ' steps with C2C distance less than or equal to ', max_c2c_dist
    close( out_f )
    nullify( o_record )
    
end subroutine extract_dist


function get_index( val, array ) result ( idx )
    IMPLICIT NONE
    
    integer :: val, idx
    integer, dimension(:) :: array
    
    do idx = 1, size( array )
        if( array( idx ).eq.val ) then
            exit
        end if
    end do
    
    if( idx.gt.size( array ) ) then
        idx = -1
    end if
    
!    return idx
end function get_index
