!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> @file trans_diffusion.f90
!!
!! Compute the autocorrelation function and translational diffusion coefficients

!> Program trans_diffusion
!!
!! \ingroup tools
!! Need box information to retrieve the correct position of the solutes\n
!! Correlation functions computed in parrallel with OpenMP\n
!! Different methods are implemented for fitting the diffusion coefficients
!!
!! Common function are implemented in utils_diffusion.f90
!!
!! sdamm tool : acf_multyprot.f , use for rotational diffusion constant
!!              msd_timeensavg_multyprot.f, use for translational diffusion coefficient
!!
!! 2 versions in sdamm :
!!    - space dependent translational diffusion, use of bins to discretize space
!!          msd_klm_ensavg
!!    - multyprot, average over type of proteins, option z only
!!
program trans_diffusion

    USE mod_record
    USE mod_input_parameter
    USE mod_sda_input
    USE AllTimer
!$  USE omp_lib
    IMPLICIT NONE

    interface
        subroutine all_msd_translation ( position, tab_prot, ACF, nchunks, b_type, value_type, nb_region, dt, io_dt )
                                         ! D_COEFF_DERIV, D_COEFF_SDERIV, &
            USE mod_array_protein
            IMPLICIT NONE
            ! x, y, z, nb_prot, time
            real ( kind=8 ), dimension (:,:,:), intent ( in ) :: position
            type ( array_protein_type ), intent ( in ) :: tab_prot
            ! Auto correlation function ( nchunk, ntype, tau )
            real ( kind=8 ), dimension (:,:,:), allocatable :: ACF !, D_COEFF_DERIV, D_COEFF_SDERIV
            real ( kind=4 ), intent ( in ) :: dt
            integer, intent ( in ) :: nchunks, value_type, nb_region, io_dt
            logical, intent ( in ) :: b_type
        end subroutine all_msd_translation

        !> print all ACF in one file
        subroutine print_acf ( ACF, time_snapshot, b_type, value_type, title )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:), intent ( in ) :: ACF
            character, intent( in ) :: title*(*)
            real (kind=4), intent ( in ) :: time_snapshot
            integer, intent ( in ) :: value_type
            logical, intent ( in ) :: b_type
        end subroutine print_acf

        !> format compatible to sdamm, not used here
        !subroutine print_acftrans_sdamm ( ACF, time_snapshot, b_type, value_type )
        !    IMPLICIT NONE
        !    real ( kind=8 ), dimension (:,:,:), intent ( in ) :: ACF
        !    integer :: value_type
        !    logical :: b_type
        !    real (kind=4) :: time_snapshot
        !end subroutine print_acftrans_sdamm

    end interface

    type ( record ) :: record_traj
    type ( array_protein_type ) :: tab_protein

    ! for reading record, even if not used
    !! keep 8 because binary wrote with 8, should change all to 4
    real ( kind = 8 ), dimension ( 2, 3 ) :: cm
    integer :: nb_line

    ! for reading input file, initialized in read_input_file
    type ( param_list ) :: list_parameter
    character*64, dimension(:,:), allocatable :: all_filename_pdb
    logical :: iflex

    !! for log files
    integer :: io_dt = 41

    type ( mytimer ) :: timer_read, timer_acf, timer_read2

    ! for reading argument
    character*64 str_arg
    character*64 fname_traj_input, filename_input
    integer :: nb_argument, count_arg
    integer :: input_traj_io
    logical :: bool_input
    ! option space dependent, not implemented yet
    !logical :: space_dependent

    ! for loop over record
    !type ( one_complexe ), pointer :: p_tmp_onecomplexe
    ! number proteins and number of record (= total_line / nb_protein)
    integer :: nb_prot, nb_record
    integer :: nloop, tmp_prot
    integer :: ierr
    real ( kind=4 ) :: time_snapshot

    ! Auto correlation function ( nchunks, ntype, tau )
    real ( kind=8 ), dimension (:,:,:), allocatable :: ACF
    ! for storing mean and variance of diffusion coefficient. Could be intern to all_msd, to do later
    !real ( kind=8 ), dimension (:,:,:), allocatable :: D_COEFF_DERIV, D_COEFF_SDERIV

    ! array to store position of all proteins, at all time. correct for cell
    ! ( tranlsation, nb_prot, time )
    real ( kind=8 ), dimension (:,:,:), allocatable :: position
    real ( kind=8 ), dimension(3) :: tmp_position
    ! to get cell position
    integer, dimension(3) :: cell
    ! box dimension read from sda_input
    real ( kind=8 ), dimension ( 3 ) :: box

    ! arguments
    integer :: nchunks, value_type, nb_region
    logical :: b_type

    !nullify ( p_tmp_onecomplexe )

    ! default initialisation
    fname_traj_input = ''
    filename_input = ''
    bool_input = .false.
    count_arg = 1
    input_traj_io = 155
    nb_line = 0
    ! to add in input, default is 3
    nb_region = 3

    ! default arguments
    nchunks = 1
    b_type = .false.
    ! space_dependent = .false.
    ! 0 by default means all
    value_type = 0

    !! initialise timer
    call init_mytimer ( timer_read )
    call init_mytimer ( timer_acf )
    call init_mytimer ( timer_read2 )

    !$  write(*,*) 'OpenMP activated'

    !! give available on hte machine, but not used here, only thread used
    !$  write(*,*) 'nb processor on this machine ', OMP_get_num_procs()

    !! give only the thread at this point
    !$  write(*,*) 'nb of threads ', OMP_get_num_threads()
    !$  write(*,*) 'max nb of threads ', OMP_get_max_threads()
    !$  write(*,*) 'dynamic activated ', OMP_get_dynamic()

    ! init record_traj
    nullify ( record_traj % pcomplexe, record_traj % ptrajectory, record_traj % ptab_prot )
    record_traj % io_compl = input_traj_io
    record_traj % header_str = ''

!!!!!!!!!!!!! Read arguments !!!!!!!!!!!!!!!!!!!!!!

    !nb_argument = IARGC ()
    nb_argument = command_argument_count()

    if ( nb_argument == 0 ) then
        write(*,*) 'trans_diffusion   - calculate the translational diffusion from SDAMM trajectory'
        write(*,*) ''
        write(*,*) 'USAGE:'
        write(*,*) ' trans_diffusion  trajectory_input [ -input sda.in ]  [ -nchunks 1 ]  [ -nblocks 3 ]  -solute_type X'
        write(*,*)
        !write(*,*) '-method : msd or acf, rotational mean square displacement or autocorrelation function'
        !write(*,*) '          default both'
        write(*,*) '-nchunks X       : split the trajectory into X equal part, and compute for the different parts; default = 1'
        write(*,*)
        write(*,*) '-solute_type [X] : optional value;'
        write(*,*) '                   if X absent, compute for all types of solutes;'
        write(*,*) '                   if X given, compute only for type X ( begin at 1 )'
        write(*,*)
        write(*,*) '-nblocks X       : will split the msd in X blocks and fit the coefficient for each separately; '
        write(*,*) '                   default = 3'
        stop
    end if

    do while ( count_arg <= nb_argument )

      call getarg( count_arg, str_arg )

      if ( count_arg == 1 ) then
          !write ( *, * ) 'read file ', trim ( str_arg )
          fname_traj_input = str_arg

      else if ( str_arg ( 1:6 ) == "-input" ) then
          bool_input = .true.
          call getarg( count_arg +1, str_arg )
          filename_input = trim( str_arg )
          count_arg = count_arg + 1

      else if ( str_arg ( 1:8 ) == "-nchunks" ) then
          call getarg(  count_arg +1, str_arg )
          read ( str_arg, fmt='(i3)' ) nchunks
          count_arg = count_arg + 1

      else if  ( str_arg ( 1:12 ) == "-solute_type" ) then
          b_type = .true.
          call getarg(  count_arg +1, str_arg )
          read ( str_arg, fmt='(i3)', iostat = ierr ) value_type
          if ( ierr == 0 ) then
              count_arg = count_arg + 1
          !else
          !    write(*,*) 'use default value_type 0'
          !    stop 1
          end if

      else if  ( str_arg ( 1:8 ) == "-nblocks" ) then
          call getarg(  count_arg +1, str_arg )
          read ( str_arg, fmt='(i3)' ) nb_region
          count_arg = count_arg + 1

      end if

      count_arg = count_arg + 1

    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    write(*,*)
    write(*,*) 'Entry Translational_diffusion '
    write(*,*)

    if ( ( fname_traj_input == '' ) .or. ( filename_input == '' ) ) then
        write(*,*) 'arguments missing'
        stop 1
    end if

    write(*,*) 'filename ', trim(fname_traj_input)
    write(*,*) 'nchunks ', nchunks
    write(*,*) 'nblock ', nb_region
    if ( b_type ) then
        write(*,*) 'solute_type ON value=', value_type
    else
        write(*,*) 'solute_type OFF'
    endif
    write(*,*)

    if ( bool_input ) write(*,*) 'sda_input=',trim( filename_input )

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                                                               ! all_filename_pdb, iflex, load_pdb, extend_array
    call read_input_sda ( tab_protein, filename_input, list_parameter, all_filename_pdb, iflex, .false., .false. )
    !write(*,*) 'list_parameter % total_solute ', list_parameter % total_solutes
    !write(*,*) 'list_parameter % dt ', list_parameter % dt1
    !write(*,*) 'list % freq ', list_parameter % freq_print
    !write(*,*) 'list_parameter % xmin ', list_parameter % xmin !xmax, ymin, ymax, zmin, zmax
    !write(*,*) 'list_parameter % xmax ', list_parameter % xmax

    box ( 1 ) = list_parameter % xmax - list_parameter % xmin
    box ( 2 ) = list_parameter % ymax - list_parameter % ymin
    box ( 3 ) = list_parameter % zmax - list_parameter % zmin
    write(*,*) 'box ', box

    ! time in nanosecond
    time_snapshot = real ( list_parameter % dt1 * list_parameter % freq_print / 1000.0 ,4)
     write(*,*) 'time_snapshot in ns. ', time_snapshot
    ! false, do not test if p_sogrid is associated
    call print_info_array_protein ( tab_protein, .false. )

  call start_mytimer( timer_read )
    ! read header, initialize record
    !! if option nb_line will compute number of record and setup variable field
    call read_header_record ( record_traj, fname_traj_input, cm, nb_line )
    write(*,*) 'nb_line ', nb_line

    !! check if sdamm and trajectory, may be less strict later ?
    if ( ( record_traj % record_type_calc % enum_type_calc /= sdamm ) .or. &
         ( record_traj % type_record /= trajectory_enum ) ) then
        write(*,*) 'not a sdamm trajectory, not implemented '
        stop 1
    end if

    !! initialize trajectory
    if ( record_traj % type_record == trajectory_enum ) then
        allocate ( record_traj % ptrajectory )
        !! allocate array of new complexe with size array = record_traj % nb_max_complexe
        !! could add optional max_size, read conf by conf
        call initialize_trajectory ( record_traj % ptrajectory, record_traj )
        !write(*,*) 'Trajectory, init array with size ', record_traj % ptrajectory % size_array
        !write (*,*) 'size array for trajectory ', record_traj % ptrajectory % size_array

    ! a pririo, not used
    !else if ( record_traj % type_record == complexe_enum ) then
        !! creates the list, no data yet
        !call new_listcomplexe ( this % pcomplexe, rmsd_min, one_run, one_conf )
        !call new_listcomplexe ( record_traj % pcomplexe, 0.d0, .false., .false. )
    !    write(*,*) ' not implemented for complexes file'
    !    stop 1
    else
        write(*,*) 'problem with record_traj % type_record, stop '
        stop 1
    end if

    !! can read both trajectory or complexe, read all the trajectory into memory
    call read_file_record ( record_traj )
    ! rewind to the first element
    call reset_position ( record_traj, ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'error in allocating orient_axis'
        stop 1
    end if
    ! point to the first
    !p_tmp_onecomplexe => record_traj % p_onecompl

  call stop_mytimer( timer_read )

    ! make loop
    nb_prot = record_traj % nb_protein
    !write(*,*) 'nb_prot ', nb_prot
    nb_record = record_traj % nb_max_complexe  / nb_prot
    !write(*,*) 'nb_record ', nb_record

    ! need global array position, efficient for storage and for parrallelisation
    ! vector pso,nb_prot,time
    allocate ( position ( 3, nb_prot, nb_record ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'error in allocating position'
        stop 1
    end if
    position = 0.d0

    ! global array, avoids to recompute cross many, many times !!
    nloop = 1
    ierr = 0

  call start_mytimer( timer_read2 )
    ! loop over all trajectory, do nloop = 1, nb_record, better for OpenMP
    do while ( ( ierr == 0 ) .and. ( nloop <= nb_record ) )
        !write(*,*) 'new record ', nloop, ' tmp_pos_in ', tmp_pos_in

        ! loop over all proteins
        do tmp_prot = 1, nb_prot

            ! read position in the box
            tmp_position = record_traj % p_onecompl % position
            cell = record_traj % p_onecompl % box

            ! compute absolute position, needed only once
            tmp_position = tmp_position + dble ( cell * box )

            ! assign to global array
            position (:, tmp_prot, nloop ) = tmp_position

            ! add option timestep for jumping over unused lines
            call getNext(  record_traj, ierr )
            ! test ierr ??

        ! end do loop over proteins
        end do

        nloop = nloop + 1
    ! end loop over record
    end do

  call stop_mytimer( timer_read2 )

  ! can delete record, save memory
  call delete_record ( record_traj, .true. )

!!!!!!!!!!!!!!!!!!!!!!!!!! Compute autocorrelation for the MSD

  call start_mytimer ( timer_acf )

  ! io_dt main log file
  call all_msd_translation ( position, tab_protein, ACF, & !D_COEFF_DERIV, D_COEFF_SDERIV, &
                             nchunks, b_type, value_type, nb_region, time_snapshot, io_dt )

  call stop_mytimer ( timer_acf )

  ! close main log file
  close ( io_dt )

  call delete_array_protein ( tab_protein )

  write(*,*) 'timer read'
  call print_mytimer ( timer_read )
  write(*,*) 'timer read2'
  call print_mytimer ( timer_read2 )
  write(*,*) 'timer computation'
  call print_mytimer ( timer_acf )


end program trans_diffusion

!> Compute all msd, autocorrelation function of the position
!! @param position : 3 dimensional array ( x, y, z, nb_prot, time ) with the position of the solutes
!! @param tab_prot : instance of \ref mod_array_protein::array_protein_type "array_protein_type"
!! @param ACF : return, all autocorrelation function
!! @param nchunks : number of part the trajectory must be split
!! @param b_type : logical, compute for different type of proteins if .true.
!! @param value_type if 0 compute for all proteins AND for all types of proteins, if > 0 compute only for this type,
!!                        could improve, read in orient_axis only for this type and forget about this argument
!! @param nb_region : number of regions ??
!! @param dt : timestep
!! @param io_dt : file descriptor for output
subroutine all_msd_translation ( position, tab_prot, ACF, &
                                 nchunks, b_type, value_type, nb_region, dt, io_dt )
    USE mod_array_protein
    USE Alltimer
    IMPLICIT NONE

    interface
        subroutine msd_translat ( input_data, output_data )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:), intent ( in ) :: input_data
            real ( kind=8 ), dimension (:), intent ( out ) :: output_data
        end subroutine msd_translat

        ! defined in math.f
        subroutine derivative ( input, output, dt, opt_factor )
          IMPLICIT NONE
          real ( kind=8 ), dimension(:), intent ( in ) :: input
          real ( kind=8 ),dimension(:),intent( out ) :: output
          real ( kind=8 ) :: dt
          real ( kind=8 ), optional :: opt_factor
        end subroutine derivative

        subroutine simple_derivative ( input, output, dt, opt_factor )
          IMPLICIT NONE
          real ( kind=8 ), dimension(:), intent ( in ) :: input
          real ( kind=8 ),dimension(:),intent( out ) :: output
          real ( kind=8 ) :: dt
          real ( kind=8 ), optional :: opt_factor
        end subroutine simple_derivative

        subroutine fit_msd_by_regions ( input, nb_region, dt, b_trans, io_dt )
            IMPLICIT NONE
            real ( kind=8 ), dimension(:), intent ( in ) :: input
            real ( kind=4 ), intent ( in ) :: dt
            integer, intent ( in ) :: nb_region, io_dt
            logical, intent ( in ) :: b_trans
        end subroutine fit_msd_by_regions

        subroutine fit_msd_derivative ( input, D_COEFF, shift_step, array_limit )
            IMPLICIT NONE
            real ( kind=8 ), dimension(:), intent ( in ) :: input
            real ( kind=8 ), dimension(4) :: D_COEFF
            integer, dimension(4) :: array_limit
            integer :: shift_step
        end subroutine fit_msd_derivative

        subroutine print_coeffd_msd ( D_COEFF, region_fit, b_type, value_type, dt, b_trans, b_deriv, io_msd )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:) :: D_COEFF
            integer, dimension(:,: ) :: region_fit
            real ( kind=4 ) :: dt
            integer :: value_type, io_msd
            logical :: b_type, b_trans, b_deriv
        end subroutine print_coeffd_msd

        ! defined in utils.h
        subroutine print_acf ( ACF, time_snapshot, b_type, value_type, title )
            IMPLICIT NONE
            real ( kind=8 ), dimension (:,:,:), intent ( in ) :: ACF
            character, intent( in ) :: title*(*)
            real ( kind=4 ), intent ( in ) :: time_snapshot
            integer, intent ( in ) :: value_type
            logical, intent ( in ) :: b_type
        end subroutine print_acf

    end interface

    ! x, y, z, nb_prot, time
    real ( kind=8 ), dimension (:,:,:), intent ( in ) :: position
    type ( array_protein_type ), intent ( in ) :: tab_prot
    ! Auto correlation function ( ntype, tau )
    real ( kind=8 ), dimension (:,:,:), allocatable :: ACF
    real ( kind=4 ), intent ( in ) :: dt
    integer, intent ( in ) :: nchunks, value_type, nb_region, io_dt
    logical, intent ( in ) :: b_type

    ! to store diffusion results
    real ( kind=8 ), dimension (:,:,:), allocatable :: D_COEFF_DERIV, D_COEFF_SDERIV
    !! contains the simple derivative of acf / time, and "exact " derivative
    real ( kind=8 ), dimension (:,:,:), allocatable :: deriv_acf, deriv_acf2
    ! to store regions where deriv are fitted, easier for output
    ! not dependent of type !nchunks,4
    integer, dimension(:,:), allocatable :: region_deriv_fit

    integer :: startstep, endstep, nacf, nstep_acf
    integer :: first_protein, last_protein, first_type, last_type

    ! local variable
    integer :: nb_prot, nb_type, size_input
    integer :: ierr, io_dt_log, itype, shift

    type ( mytimer ) :: timer_acf

    call init_mytimer ( timer_acf )

    io_dt_log = 31

    nb_type = -1
    write(*,*) 'Entry all_msd_translation '
    nb_prot = size ( tab_prot % array )
    size_input = size ( position, 3 )
    !write(*,*) 'nb_prot ', nb_prot
    !write(*,*) 'nchunks ', nchunks
    !write(*,*) 'size_input ', size_input


    ! want for all proteins and for the different types
    if ( ( b_type ) .and. ( value_type == 0 ) ) then
        nb_type = tab_prot % nb_grids + 1
    ! only one computation
    else if ( ( ( b_type ) .and. ( value_type > 0 ) ) .or. &
              ( b_type .eqv. .false. ) ) then
        nb_type = 1
    end if
    write(*,*) 'nb_type ', nb_type
    write(*,*) 'nchunks ', nchunks
    ! account for nchunks
    nstep_acf = int ( size_input / nchunks )
    write(*,*) 'nstep_acf ', nstep_acf

    ! allocate to the correct size
    ! same size as input, because 0 is added
    allocate ( ACF ( nchunks, nb_type, nstep_acf ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'error in allocating ACF'
        stop 1
    end if
    ACF = 0.d0
    ! size - 1 for derivative
    allocate ( deriv_acf ( nchunks, nb_type, nstep_acf -1 ), &
               deriv_acf2 ( nchunks, nb_type, nstep_acf -1 ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'error in allocating deriv_acf '
        stop 1
    end if
    deriv_acf = 0.d0
    deriv_acf2 = 0.d0

    ! used if fit of derivative of rmsd
    allocate ( D_COEFF_DERIV ( nchunks, nb_type, 4 ), &
               D_COEFF_SDERIV ( nchunks, nb_type, 4 ), stat = ierr )
    if ( ierr /= 0 ) then
        write(*,*) 'error in allocating D_COEFF'
        stop 1
    end if
    D_COEFF_DERIV = 0.d0
    D_COEFF_SDERIV = 0.d0

    !! not dependent of type ! nchumks,4
    allocate ( region_deriv_fit ( nchunks, 4 ), stat = ierr )
     if ( ierr /= 0 ) then
        write(*,*) 'error in allocating ACF'
        stop 1
    end if
    region_deriv_fit = 0

    ! main log file
    open ( unit=io_dt,file='Translational_Diffusion.log', status='replace' )
    write( io_dt, fmt='(a)' )
    write( io_dt, fmt='(a)' ) 'Translational diffusion coefficient by mean-square displacement'
    write( io_dt, fmt='(a,i6)' ) 'Total number of steps ', size_input
    write( io_dt, fmt='(a,f5.2,a)' ) 'dt =', dt, ' ns'

    shift = 0
    ! default, loop over all types
    first_type = 1
    last_type = tab_prot % nb_grids

LOOP_CHUNK: do  nacf = 1, nchunks

        startstep = nstep_acf * ( nacf - 1 ) + 1
        endstep = nstep_acf * ( nacf )

        write(*,*) 'startstep ', startstep
        write(*,*) 'endstep ', endstep
        !write(*,*) 'nstep_acf ', nstep_acf

! all in output, may change again after
!        write( io_dt, fmt='(a)' )
!        write( io_dt, fmt='(a)' ) 'Definition of short and long time scale when fitting the derivative of the MSD'
!        write( io_dt, fmt='(a,i10)' ) 'For chunk ', nacf
!        write( io_dt, fmt='(a)' ) 'Short time scale '
!        write( io_dt, fmt='(a,i10,a,i10)' ) &
!                   'Average between step ',startstep,' and step ',startstep+(nstep_acf/5)
!        write( io_dt, fmt='(a,f10.2,a,f10.2,a)' ) &
!                   'Average between ',startstep*dt,' (ns) and ',(startstep+(nstep_acf/5))*dt,' (ns)'
!
!        write( io_dt, fmt='(a)' )
!        write( io_dt, fmt='(a)' ) 'Long time scale '
!        write( io_dt, fmt='(a,i10,a,i10)' ) &
!               !'Average between step ',startstep+(nstep_acf/5),' and step ',endstep
!               'Average between step ',startstep+(nstep_acf/5),' and step ',endstep-(nstep_acf/3)
!        write( io_dt, fmt='(a,f10.2,a,f10.2,a)' ) &
!               !'Average between ',(startstep+(nstep_acf/5))*dt,' (ns) and ',endstep*dt,' (ns)'
!               'Average between ',(startstep+(nstep_acf/5))*dt,' (ns) and ',(endstep-(nstep_acf/3))*dt,' (ns)'


        ! compute acf for all proteins, send full array
        if ( (( b_type ) .and. (value_type == 0)) .or. &
              ( b_type .eqv. .false. ) ) then

             !write(*,*) 'b_type is true and compute for all protein types'
             call start_mytimer ( timer_acf )
             call msd_translat ( position(:,:,startstep:endstep), ACF( nacf, 1,: ) )

             ! make a least_square fit on blocks
             write( io_dt, fmt="(a)" )
             write( io_dt, fmt="(a)" ) 'Fit msd of the average over all solutes '
             call fit_msd_by_regions ( ACF ( nacf, 1, : ), nb_region, dt, .true., io_dt )

             ! compute and fit simplified derivative, divide by 6 for translation
             call simple_derivative ( ACF ( nacf, 1, : ), deriv_acf( nacf, 1, : ), real(dt,8), 6.d0 )
             ! fit and store mean and std err and short and long-time
             call fit_msd_derivative ( deriv_acf( nacf, 1, : ), D_COEFF_SDERIV( nacf, 1, : ), &
                                       startstep, region_deriv_fit(nacf,:) )

             ! compute and fit real derivative, rotation need to divide by 4
             call derivative ( ACF ( nacf, 1, : ), deriv_acf2( nacf, 1, : ), real(dt,8), 6.d0 )
             call fit_msd_derivative ( deriv_acf2( nacf, 1, : ), D_COEFF_DERIV( nacf, 1, : ), &
                                       -1, region_deriv_fit(nacf,:) )

             call stop_mytimer ( timer_acf )
             ! if no type calculation required
             if ( b_type .eqv. .false. ) then
                CYCLE LOOP_CHUNK
             end if

            ! if add type acf, will be shifted by one
            shift = 1

        ! else only acf for one type
        else if ( ( b_type ) .and. ( value_type > 0 ) ) then

            ! will limit the loop over the grid
            first_type = value_type
            last_type = value_type

            ! assur than ACF(nacf,1,:) is filled
            shift = 1 - value_type

        end if ! end of tests


        !! calculations of the different types
        do itype = first_type, last_type

                first_protein = tab_prot % set_of_protein ( itype, 1 )
                last_protein  = tab_prot % set_of_protein ( itype, 2 )
                !write(*,*) 'first last prot ', first_protein, last_protein
                call start_mytimer ( timer_acf )

                call msd_translat ( position(:, first_protein:last_protein, startstep:endstep), &
                                    ACF( nacf, itype + shift,: ) )

                write( io_dt, fmt="(a)" )
                write( io_dt, fmt="(a,i3)" ) 'Least square fit of the MSD for solutes type ', itype
                call fit_msd_by_regions ( ACF ( nacf, itype + shift, : ), nb_region, dt, .true., io_dt )

                ! compute and fit simplified derivative, rotation need to divide by 6 for translation
                call simple_derivative ( ACF ( nacf, itype + shift, : ), deriv_acf( nacf, itype + shift, : ), real(dt,8), 6.d0 )
                ! fit and store mean and std err and short and long-time

                ! if itype == 1 startstep, otherwise -1, do not overwrite
                call fit_msd_derivative ( deriv_acf( nacf, itype + shift, : ), D_COEFF_SDERIV( nacf, itype + shift, : ), &
                                          startstep, region_deriv_fit(nacf,:) )

                ! compute and fit real derivative
                call derivative ( ACF ( nacf, itype + shift, : ), deriv_acf2( nacf, itype + shift, : ), real(dt,8), 6.d0 )
                call fit_msd_derivative ( deriv_acf2( nacf, itype + shift, : ), D_COEFF_DERIV( nacf, itype + shift, : ), &
                                          -1, region_deriv_fit(nacf,:) )

                call stop_mytimer ( timer_acf )
        end do

    ! end do nchunks
    end do LOOP_CHUNK

    ! print derivative
    call print_acf ( acf, dt, b_type, value_type, 'trans_msd.dat' )
    call print_acf ( deriv_acf, dt, b_type, value_type, 'Dt_simpderiv.dat' )
    call print_acf ( deriv_acf2, dt, b_type, value_type, 'Dt_deriv.dat' )
    deallocate ( deriv_acf, deriv_acf2 )

    ! used in previous version                                                      b_trans, b_deriv
    call print_coeffd_msd ( D_COEFF_DERIV,  region_deriv_fit, b_type, value_type, dt, .true., .true., io_dt )
    call print_coeffd_msd ( D_COEFF_SDERIV, region_deriv_fit, b_type, value_type, dt, .true., .false., io_dt )

    deallocate ( D_COEFF_DERIV, D_COEFF_SDERIV, region_deriv_fit )

    write(*,*)"timer autocorrelation"
    call print_mytimer ( timer_acf )

end subroutine all_msd_translation

!> Version using the maximum number of data points
!!
!! Large tau values have a poor statistic
!! @param input_data : position of the solute
!! @param output_data : autocorrelation function, 1 dimensional array
subroutine msd_translat ( input_data, output_data )
!$  USE omp_lib
    IMPLICIT NONE
    ! force to have the same shape
    real ( kind=8 ), dimension (:,:,:), intent ( in ) :: input_data
    ! only time
    real ( kind=8 ), dimension (:), intent ( out ) :: output_data

    !integer ( kind=8 ) :: tmp_large_integer
    integer :: nprot, size_input, tau, t0, iprot, tmax, max_tau
    real ( kind=8 ), dimension ( 3 ) :: displ
    real ( kind=8 ) :: tmp_sum


    !write(*,*) 'msd_translat '
    nprot = size ( input_data, 2 )
    size_input = size ( input_data, 3 )
    max_tau = size_input - 1
    write(*,*) 'size_input ', size_input
    write(*,*) 'max_tau ', max_tau
    write(*,*) 'nprot ', nprot

    tmax = 0

    !! can parrallelize this loop
    !$OMP PARALLEL DEFAULT ( NONE ) SHARED ( input_data, size_input, output_data, nprot, max_tau ) &
    !$OMP PRIVATE ( tau, t0, iprot, tmax, tmp_sum, displ ) !, tmp_large_integer )

    !$OMP DO SCHEDULE( DYNAMIC )
    do tau = 0, max_tau !- 1

        tmax = size_input - tau
        tmp_sum = 0.d0

        !do t0 = 1, size_input / 2
        do t0 = 1, tmax

            !! loop over the proteins
            do iprot = 1, nprot

                !! dot_product of the displacement after a time tau
                displ = input_data(:,iprot,t0+tau ) - input_data(:,iprot,t0 )
                tmp_sum = tmp_sum + dot_product ( displ, displ )

            ! end do iprot
            end do

        ! end do t0
        end do

        if ( MOD(tau, 2000) == 0 ) &
          write(*,*) 'computed integral tau', tau

        ! assign value to ACF at tau, IMPORTANT nprot*tmax*tau NEED integer*4
        !tmp_large_integer = nprot * tmax ! not needed anymore, derive in another function

        !output_data ( tau+1 ) = tmp_sum / real ( tmp_large_integer, 8 )
        output_data ( tau + 1 ) = tmp_sum / real ( int(nprot*tmax,4), 8 )

    ! end do tau
    end do
    !$OMP END DO
    !$OMP END PARALLEL

    !write(*,*) 'output_data(1) ', output_data ( 1 )
    !write(*,*) 'output_data(size_inptut -1) ', output_data ( size_input - 1 )
    !write(*,*) 'output_data(size_inptut) ', output_data ( size_input )

end subroutine msd_translat

!> Version with output similar to sdamm, need to produce more files
!!
!! For use with octave and rotation, here not used !!
subroutine print_acftrans_sdamm ( ACF, time_snapshot, b_type, value_type )
    IMPLICIT NONE
    ! Auto correlation function ( nchunks, ntype, tau )
    real ( kind=8 ), dimension (:,:,:), intent ( in ) :: ACF
    integer :: value_type
    logical :: b_type

    character*64 :: str_format
    character*3 :: tmp_str
    character*64 :: title_file
    integer :: nchunks
    integer :: ierr, io_unit = 55
    integer :: size_acf, i, j, nb_type
    real (kind=4) :: time_snapshot
    nchunks = size ( ACF, 1 )
    nb_type = size ( ACF, 2 )
    size_acf = size ( ACF, 3 )

    write(*,*) 'nb_type ', nb_type
    write(*,*) 'size_acf ', size_acf
    write(*,*) 'value_type ', value_type

    write ( str_format, '("(f10.2,",i2,"f15.7)")' ) nchunks

    do i = 1, nb_type

        ! will print only one type
        if ((b_type) .and. (value_type /= 0)) then
            write ( tmp_str, fmt='(i2)' ) value_type
        else
            write ( tmp_str, fmt='(i2)' ) i-1
        end if

        title_file = "MSDtrans_type"//trim(adjustl(tmp_str))//".dat"
        write(*,*) 'title_file', trim ( title_file )

        if ( ( (b_type) .and. (value_type == 0) .and. (i==1) ) .or. &
             ( b_type .eqv. .false. ) ) then
            title_file="MSDtrans_average_all_solutes.dat"
        end if
        open ( io_unit, file=title_file, form='formatted', iostat = ierr )

        write(io_unit,*) '# MSD translation'
        !write(io_unit,*) '# Header of the column to create, include timestep '

        ! case of all proteins, first is average
        if ( (b_type) .and. (value_type == 0) .and. (i==1) ) then
            write (io_unit,*) '#    Average all solutes'
        else if ( (b_type) .and. (value_type > 0) ) then
            write (io_unit,*) '#    Solute ', i
        else
            write (io_unit,*) '#    Solute ', i-1 !value_type
        end if

        do j = 1, size_acf
            write ( io_unit, str_format ) time_snapshot*(j-1), ACF(:,i,j)
        end do

        write(io_unit,*)

        close ( io_unit )

    end do

end subroutine print_acftrans_sdamm

