!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> @file utils_diffusion.f90
!! \ingroup tools
!! Common functions used by diffusion program tools

!> General function for printing ACFs type, in a single file
!! @param ACF : 3 diemsional array for autocorrelation function ( "x,y,z", protein_nb, time )
!! @param time_snapshot : timestep between the frames
!! @param b_type : if true, there is different types of solutes in the simulation
!! @param value_type : indicates which type if b_type is true
!! @param title : filename for the output
subroutine print_acf ( ACF, time_snapshot, b_type, value_type, title )

    IMPLICIT NONE
    ! Auto correlation function ( nchunks, ntype, tau )
    real ( kind=8 ), dimension (:,:,:), intent ( in ) :: ACF
    character, intent( in ) :: title*(*)
    real ( kind=4 ), intent ( in ) :: time_snapshot
    integer, intent ( in ) :: value_type
    logical, intent ( in ) :: b_type


    character*64 :: str_format
    integer :: nchunks !, value_type
    integer :: ierr, io_unit = 55
    integer :: size_acf, i, j, nb_type, m

    nchunks = size ( ACF, 1 )
    nb_type = size ( ACF, 2 )
    size_acf = size ( ACF, 3 )

    !write(*,*) 'nchunks ', nchunks
    !write(*,*) 'nb_type ', nb_type
    !write(*,*) 'size_acf ', size_acf
    !write(*,*) 'value_type ', value_type
    !write(*,*) 'time_snap ', time_snapshot
    !write(*,*) 'title ', trim( title )
    !write(*,*) 'first value ', ACF ( 1, 1, 1 ), ACF ( nchunks, nb_type, size_acf )

    open ( io_unit, file=trim(title), form='formatted', iostat = ierr ) !position='append'

    write(io_unit,*) '# ACF '
    write(io_unit,*) '# Header of the column to create, include timestep '

    if (b_type .eqv. .false. ) then
        write (io_unit,*) '#        Average all solutes'
    end if

    if ((b_type) .and. (value_type > 0)) then
        write (io_unit,*) '#        Solute ',value_type
    end if

    if ((b_type) .and. (value_type == 0)) then
        write (io_unit,fmt='(a29)',advance='no') '#        Average all solutes'
        do m= 1, nb_type - 1
            write (io_unit,fmt='(a10,i3)',advance='no') 'Solute',m
        end do
        write(io_unit,*)
    end if

    !! create format
    write ( str_format, '("(f10.2,",i2,"e15.5)")' ) nb_type
    write (*,*)  str_format

    !! print data
    do j = 1, nchunks
        write(io_unit,*)
        write(io_unit,*) '# Chunks ',j

        do i = 1, size_acf
            write ( io_unit, str_format ) time_snapshot*(i-1), ACF(j,:,i)
        end do

    end do

    close ( io_unit )

end subroutine print_acf

!> Print coefficient from the msd fit method
!!
!! Give io_msd to append the data to the main log file
!! @param D_COEFF : contains the computed values
!! @param region_fit : info about the fitting
!! @param b_type : if true, there is different types of solutes in the simulation
!! @param value_type : indicates which type if b_type is true
!! @param dt : timestep
!! @param b_trans : true if translation, false rotation
!! @param b_deriv : true if exact derivation, false simplified derivative ( MSD/time )
!! @param io_msd : file descriptor
subroutine print_coeffd_msd ( D_COEFF, region_fit, b_type, value_type, dt, b_trans, b_deriv, io_msd )
    IMPLICIT NONE
    real ( kind=8 ), dimension (:,:,:) :: D_COEFF
    integer, dimension(:,:) :: region_fit
    real ( kind=4 ) :: dt
    integer :: value_type, io_msd
    logical :: b_type, b_trans, b_deriv

    character*32 :: tmp_str
    integer :: i, j, nb_type, nb_chunks

    nb_chunks = size ( D_COEFF, 1 )
    nb_type = size ( D_COEFF, 2 )
    write(*,*) 'print_coeffd_msd nb_type ', nb_type, 'nb chunks ', nb_chunks
    write(*,*) 'io_msd ', io_msd

    ! some data already printed before, just append to the main log file
    write ( io_msd, fmt='(a)' )
    if ( b_deriv .eqv. .true. ) then
        write ( io_msd, fmt='(a)' ) '///////////////// Results of Fitting the exact derivative of the MSD /////////////'
    else
         write ( io_msd, fmt='(a)' ) '///////////////// Results of Fitting the simplified derivative of the MSD /////////////'
    end if

    do j = 1, nb_chunks

      write( io_msd, fmt='(a,i3)' ) 'For Chunk ', j
      do i = 1, nb_type

        if ( ( b_type .eqv. .false. ) .or. &
             ( ( b_type .eqv. .true. ) .and. ( value_type == 0 ) .and. ( i == 1 ) ) ) then
            write ( tmp_str,*) 'average all solutes'
        else if ( ( b_type .eqv. .true. ) .and. ( value_type /= 0 ) ) then
            write ( tmp_str,fmt='(a,i3)' ) 'solute type ',value_type
        else
            write ( tmp_str,fmt='(a,i3)' ) 'solute type ',i-1
        end if

        write( io_msd, fmt='(a)' )

        if ( b_trans .eqv. .true. ) then
            write( io_msd,'(a,a)') '  Short-time translational diffusion for ',trim(tmp_str)
            write( io_msd,'(a,f9.2,a,f9.2,a)' ) 'fitted in region ', real ( region_fit(j,1)*dt),' ns to ', &
                                              real ( region_fit(j,2)*dt),' ns'
            write( io_msd,'(a,e12.4,a,a,e12.4,a)') '  <Dt^s>',D_COEFF(j,i,1)/1000.d0,' (A**2/ps) ', &
               'St. dev',sqrt(D_COEFF(j,i,2))/1000.d0,' (A**2/ps)'
            write( io_msd,'(a)')

            write( io_msd,'(a,a)') '  Long-time translational diffusion for ',trim(tmp_str)
            write( io_msd,'(a,f9.2,a,f9.2,a)' ) 'fitted in region ', real ( region_fit(j,3)*dt),' ns to ', &
                                               real ( region_fit(j,4)*dt), ' ns'
            write( io_msd,'(a,e12.4,a,a,e12.4,a)') '  <Dt^l>',D_COEFF(j,i,3)/1000.d0,' (A**2/ps) ',&
               'St. dev',sqrt(D_COEFF(j,i,4))/1000.d0,' (A**2/ps)'
            write( io_msd,'(a)')

        else
            write( io_msd,'(a,a)') '  Short-time rotational diffusion for ',trim(tmp_str)
            write( io_msd,'(a,f9.2,a,f9.2)' ) '  fitted in region ', real ( region_fit(j,1)*dt ),' ns to ', &
                                              real ( region_fit(j,2)*dt),' ns'
            write( io_msd,'(a,e12.4,a,a,e12.4,a)') '  <Dr^s>',D_COEFF(j,i,1)/1000.d0,' (rad**2/ps) ', &
               'St. dev',sqrt(D_COEFF(j,i,2))/1000.d0,' (rad**2/ps)'
            write( io_msd,'(a)')

            write( io_msd,'(a,a)') '  Long-time rotational diffusion for ',trim(tmp_str)
            write( io_msd,'(a,f9.2,a,f9.2,a)' ) '  fitted in region ', real (region_fit(j,3)*dt), ' ns to ', &
                                                 real ( region_fit(j,4)*dt ), ' ns'
            write( io_msd,'(a,e12.4,a,a,e12.4,a)') '  <Dr^l>',D_COEFF(j,i,3)/1000.d0,' (rad**2/ps) ',&
               'St. dev',sqrt(D_COEFF(j,i,4))/1000.d0,' (rad**2/ps)'
            write( io_msd,'(a)')
        end if

      end do
    end do

    ! do not close the file
    !close ( io_msd )

end subroutine print_coeffd_msd

!> Fit the MSD graph
!!
!! @param input : 1 dimension array
!! @param nb_region : how many blocks are divided
!! @param dt : time between each frame
!! @param io_dt : file unit of the output file. Should be opened / close by calller
!! @param b_trans : parameter trans / rot for factor 4 or 6
subroutine fit_msd_by_regions(input, nb_region, dt, b_trans, io_dt)
      IMPLICIT NONE

      interface
        subroutine meanvar( input, mean, var )
          IMPLICIT NONE
          real(kind=8), dimension(:), intent ( in ) :: input
          real(kind=8), intent ( out ) :: mean, var
        end subroutine meanvar

        ! same but shift array reference from 1 to n
        subroutine least_square2(n,X,Y,a,b,d,r2)
            implicit none
            integer, intent ( in ) :: n
            real ( kind=8 ), dimension (n), intent ( in ) :: X,Y
            real ( kind=8 ), intent ( out ) ::  a, b, d, r2
        end subroutine least_square2

      end interface

      real ( kind=8 ), dimension(:), intent( in ) :: input
      real ( kind=4 ), intent ( in ) :: dt
      integer, intent( in ) :: nb_region, io_dt
      logical, intent ( in ) :: b_trans

      real ( kind=8 ), dimension(:), allocatable :: tsteptofit
      real ( kind=8 ), dimension(:), allocatable :: Dif_fit
      real ( kind=8 ) :: factor, afit, bfit, dfit, r2fit, mean, var

      integer :: size_input, block
      integer :: i, start, end

      if ( b_trans .eqv. .true. ) then
        factor = 6.0d0
        write(*,*) 'tranlsation factor = 6. '
      else
        factor = 4.0d0
        write(*,*) 'tranlsation factor = 4. '
      end if

      write(*,*) 'Fit msd_by_region nb_region ', nb_region
      !write(*,*) 'io_dt ', io_dt
      !write(*,*) 'dt ', dt

      size_input = size ( input, 1 )
      write(*,*) 'size input ', size_input
      !write(*,*) 'test input ( size_input ) ', input ( size_input )

      block = size_input / nb_region
      write(*,*) 'block ', block

      ! allocate for computing average of Dfit, could just sum up
      allocate ( Dif_fit ( block ) )
      Dif_fit = 0.d0

      !allocate ( tsteptofit ( 0:size_input-1 ) )
      allocate ( tsteptofit ( size_input ) )
      tsteptofit = 0.d0
      do i = 1, size_input !-1
        tsteptofit ( i ) = real(real(i-1,8)*dt,8)
      end do
      write(*,*) 'tsteptofit(size_intput) ', tsteptofit ( size_input )

      do i = 1, nb_region

        if ( i > 1 ) then
            start = start + block
        else
            start = 1
        end if

        end = start + ( block - 1 )
        write(*,*) 'start end', start, end
        write(*,*) 'end-start+1 ', end-start+1

        call least_square2 ( end-start+1, tsteptofit(start:end), input(start:end), &
                             afit, bfit, dfit, r2fit )

        !write(*,*) 'afit, bfit ', afit, bfit, dfit, r2fit
        !write(*,*) 'bfit/factor ', bfit/factor

        if ( b_trans ) then
          write( io_dt, fmt="('Region',i8,2f15.3,'(ns)',1e10.3,' (A**2/ps), R2:',1e10.3)") &
                i, tsteptofit(start), tsteptofit(end), bfit/(factor*1000.d0), r2fit
        else
           write( io_dt, fmt="('Region',i8,2f15.3,'(ns)',1e10.3,' (R**2/ps), R2:',1e10.3)") &
                  i, tsteptofit(start), tsteptofit(end), bfit/(factor*1000.d0), r2fit
        end if

        Dif_fit( i ) = bfit / factor !6.d0 for trans, 4.d0 for rot
      ! end do loop nb_region
      end do

      ! need to test nb_region here ...

      if ( nb_region >= 5 ) then
        call meanvar( Dif_fit(2:nb_region-1), mean, var )
        write ( io_dt, fmt="(a,f10.3,a,a,f10.3,a,i3,a,i3,a)" ) &
            'Average ', mean/1000.d0, &
            ' (A**2/ps) ', 'St. dev',sqrt(var)/1000.d0, &
            ' (A**2/ps) [Avg on region ',2,'-',nb_region-1,']'
      end if

      deallocate ( Dif_fit )

end subroutine fit_msd_by_regions

!> Fit the derivative of the MSD
!!
!! Different method\n
!! To add array for storing limit of fit, will be easier to print out
!! @param input : 1D array with values to fit
!! @param D_COEFF : array to store the results
!! @param shift_step : ??
!! @param array_limit : store the fitted region
subroutine fit_msd_derivative ( input, D_COEFF, shift_step, array_limit )
    IMPLICIT NONE

    interface
        ! defined in math.f, used only if fit of Dr.dat, seems very sensible
        subroutine meanvar ( input, mean, var )
            IMPLICIT NONE
            real(kind=8), dimension(:), intent ( in ) :: input
            real(kind=8), intent ( out ) :: mean, var
        end subroutine meanvar

    end interface

    real ( kind=8 ), dimension(:), intent ( in ) :: input
    real ( kind=8 ), dimension(4) :: D_COEFF
    integer, dimension(4) :: array_limit
    integer :: shift_step

    integer :: size_input

    size_input = size ( input )
    write(*,*) 'Entry fit_msd_derivative size_input ', size_input
    write(*,*) 'Fit short from 1 to ', size_input/5, &
               ' and long from ', 1+(size_input/5), ' to ', (size_input*2)/3

    ! Short time from 1 to 1/5 of the input
    call meanvar ( input(1:(size_input/5)), D_COEFF( 1 ), D_COEFF( 2 ) )

    !! long nstep / 5 to end - 1 ! real derivative nstep_acf - 2 ??
    !! Long time, should never fit the end, poor statistics, use 2/3 by default
    call meanvar( input(1+(size_input/5): (size_input*2)/3 ), D_COEFF( 3 ), D_COEFF( 4 ) )

    ! not needed
    !array_limit ( 1 ) = 1
    if ( shift_step >= 0 ) then
        write(*,*) ' shift_step ', shift_step
        array_limit ( 1 ) = shift_step
        array_limit ( 2 ) = size_input/5 + shift_step
        array_limit ( 3 ) = array_limit ( 2 ) + 1 !(size_input*2)/3 + shift_step
        array_limit ( 4 ) = (size_input*2)/3 + shift_step
        write(*,*) 'array_limit ', array_limit
    end if


end subroutine fit_msd_derivative

