!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<

!> \file mod_array_protein.f90
!!
!! Define an array where all proteins are stored

! Use of new variable count_type : defined as the number of BLOCK/GROUP Solute_Grid
! Should be in input_file, later may delete and extend array and array_type on the fly

!> Module to define an array of proteins
MODULE mod_array_protein

    USE mod_protein

    !> Main class which owns all subgroups : protein and sogrid, see UMLshema_archi
    !!
    !! Its copy/deletion will assure correct copy/deletion of the subgroups
    !!
    !! Store pointers to all proteins\n
    !! Store the type of the solute in set_of_protein eg : prot 1 - 5 type 1 ( associated to sogrid/plist_flex 1 ), prot 6 10 type 2 ( with sogrid/plist_flex 2 )\n
    type array_protein_type
        !> array of pointers to proteins, not so convenient in case of muVT extension
        type ( protein ), dimension ( : ), pointer :: array

        !> Total number of grids ( as given in sda_input )\n
        !! Count for different types, flexible conformations not accounted here but in \ref mod_flexibility::flexible "flexible"
        integer :: nb_grids
        !> store relation prot-grid in an array of size ( nb_grids, 2 ) : type number, [first prot, last prot]
        integer, dimension ( :, : ), pointer :: set_of_protein
        integer :: size_metal

    end type array_protein_type

    contains

    !> Allocate the arrays
    !! @param this : instance of array_protein_type
    !! @param size : total number of solutes
    !! @param nb_grids : total number of grids ( in fact nb of \ref mod_setofgrid::sogrid "sogrid"), a flexible solute accounts for only one grid ( it uses only one \ref mod_flexibility::flexible "flexible" )
    subroutine new_array_protein ( this, size, nb_grids )
        IMPLICIT NONE
         type ( array_protein_type ) :: this
         integer :: size, nb_grids

         integer :: ierr

         nullify ( this % array, this % set_of_protein )

         !write(*,*) 'Entry new_array_protein size ', size, nb_grid

         allocate ( this % array ( size ), stat = ierr )
         if ( ierr /= 0 ) then
            write ( *, * ) 'Cannot allocate array protein size ', size
         end if

         this % nb_grids = nb_grids

         allocate ( this % set_of_protein ( nb_grids, 2 ), stat = ierr )
         if ( ierr /= 0 ) then
             write ( *, * ) 'Cannot allocate set of protein ', nb_grids, 2
         end if
         ! initialize set_of_proteins to all zero, should be overwritten
         this % set_of_protein = 0

    end subroutine new_array_protein

    !> Deallocate arrays \ref array_protein_type::array "array" and \ref array_protein_type::set_of_protein "set_of_protein"
    !!
    !! By default, delete all proteins and associated flexible and/or \ref mod_setofgrid::sogrid "sogrid" associated\n
    !! Deletion of the \ref mod_setofgrid::sogrid "sogrid" is optional, needed only with sda_2proteins and OpenMP
    !! @param this : instance of array_protein_type
    !! @param opt_copy_omp : optionaly do not delete the associated \ref mod_setofgrid::sogrid "sogrid" and \ref mod_flexibility::flexible "flexible". Default delete all
    subroutine delete_array_protein ( this, opt_copy_omp )
        IMPLICIT NONE
        type ( array_protein_type ) :: this
        logical, optional :: opt_copy_omp

        integer :: igrid, iprot, ierr
        ! tmp pointer to delete at the end of loop, allow check of association for tool
        type ( sogrid ), pointer :: psgrid
        type ( list_flex_type ), pointer :: plist_flex_tmp
        logical :: delete_grid

        nullify ( psgrid, plist_flex_tmp )
        delete_grid = .true.

        if ( present ( opt_copy_omp ) ) then
            delete_grid = .false.
        end if

        ! added for tool, should be one check only if no grid associated, maybe logical value
        ! write(*,*) 'Entry delete array_protein delete_grid ', delete_grid
        ! write( *,* ) 'this % nb_grids ', this % nb_grids

        !!use set of grid, delete all proteins, and then tmp_grid for each set
 GRID:  do igrid = 1, this % nb_grids

            !! could make a function get_nb_prot( igrid,nb_prot_in_set )

            ! if not flexible, one set of grid to delete, use first one by default
            if ( this % array ( this % set_of_protein ( igrid, 1 ) ) % iflex .eqv. .false. ) then

                !write(*,*) 'not flexible '
                ! keep pointer on the associated p_sogrid
                if ( associated ( this % array ( this % set_of_protein ( igrid, 1 ) ) % p_sogrid ) ) then
                    psgrid => this % array ( this % set_of_protein ( igrid, 1 ) ) % p_sogrid
                else
                    nullify ( psgrid )
                end if

                ! delete the proteins, explicit now because local_copy has a pointer
                ! to do before deleting sogrid for test of local_copy
                do iprot = this % set_of_protein ( igrid, 1 ), this % set_of_protein ( igrid, 2 )
                    !write(*,*) 'call delete protein iprot ', iprot
                    call delete_protein ( this % array ( iprot ) )
                end do

                if ( associated ( psgrid ) .and. delete_grid ) then
                    !write(*,*) 'call delete sogrid'
                    call delete_sogrid ( psgrid )
                    deallocate ( psgrid , stat = ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'Cannot deallocate psgrid in delete array'
                    end if
                !else
                !    write(*,*) 'psgrid not associated, no grid to delete'
                end if

            ! TO DO Add a check if exist or not, if used by tools
            ! else flexible, delete plist_flex, will delete correctly all set of grid depending of optional argument
            else
                !write(*,*) 'flexible '

                !! cannot test more
                if ( associated ( this % array ( this % set_of_protein ( igrid, 1 ) ) % p_flex ) ) then
                    plist_flex_tmp => this % array ( this % set_of_protein ( igrid, 1 ) ) % p_flex % plist_flex
                    !write(*,*) 'associated'
                else
                    nullify ( plist_flex_tmp )
                end if
                !delete the proteins, explicit now because local_copy has a pointer
                do iprot = this % set_of_protein ( igrid, 1 ), this % set_of_protein ( igrid, 2 )
                    call delete_protein ( this % array ( iprot ) )
                end do

                if ( associated ( plist_flex_tmp ) .and. delete_grid ) then
                    call delete_listflex ( plist_flex_tmp )
                    !! deallocate flexible for every proteins
                    deallocate ( plist_flex_tmp , stat = ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'Cannot deallocate p_flex in delete array'
                    end if
                end if

            end if

        end do GRID

        ! deallocate set_of_protein
        deallocate ( this % set_of_protein, stat = ierr )
        if ( ierr /= 0 ) then
            write(*,*) 'Cannot deallocate set_of_protein in delete array'
        end if
        this % nb_grids = 0

        ! finally after all grids deleted, delete array of proteins
        deallocate ( this % array, stat = ierr )
        if ( ierr /= 0 ) then
            write ( *, * ) 'Cannot deallocate array protein'
        end if

        nullify ( this % array, this % set_of_protein )

    end subroutine delete_array_protein

    !> Make a copy of the array
    !!
    !! Used with OpenMP, makes pointer easier with older specification\n
    !! Parameter type_calc to force sda_2proteins and solute 2 with local copy\n
    !! Sdamm not used yet, to see later\n
    !! @param this : instance of array_protein_type, source of the copy
    !! @param ptab_dest : pointer to the target of the copy
    !! @param type_calc : instance of \ref mod_type_calc::type_calculation "type_calculation"
    !! @param param_metaldesolv : instance of type com_sda::parameter_metaldesolv
    subroutine copy_array_protein ( this, ptab_dest, type_calc, param_metaldesolv )
        USE mod_type_calc
        USE com_sda
        IMPLICIT NONE
        type ( array_protein_type ) :: this
        type ( array_protein_type ), pointer :: ptab_dest
        type ( type_calculation ), intent ( in ) :: type_calc
        type ( parameter_metaldesolv ), intent ( in ) :: param_metaldesolv

        integer :: i, ierr

        !write ( *,* ) 'Entry copy_array_protein '

        !write(*,*) 'original size ', size ( this % array )
        !write(*,*) 'original nb_grids ', this % nb_grids
        !write(*,*) 'set_of_protein ', this % set_of_protein
        allocate ( ptab_dest,  stat = ierr )
        if ( ierr /= 0 ) then
            write (*,*) 'ERROR In allocating ptab_dest'
            stop 1
        end if

        ! copy set of gird, necessary if OpenMP
        call new_array_protein ( ptab_dest, size ( this % array ), this % nb_grids )
        ptab_dest % set_of_protein = this % set_of_protein


        !! copy all proteins, could be a function in mod_protein copy_protein
        do i = 1, size ( this % array )

            call new_protein ( ptab_dest % array ( i ), this %  array ( i ) %  rotate, this %  array ( i ) % surface, &
                               this %  array ( i ) % iflex, this %  array ( i ) % iimg_chg, this %  array ( i ) % id, &
                               this % array ( i ) %  imetaldesolv )

            !!other association
            ptab_dest % array ( i ) % center_protein = this % array ( i ) % center_protein
            ! should work whatever flex or not if p_sogrid is correclty associated
            ptab_dest % array ( i ) % p_sogrid => this % array ( i ) % p_sogrid

            ! should be already done in init_movable_array
            !ptab_dest % array ( i ) % diff_trans = this %  diff_trans
            !ptab_dest % array ( i ) % diff_rot   = this % diff_rotat

            !  general if not sdamm, protein 1 never moves !
            if ( ( i == 1 ) .and. ( type_calc % enum_type_calc /= sdamm ) ) then

                !! if flex and protein 1 need to create the pflexible
                if ( this % array ( i ) % iflex ) then

                    !! allocate p_flex, assign correct plist_flex and allocate visit conf. initial value by default
                    call set_flexible ( ptab_dest % array ( i ), this % array ( i ) % p_flex % plist_flex )
                    !write (*,*) 'iflex and protein 1: create p_flex but do not make_local_copy '

                end if
                ! never move, false
                call init_movable_array ( ptab_dest % array ( i ), this % array ( i ) % p_sogrid, .false. )

            !if flex, plist_flexible is unique, but need to allocate pflex and initialize correctly
            else if ( this % array ( i ) % iflex ) then

                !! allocate p_flex, assign correct plist_flex and allocate visit conf. initial value by default
                call set_flexible ( ptab_dest % array ( i ), this % array ( i ) % p_flex % plist_flex )

                !! if sda_2proteins prot 2 has no local copy, if OpenMP need a local copy, even if master use the global one
                !! only case where origianl local_copy is false, but OpenMP copy need a local_copy
                if ( ( this % array ( i ) % local_copy ) .or. ( type_calc % enum_type_calc == sda_2proteins ) ) then
                    !write(*,*) 'flex and local_copy init_movable id', i, ' true_rotate ', this % array ( i ) % true_rotate
                    call init_movable_array ( ptab_dest % array ( i ), this % array ( i ) % p_sogrid, .true. )
                !! else flex, use original copy
                !! iflex and ( sda_2proteins type but not sda_2proteins )
                else
                    !write(*,*) 'flex and not local_copy init_movable id', i, ' true_rotate ', this % array ( i ) % true_rotate
                    !! keep default local_copy from the original, not really tested
                    call init_movable_array ( ptab_dest % array ( i ), this % array ( i ) % p_sogrid, &
                                      this % array ( i ) % local_copy )
                end if
            !! not flexible, should group new copy always local_copy
            else
                !write (*,*) 'not iflex '

                !! if protein2 and sda_2protein need local copy for OpenMP
                if ( type_calc % enum_type_calc == sda_2proteins ) then
                    call init_movable_array ( ptab_dest % array ( i ), this % array ( i ) % p_sogrid, .true. )

                !! if protein2 and sda_pmf need local copy for OpenMP
                else if ( type_calc % enum_type_calc == sda_pmf ) then
                    call init_movable_array ( ptab_dest % array ( i ), this % array ( i ) % p_sogrid, .true. )

                ! else copy default
                else !!if ( this % array ( i ) % local_copy ) then
                    !write(*,*) 'not flex and local_copy init_movable id', i, ' true_rotate ', this % array ( i ) % true_rotate
                    call init_movable_array ( ptab_dest % array ( i ), this % array ( i ) % p_sogrid, &
                                              this % array ( i ) % local_copy )
                end if

            end if
            ! if correctly initialized make the copy here
            ! necessary for correct pointer ?? and to make the copy effective
            call copy_movable_array ( ptab_dest % array ( i ) )

            ! moved at end, when fully initilaized
            ! ask allocation of param_desolv if needed only
            if ( this % array ( i ) %  imetaldesolv ) then
                ! need to make manual copy, all size should be saved
                ! or pass, maybe arguement not needed
                !write(*,*) 'prot i=', i, ' call allocate_array_mdesolv '
                call allocate_array_mdesolv ( ptab_dest % array ( i ), param_metaldesolv )
            end if

        end do

    end subroutine copy_array_protein

    !> Function to get the protein type
    !!
    !! Should not be used inside heavy computational code\n
    !! Usefull for tools and initialization\n
    !! If too slow make an array during initilisation, this choice was prefered with \ref array_protein_type::set_of_protein "set_of_protein"
    function get_type ( this, prot_nb ) result ( itype )
        IMPLICIT NONE
        type ( array_protein_type ), intent ( in ) :: this
        integer, intent ( in ) :: prot_nb

        integer :: itype

        itype = 1
        do while ( itype <=  this % nb_grids )
            if ( ( prot_nb >= this % set_of_protein ( itype, 1 ) ) .and. &
                 ( prot_nb <= this % set_of_protein ( itype, 2 ) )  ) return !itype

            itype = itype + 1
        end do
        ! should never be
        write(*,*)

    end function get_type

    !> Print array information
    !!
    !! Set opt false for tools for instance, in sda stop because there is an error in initialization
    !! @param this : instance of array_protein_type, source of the copy
    !! @param opt_check_associated optional, stop if so_grid of proteins are not associated
    subroutine print_info_array_protein ( this, opt_check_associated )
        IMPLICIT NONE
        type ( array_protein_type ) :: this
        logical, optional :: opt_check_associated

        type ( list_flex_type ), pointer :: plist_flex_tmp
        integer :: igrid, iprot

        nullify ( plist_flex_tmp )

        write(*,*)
        write(*,*) '========================================================='
        write(*,*) '  Print info array protein, nb_grids ', this % nb_grids

        do igrid = 1, this % nb_grids

            write (*,*) 'set_of_protein grid ', igrid, 'from solute ', this % set_of_protein ( igrid, 1 ), ' to ', &
                         this % set_of_protein ( igrid, 2 )

            ! just the first one is needed
            do iprot = this % set_of_protein ( igrid, 1 ), this % set_of_protein ( igrid, 2 )
                !write(*,*) 'trans. diff coeff ', this % array ( iprot ) % p_sogrid % diffusion_trans
                !write(*,*) 'rot.   diff coeff ', this % array ( iprot ) % p_sogrid % diffusion_rotat
                !write(*,*) 'stoke_radius ', this % array ( iprot ) % p_sogrid % stoke_radius
                !write(*,*) 'trans. diff prot ', this % array ( iprot ) % diff_trans
                !write(*,*) 'rot.   diff prot ', this % array ( iprot ) % diff_rotat

                ! print info only for one protein, the first one, limit output
                if ( iprot == this % set_of_protein ( igrid, 1 ) ) then

                    ! if flex
                    if ( this % array ( iprot ) % iflex .eqv. .true. ) then
                        write (*,*) 'Solute(s) are flexible'

                        plist_flex_tmp =>  this % array ( iprot ) % p_flex % plist_flex

                        write( *, fmt = 401 ) plist_flex_tmp % method_flex, &
                            plist_flex_tmp % nearest_ngb, &
                            trim (plist_flex_tmp % list_conf_fname), &
                            plist_flex_tmp % frequence, plist_flex_tmp % std_freq_random, plist_flex_tmp % total_conf, &
                            plist_flex_tmp % initial_conf, plist_flex_tmp % max_size_movable_atom, &
                            plist_flex_tmp % max_size_movable_charge

                    ! test correct association of pflex and sogrid to do
                    else
                        write (*,*) 'Solute(s) are NOT flexible'
                    end if

                    ! can make local variable test_associated_sogrid
                    ! error with pgfortran only, cannot in 1 line !
                    if ( present ( opt_check_associated ) ) then
                        if ( opt_check_associated .eqv. .false. ) continue
                    !else
                    !    write(*,*) 'trans. diff coeff ', this % array ( iprot ) % p_sogrid % diffusion_trans
                    !    write(*,*) 'rot.   diff coeff ', this % array ( iprot ) % p_sogrid % diffusion_rotat
                    end if

                    ! local copy or not
                    if ( this % array ( iprot ) % local_copy .eqv. .true. ) then
                        write (*,*) 'Protein(s) of this group have a local copy of movable array for each protein'
                    ! test pflex and sogrid to do
                    else
                        write (*,*) 'Protein(s) of this group points to the same movable array'
                    end if

                    if ( this % array ( iprot ) % rotate ) then
                        write (*,*) 'Protein(s) of this group rotate'
                    else
                        write (*,*) 'Protein(s) of this group do not rotate (tricky for sda_2proteins and derived)'
                    end if

                    ! maybe too much information, maybe in debug mode
                    ! call info_protein( this % array ( iprot ) )
                end if

                ! test for error, default no optional argument make the test
                if ( present ( opt_check_associated ) ) then
                    if ( opt_check_associated .eqv. .false. ) continue
                else
                    if ( associated ( this % array ( iprot ) % p_sogrid ) .eqv. .false. ) then
                        write(*,*) 'ERROR p_sogrid not associated, stop in print_info_array_protein'
                        stop 1
                    end if
                end if

            end do ! first protein of each group
            write(*,*)

        end do
        write(*,*) '========================================================='

 ! format data for list flexible
 401 FORMAT ( 'method:            ',i8,/, &
              'nearest neighbour: ',b8,/,&
              'list_conf_fname:   ',(a),/, &
              'frequency:         ',f8.3,/,&
              'std frequency:     ',f8.3,/,&
              'total_conf:        ',i8,/, &
              'initial_conf:      ',i8,/, &
              'max_atom:          ',i8,/, &
              'max charge:        ',i8 )

    end subroutine print_info_array_protein

END MODULE
