!>  \version{version 7.1 (2015)}
!!
!>  Copyright (c) 2009, 2010, 2015
!>  Heidelberg Institute of Theoretical Studies (HITS, www.h-its.org)
!>        formerly EML Research gGmbH (EML-R )
!>  Schloss-Wolfsbrunnenweg 35
!>  69118 Heidelberg, Germany
!>
!>  Copyright (c) 2000, 2003
!>  European Molecular Biology Laboratory
!>  Meyerhofstr. 1, Postfach 10.2209
!>  D-69012, Heidelberg, Germany 
!>
!>  Please send your contact address to get information on updates and
!>  new features to "mcmsoft@h-its.org". Questions will be
!>  answered as soon as possible.
!>
!>  References (see also http://mcm.h-its.org/sda7/doc/doc_sda7/references.html):
!>
!>  Brownian dynamics simulation of protein-protein diffusional encounter.
!>
!>  (1998) Methods, 14, 329-341.
!>
!> Authors: M.Martinez, N.J.Bruce, J.Romanowska, D.B.Kokh, P.Mereghetti, R.R.Gabdoulline, M. Ozboyaci, S.Richter and R.C.Wade
!>
!>------------------------------------------------------------------------
!<
!> \file mod_protein.f90
!! Define one protein or solute

!> Module to define one protein (position, orientation, is surface?)\n
MODULE mod_protein

    USE mod_flexibility

    !> Define properties of one solute / protein
    !!
    !! Contains :
    !!  - geometric center
    !!  - position and orientation
    !!  - pointer to one \ref mod_setofgrid::sogrid "sogrid", if the solute is not flexible is associated\n
    !!  - pointeur to one \ref mod_flexibility::flexible "flexible", if the solute is flexible ( one flexible by protein)
    !!  - optional an internal copy of the movable_atoms ( effective charges, accessibility atoms and reaction criteria ), when an independent copy from the sogrid is needed ( sdamm, sda_2proteins with OpenMP )
    !!  - new added a copy of type_ljgrid, contains movable atoms and points to the unique array of grid
    type protein

    !> keep trace of the box position, needed for sdamm
    !! \todo check if still needed
    integer, dimension ( 3 ) :: box
    !> real position for sdamm, relative position for sda_2proteins
    real ( kind=8 ), dimension(3) :: position
    !> For langevin integrator, not implementd
    !real ( kind=8  ), dimension(3) ::velocity

    !> orientation of the solute, store vector for the 3 axis
    !! @{
    real ( kind=8 ), dimension(3) :: orientation_x
    real ( kind=8 ), dimension(3) :: orientation_y
    real ( kind=8 ), dimension(3) :: orientation_z
    !> @}
    !> center of geometry from the pdb file
    real ( kind=8 ), dimension(3) :: center_protein

    !> may be used by mod_record, keep it
    integer :: id

    !> local copy of diffusion coefficient. Maybe be modified by Hydrodynamic Interactions
    !! @{
    real ( kind=8 ) :: diff_trans, diff_rotat
    !> @}

    !> pointer to the associated set of grid (\ref mod_setofgrid::sogrid "sogrid")
    type ( sogrid ), pointer :: p_sogrid
    !> pointer to a \ref mod_flexibility::flexible "flexible", only one of these two pointers should be initalized
    type ( flexible ), pointer :: p_flex

    !! iimg_chg ( and other ) in protein, but could be in sogrid, less copy but easier to understand ??
    !> General variable of the properties of the solute
    !! @{
    logical :: surface, iflex, rotate, iimg_chg
    !> @}

    !> If a local copy of the movable atoms is needed
    logical :: local_copy

    !> In case of local copy, size of the arrays
    !! @{
    integer :: nb_charge, nb_atom_surface, nb_rcriteria
    !> @}

    !> Parameter specific to metaldesolvation, grid owned by each protein
    !! @{
    logical :: imetaldesolv
    real ( kind=8 ), dimension(:,:), pointer :: array_mdesolv
    integer :: size_metal
    !> @}


    ! new version with pointer, array allocated if multi-protein rotates / or if need to initial position ( flex )
    ! otherwise points to sogrid
    ! This data need to be rotated, change to pointer because : \n
    ! can point to sogrid array if protein does not rotate ( sda_2protein prot1 )\n
    ! ( flexible or not ) ) \n
    ! need to create a local copy of the array if sdamm with more than one\n
    ! protein by sogrid, sda_2protein with prot2 flex \n
    ! need to keep initial position in sogrid if flex prot2 or sdamm flex,\n
    ! otherwise can delete array in sogrid ( sdamm not flexible )\n
    !
    ! Design, allocate first to sogrid when reading Solute_Grid,
    ! Copy local copy only if needed, may delete origial copy ( sdamm not flex )
    ! pointer/ creation of copy / delete of sogrid array just after

    !> pointers to the movable arrays
    !! if there is no local copy, they point to sogrid arrays\n
    !! if there is a local copy needed, they allocate their own arrays\n
    !! @{
    real ( kind = 8 ), dimension(:,:), pointer :: p_pos_charge
    real ( kind = 8 ), dimension(:,:), pointer :: p_list_atom_surface
    real ( kind = 4 ), dimension(:,:), pointer :: p_pos_rcriteria
    !> @}

    !> one-dimensional array listing the atoms which are on the surface (order as in the orig.PDB file)
    integer, dimension(:), pointer :: p_which_atom_surface
    
    !> pointer to movable arrays for lennard-jones, contains all grids and position
    type ( type_ljgrid ), pointer :: p_pos_lenjones

    end type protein

    contains

        !> Allocate one protein
        !! @param this : instance of protein
        !! @param irot : if the solute rotates
        !! @param isurf : if the solute is a surface ( no rotation, no translation, different PBC )
        !! @param iflex : if the solute is flexible
        !! @param iimg_chg : if solute should compute image_charge forces and energies
        !! @param id : identifier, depreceated
        !! @param opt_imdesolv : optional, global object mdesolv
        subroutine new_protein ( this, irot, isurf, iflex, iimg_chg, id, opt_imdesolv  )
            IMPLICIT NONE
            type ( protein ) :: this
            logical, intent ( in ) :: irot, isurf, iflex, iimg_chg
            integer, intent ( in ) :: id
            logical, optional :: opt_imdesolv

            real ( kind=8 ), dimension( 3 ) :: null_vec=(/0., 0., 0./)

            ! nullify pointers
            nullify ( this % p_sogrid, this % p_flex )
            nullify ( this % p_pos_charge )
            nullify ( this % p_list_atom_surface )
            nullify ( this % p_pos_rcriteria )
            nullify ( this % p_pos_lenjones )
            nullify ( this % p_which_atom_surface )

            this % position = null_vec
            ! initialize idendity
            this % orientation_x = (/1., 0., 0./)
            this % orientation_y = (/0., 1., 0./)
            this % orientation_z = (/0., 0., 1./)
            this % box = 0
            this % center_protein = null_vec
            this % id = id
            this % rotate = irot
            this % surface = isurf
            this % iflex = iflex
            this % iimg_chg = iimg_chg
            this % diff_trans = 0.d0
            this % diff_rotat = 0.d0
            !! for movable array
            this % local_copy = .false.
            !this % true_rotate = .false.
            this % nb_charge = 0
            this % nb_atom_surface = 0
            this % nb_rcriteria = 0
            ! default is false
            this % imetaldesolv = .false.
            nullify( this % array_mdesolv )
            this % size_metal = 0
!            this % size_y_metal = 0
            this % imetaldesolv = .false.

            !write(*,*) 'new prot id', id
            if ( present ( opt_imdesolv ) ) then
                if ( opt_imdesolv .eqv. .true. ) then

                    ! apply only in case it is not a surface
                    !if ( this % surface .eqv. .true. ) then
                    !    write(*,*) 'imetaldesolv will not apply to this protein, it is a surface'
                    !else
                    if ( this % surface .eqv. .false. ) then
                        !write(*,*) 'apply imetaldesolv '
                        this % imetaldesolv = .true.
                        !! need to allocate array at a point later, because not associated to grid yet
                    end if
                end if
            endif

!            write(*,*) 'after new_protein id', id, ' imetal= ', this % imetaldesolv

        end subroutine new_protein

        !> Allocate a \ref mod_flexibility::flexible "flexible" and assign the plist_flexible
        !!
        !! @param this : instance of protein
        !! @param plist_flexible : pointer to a \ref mod_flexibility::list_flex_type "list_flex_type"
        subroutine set_flexible ( this, plist_flexible )
            IMPLICIT NONE
            type ( protein ) :: this
            type ( list_flex_type ) :: plist_flexible

            integer :: ierr

            !! create object flexible and initialize
            allocate ( this % p_flex, stat = ierr )
            if ( ierr /= 0 ) then
                write ( *, * ) 'ERROR In allocating p_flex'
            end if

            ! here use inital_conf, first node, no data
            ! init conf could be done later, in MainLoop
            !write(*,*) 'before new_flexible ', plist_flexible % nearest_ngb
            call new_flexible ( this % p_flex, plist_flexible )

        end subroutine set_flexible

        !!! maybe param is needed, sure distance max is enough
        subroutine allocate_array_mdesolv ( this, param_metaldesolv )
            USE com_sda
            IMPLICIT NONE
            type ( protein ) :: this
            type ( parameter_metaldesolv ) :: param_metaldesolv
            real ( kind=8 )  :: RSell

            integer :: ierr

            !write(*,*) 'Entry allocate_array_mdesolv id ', this % id

            if ( this % imetaldesolv .eqv. .true. ) then
                !! allocate the array, size will depend of p_sogrid
                !write(*,*) 'associated p_sogrid ', associated ( this % p_sogrid )
                !write(*,*) 'maximum distance', this % p_sogrid % distance_max

                !! we devide contact (unit)area for each molecule to 9 parts (3x3) for more accurate computations
                RSell = (param_metaldesolv % Radw)/3.0! RSell=Radw/3.0; 
                ! maxx_S=int(2*dismx2/RSell)+6
                ! and we select anarray a bit larger than the protein size to treat correctly boundary
                this % size_metal = int(2*(this % p_sogrid % distance_max)/RSell)+6 ! maxx_S=int(2*dismx2/RSell)+6
                allocate ( this % array_mdesolv ( this % size_metal, this % size_metal ), stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In allocating array_mdesolv'
                    stop 1
                end if
                this % array_mdesolv = 0.d0
            end if

        end subroutine

        !> Destructor, normally not called directly but by mod_array_protein
        !! add new variable local_copy, simpler and allow easier for flex
        subroutine delete_protein ( this )
            IMPLICIT NONE
            type ( protein ) :: this

            integer :: ierr
            !default error
            ierr = 1

            !write(*,*) 'entry delete protein'

            if ( associated ( this % p_pos_charge ) ) then
                !! check than they are different => is a local copy
                !if (  associated ( this % p_pos_charge, this % p_sogrid % position_charge ) .eqv. .false. ) then
                if ( this % local_copy ) then
                    !write(*,*) 'delete local copy of position charge'
                    deallocate ( this % p_pos_charge, stat= ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'ERROR In deallocation of p_pos_charge'
                        stop 1
                    end if
                end if
                nullify ( this % p_pos_charge )
                this % nb_charge = 0
            end if

            if ( associated ( this % p_list_atom_surface ) ) then
                !! check than they are different => is a local copy
                !if (  associated ( this % p_list_atom_surface, this % p_sogrid % list_atom_surface ) .eqv. .false. ) then
                if ( this % local_copy ) then
                    !write(*,*) 'delete local copy of position atom surface'
                    deallocate ( this % p_list_atom_surface, stat= ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'ERROR In deallocation of p_list_atom'
                        STOP 1
                    end if
                end if
                nullify ( this % p_list_atom_surface )
                this % nb_atom_surface = 0
            end if

            if ( associated( this % p_which_atom_surface ) ) then
                if ( this % local_copy ) then
                    deallocate( this % p_which_atom_surface, stat = ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'ERROR In deallocating which_atom_surface'
                        stop 1
                    end if
                end if
                nullify( this % p_which_atom_surface )
            end if

            if ( associated ( this % p_pos_rcriteria ) ) then
                !! check than they are different => is a local copy
                !if ( associated ( this % p_pos_rcriteria, this % p_sogrid % reaction_criteria_pos ) .eqv. .false. ) then
                if ( this % local_copy ) then
                    !write(*,*) 'delete local copy of position rcriteria'
                    deallocate ( this % p_pos_rcriteria, stat = ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'ERROR In deallocation of p_pos_rcriteria'
                        STOP 1
                    end if
                end if
                nullify ( this % p_pos_rcriteria )
                this % nb_rcriteria = 0
            end if

            ! more tricky here, pointer of type_lenjones
            if ( associated ( this % p_pos_lenjones ) ) then
                !! check than they are different => is a local copy
                !if ( associated ( this % p_pos_rcriteria, this % p_sogrid % reaction_criteria_pos ) .eqv. .false. ) then
                if ( this % local_copy ) then
                    !write(*,*) 'delete local copy of position lennard jones of a local copy '

                    !! here copy_omp is true, real copy of the atoms
                    call deallocate_ljgrid ( this % p_pos_lenjones, .true. )
                    if ( ierr /= 0 ) then
                        write(*,*) 'Deallocation of p_pos_lenjones in mod_protein to write correctly'
                        stop 1
                    end if
                    ! it is a pointer, need to deallocate it
                    deallocate ( this % p_pos_lenjones, stat = ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'ERROR In deallocating p_pos_lenjones '
                        stop 1
                    end if
                    nullify ( this % p_pos_lenjones )

                ! just for testing
                !else
                !    write(*,*) 'not a local_copy, can just nullify the pointer'
                end if
                ! need a loop anyway
                !write(*,*) 'Deallocation of p_pos_lenjones in mod_protein to write correctly'
                nullify ( this % p_pos_lenjones )
            end if

            !!should be ok, if flex or not, before delete only local_copy
            if ( associated ( this % p_flex ) ) then
                !write(*,*) 'delete flexible '
                call delete_flexible ( this % p_flex )
                deallocate ( this % p_flex, stat = ierr )
                if ( ierr /= 0 ) then
                    write(*,*) 'ERROR In deallocation of p_flex'
                    STOP 1
                end if
            end if

            if ( this % imetaldesolv .eqv. .true. ) then
                if ( associated ( this % array_mdesolv ) ) then
                    !write(*,*) 'deallocate array_mdesolv'
                    deallocate ( this % array_mdesolv, stat = ierr )
                    if ( ierr /= 0 ) then
                        write(*,*) 'ERROR In deallocation of array_mdesolv'
                        STOP 1
                    end if
                !else
                !    write(*,*) 'imetaldesolv is true, but no array associated !, i am certainly the surface'
                    !stop 1
                endif
            endif

        end subroutine delete_protein

        !> rotate orientation axis version with explicit terms, faster
        subroutine update_rot_orientation( this, mrot )
            IMPLICIT NONE
            type ( protein ) :: this
            real ( kind=8 ), dimension(3,3), intent ( in ) :: mrot

            real ( kind=8 ) :: xold, yold, zold

            xold = this % orientation_x(1)
            yold = this % orientation_x(2)
            zold = this % orientation_x(3)
            this % orientation_x(1) = mrot(1,1) * xold + mrot(1,2) * yold + mrot(1,3) * zold
            this % orientation_x(2) = mrot(2,1) * xold + mrot(2,2) * yold + mrot(2,3) * zold
            this % orientation_x(3) = mrot(3,1) * xold + mrot(3,2) * yold + mrot(3,3) * zold

            xold = this % orientation_y(1)
            yold = this % orientation_y(2)
            zold = this % orientation_y(3)
            this % orientation_y(1) = mrot(1,1) * xold + mrot(1,2) * yold + mrot(1,3) * zold
            this % orientation_y(2) = mrot(2,1) * xold + mrot(2,2) * yold + mrot(2,3) * zold
            this % orientation_y(3) = mrot(3,1) * xold + mrot(3,2) * yold + mrot(3,3) * zold

            xold = this % orientation_z(1)
            yold = this % orientation_z(2)
            zold = this % orientation_z(3)
            this % orientation_z(1) = mrot(1,1) * xold + mrot(1,2) * yold + mrot(1,3) * zold
            this % orientation_z(2) = mrot(2,1) * xold + mrot(2,2) * yold + mrot(2,3) * zold
            this % orientation_z(3) = mrot(3,1) * xold + mrot(3,2) * yold + mrot(3,3) * zold

        end subroutine update_rot_orientation

        !> @brief Rotate the position of atoms/ charges / etc.
        !! version with explicit rotation, faster
        !! now use the pointer, either points to a sogrid copy, or rotate the local copy
        subroutine update_rotation_protein ( this, matrix_rot )
            IMPLICIT NONE
            type ( protein ) :: this
            real ( kind=8 ), dimension(3,3), intent( in ) :: matrix_rot

            real ( kind=8 ) :: xold, yold, zold
            integer :: ia, j

            !write(*,*) 'update rotation_protein '
            !write(*,*) 'matrix_rot1 ',matrix_rot(:,1)

            ! update charge_pos, faster to test nb_charge ??
            if ( associated (  this % p_pos_charge ) ) then
                !write(*,*) 'charge before rotation '
                !write(*,*) this % p_pos_charge (:, 1),  this % p_pos_charge (:,this % nb_charge)
                do ia = 1, this % nb_charge
                xold = this % p_pos_charge (1, ia )
                yold = this % p_pos_charge (2, ia )
                zold = this % p_pos_charge (3, ia )
                this % p_pos_charge (1, ia) = matrix_rot(1,1)*xold + &
                                              matrix_rot(1,2)*yold + &
                                              matrix_rot(1,3)*zold
                this % p_pos_charge (2, ia) = matrix_rot(2,1)*xold + &
                                              matrix_rot(2,2)*yold + &
                                              matrix_rot(2,3)*zold
                this % p_pos_charge (3, ia) = matrix_rot(3,1)*xold + &
                                              matrix_rot(3,2)*yold + &
                                              matrix_rot(3,3)*zold
                end do
            endif

            if ( associated ( this % p_list_atom_surface ) ) then
                do ia = 1, this % nb_atom_surface
                xold = this % p_list_atom_surface (1, ia)
                yold = this % p_list_atom_surface (2, ia)
                zold = this % p_list_atom_surface (3, ia)
                this % p_list_atom_surface (1, ia) = matrix_rot(1,1)*xold + &
                                                     matrix_rot(1,2)*yold + &
                                                     matrix_rot(1,3)*zold
                this % p_list_atom_surface (2, ia) = matrix_rot(2,1)*xold + &
                                                     matrix_rot(2,2)*yold + &
                                                     matrix_rot(2,3)*zold
                this % p_list_atom_surface (3, ia) = matrix_rot(3,1)*xold + &
                                                     matrix_rot(3,2)*yold + &
                                                     matrix_rot(3,3)*zold
               end do
            endif

            if ( associated ( this % p_pos_rcriteria ) ) then
               do ia = 1, this % nb_rcriteria
               xold = this % p_pos_rcriteria (1, ia)
               yold = this % p_pos_rcriteria (2, ia)
               zold = this % p_pos_rcriteria (3, ia)
               this % p_pos_rcriteria (1, ia) = real ( matrix_rot(1,1)*xold + &
                                                       matrix_rot(1,2)*yold + &
                                                       matrix_rot(1,3)*zold, 4 )
               this % p_pos_rcriteria (2, ia) = real ( matrix_rot(2,1)*xold + &
                                                       matrix_rot(2,2)*yold + &
                                                       matrix_rot(2,3)*zold, 4 )
               this % p_pos_rcriteria (3, ia) = real ( matrix_rot(3,1)*xold + &
                                                       matrix_rot(3,2)*yold + &
                                                       matrix_rot(3,3)*zold, 4 )
               end do
            endif

            ! here loop over the atom position,
            if ( associated ( this % p_pos_lenjones ) ) then

              ! make test on first, should be better, ok solve bug valgrind
              !if ( associated ( this % p_pos_lenjones % array_position_atom_lj ( 1 ) %  p_position_atom ) ) then
              if ( allocated ( this % p_pos_lenjones % array_position_atom_lj ) ) then
                write(*,*) 'protein id ', this % id
                write(*,*) 'update rotation protein size of the array of lennard-jones ', &
                    size ( this % p_pos_lenjones % array_position_atom_lj )

                !! make a loop, only if array_position_atom_lj associated, for the moment protein2
                do j = 1,  this % p_pos_lenjones % nb_ljgrid
                    ! cetainly an other test here, yes needed
                    !write(*,*) 'size of this array position ', &
                    !    this % p_pos_lenjones % array_position_atom_lj( j ) % nb_pos_atom

                    do ia = 1, this % p_pos_lenjones % array_position_atom_lj( j ) % nb_pos_atom
                    xold = this % p_pos_lenjones % array_position_atom_lj ( j ) % p_position_atom (1, ia)
                    yold = this % p_pos_lenjones % array_position_atom_lj ( j ) % p_position_atom (2, ia)
                    zold = this % p_pos_lenjones % array_position_atom_lj ( j ) % p_position_atom (3, ia)

                    this % p_pos_lenjones % array_position_atom_lj ( j ) % p_position_atom (1, ia) = &
                                                          matrix_rot(1,1)*xold + &
                                                           matrix_rot(1,2)*yold + &
                                                           matrix_rot(1,3)*zold
                    this % p_pos_lenjones % array_position_atom_lj ( j ) % p_position_atom (2, ia) = &
                                                           matrix_rot(2,1)*xold + &
                                                           matrix_rot(2,2)*yold + &
                                                           matrix_rot(2,3)*zold
                    this % p_pos_lenjones % array_position_atom_lj ( j ) % p_position_atom (3, ia) = &
                                                           matrix_rot(3,1)*xold + &
                                                           matrix_rot(3,2)*yold + &
                                                           matrix_rot(3,3)*zold
                    end do
                end do
              end if
            end if


       end subroutine update_rotation_protein

       !> Print info
       subroutine info_protein( this )
            IMPLICIT NONE
            type ( protein ), intent(in) :: this
            integer :: i

            write(*,*) 'protein id', this % id
            if ( associated(this % p_sogrid) ) then
                write(*,*) ' is associated to a set_ofgrid '
                call set_ofgrid_info ( this % p_sogrid )
                write(*,*) 'number of surface atom ', this % p_sogrid % nb_atom_surface
                write(*,*) 'number of charge ', this % p_sogrid % nb_charge
                write(*,*) 'center: ', ( this % center_protein( i ),i=1,3 )
            else
                 write(*,*) ' is NOT associated to a set_ofgrid '
                 write(*,*)
            endif
       end subroutine info_protein

!!!! new added for version 2, here make a copy of "movable" array for each protein
!!!! reaction criteria read pos_atom1 and pos_atom2 opt_nb_prot indicates which pos_atom to associate
!!!! if sdamm and identical proteins need both pos_atom1 and pos_atom2

!! Cannot allocate each time, need an init with max size ( to see later for dynamical allocation,
!! certainly need to extend on the fly, like C++ vector
!! and a copy_movable array which overwrites the data

!! here may want a max size given by list_flex, use size of sogrid for normal ( not flex ) runs
!! no need max size here, better if setup once ?? for every protein in listflex
!! only for flex, for sdamm also

        !> Make the copy of the atomic position to protein if make_local_copy is true
        !! Otherwise points to sogrid
        !! why sgrid ? already associated ? Not if flex to check

        !! sgrid could be deleted, if prot already points to the correct so_grid, seems to be the case
        !! if flex p_sogrid should also point to the correct, only available for force, energy....

        !! could test for flex here, if flex look at p_flexible and max_size, to see later with tool
        !! will simplify a lot the call !! for tool and sda_koff !!

        !! sgrid could be deleted if already associated, ?? sogrid contains the grids, need to keep ??
        subroutine init_movable_array ( this, sgrid, make_local_copy )
            IMPLICIT NONE
            type ( protein ) :: this
            type ( sogrid ), pointer :: sgrid
            logical :: make_local_copy

            integer :: max_size_charge, max_size_surface, max_size_rcriteria
            integer :: ierr

            !write(*,*) 'Entry init_movable_array local_copy ', make_local_copy !, ' true_rotation ', true_rotation
            !write(*,*) 'associated ( protein % p_sogrid ) ', associated ( this % p_sogrid )

            !! initialize local diffusion coefficient
            !write(*,*) 'update diffusion from sgrid prot id=', this % id
            this % diff_trans = sgrid %  diffusion_trans
            this % diff_rotat = sgrid % diffusion_rotat

            if ( associated ( sgrid ) .eqv. .false. ) then
                write (*,*) 'ERROR sgrid not associated in init_movable_array'
                stop 1
            end if

            !! move test iflex here, much simpler for calling function
            !! not best for tools..
            if ( this % iflex ) then
                !write(*,*) 'protein is flexible, need max size of array '
                max_size_charge = this % p_flex % plist_flex % max_size_movable_charge
                max_size_surface = this % p_flex % plist_flex % max_size_movable_atom
                !write(*,*) 'max charge, max atom surf', max_size_charge, max_size_surface
            else
                ! default value the associated grid, if no opt data
                ! just here need to check if flex !! and appropriate size
                max_size_charge = sgrid % nb_charge
                max_size_surface = sgrid % nb_atom_surface
                !write(*,*) 'protein not flexible, size of array ', max_size_charge, max_size_surface
            end if

            !! here pointer, need to test, should be fixed for all grids ??, no max
            if ( allocated ( sgrid % reaction_criteria_pos ) ) then
                max_size_rcriteria = sgrid % nb_reaction_criteria
            else
                max_size_rcriteria = 0
            end if


            !! aplly to all movable array, if they are defined
            if ( make_local_copy .eqv. .true. ) this % local_copy = .true.

            if ( allocated ( sgrid % position_charge ) ) then
                !write(*,*) 'set protein nb_charge ', sgrid % nb_charge
                this % nb_charge = sgrid % nb_charge

                if ( make_local_copy ) then
                    allocate ( this % p_pos_charge ( 3, max_size_charge ), stat = ierr )
                    if ( ierr /= 0 ) then
                        write (*,*) 'ERROR Couldn t allocate memory for copying movable array charge', &
                                     max_size_charge
                        stop 1
                    end if
                    ! init to zero
                    this % p_pos_charge = 0.
                !! only points to sogrid, no need to copy data
                else
                    this % p_pos_charge => sgrid % position_charge
                endif

            end if

            !! atom surface always, maybe not ?? analytic or else ??
            this % nb_atom_surface = sgrid % nb_atom_surface

            if ( make_local_copy ) then
                allocate ( this % p_list_atom_surface ( 3, max_size_surface ), stat = ierr )
                if ( ierr /= 0 ) then
                    write (*,*) 'ERROR Couldn t allocate memory for copying movable array ', &
                                 max_size_surface  !this % nb_atom_surface
                    stop 1
                endif
                this % p_list_atom_surface = 0.
                ! miss an association here ? this % p_which_atom_surface =>
            else
                this % p_list_atom_surface => sgrid % list_atom_surface
            end if
            if ( sgrid % alloc_which_at_surf ) then
                this % p_which_atom_surface => sgrid % which_atom_surface
            end if

            ! new added lenjones
            if ( associated ( sgrid % p_ljgrid  ) ) then
                !write(*,*) 'sgrid % p_ljgrid is associated '
                if ( make_local_copy ) then
                    ! copy the type type_ljgrid, array of grids points to the original ( no copy )
                    ! atom_position are fully copied.
                    call copy_ljgrid ( sgrid % p_ljgrid, this % p_pos_lenjones )
                    !write(*,*) 'after the copy'
                else
                    !write(*,*) 'Now local copy points to sogrid % p_ljgrid'
                    this % p_pos_lenjones => sgrid % p_ljgrid
                end if

            !else
                !write(*,*) 'sgrid % p_ljgrid associated '
            end if

            !! local copy use if initial conf is needed
            if ( allocated ( sgrid % reaction_criteria_pos ) ) then
                !write(*,*) 'reaction criteria associated to sogrid'
                !! both should be valid
                !this % nb_rcriteria = sgrid % p_reaction_criteria % ntotal
                this % nb_rcriteria = max_size_rcriteria

                if ( make_local_copy ) then
                    allocate ( this % p_pos_rcriteria ( 3, this % nb_rcriteria ), stat = ierr )
                    if ( ierr /= 0 ) then
                        write (*,*) 'ERROR Couldn t allocate memory for copying movable reaction_criteria ', &
                                     this % nb_rcriteria
                        stop 1
                    end if
                    this % p_pos_rcriteria = 0.
                else
                    ! nb1 or nb2, defined during prepare grid and protein
                    this % p_pos_rcriteria => sgrid % reaction_criteria_pos
                end if
            !else
            !    write(*,*) 'No reaction criteria associated'
            end if

            !write(*,*) 'init_movable nb charge, nb atom surf, nb_rcriteria', &
            !            this % nb_charge, this % nb_atom_surface, this % nb_rcriteria

        end subroutine init_movable_array

        !> @brief Copy movable atoms from sogrid arrays to a local copy in the protein.
        !! Needed in case of flexibility (keep a copy of the initial position), certianly sda_pmf (with option opt_mat)\\
        !! Check if local copy is true, otherwise do nothing\n
        !! Optional opt_mat will load the initial position locally and apply the matrix of rotation.\n
        !!
        !! \pre To be called after init_movable_array
        subroutine copy_movable_array ( this, opt_mat )
            IMPLICIT NONE
            type ( protein ) :: this
            real(kind=8), dimension(3,3), optional :: opt_mat

            real ( kind=8 ), dimension(3,3) :: matrix_rot
            logical :: set_matrix

#ifdef DEBUG
            write(*,*) 'Entry copy_movable_array local_copy ', this % local_copy
#endif
            ! default value
            set_matrix = .false.
            if ( present ( opt_mat ) ) then
                set_matrix = .true.
            endif

            if (set_matrix .eqv. .true.) then
                if ( this % local_copy .eqv. .false. ) then
                    write(*,*) 'ERROR option opt_matrix with local_copy false is an error. Initial configuration is lost.'
                    stop 1
                endif
            endif

            ! just copy the data, here works because same array but size could differ
            ! necessary to test ? normally not...
            if ( allocated ( this  % p_sogrid % position_charge ) ) then

                ! update size, whatever local copy or not
                this % nb_charge = this % p_sogrid % nb_charge
                if ( this % local_copy ) then
                    !write(*,*) 'Copy position_charge '
                    !write(*,*) 'copy pos_charge prot', this % nb_charge, size ( this % p_sogrid % position_charge )
                    this % p_pos_charge =  this % p_sogrid % position_charge
                else
                    this % p_pos_charge => this % p_sogrid % position_charge
                end if
            end if

            if ( allocated ( this  % p_sogrid % list_atom_surface ) ) then
                this % nb_atom_surface = this % p_sogrid % nb_atom_surface

                if ( this % local_copy ) then
                    this % p_list_atom_surface = this % p_sogrid % list_atom_surface
                !! just points to the correct array
                else
                    this % p_list_atom_surface => this % p_sogrid % list_atom_surface
                end if
            end if

            ! should work like this, correct association 1 / 2 and center of mass shift done in read_input_file
            ! because no more reference to atom_pos1/2, already associated during read_input
            if ( allocated ( this % p_sogrid % reaction_criteria_pos ) ) then
               ! here do not consider change in size, easy to add...
               if ( this % local_copy ) then
                   !write(*,*) 'Copy reaction criteria'
                   this % p_pos_rcriteria = this % p_sogrid % reaction_criteria_pos
               else
                      this % p_pos_rcriteria => this % p_sogrid % reaction_criteria_pos
               end if
            end if

            if ( this % local_copy .eqv. .false. ) then
                !write (*,*) 'protein local_copy is false, no need to explicit rotation return '
                return
            end if

            ! original method, apply the rotation corresponding to the last position of the solute, identity at start up
            ! case flexibility
            if ( set_matrix .eqv. .false. ) then
                ! write(*,*) 'set matrix false'
                ! Need to rotate the atom to the correct position
                ! matrix of rotation, easy given by the new orientation
                ! because orientation in so_grid ( global_copy read in input ) is unity !!!
                matrix_rot (:,1) = this %  orientation_x
                matrix_rot (:,2) = this %  orientation_y
                matrix_rot (:,3) = this %  orientation_z

            ! give a new rotation matrix. case sda_pmf.
            else
            	!write(*,*) 'set matrix is true'
                this %  orientation_x = opt_mat(:,1)
                this %  orientation_y = opt_mat(:,2)
                this %  orientation_z = opt_mat(:,3)
                matrix_rot (:,1) = opt_mat(:,1)
                matrix_rot (:,2) = opt_mat(:,2)
                matrix_rot (:,3) = opt_mat(:,3)
            endif

            ! apply the rotation to the movable arrays
            call update_rotation_protein ( this, matrix_rot )

            ! atom have been copied to their initial position
            !else
!#ifdef DEBUG
!#endif
!                this % orientation_x = (/1., 0., 0./)
!                this % orientation_y = (/0., 1., 0./)
!                this % orientation_z = (/0., 0., 1./)
!            endif


        end subroutine copy_movable_array

        !> First version, fastest version, do not check for asymetry.\n
        !!
        !! Can be called by sdamm, because an additional check is done in compute_force/energy
        !!
        !! Assign pointer to the correct grid / charge.\n
        !! @param grid_number : grid to select
        !! @param prot1, prot2 : protein
        !! @param nq1, nq2 : number of "charges"
        !! @param p_charge1, p_charge2 : pointer to the "charge" value
        !! @param p_charge_position1, p_charge_position2 : pointer to the "charge" positions
        !! @param dummy_array : sned the dummy array adress, used by hydrophobic and lennard-jones
        subroutine select_charge ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                    p_charge_position1, p_charge_position2, dummy_array )
            IMPLICIT NONE
            type ( protein ), intent ( in ), target :: prot1, prot2
            real ( kind=8 ), dimension(:,:), pointer, intent(out) :: p_charge_position1, p_charge_position2
            real ( kind = 8 ), dimension ( : ), pointer, intent(out) :: p_charge1, p_charge2
            real ( kind = 8 ), dimension ( : ), target :: dummy_array
            integer, intent(out) :: nq1, nq2
            integer, intent(in) :: grid_number

            ! if moved here, can be only intent(out), no care to take from the caller
            nullify ( p_charge_position1, p_charge_position2 )
            nullify ( p_charge1, p_charge2 )
            nq1 = 0
            nq2 = 0

            select case ( grid_number )
        ! electrostatic
        case ( 1 )
            nq1 = prot1 % nb_charge
            nq2 = prot2 % nb_charge
            p_charge2 => prot2 % p_sogrid % charge_grid
            p_charge1 => prot1 % p_sogrid % charge_grid
            p_charge_position2 =>  prot2 % p_pos_charge
            p_charge_position1 =>  prot1 % p_pos_charge
        ! desolvation charge square
        case ( 2 )
            nq1 = prot1 % nb_charge
            nq2 = prot2 % nb_charge
            p_charge2 => prot2 % p_sogrid % charge_square_grid
            p_charge1 => prot1 % p_sogrid % charge_square_grid
            p_charge_position2 =>  prot2 % p_pos_charge
            p_charge_position1 =>  prot1 % p_pos_charge
        ! hydrophobic
        ! accessibility list_atom_surface and accessibility are the "charges"
        case ( 3 )
            nq1 = prot1 % nb_atom_surface
            nq2 = prot2 % nb_atom_surface
            p_charge2 => prot2 % p_sogrid % accessibility
            p_charge1 => prot1 % p_sogrid % accessibility
            p_charge_position2 =>  prot2 % p_list_atom_surface
            p_charge_position1 =>  prot1 % p_list_atom_surface
        !! can add more case
        !! lj repulsive
        case ( 4 )
            nq1 = prot1 % nb_atom_surface
            nq2 = prot2 % nb_atom_surface
            p_charge2 => dummy_array
            p_charge1 => dummy_array
            p_charge_position2 =>  prot2 % p_list_atom_surface
            p_charge_position1 =>  prot1 % p_list_atom_surface

        ! case lennard-jones
        case ( 5 )
            !! need additional subgrid and need  asymetry
            write(*,*) 'Select_charge should not be called  with lennard-jones, call select_charge2 '

        end select

        end subroutine select_charge

        !> Select correct pointer for charges ( and optionaly subgrids : only lennard-jones now ).\n
        !!
        !! Include lennard-jones grids\n
        !! Check if associated, allow asymetric computation\n
        !! Could replace select_charge, need benchmark.
        !!
        !! @param grid_number : grid to select
        !! @param prot1, prot2 : protein
        !! @param nq1, nq2 : number of "charges"
        !! @param p_charge1, p_charge2 : pointer to the "charge" value
        !! @param p_charge_position1, p_charge_position2 : pointer to the "charge" positions
        !! @param dummy_array : sned the dummy array adress, used by hydrophobic and lennard-jones
        !! @param opt_pgrid1, opt_pgrid2 : optional, pointer to the grid, can be lennard-jones
        !! @param opt_subgrid : optional, the subgrid number ( only lj now )
        subroutine select_charge2 ( grid_number, prot1, prot2, nq1, nq2, p_charge1, p_charge2, &
                    p_charge_position1, p_charge_position2, dummy_array, &
                    opt_pgrid1, opt_pgrid2, opt_subgrid )
            IMPLICIT NONE
            type ( protein ), intent ( in ), target :: prot1, prot2
            type ( type_grid ), optional, pointer, intent(out) :: opt_pgrid1, opt_pgrid2
            real ( kind=8 ), dimension(:,:), pointer, intent(out) :: p_charge_position1, p_charge_position2
            real ( kind = 8 ), dimension ( : ), pointer,intent(out) :: p_charge1, p_charge2
            real ( kind = 8 ), dimension ( : ), target :: dummy_array
            integer,intent(out) :: nq1, nq2
            integer, intent (in) :: grid_number
            integer, intent(in), optional :: opt_subgrid

            !write(*,*) 'Entry select_charge2 grid ', grid_number

            ! if moved here, can be only intent(out), no care to take from the caller
            nullify ( opt_pgrid1, opt_pgrid1 )
            nullify ( p_charge_position1, p_charge_position2 )
            nullify ( p_charge1, p_charge2 )
            nq1 = 0
            nq2 = 0

            select case ( grid_number )
        ! electrostatic
        case ( 1 )
            nq1 = prot1 % nb_charge
            nq2 = prot2 % nb_charge
            p_charge2 => prot2 % p_sogrid % charge_grid
            p_charge1 => prot1 % p_sogrid % charge_grid
            p_charge_position2 =>  prot2 % p_pos_charge
            p_charge_position1 =>  prot1 % p_pos_charge

            ! here opt_pgrid called for sure
            if ( is_associated ( prot1 % p_sogrid % pset_grid ( 1 ) ) ) then
                opt_pgrid1 => prot1 % p_sogrid % pset_grid( 1 )
            !! not necessary if sure it is not used
            else
                nullify ( opt_pgrid1 )
            endif

            if ( is_associated ( prot2 % p_sogrid % pset_grid ( 1 ) ) ) then
                opt_pgrid2 => prot2 % p_sogrid % pset_grid( 1 )
            else
                nullify ( opt_pgrid2 )
            endif

        ! desolvation charge square
        case ( 2 )
            nq1 = prot1 % nb_charge
            nq2 = prot2 % nb_charge
            p_charge2 => prot2 % p_sogrid % charge_square_grid
            p_charge1 => prot1 % p_sogrid % charge_square_grid
            p_charge_position2 =>  prot2 % p_pos_charge
            p_charge_position1 =>  prot1 % p_pos_charge
            !! add for optional asymetric

            ! could test or just associate ??

            if ( present ( opt_pgrid1 ) ) then
            if ( is_associated ( prot1 % p_sogrid % pset_grid ( 2 ) ) ) then
                !write(*,*) ' prot1 desolv is associated '
                opt_pgrid1 => prot1 % p_sogrid % pset_grid( 2 )
            else
                nullify ( opt_pgrid1 )
            end if
            if ( is_associated ( prot2 % p_sogrid % pset_grid ( 2 ) ) ) then
                !write(*,*) ' prot2 desolv is associated '
                opt_pgrid2 => prot2 % p_sogrid % pset_grid( 2 )
            else
                nullify ( opt_pgrid2 )
            end if
            end if ! opt present


        ! hydrophobic
        ! accessibility list_atom_surface and accessibility are the "charges"
        case ( 3 )
            nq1 = prot1 % nb_atom_surface
            nq2 = prot2 % nb_atom_surface
            p_charge2 => prot2 % p_sogrid % accessibility
            p_charge1 => prot1 % p_sogrid % accessibility
            p_charge_position2 =>  prot2 % p_list_atom_surface
            p_charge_position1 =>  prot1 % p_list_atom_surface
            !! could associate to psgrid even if not associated
            if ( present ( opt_pgrid1 ) ) then
            if ( is_associated ( prot1 % p_sogrid % pset_grid ( 3 ) ) ) then
                !write(*,*) ' prot1 desolv is associated '
                opt_pgrid1 => prot1 % p_sogrid % pset_grid( 3 )
            else
                nullify ( opt_pgrid1 )
            end if
            if ( is_associated ( prot2 % p_sogrid % pset_grid ( 3 ) ) ) then
                !write(*,*) ' prot2 desolv is associated '
                opt_pgrid2 => prot2 % p_sogrid % pset_grid( 3 )
            else
                nullify ( opt_pgrid2 )
            end if
            end if ! opt present

        !! can add more case
        !! lj repulsive
        case ( 4 )
            nq1 = prot1 % nb_atom_surface
            nq2 = prot2 % nb_atom_surface
            p_charge2 => dummy_array
            p_charge1 => dummy_array
            p_charge_position2 =>  prot2 % p_list_atom_surface
            p_charge_position1 =>  prot1 % p_list_atom_surface

            if ( present ( opt_pgrid1 ) ) then
            if ( is_associated ( prot1 % p_sogrid % pset_grid( 4 ) ) ) then
                !write(*,*) ' prot1 desolv is associated '
                opt_pgrid1 => prot1 % p_sogrid % pset_grid( 4 )
            else
                nullify ( opt_pgrid1 )
            end if
            if ( is_associated ( prot2 % p_sogrid % pset_grid( 4 ) ) ) then
                !write(*,*) ' prot2 desolv is associated '
                opt_pgrid2 => prot2 % p_sogrid % pset_grid( 4 )
            else
                nullify ( opt_pgrid2 )
            end if
            end if ! opt present

        ! case lennard-jones, all other no problem with assymetry, assign null or 0
        ! but here need to test
        case ( 5 )
            !! need additional subgrid and need  asymetry
            !write(*,*) 'select charge, case 5 for lennard jones opt_subgrid', opt_subgrid

            ! first test working only for sda_2proteins, now need to consider both protein
            !if ( allocated ( prot1 % p_pos_lenjones % array_position_atom_lj ) ) then
            if ( allocated ( prot1 % p_pos_lenjones % array_position_atom_lj ) .and. &
                associated ( prot2 % p_sogrid % p_ljgrid % array_ljgrid ) ) then
                !write(*,*) 'prot1 array_pos allocated'
                nq1 = prot1 % p_pos_lenjones % array_position_atom_lj ( opt_subgrid ) % nb_pos_atom
                p_charge_position1 =>  prot1 % p_pos_lenjones % array_position_atom_lj ( opt_subgrid ) % p_position_atom
                ! could make an other test
                opt_pgrid2 => prot2 % p_sogrid % p_ljgrid % array_ljgrid ( opt_subgrid )
            end if
            p_charge1 => dummy_array
            p_charge2 => dummy_array

            !if ( allocated ( prot2 % p_pos_lenjones % array_position_atom_lj ) ) then
            if ( allocated ( prot2 % p_pos_lenjones % array_position_atom_lj ) .and. &
                 associated ( prot1 % p_sogrid % p_ljgrid % array_ljgrid ) ) then
                !write(*,*) 'prot2 array_pos allocated'
                nq2 = prot2 % p_pos_lenjones % array_position_atom_lj ( opt_subgrid ) % nb_pos_atom
                p_charge_position2 =>  prot2 % p_pos_lenjones % array_position_atom_lj ( opt_subgrid ) % p_position_atom
                opt_pgrid1 => prot1 % p_sogrid % p_ljgrid % array_ljgrid ( opt_subgrid )
            end if

        end select

        end subroutine select_charge2

        !> Initialize the conformation in case of flexible.\n
        !! replace change_conformation::init_conformation\n
        !! @param this : instance of protein
        !! @param opt_conf_nb :: optional, indicates which conformation. If zero chosses a random one
        subroutine initialize_conformation ( this, opt_conf_nb )
            IMPLICIT NONE
            type ( protein ) :: this
            integer, optional :: opt_conf_nb

            type ( NodeFlex ), pointer :: p_new_node
            real ( kind=8 ) :: unif
            integer :: new_conf, conf_nb
            integer :: counter, status

            !write(*,*) 'Entry protein :: initialize conformation '

            counter = 0
            conf_nb = 0
            nullify ( p_new_node )

            ! if present should force it, better for sdamm
            if ( present ( opt_conf_nb ) ) then
                conf_nb = opt_conf_nb
            end if

            ! if init_conf = -1 OR opt_conf_nb == -1
            if ( ( this % p_flex % plist_flex % initial_conf == -1 ) .or. &
                 ( conf_nb == -1 ) ) then
                 !( present ( opt_conf_nb ) .eqv. .false. ) ) then
                call ggubs_thread_safe ( unif )
                new_conf = int ( unif * ( this % p_flex % plist_flex % total_conf ) ) + 1
                !write(*,*) 'new random conf ', new_conf

            ! else forced into a conformation number, send 0 to get initial_conf
            else if ( conf_nb > 0  ) then
                !write(*,*) 'opt_conf_nb is present, force it to ', conf_nb
                new_conf = conf_nb
            ! finally use initial conf
            else
                !write(*,*) 'by default use initial ', this % p_flex % plist_flex % initial_conf
                new_conf = this % p_flex % plist_flex % initial_conf
            end if

            status = associate_to_conformation ( this % p_flex % plist_flex % p_first, p_new_node, new_conf, counter, 0 )
            ! comment, only for test.
            !if ( associated ( p_new_node ) ) then
                !! if associated, modify p_current_node
                !write (*,*) ' p_new_node has been associated counter ', counter
                this % p_flex % p_current_node => p_new_node
                this % p_flex % current_conf = counter
            !else
            !    write(*,*) 'error in mod_protein::initialize_conformation '
            !end if

            !! associate the pointer p_sogrid
            this % p_sogrid => this % p_flex % p_current_node % p_data
            !write(*,*) 'end initialize_conformation current_conf ', this % p_flex % current_conf

        end subroutine initialize_conformation

!!!!!!!!!!!!!!!!!!!!!!!!!!!! function to deal with flexibility !!!!!!!!!

!! Could in a separate file change-conformation.f90, and not belong to a module, no problems with order of module
!! Could simpler deal with type_calc, array protein and flexible function modules

        !> update array statistics for sda_flex, use timestep rather than integer, may change
        subroutine update_visited_conf ( this, timestep )
            IMPLICIT NONE
            type ( protein ) :: this
            real ( kind=4 ) :: timestep

            !write(*,*) 'update visited conf current', this % p_flex % current_conf
            this % p_flex % visited_conf ( this % p_flex % current_conf ) = &
                this % p_flex % visited_conf ( this % p_flex % current_conf ) + timestep

        end subroutine update_visited_conf

        subroutine merge_visited_conf ( this, to_merge )
            IMPLICIT NONE
            type ( protein ) :: this, to_merge
            integer :: i

            !write(*,*) 'entry merge_visited_conf'

            do i = 1, size ( this % p_flex % visited_conf )
                !$OMP ATOMIC
                this % p_flex % visited_conf ( i ) = this % p_flex % visited_conf ( i ) + to_merge % p_flex % visited_conf ( i )
            end do

        end subroutine

        subroutine print_visited_conf ( this )
            IMPLICIT NONE
            type ( protein ) :: this

            real ( kind=4 ) :: sum_conf
            integer :: i

            sum_conf = sum ( this % p_flex % visited_conf )

            write ( *, * ) 'Statistic of visited conf '
            write ( *, * ) 'total ', sum_conf
            do i = 1, size ( this % p_flex % visited_conf )
                write (*,*) i, this % p_flex % visited_conf ( i ), &
                            this % p_flex % visited_conf ( i ) / sum_conf * 100., '%'
            end do

        end subroutine print_visited_conf

END MODULE mod_protein
